; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ARITH.ASM						!
; 									!
; Description:	Definitions for playing with integer arithmetic.	!
; 									!
; List of FROTH definitions defined in this module:			!
;	VALSE		VFALSE						!
;	TRUE		VTRUE						!
;	0==		ZEROEQU						!
;	0!=		ZERONEQU					!
;	0<		ZEROLESS					!
;	0<=		ZEROLESSEQ					!
;	0>		ZEROGREAT					!
;	0>=		ZEROGREATEQ					!
;	NEGATE		NEGATE						!
;	MIN		MIN						!
;	MAX		MAX						!
;	UMIN		UMIN						!
;	UMAX		UMAX						!
;	+		PLUS						!
;	1+		ONEPLUS						!
;	2+		TWOPLUS						!
;	WordSize+	WORDSIZEPLUS					!
;	-		MINUS						!
;	1-		ONEMINUS					!
;	2-		TWOMINUS					!
;	WordSize-	WORDSIZEMINUS					!
;	*		STAR						!
;	2*		TWOSTAR						!
;	WordSize*	WORDSIZESTAR					!
;	/		SLASH						!
;	2/		TWOSLASH					!
;	WordSize/	WORDSIZESLASH					!
;	/MOD		SLASHMOD					!
;	MOD		IMOD						!
;	*/		STARSLASH					!
;	*/MOD		STARSLASHMOD					!
;	ABS		IABS						!
;	==		EQUAL						!
;	!=		NEQUAL						!
;	<		LESS						!
;	<=		LESSEQ						!
;	>		GREAT						!
;	>=		GREATEQ						!
;	><		GREATLESS					!
;	>=<		GREATEQLESS					!
;	><=		GREATLESSEQ					!
;	>=<=		GREATEQLESSEQ					!
;	U<		ULESS						!
;	U<=		ULESSEQ						!
;	U>		UGREAT						!
;	U>=		UGREATEQ					!
;	U><		UGREATLESS					!
;	U>=<		UGREATEQLESS					!
;	U><=		UGREATLESSEQ					!
;	U>=<=		UGREATEQLESSEQ					!
;	AND		IAND						!
;	OR		IOR						!
;	XOR		IXOR						!
;	SHR		ISHR						!
;	SHL		ISHL						!
;	SETBIT		SETBIT						!
;	RESETBIT	RESETBIT					!
;	COMPLBIT	COMPLBIT					!
;	GETBIT		GETBIT						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions


; ===============================================================
; = VFALSE							=
; =                                                 		=
; = Froth word:		FALSE					=
; =                     	                            	=
; = Stack:		( --- ff )				=
; ===============================================================

		align	4
VFALSE		proc

		; Push value on stack
		IPUSH	FALSE

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_VFALSE:
VFALSE		endp


; ===============================================================
; = VTRUE							=
; =                                                 		=
; = Froth word:		TRUE					=
; =                     	                            	=
; = Stack:		( --- 1 )				=
; ===============================================================

		align	4
VTRUE		proc

		; Push value on stack
		IPUSH	TRUE

		; Verify stack frame overflow, then return to caller
		OVSTACK
		
		; Return to caller
		ret

E_VTRUE:
VTRUE		endp


; ===============================================================
; = ZEROEQU							=
; =                                                 		=
; = Froth word:		0==					=
; =                     	                            	=
; = Stack:		( f --- ~f )				=
; ===============================================================

		align	4
ZEROEQU		proc

		; Verify stack frame
		VSTACK	1

		; Compare value with 0 or not 0
		cmp	PARAM0, FALSE
		je	@@Zero
		mov	PARAM0, FALSE
		ret						; jmp	@@Exit
@@Zero:		mov	PARAM0, TRUE

		; Return to caller
@@Exit:		ret

E_ZEROEQU:
ZEROEQU		endp


; ===============================================================
; = ZERONEQU							=
; =                                                 		=
; = Froth word:		0!=					=
; =                     	                            	=
; = Stack:		( f --- f )				=
; ===============================================================

		align	4
ZERONEQU	proc

		; Verify stack frame
		VSTACK	1

		; Compare value with 0 or not 0
		cmp	PARAM0, FALSE
		je	@@Zero
		mov	PARAM0, TRUE
		ret						; jmp	@@Exit
@@Zero:		mov	PARAM0, FALSE

		; Return to caller
@@Exit:		ret

E_ZERONEQU:
ZERONEQU	endp


; ===============================================================
; = ZEROLESS							=
; =                                                 		=
; = Froth word:		0<					=
; =                     	                            	=
; = Stack:		( n --- f )				=
; =                     	                            	=
; ===============================================================

		align	4
ZEROLESS		proc

		; Verify stack frame
		VSTACK	1

		; Compare value with 0
		cmp	PARAM0, 0
		jl	@@on

@@off:		mov	PARAM0, FALSE
		ret						; jmp	@@Exit

@@on:		mov	PARAM0, TRUE

		; Return to caller
@@Exit:		ret

E_ZEROLESS:
ZEROLESS		endp


; ===============================================================
; = ZEROLESSEQ							=
; =                                                 		=
; = Froth word:		0<=					=
; =                     	                            	=
; = Stack:		( n --- f )				=
; =                     	                            	=
; ===============================================================

		align	4
ZEROLESSEQ	proc

		; Verify stack frame
		VSTACK	1

		; Compare value with 0
		cmp	PARAM0, 0
		jle	@@on

@@off:		mov	PARAM0, FALSE
		ret						; jmp	@@Exit

@@on:		mov	PARAM0, TRUE

		; Return to caller
@@Exit:		ret

E_ZEROLESSEQ:
ZEROLESSEQ	endp


; ===============================================================
; = ZEROGREAT							=
; =                                                 		=
; = Froth word:		0>					=
; =                     	                            	=
; = Stack:		( n --- f )				=
; =                     	                            	=
; ===============================================================

		align	4
ZEROGREAT	proc

		; Verify stack frame
		VSTACK	1

		; Compare value with 0
		cmp	PARAM0, 0
		jg	@@on

@@off:		mov	PARAM0, FALSE
		ret						; jmp	@@Exit

@@on:		mov	PARAM0, TRUE

		; Return to caller
@@Exit:		ret

E_ZEROGREAT:
ZEROGREAT	endp


; ===============================================================
; = ZEROGREATEQ							=
; =                                                 		=
; = Froth word:		0>=					=
; =                     	                            	=
; = Stack:		( n --- f )				=
; =                     	                            	=
; ===============================================================

		align	4
ZEROGREATEQ	proc

		; Verify stack frame
		VSTACK	1

		; Compare value with 0
		cmp	PARAM0, 0
		jge	@@on

@@off:		mov	PARAM0, FALSE
		ret						; jmp	@@Exit

@@on:		mov	PARAM0, TRUE

		; Return to caller
@@Exit:		ret

E_ZEROGREATEQ:
ZEROGREATEQ	endp


; ===============================================================
; = NEGATE							=
; =                                                 		=
; = Code for the forth word:	negate				=
; =                     	                            	=
; = Stack:	( n --- -n )					=
; ===============================================================

		align	4
NEGATE		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the addition
		; -----------------
		
		neg	PARAM0
		

		; Return to caller
		; ----------------
		
		ret

E_NEGATE:
NEGATE		endp


; ===============================================================
; = MIN								=
; =                                                 		=
; = Code for the forth word:	MIN				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; ===============================================================

		align	4
MIN		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ------------
		; Do operation
		; ------------
		
		mov	eax, PARAM0				; Get <n2>
		DISCARD	1					; drop
		cmp	eax, PARAM0				; n2 .cmp. n1
		jge	@@Exit
		mov	PARAM0, eax
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_MIN:
MIN		endp


; ===============================================================
; = MAX								=
; =                                                 		=
; = Code for the forth word:	MAX				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; ===============================================================

		align	4
MAX		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ------------
		; Do operation
		; ------------
		
		mov	eax, PARAM0				; Get <n2>
		DISCARD	1					; drop
		cmp	eax, PARAM0				; n2 .cmp. n1
		jle	@@Exit
		mov	PARAM0, eax
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_MAX:
MAX		endp


; ===============================================================
; = UMIN							=
; =                                                 		=
; = Code for the forth word:	UMIN				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; ===============================================================

		align	4
UMIN		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ------------
		; Do operation
		; ------------
		
		mov	eax, PARAM0				; Get <n2>
		DISCARD	1					; drop
		cmp	eax, PARAM0				; n2 .cmp. n1
		jge	@@Exit
		mov	PARAM0, eax
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UMIN:
UMIN		endp


; ===============================================================
; = UMAX							=
; =                                                 		=
; = Code for the forth word:	UMAX				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; ==============================================================

		align	4
UMAX		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ------------
		; Do operation
		; ------------
		
		mov	eax, PARAM0				; Get <n2>
		DISCARD	1					; drop
		cmp	eax, PARAM0				; n2 .cmp. n1
		jbe	@@Exit
		mov	PARAM0, eax
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UMAX:
UMAX		endp


; ===============================================================
; = PLUS							=
; =                                                 		=
; = Code for the forth word:	+				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n1+n2)				=
; ==============================================================

		align	4
PLUS		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; -----------------
		; Make the addition
		; -----------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1
		add	PARAM0, eax				; n1 + n2
		

		; ----------------
		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_PLUS:
PLUS		endp


; ===============================================================
; = ONEPLUS							=
; =                                                 		=
; = Code for the forth word:	1+				=
; =                     	                            	=
; = Stack:	( n --- n+1)					=
; ==============================================================

		align	4
ONEPLUS		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the addition
		; -----------------
		
		inc	PARAM0					; n++
		

		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_ONEPLUS:
ONEPLUS		endp


; ===============================================================
; = TWOPLUS							=
; =                                                 		=
; = Code for the forth word:	2+				=
; =                     	                            	=
; = Stack:	( n --- n+2)					=
; ===============================================================

		align	4
TWOPLUS		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the addition
		; -----------------
		
		add	PARAM0, 2				; n += 2
		

		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_TWOPLUS:
TWOPLUS		endp


; ===============================================================
; = WORDSIZEPLUS						=
; =                                                 		=
; = Code for the forth word:	WordSize+			=
; =                     	                            	=
; = Stack:	( n --- n + 4 )					=
; ===============================================================

		align	4
WORDSIZEPLUS	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the operation
		; ------------------
		
		add	PARAM0, @WordSize
		

		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_WORDSIZEPLUS:
WORDSIZEPLUS	endp


; ===============================================================
; = MINUS							=
; =                                                 		=
; = Code for the forth word:	-				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n1-n2)				=
; ==============================================================

		align	4
MINUS		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ---------------------
		; Make the substraction
		; ---------------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1
		sub	PARAM0, eax				; n1 + n2
		

		; ----------------
		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_MINUS:
MINUS		endp


; ===============================================================
; = ONEMINUS							=
; =                                                 		=
; = Code for the forth word:	1-				=
; =                     	                            	=
; = Stack:	( n --- n-1)					=
; ===============================================================

		align	4
ONEMINUS		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the operation
		; ------------------
		
		dec	PARAM0					; n--
		

		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_ONEMINUS:
ONEMINUS		endp


; ===============================================================
; = TWOMINUS							=
; =                                                 		=
; = Code for the forth word:	2-				=
; =                     	                            	=
; = Stack:	( n --- n-2)					=
; ===============================================================

		align	4
TWOMINUS		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the operation
		; ------------------
		
		sub	PARAM0, 2				; n -= 2
		

		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_TWOMINUS:
TWOMINUS		endp


; ===============================================================
; = WORDSIZEMINUS						=
; =                                                 		=
; = Code for the forth word:	WordSize-			=
; =                     	                            	=
; = Stack:	( n --- n - 4 )					=
; ===============================================================

		align	4
WORDSIZEMINUS	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the operation
		; ------------------
		
		sub	PARAM0, @WordSize
		

		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_WORDSIZEMINUS:
WORDSIZEMINUS	endp


; ===============================================================
; = STAR							=
; =                                                 		=
; = Code for the forth word:	*				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n1*n2)				=
; =                     	                            	=
; = Usage:	Multiply the two values on top stack.		=
; ===============================================================

		align	4
STAR		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the operation
		; ------------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1
		imul	PARAM0					; EDX:EAX = n1 * n2
		mov	PARAM0, eax
		

		; Return to caller
		; ----------------

		into						; Integer arithmetic overflow test
		ret

E_STAR:
STAR		endp


; ===============================================================
; = TWOSTAR							=
; =                                                 		=
; = Code for the forth word:	2*				=
; =                     	                            	=
; = Stack:	( n --- n * 2 )					=
; ===============================================================

		align	4
TWOSTAR		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the operation
		; ------------------
		
		sal	PARAM0, 1
		

		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_TWOSTAR:
TWOSTAR		endp


; ===============================================================
; = WORDSIZESTAR						=
; =                                                 		=
; = Code for the forth word:	WordSize*			=
; =                     	                            	=
; = Stack:	( n --- n * 4 )					=
; ===============================================================

		align	4
WORDSIZESTAR	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the operation
		; ------------------
		
		sal	PARAM0, 2
		

		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_WORDSIZESTAR:
WORDSIZESTAR	endp


; ===============================================================
; = SLASH							=
; =                                                 		=
; = Code for the forth word:	/				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n1/n2)				=
; =                     	                            	=
; = Usage:	Divide the two values on top stack.		=
; ===============================================================

		align	4
SLASH		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the operation
		; ------------------
		
		mov	eax, PARAM1				; Get n1
		cdq						; EDX:EAX
		idiv	PARAM0					; n1 / n2
		DISCARD	1
		mov	PARAM0, eax
		

		; Return to caller
		; ----------------
		
		ret

E_SLASH:
SLASH		endp


; ===============================================================
; = TWOSLASH							=
; =                                                 		=
; = Code for the forth word:	2/				=
; =                     	                            	=
; = Stack:	( n --- n / 2)					=
; ===============================================================

		align	4
TWOSLASH	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the operation
		; ------------------
		
		sar	PARAM0, 1
		

		; Return to caller
		; ----------------
		
		ret

E_TWOSLASH:
TWOSLASH	endp


; ===============================================================
; = WORDSIZESLASH						=
; =                                                 		=
; = Code for the forth word:	WordSize/			=
; =                     	                            	=
; = Stack:	( n --- n / 4 )					=
; ===============================================================

		align	4
WORDSIZESLASH	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the operation
		; ------------------
		
		sar	PARAM0, 2
		

		; Return to caller
		; ----------------
		
		ret

E_WORDSIZESLASH:
WORDSIZESLASH	endp


; ===============================================================
; = SLASHMOD							=
; =                                                 		=
; = Code for the forth word:	/MOD				=
; =                     	                            	=
; = Stack:	( n1 n2 --- remainder quotient )		=
; =                     	                            	=
; = Usage:	Divide the two values on top stack.		=
; ===============================================================

		align	4
SLASHMOD		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the operation
		; ------------------
		
		mov	eax, PARAM1				; Get n1
		cdq						; EDX:EAX
		idiv	PARAM0					; n1 / n2
		mov	PARAM1, edx				; remainder
		mov	PARAM0, eax				; quotient
		

		; Return to caller
		; ----------------
		
		ret

E_SLASHMOD:
SLASHMOD	endp


; ===============================================================
; = IMOD							=
; =                                                 		=
; = Code for the forth word:	MOD				=
; =                     	                            	=
; = Stack:	( n1 n2 --- remainder )				=
; =                     	                            	=
; = Usage:	Divide the two values on top stack.		=
; ===============================================================

		align	4
IMOD		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the operation
		; ------------------
		
		mov	eax, PARAM1				; Get n1
		cdq						; EDX:EAX
		idiv	PARAM0					; EDX:EAX / [esi]
		DISCARD	1
		mov	PARAM0, edx

		; Return to caller
		; ----------------
		
		ret

E_IMOD:
IMOD		endp


; ===============================================================
; = STARSLASH							=
; =                                                 		=
; = Code for the forth word:	*/				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- n1*n2/n3)			=
; ===============================================================

		align	4
STARSLASH	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3

		
		; Perform operation
		; -----------------

		mov	eax, PARAM2				; Get n1
		imul	PARAM1					; EDX:EAX = n1 * n2
		mov	ebx, PARAM0
		DISCARD	2					; drop drop
		idiv	ebx					; EDX:EAX / n3
		mov	PARAM0, eax


		; Return to caller
		; ----------------
		
		ret

E_STARSLASH:
STARSLASH	endp


; ===============================================================
; = STARSLASHMOD						=
; =                                                 		=
; = Code for the forth word:	*/MOD				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- mod n1*n2/n3 )			=
; ===============================================================

		align	4
STARSLASHMOD	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Perform operation
		; -----------------

		mov	eax, PARAM2				; Get n1
		mov	ebx, PARAM0
		DISCARD	1
		imul	PARAM0					; EDX:EAX = n1 * n2
		idiv	ebx					; EDX:EAX / n3
		mov	PARAM0, eax
		mov	PARAM1, edx


		; Return to caller
		; ----------------
		
		ret

E_STARSLASHMOD:
STARSLASHMOD	endp


; ===============================================================
; = IABS							=
; =                                                 		=
; = Code for the forth word:	ABS				=
; =                     	                            	=
; = Stack:	( n1 --- n2 )					=
; ===============================================================

		align	4
IABS		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the operation
		; ------------------
		
		cmp	PARAM0, 0
		jge	@@exit
		neg	PARAM0


		; Return to caller
		; ----------------
		
@@exit:		ret

E_IABS:
IABS		endp


; ===============================================================
; = EQUAL							=
; =                                                 		=
; = Code for the forth word:	==				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values.				=
; ===============================================================

		align	4
EQUAL		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the test
		; -------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1					; drop
		cmp	PARAM0, eax				; n1 == n2 ?
		jne	@@LL1
		mov	PARAM0, TRUE
		ret						; jmp	@@Exit
@@LL1:		mov	PARAM0, FALSE


		; Return to caller
		; ----------------

@@Exit:		ret

E_EQUAL:
EQUAL		endp


; ===============================================================
; = NEQUAL							=
; =                                                 		=
; = Code for the forth word:	!=				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values.				=
; ===============================================================

		align	4
NEQUAL		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the test
		; -------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1					; drop
		cmp	PARAM0, eax				; n1 == n2 ?
		jne	@@LL1
		mov	PARAM0, FALSE
		ret						; jmp	@@Exit
@@LL1:		mov	PARAM0, TRUE
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_NEQUAL:
NEQUAL		endp


; ===============================================================
; = LESS							=
; =                                                 		=
; = Code for the forth word:	<				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values signed.			=
; ===============================================================

		align	4
LESS		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the test
		; -------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1					; drop
		cmp	PARAM0, eax				; n1 == n2 ?
		jl	@@LL1
		mov	PARAM0, FALSE
		ret						; jmp	@@Exit
@@LL1:		mov	PARAM0, TRUE
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_LESS:
LESS		endp


; ===============================================================
; = LESSEQ							=
; =                                                 		=
; = Code for the forth word:	<=				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values signed.			=
; ===============================================================

		align	4
LESSEQ		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the test
		; -------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1					; drop
		cmp	PARAM0, eax				; n1 == n2 ?
		jle	@@LL1
		mov	PARAM0, FALSE
		ret						; jmp	@@Exit
@@LL1:		mov	PARAM0, TRUE
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_LESSEQ:
LESSEQ		endp


; ===============================================================
; = GREAT							=
; =                                                 		=
; = Code for the forth word:	>				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values signed.			=
; ===============================================================

		align	4
GREAT		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the test
		; -------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1					; drop
		cmp	PARAM0, eax				; n1 == n2 ?
		jg	@@LL1
		mov	PARAM0, FALSE
		ret						; jmp	@@Exit
@@LL1:		mov	PARAM0, TRUE
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_GREAT:
GREAT		endp


; ===============================================================
; = GREATEQ							=
; =                                                 		=
; = Code for the forth word:	>=				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values signed.			=
; ===============================================================

		align	4
GREATEQ		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the test
		; -------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1					; drop
		cmp	PARAM0, eax				; n1 == n2 ?
		jge	@@LL1
		mov	PARAM0, FALSE
		ret						; jmp	@@Exit
@@LL1:		mov	PARAM0, TRUE
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_GREATEQ:
GREATEQ		endp


; ===============================================================
; = GREATLESS							=
; =                                                 		=
; = Code for the forth word:	><				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 < n1 < n3				=
; ===============================================================

		align	4
GREATLESS	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jng	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnl	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_GREATLESS:
GREATLESS	endp


; ===============================================================
; = GREATEQLESS							=
; =                                                 		=
; = Code for the forth word:	>=<				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 <= n1 < n3				=
; ===============================================================

		align	4
GREATEQLESS	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jnge	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnl	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_GREATEQLESS:
GREATEQLESS	endp


; ===============================================================
; = GREATLESSEQ							=
; =                                                 		=
; = Code for the forth word:	><=				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 < n1 <= n3				=
; ===============================================================

		align	4
GREATLESSEQ	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jng	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnle	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_GREATLESSEQ:
GREATLESSEQ	endp


; ===============================================================
; = GREATEQLESSEQ						=
; =                                                 		=
; = Code for the forth word:	>=<=				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 <= n1 <= n3				=
; ===============================================================

		align	4
GREATEQLESSEQ	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jnge	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnle	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_GREATEQLESSEQ:
GREATEQLESSEQ	endp


; ===============================================================
; = ULESS							=
; =                                                 		=
; = Code for the forth word:	U<				=
; =                     	                            	=
; = Stack:	( u1 u2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values unsigned.			=
; ===============================================================

		align	4
ULESS		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the test
		; -------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1					; drop
		cmp	PARAM0, eax				; n1 == n2 ?
		jb	@@LL1
		mov	PARAM0, FALSE
		ret						; jmp @@exit
@@LL1:		mov	PARAM0, TRUE
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_ULESS:
ULESS		endp


; ===============================================================
; = ULESSEQ							=
; =                                                 		=
; = Code for the forth word:	U<=				=
; =                     	                            	=
; = Stack:	( u1 u2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values unsigned.			=
; ===============================================================

		align	4
ULESSEQ		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the test
		; -------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1					; drop
		cmp	PARAM0, eax				; n1 == n2 ?
		jbe	@@LL1
		mov	PARAM0, FALSE
		ret						; jmp @@exit
@@LL1:		mov	PARAM0, TRUE
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_ULESSEQ:
ULESSEQ		endp


; ===============================================================
; = UGREAT							=
; =                                                 		=
; = Code for the forth word:	U>				=
; =                     	                            	=
; = Stack:	( u1 u2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values unsigned.			=
; ===============================================================

		align	4
UGREAT		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the test
		; -------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1					; drop
		cmp	PARAM0, eax				; n1 == n2 ?
		ja	@@LL1
		mov	PARAM0, FALSE
		ret						; jmp @@exit
@@LL1:		mov	PARAM0, TRUE
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UGREAT:
UGREAT		endp


; ===============================================================
; = UGREATEQ							=
; =                                                 		=
; = Code for the forth word:	U>=				=
; =                     	                            	=
; = Stack:	( u1 u2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values unsigned.			=
; ===============================================================

		align	4
UGREATEQ	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the test
		; -------------
		
		mov	eax, PARAM0				; Get n2
		DISCARD	1					; drop
		cmp	PARAM0, eax				; n1 == n2 ?
		jae	@@LL1
		mov	PARAM0, FALSE
		ret						; jmp	@@Exit
@@LL1:		mov	PARAM0, TRUE
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UGREATEQ:
UGREATEQ	endp


; ===============================================================
; = UGREATLESS							=
; =                                                 		=
; = Code for the forth word:	U><				=
; =                     	                            	=
; = Stack:	( u1 u2 u3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 < n1 < n3				=
; ===============================================================

		align	4
UGREATLESS	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jna	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnb	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret						; jmp	@@Exit
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UGREATLESS:
UGREATLESS	endp


; ===============================================================
; = UGREATEQLESS						=
; =                                                 		=
; = Code for the forth word:	U>=<				=
; =                     	                            	=
; = Stack:	( u1 u2 u3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 <= n1 < n3				=
; ===============================================================

		align	4
UGREATEQLESS	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jnae	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnb	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret						; jmp	@@Exit
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UGREATEQLESS:
UGREATEQLESS	endp


; ===============================================================
; = UGREATLESSEQ						=
; =                                                 		=
; = Code for the forth word:	U><=				=
; =                     	                            	=
; = Stack:	( u1 u2 u3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 < n1 <= n3				=
; ===============================================================

		align	4
UGREATLESSEQ	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jna	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnbe	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret						; jmp	@@Exit
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UGREATLESSEQ:
UGREATLESSEQ	endp


; ===============================================================
; = UGREATEQLESSEQ						=
; =                                                 		=
; = Code for the forth word:	U>=<=				=
; =                     	                            	=
; = Stack:	( u1 u2 u3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 <= n1 <= n3				=
; ===============================================================

		align	4
UGREATEQLESSEQ	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jnae	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnbe	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret						; jmp	@@Exit
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UGREATEQLESSEQ:
UGREATEQLESSEQ	endp


; ===============================================================
; = IAND							=
; =                                                 		=
; = Code for the forth word:	AND				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; =                     	                            	=
; = Usage:	n3 = n1 AND n2					=
; ===============================================================

		align	4
IAND		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the logical operation
		; --------------------------
		
		mov	eax, PARAM0				; Get n1
		DISCARD	1
		and	PARAM0, eax				; n1 AND n2


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_IAND:
IAND		endp


; ===============================================================
; = IOR								=
; =                                                 		=
; = Code for the forth word:	OR				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; =                     	                            	=
; = Usage:	n3 = n1 OR n2					=
; ===============================================================

		align	4
IOR		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the logical operation
		; --------------------------
		
		mov	eax, PARAM0				; Get n1
		DISCARD	1
		or	PARAM0, eax				; n1 OR n2


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_IOR:
IOR		endp


; ===============================================================
; = XOR								=
; =                                                 		=
; = Code for the forth word:	XOR				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; =                     	                            	=
; = Usage:	n3 = n1 XOR n2					=
; ===============================================================

		align	4
IXOR		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the logical operation
		; --------------------------
		
		mov	eax, PARAM0				; Get n1
		DISCARD	1
		xor	PARAM0, eax				; n1 XOR n2


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_IXOR:
IXOR		endp


; ===============================================================
; = ISHL							=
; =                                                 		=
; = Code for the forth word:	shl				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; =                     	                            	=
; = Usage:	n3 = n1 SHL n2					=
; ===============================================================

		align	4
ISHL		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the logical operation
		; --------------------------
		
		mov	ecx, PARAM0				; Get n1
		DISCARD	1
		shl	PARAM0, cl				; n1 shl n2


		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_ISHL:
ISHL		endp


; ===============================================================
; = ISHR							=
; =                                                 		=
; = Code for the forth word:	shr				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; =                     	                            	=
; = Usage:	n3 = n1 SHR n2					=
; ===============================================================

		align	4
ISHR		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the logical operation
		; --------------------------
		
		mov	ecx, PARAM0				; Get n1
		DISCARD	1
		shr	PARAM0, cl				; n1 shr n2


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_ISHR:
ISHR		endp


; ===============================================================
; = SetBit							=
; =                                                 		=
; = Code for the forth word:	SetBit				=
; =                     	                            	=
; = Stack:	( n addr --- )					=
; ===============================================================

		align	4
SETBIT		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the logical operation
		; --------------------------
		
		mov	eax, PARAM1				; Get <n>
		mov	ebx, PARAM0				; Get <adr>
		bound	ebx, qword ptr DDA_BOUND		; Check addr against bound
		bts	[ebx], eax				; Set bit
		DISCARD	2					; ddrop


		; Return to caller
		; ----------------
		
		ret

E_SETBIT:
SETBIT		endp


; ===============================================================
; = ResetBit							=
; =                                                 		=
; = Code for the forth word:	ResetBit			=
; =                     	                            	=
; = Stack:	( n addr --- )					=
; ===============================================================

		align	4
RESETBIT	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the logical operation
		; --------------------------
		
		mov	eax, PARAM1				; Get <n>
		mov	ebx, PARAM0				; Get <adr>
		bound	ebx, qword ptr DDA_BOUND		; Check addr against bound
		btr	[ebx], eax				; Set bit
		DISCARD	2					; ddrop


		; Return to caller
		; ----------------
		
		ret

E_RESETBIT:
RESETBIT	endp


; ===============================================================
; = ComplBit							=
; =                                                 		=
; = Code for the forth word:	ComplBit			=
; =                     	                            	=
; = Stack:	( n addr --- )					=
; ===============================================================

		align	4
COMPLBIT	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the logical operation
		; --------------------------
		
		mov	eax, PARAM1				; Get <n>
		mov	ebx, PARAM0				; Get <adr>
		bound	ebx, qword ptr DDA_BOUND		; Check addr against bound
		btc	[ebx], eax				; Set bit
		DISCARD	2					; ddrop


		; Return to caller
		; ----------------
		
		ret

E_COMPLBIT:
COMPLBIT	endp


; ===============================================================
; = GetBit							=
; =                                                 		=
; = Code for the forth word:	GetBit				=
; =                     	                            	=
; = Stack:	( n addr --- f )				=
; ===============================================================

		align	4
GETBIT		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the logical operation
		; --------------------------
		
		mov	ebx, PARAM0				; Get <adr>
		bound	ebx, qword ptr DDA_BOUND		; Check addr against bound
		DISCARD	1					; drop
		mov	eax, PARAM0				; Get <n>
		bt	[ebx], eax				; Set bit
		jc	@@On
@@Off:		mov	PARAM0, FALSE
		ret						; jmp	@@Exit
@@On:		mov	PARAM0, TRUE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_GETBIT:
GETBIT		endp


code32		ends
		end

