; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ASM.ASM							!
; 									!
; Description:	Definitions for playing with assembler.			!
; 									!
; List of FROTH definitions defined in this module:			!
;	OPENASMBRACKET	[:						!
;	CLOSEASMBRACKET	:]						!
;	ASM		ASM						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc


; ===============================================================
; = OPENASMBRACKET						=
; =                                                         	=
; = Code for the forth word:    [:                		=
; =                                                         	=
; = Stack:  ( --- )                 				=
; ===============================================================

		align	4
OPENASMBRACKET	proc
	
		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		SUMMON	QCOMPILE
		

		; -----------------------------
		; Is the [] mode already open ?
		; -----------------------------
		
		cmp	[CDA_BRACKET_MODE], FALSE
		je	@@Ok
		
		mov	ah, ERR_BRACKETALOPEN			; [ mode already open.
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		
@@Ok:


		; ---------------		
		; Open the [ mode
		; ---------------
		
		mov	[CDA_BRACKET_MODE], TRUE


		; ----------------	
		; Return to caller
		; ----------------	
		
		ret

E_OPENASMBRACKET:
OPENASMBRACKET	endp


; ===============================================================
; = CLOSEBRACKET						=
; =                                                         	=
; = Code for the forth word:   :]                		=
; =                                                         	=
; = Stack:  ( --- )                 				=
; ===============================================================

		align	4
CLOSEASMBRACKET	proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		SUMMON	QCOMPILE
		
		
		; -------------------------
		; Is the [] mode not open ?
		; -------------------------
		
		cmp	[CDA_BRACKET_MODE], FALSE
		jne	@@Ok
		
		mov	ah, ERR_BRACKETNOTOPEN			; [ mode not open.
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		
@@Ok:


		; ----------------		
		; Close the [ mode
		; ----------------
		
		mov	[CDA_BRACKET_MODE], FALSE


		; ----------------	
		; Return to caller
		; ----------------	
		
		ret

E_CLOSEASMBRACKET:
CLOSEASMBRACKET	endp


; ===============================================================
; = ASM                            				=
; =                                                         	=
; = Code for the froth word:    ASM                		=
; =                                                         	=
; = Stack:  ( --- )                 				=
; ===============================================================

		align	4
ASM		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		SUMMON	QCOMPILE
		
		
		; ----------------
		; Return to caller
		; ----------------

		ret

E_ASM:
ASM		endp


code32		ends
		end
