; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	BIO.ASM							!
; 									!
; Description:	Definitions for playing with basic I/O.			!
; 									!
; List of FROTH definitions defined in this module:			!
;	EMIT		EMIT						!
;	PEMIT		PEMIT						!
;	PTYPE		PTYPE						!
;	PCR		PCR						!
;	PSPACE		PSPACE						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Ed.inc                      		; Constants, Structurs, Variables used by FROSTED

EKEY_ALT_1	equ	0802h				; Alt-1 key (used into ReadLine)
EKEY_ALT_8	equ	EKEY_ALT_1+7			; Alt-8 key (used into ReadLine)


; ===============================================================
; = COUT							=
; =								=
; = Low level function to print a character.			=
; =								=
; = I) AL : the character to print				=
; =								=
; = O) none							=
; =								=
; = M) none							=
; ===============================================================

		align	4
COUT		proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; Save regs which will be modified
		push	ax

		; Call BIOS function
		mov 	v86r_al, al				; Char to print
		mov 	v86r_ah, 0Eh				; Display char function
		mov 	al, 10h					; Call INT 10h
		int 	33h
		
		; Restore regs
		pop 	ax

		; CTRL-BREAK allowed
		SUMMON	CTRLBREAK_REST

		; return
		ret

COUT		endp


; ===============================================================
; = xWhereXY							=
; =								=
; = return where is the cursor (x is the column, y the row).	=
; =								=
; = I)	none							=
; =								=
; = O)	AH  : row (from 0 to ..)				=
; =	AL  : column (from 0 to ..)				=
; =								=
; = M)	none							=
; =								=
; ===============================================================

		align	4
xWhereXY		proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Call BIOS function
		; ------------------

		call	doGetVideoPage				; PAGE#
		mov 	v86r_ah, 03h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		mov	ah, v86r_dh				; Output: row #
		mov	al, v86r_dl				; Output: col #


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

xWhereXY		endp


; ===============================================================
; = xGotoXY							=
; =								=
; = Set the cursor at (x, y) (x is the column, y the row).	=
; =								=
; = I)	AH  : row (from 0 to ..)				=
; =	AL  : column (from 0 to ..)				=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; =								=
; ===============================================================

		align	4
xGotoXY		proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; -------
		; Save AL
		; -------
		
		push	ax
		

		; ------------------
		; Call BIOS function
		; ------------------

		call	doGetVideoPage			; PAGE#
		mov	v86r_dh, ah			; Input: row #
		mov	v86r_dl, al			; Input: col #
		mov 	v86r_ah, 02h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------

		pop	ax		
		ret

xGotoXY		endp


; ===============================================================
; = ReadLine							=
; =								=
; = Low level function to read a line of characters from the	=
; = terminal.							=
; =								=
; = I) EDX : address of buffer to store line			=
; =    ECX : max nb of char					=
; =								=
; = O) none							=
; =								=
; = M) EAX and ECX only						=
; =								=
; ===============================================================

		align	4
ReadLine		proc

		; --------------
		; Save used regs
		; --------------

		mov	[@@ebp], ebp
		mov	[@@edi], edi
		mov	[@@esi], esi
		mov	[@@edx], edx
		mov	[@@ecx], ecx
		mov	[@@ebx], ebx
		mov	esi, ecx
		mov	edi, edx


		; ----------------------------
		; Read cursor initial position
		; ----------------------------

		call	xWhereXY
		mov	bp, ax					; Save position
		mov	cl, al					; Save 1st column position


		; ---------------
		; Initializations
		; ---------------
		
		clr	ebx					; Nb of char read
		clr	edx					; Position
		jmp	@@ReadChar


		; ---------------
		; Redraw the line
		; ---------------


@@RedrawLine:	mov	ax, bp
		call	xGotoXY
		push	ecx
		clr	ecx
@@RL2:		mov	al, [edi + ecx]
		call	COUT
		inc	cl
		cmp	cl, bl
		jle	@@RL2
		pop	ecx
		mov	ax, bp
		mov	al, dl					; Column #
		add	al, cl					; Add 1st column position
		call	xGotoXY


		; ------------------------------
		; Read a character (without echo)
		; ------------------------------
		
@@ReadChar:	
		SUMMON	CharIn

		if UNDERDEV			; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

		; Exit to DOS ?
		cmp	ax, 81Eh				; Alt-Q
		je	_exit

		endif				; >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


		; ----------
		; Exit on CR
		; ----------
		
		cmp	ax, EKEY_CR				; CR ?
		je	@@ReadDone


		;-------------------------------------------
		; Select a given video page (Alt-1 to Alt-8)
		;-------------------------------------------

		cmpz	edx, 0
		jne	@@NoAlt_1_8		
		cmp	ax, EKEY_ALT_1
		jb	@@NoAlt_1_8
		cmp	ax, EKEY_ALT_8
		ja	@@NoAlt_1_8

		push	esi edi
		mov	edi, [DDIR_START]       		; Data table: DDIR
		mov	esi, [PStackStart]			; Stack pointer for thread #0
		GETFVAR	ebx, D_NBPAGES
		sub	ax, EKEY_ALT_1
		cmp	ax, bx
		jge	@@ChangePageDone
		cwde						; movzx eax, ax
		pushad
		sub	esi, @WordSize
		mov	PARAM0, eax
		SUMMON	CHANGEPAGE
		popad

		clr	ebx					; No char entered
		call	xWhereXY
		mov	bp, ax					; Save position
		mov	cl, al					; Save 1st column position

@@ChangePageDone:	pop	edi esi
		jmp	@@ReadChar


@@NoAlt_1_8:	


		; ----------
		; Home key ?
		; ----------
		
		cmp	ax, EKEY_HOME
		je	@@Home
		cmp	ax, EKEY_HOME + 80h
		jne	@@NotHome
@@Home:		clr	dl
		mov	ax, bp
		mov	al, cl					; Goto 1st column position
		call	xGotoXY
		jmp	@@ReadChar
@@NotHome:
		
		
		; ---------
		; End key ?
		; ---------
		
		cmp	ax, EKEY_END
		je	@@End
		cmp	ax, EKEY_END + 80h
		jne	@@NotEnd
@@End:		mov	dl, bl
		mov	ax, bp
		mov	al, bl
		add	al, cl					; Add 1st column position
		call	xGotoXY
		jmp	@@ReadChar
@@NotEnd:
		
		
		; ----------
		; Left key ?
		; ----------
		
		cmp	ax, EKEY_LEFT
		je	@@Left
		cmp	ax, EKEY_LEFT + 80h
		jne	@@NotLeft
@@Left:		cmp	dl, 0
		je	@@ReadChar
		mov	ax, bp
		dec	dl
		mov	al, dl					; Column #
		add	al, cl					; Add 1st column position
		call	xGotoXY
		jmp	@@ReadChar
@@NotLeft:


		; -----------
		; Right key ?
		; -----------
		
		cmp	ax, EKEY_RIGHT
		je	@@Right
		cmp	ax, EKEY_RIGHT + 80h
		jne	@@NotRight
@@Right:	cmp	dl, bl
		je	@@ReadChar
		mov	ax, bp
		inc	dl
		mov	al, dl					; Column #
		add	al, cl					; Add 1st column position
		call	xGotoXY
		jmp	@@ReadChar
@@NotRight:


		; -------------------------
		; DEL ? (delete under char)
		; -------------------------
		
		cmp	ax, EKEY_DEL
		je	@@Del
		cmp	ax, EKEY_DEL + 80h
		jne	@@NotDEL
@@Del:		push	ecx esi edi
		mov	ecx, ebx
		sub	ecx, edx				; Nb of chars to move
		add	edi, edx				; Destination
		mov	esi, edi
		inc	esi					; Source
		cld						; Inc ESI and EDI
		rep	movsb  		          	    	; Move the block
		pop	edi esi ecx
		dec	bl
		mov	byte ptr [edi + ebx], ' '
		jmp	@@RedrawLine
@@NotDEL:
		
		
		; -------------------------------
		; INS ? (insert space under char)
		; -------------------------------
		
		cmp	ax, EKEY_INS
		je	@@Ins
		cmp	ax, EKEY_INS + 80h
		jne	@@NotINS
@@Ins:		cmp	ebx, esi
		je	@@NotINS
		push	ecx esi edi
		mov	ecx, ebx
		sub	ecx, edx				; Nb of chars to move
		add	edi, ebx				; Destination
		mov	esi, edi
		dec	esi					; Source
		std						; Dec ESI and EDI
		rep	movsb  		          	    	; Move the block
		cld
		pop	edi esi ecx
		mov	byte ptr [edi + edx], ' '
		inc	bl
		jmp	@@RedrawLine
@@NotINS:
		
		
		; -----------------------
		; BS ? (delete left char)
		; -----------------------
		
		cmp	ax, EKEY_BS
		jne	@@NotBS
		cmp	dl, 0
		je	@@ReadChar
		dec	dl
		push	ecx esi edi
		mov	ecx, ebx
		sub	ecx, edx				; Nb of chars to move
		add	edi, edx				; Destination
		mov	esi, edi
		inc	esi					; Source
		cld						; Inc ESI and EDI
		rep	movsb  		          	    	; Move the block
		pop	edi esi ecx
		dec	bl
		mov	byte ptr [edi + ebx], ' '
		jmp	@@RedrawLine
@@NotBS:
		
		
		; ----------------
		; Ctrl-Right key ?
		; ----------------
		
		cmp	ax, EKEY_CTRL_RIGHT
		je	@@CtrlRight
		cmp	ax, EKEY_CTRL_RIGHT + 80h
		jne	@@NotCtrlRight
@@CtrlRight:	cmp	dl, bl
		je	@@ReadChar
		cmp	byte ptr [edi + edx], ' '
		jne	@@CR2
@@CR4:		inc	dl
		cmp	dl, bl
		je	@@CR3
		cmp	byte ptr [edi + edx], ' '
		je	@@CR4
		jmp	@@CR3
@@CR2:		inc	dl
		cmp	dl, bl
		je	@@CR3
		cmp	byte ptr [edi + edx], ' '
		jne	@@CR2
		jmp	@@CR4
@@CR3:		mov	ax, bp
		mov	al, dl					; Column #
		add	al, cl					; Add 1st column position
		call	xGotoXY
		jmp	@@ReadChar
@@NotCtrlRight:


		; ---------------
		; Ctrl-Left key ?
		; ---------------
		
		cmp	ax, EKEY_CTRL_LEFT
		je	@@CtrlLeft
		cmp	ax, EKEY_CTRL_LEFT + 80h
		jne	@@NotCtrlLeft
@@CtrlLeft:	cmp	dl, 0
		je	@@ReadChar
		dec	dl
		cmp	dl, 0
		je	@@ReadChar
		cmp	byte ptr [edi + edx], ' '
		jne	@@CL2
@@CL4:		dec	dl
		cmp	dl, 0
		je	@@CL3
		cmp	byte ptr [edi + edx], ' '
		je	@@CL4
@@CL2:		dec	dl
		cmp	dl, 0
		je	@@CL3
		cmp	byte ptr [edi + edx], ' '
		jne	@@CL2
		inc	dl
@@CL3:		mov	ax, bp
		mov	al, dl					; Column #
		add	al, cl					; Add 1st column position
		call	xGotoXY
		jmp	@@ReadChar
@@NotCtrlLeft:


		; ------------------
		; Char not allowed ?
		; ------------------
		
		cmp	ax, 0100h
		jae	@@ReadChar
		cmp	ax, 020h
		jl	@@ReadChar

		
		; ------------------
		; Max size reached ?
		; ------------------
		
		cmp	ebx, esi
		je	@@ReadChar
		

		; ---------
		; Show char
		; ---------
		
		call	COUT


		; ----------
		; Store char
		; ----------

		mov	[edi + edx], al
		inc	edx				; Update position
		cmp	edx, ebx
		jle	@@NotUpdtRead
		mov	ebx, edx			; Update nb of char read
@@NotUpdtRead:


		; --------
		; Do again
		; --------
		
		jmp	@@ReadChar
		


@@ReadDone:

		; ---------------------
		; End buffer with '\0\0'
		; ---------------------

		mov	word ptr [edi + ebx], 0


		; ----------------------
		; Put cursor on end text
		; ----------------------

		mov	ax, bp
		mov	al, bl
		add	al, cl					; Add 1st column position
		call	xGotoXY					; Row in AH,  Col in AL


		; ----------------
		; Return to caller
		; ----------------

		mov	ebp, [@@ebp]
		mov	edi, [@@edi]
		mov	esi, [@@esi]
		mov	edx, [@@edx]
		mov	ecx, [@@ecx]
		mov	ebx, [@@ebx]
		ret
	
		align	4
@@ebp		dd	?
@@edi		dd	?
@@esi		dd	?
@@edx		dd	?
@@ecx		dd	?
@@ebx		dd	?

E_ReadLine:
ReadLine		endp


; ===============================================================
; = POUT							=
; =								=
; = Low level function to print a character (to printer).	=
; =								=
; = I) AL : the character to print				=
; =								=
; = O) none							=
; =								=
; = M) none							=
; =								=
; ===============================================================

		align	4
POUT		proc

jmp	COUT	; !!!

		; Save regs which will be modified
		push	ax

		; Call DOS function
		mov 	v86r_dl, al			; Char to print
		mov 	v86r_ah, 05h			; Print char function
		mov 	al, 21h				; Call INT 21h
		int 	33h
		
		; Restore regs
		pop 	ax

		; return
		ret

POUT		endp


; ===============================================================
; = EMIT							=
; =								=
; = Code for the froth word:	EMIT				=
; =								=
; = Stack:	( c --- )					=
; =								=
; ===============================================================

		align	4
EMIT		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Character to print
		mov	eax, [esi]
		add	esi, @WordSize
		call	COUT

		; Return to caller
		ret
E_EMIT:

EMIT		endp


; ===============================================================
; = PEMIT							=
; =								=
; = Code for the froth word:	PEMIT				=
; =								=
; = Stack:	( c --- )					=
; =								=
; ===============================================================

		align	4
PEMIT		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Character to print
		mov	eax, [esi]
		add	esi, 4
		call	POUT

		; Return to caller
		ret

E_PEMIT:
PEMIT		endp


; ===============================================================
; = PTYPE							=
; =								=
; = Code for the froth word:	PTYPE				=
; =								=
; = Stack:	( adr --- )					=
; =								=
; ===============================================================

		align	4
PTYPE		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; String address
		mov	ebx, [esi]
		add	esi, 4

		; Loop until end of string		
@@LL1:		mov	al, [ebx]
		cmp	al, 0
		je	@@LL2
		call	POUT
		inc	ebx
		jmp	@@LL1

@@LL2:
		; Return to caller
		ret

E_PTYPE:
PTYPE		endp


; ===============================================================
; = PCR								=
; =								=
; = Code for the forth word:	PCR				=
; =								=
; = Stack:	( --- )        	                           	=
; =								=
; ===============================================================

		align	4
PCR		proc

		; Print CR LF
		mov	al, 0Dh
		call	POUT
		mov	al, 0Ah
		call	POUT

		; Return to caller		
		ret
E_PCR:

PCR		endp


; ===============================================================
; = PSPACE							=
; =								=
; = Code for the forth word:	PSPACE				=
; =								=
; = Stack:	( --- )        	                           	=
; =								=
; ===============================================================

		align	4
PSPACE		proc

		; printf("%c", ' ')
		mov	al, ' '
		call	POUT

		; Return to caller		
		ret

E_PSPACE:
PSPACE		endp


; ===============================================================
; = RdLoadFileLine						=
; =								=
; = Low level function to read lines from file loaded.		=
; =								=
; = O)	EDX	Buffer line address.				=
; =								=
; ===============================================================

		align	4
RdLoadFileLine	proc

		; EBX is set to the start of the next line into the buffer
@@FirstLine:	mov	eax, [LoadFilesNb]
		dec	eax
		imul	ecx, eax, size loadfileentry
		add	ecx, [LoadFilesInfos]
		mov	ebx, [ecx.loadfile_inBuff]		; Current position in the buffer		
		cmp	ebx, LOADFILEBUFFSZ
		jge	@@LL10					; Read next block from file in this case
		mov	edx, [ecx.loadfile_buff]		; Pointer to buffer

		mov	ecx, ebx
		dec	ecx
@@LL1:		inc	ecx
		mov	al, [edx + ecx]				; Next char into the buffer
		cmpz	al, 0
		je	@@LL10					; Read next block from the file
		cmp	al, 0Dh					; CR ?
		jne	@@LL1

		inc	ecx
		mov	al, [edx + ecx]
		cmpz	al, 0
		je	@@LL10					; Read next block from the file
		cmp	al, 0Ah					; LF ?
		je	@@LL20

		mov	al, '?'					; !!!
		call	COUT
		jmp	_exit	

		
		; Read next block from the file
		; -----------------------------

@@LL10:		; LoadFile infos address in ECX
		mov	eax, [LoadFilesNb]
		dec	eax
		imul	edx, eax, size loadfileentry
		add	edx, [LoadFilesInfos]

		; Move the file pointer
		mov	al, SEEK_SET				; Move file pointer from beginning of file
		mov	bx, [edx.loadfile_hdl]			; File handle
		mov	ecx, [edx.loadfile_inFile]		; Current position in the file
		add	ecx, [edx.loadfile_inBuff]		; Add current position in the buffer
		call	doSeekFile
		jc	QuoteLoadErr

		; Reset the pointer into the buffer
		mov	eax, [edx.loadfile_inFile]		; Current position in the file
		add	eax, [edx.loadfile_inBuff]		; Add current position in the buffer
		mov	[edx.loadfile_inFile], eax		; Current position in the file
		mov	dword ptr [edx.loadfile_inBuff], 0	; Current position in the buffer

		; Read the file into buffer
		mov	bx, [edx.loadfile_hdl]			; File handle
		mov	ecx, LOADFILEBUFFSZ + IBUFFMAXS		; Bytes to read
		mov	edx, [edx.loadfile_buff]		; Buffer address
		call	doReadFile				; in EBX: bytes read from the file
		jc	QuoteLoadErr

		; If zero bytes are read, end-of-file is reached
		cmpz	ebx, 0
		je	@@EndOfFile

		; Put a '\0' at the end of the buffer
		mov	ecx, [LoadFilesNb]
		dec	ecx
		imul	edx, ecx, size loadfileentry
		add	edx, [LoadFilesInfos]
		mov	ecx, [edx.loadfile_buff]		; Pointer to buffer
		mov	byte ptr [ecx + ebx], 0			; '\0'
		
		; Find first line into the buffer
		jmp	@@FirstLine


		; End of file is reached
		; ----------------------
		
@@EndOfFile:	; Update the 'depth' of files loading
		dec	[LoadFilesNb]
		mov	eax, [LoadFilesNb]
		imul	edx, eax, size loadfileentry
		add	edx, [LoadFilesInfos]
		
		; Close the file
		mov	bx, [edx.loadfile_hdl]			; File handle
		call	doCloseFile
		jc	QuoteLoadErr
		mov	[edx.loadfile_hdl], 0			; Reset handle stored for this level of include
		
		; Return a null buffer
		mov	ebx, [edx.loadfile_buff]
		mov	byte ptr [ebx], 0

		; If return to interpret (level #0) with compile state, this is an error (missing the ;)
		cmpz	ax, 0
		jne	@@Exit
		cmp	[CompileMode], FALSE
		je	@@Exit

		mov	ah, ERR_FORGETSEMIC			; Missing ;
		movzx	ebx, [edx.loadfile_colnb]		; Line #
		mov	ecx, [edx.loadfile_linenb]		; Line #
		mov	edx, [VOC_CURRENT]
		add	edx, [VOC_START]
		mov	edx, [edx.voc_lastword]
		add	edx, [NIA_START]
		add	edx, word_name				; Pointer to the name of the last definition
		jmp	doError


		; Found end of line (CR / LF)
		; ---------------------------

@@LL20:		; Put a '\0\0' at the end of the line
		mov	byte ptr [edx + ecx - 1], 0
		mov	byte ptr [edx + ecx], 0

		; Update buffer pointer
		mov	eax, [LoadFilesNb]
		dec	eax
		imul	edx, eax, size loadfileentry
		add	edx, [LoadFilesInfos]
		inc	ecx
		mov	[edx.loadfile_inBuff], ecx			; Current position in the buffer		

		; Update the line nb into the file
		inc	[edx.loadfile_linenb]

		; Update the starting address of this line into buffer (used to compute col position)
		mov	eax, [edx.loadfile_buff]
		add	eax, ebx
		mov	[edx.loadfile_linsaddr], eax

		; Buffer line address in EDX, then return
		mov	edx, [edx.loadfile_buff]			; Pointer to buffer
		add	edx, ebx


		; Return to caller		
		; ----------------
		
@@Exit:		ret

E_RdLoadFileLine:
RdLoadFileLine	endp


; ===============================================================
; = doOpenFile							=
; =								=
; = Low level function to open a file.				=
; =								=
; = I) EBX : pointer to filename				=
; =    AL  : access mode					=
; =								=
; = O) AX  : ? if ok, else error code				=
; =    BX  : File handle.					=
; =    CF  : 1 if error, else 0					=
; =								=
; = M) 								=
; =								=
; ===============================================================

		align	4
doOpenFile	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; -----------------------------------
		; Call DOS function 3Dh (Open handle)
		; -----------------------------------
		
		mov 	v86r_al, al			; Access mode
		HILO	v86r_ds, v86r_dx, ebx		; Pointer to filename
		mov 	v86r_ah, 3Dh			; Open handle function
		mov 	al, 21h				; Call INT 21h
		int 	33h
		jnc	@@NoError


		; --------
		; An error
		; --------
		
		mov	ax, v86r_ax
		jmp	@@Exit


		; --------
		; No error
		; --------

@@NoError:	mov	bx, v86r_ax			; File handle
		clr	ax
				

		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
@@Exit:		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

doOpenFile	endp


; ===============================================================
; = doWideOpenFile						=
; =								=
; = Low level function to open a file.				=
; =								=
; = I) EBX : pointer to filename				=
; =    AX  : access mode					=
; =    CL  = 1 authrorize new file creation, else not.		=
; =								=
; = O) AX  : ? if ok, else error code				=
; =    BX  : File handle.					=
; =    CF  : 1 if error, else 0					=
; =    CX  : 0:file open  1:new file				=
; =								=
; = M) all							=
; =								=
; ===============================================================

		align	4
doWideOpenFile	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; -----------------------------
		; Save registers which are used
		; -----------------------------
		
		push	edx esi ebp

	
		; -------------------
		; Update DOS reaction
		; -------------------

		cmp	cl, 0
		je	@@DosReact1
		mov	v86r_dx, 11h
		jmp	@@DosReact2
@@DosReact1:	mov	v86r_dx, 01h
@@DosReact2:

		; --------------------------------
		; Move filename string below 1Mb ?
		; --------------------------------

		ifdef	DOSREAL
		
		; Copy the string below 1 MB ?
		cmp	ebx, 10000h
		jb	@@Below1MB
		
		; Move string
		push	edx ecx ax
		mov	edx, [QuoteStrBuff]
		clr	ecx
@@MovStr1:	mov	al, [ebx + ecx]
		mov	[edx + ecx], al
		cmp	al, 0
		je	@@MovStr2
		inc	ecx
		jmp	@@MovStr1
@@MovStr2:	mov	ebx, edx
		pop	ax ecx edx
		
@@Below1MB:

		endif
		
		
		; Call DOS function 6Ch (Open handle)
		mov 	v86r_bx, ax				; Access mode
		mov	v86r_cx, 0				; Attribute
		HILO	v86r_ds, v86r_si, ebx			; Pointer to filename
		mov 	v86r_ax, 6C00h				; Open handle function
		mov 	al, 21h					; Call INT 21h
		int 	33h

		; Is there an error ?
		jnc	@@NoErr

		; An error
		mov	ax, v86r_ax
		jmp	@@Exit

		; No error
@@NoErr:		mov	bx, v86r_ax			; File handle
		mov	cx, v86r_cx				; DOS action
		clr	ax
				
		; Return to caller
@@Exit:		pop	ebp esi edx


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST

		ret

doWideOpenFile	endp


; ===============================================================
; = doCreateFile						=
; =								=
; = Low level function to create a file.			=
; =								=
; = I) EBX : pointer to filename				=
; =    AX  : file attribute					=
; =								=
; = O) AX  : 0 if ok, else error code				=
; =    BX  : File handle.					=
; =								=
; = M) 								=
; =								=
; ===============================================================

		align	4
doCreateFile	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; Call DOS function 3Ch (Create handle)
		mov 	v86r_cx, ax			; File attribute.
		HILO	v86r_ds, v86r_dx, ebx		; Pointer to filename
		mov 	v86r_ah, 3Ch			; Create handle function	
		mov 	al, 21h				; Call INT 21h
		int 	33h

		; Is there an error ?
		jnc	@@NoErr

		; An error
		mov	ax, v86r_ax
		jmp	@@Exit

		; No error
@@NoErr:	clr	ax				; No error
		mov	bx, v86r_ax			; File handle

				
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
@@Exit:		SUMMON	CTRLBREAK_REST


		; Return to caller
		ret

doCreateFile	endp


; ===============================================================
; = doReadFile							=
; =								=
; = Low level function to read data from a file.		=
; =								=
; = I) BX  : File handle					=
; =    ECX : Bytes to read					=
; =    EDX : Pointer to buffer					=
; =								=
; = O) AX  : 0 if ok, else error code				=
; =    EBX : Bytes read						=
; =    CF  : 1 if error, else 0					=
; =								=
; = M) AX EBX CF						=
; =								=
; ===============================================================

		align	4
doReadFile	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; -----------------------------
		; Save registers which are used
		; -----------------------------
		
		push	ecx edx esi ebp

	
		; --------------------
		; Loop initializations
		; --------------------
		
		mov	esi, edx			; Current addr to store data
		clr	ebp				; Count the total of bytes read


		; ---------
		; Main loop
		; ---------

		; In ECX: nb of bytes left to read
@@Loop:		mov	ax, cx
		cmp	ecx, FILEATOMSIZE
		jle	@@SizeOk
		mov	ax, FILEATOMSIZE
@@SizeOk:						; in EAX, nb of bytes to read

		; Bring data in a buffer under 1mb ?
		mov	esi, edx
		cmp	edx, 10000h
		jb	@@Below1MB
		mov	esi, [FileIOBuff]
@@Below1MB:

		; Call DOS function 3Fh (Read handle)
		mov	[v86r_bx], bx			; File handle
		mov	[v86r_cx], ax			; Bytes to read
		HILO	v86r_ds, v86r_dx, esi		; Pointer to buffer where to read
		mov 	[v86r_ah], 3Fh			; Read handle function
		mov 	al, 21h				; Call INT 21h
		int 	33h
		jc	@@Error

		; Copy the buffer above 1mb ?
		cmp	edx, 10000h
		jb	@@Below1MB_2
		push	ecx edi
							; Source addr is in ESI
		mov	edi, edx			; Destination addr
		movzx	ecx, [v86r_ax]			; Nb of bytes to move
		cld                     		; Inc SI and DI
		rep	movsb  		            	; Move the block
		pop	edi ecx
@@Below1MB_2:


		; -------------------------------------------
		; Nb of bytes read is less than bytes asked ?
		; -------------------------------------------

		movzx	eax, [v86r_ax]			; Nb of bytes read
		add	ebp, eax
		cmp	ax, [v86r_cx]			; Compare with nb of bytes asked
		jb	@@NoError			; Exit in this case, no more byte to read
		
		
		; -------------------
		; More data to read ?
		; -------------------

		add	edx, eax		
		sub	ecx, eax			; Update left nb bytes to read
		jnz	@@Loop
		jmp	@@NoError


		; --------
		; An error
		; --------
		
@@Error:		mov	ax, v86r_ax
		jmp	@@Exit


		; --------
		; No error
		; --------
		
@@NoError:	clr	ax				; No error
		mov	ebx, ebp			; Bytes read


		; -----------------------------------------------------
		; Restore registers which are used and return to caller
		; -----------------------------------------------------
		
@@Exit:		pop	ebp esi edx ecx


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		ret

doReadFile	endp


; ===============================================================
; = doWriteFile							=
; =								=
; = Low level function to write data to a file.			=
; =								=
; = I) BX  : File handle					=
; =    ECX : Bytes to write					=
; =    EDX : Pointer to buffer					=
; =								=
; = O) AX  : 0 if ok, else error code				=
; =    EBX : Bytes writen					=
; =    CF  : 1 if error, else 0					=
; =								=
; = M) AX EBX CF						=
; =								=
; ===============================================================

		align	4
doWriteFile	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; -----------------------------
		; Save registers which are used
		; -----------------------------
		
		push	ecx edx esi ebp

		; --------------------
		; Loop initializations
		; --------------------
		
		mov	esi, edx			; Current addr to store data
		clr	ebp				; Count the total of bytes writen


		; ---------
		; Main loop
		; ---------

		; In ECX: nb of bytes left to read
@@Loop:		mov	ax, cx
		cmp	ecx, FILEATOMSIZE
		jle	@@SizeOk
		mov	ax, FILEATOMSIZE
@@SizeOk:						; in EAX, nb of bytes to read

		; Bring data in a buffer under 1mb ?
		mov	esi, edx
		cmp	edx, 10000h
		jb	@@Below1MB
		push	ecx edi
		mov	esi, edx			; Source addr
		mov	edi, [FileIOBuff]		; Destination addr
		movzx	ecx, ax				; Nb of bytes to move
		cld                     			; Inc SI and DI
		rep	movsb  		            	; Move the block
		pop	edi ecx
		mov	esi, [FileIOBuff]		; Address to transfer data (memory -> file)
@@Below1MB:

		; Call DOS function 40h (Write handle)
		mov	[v86r_bx], bx			; File handle
		mov	[v86r_cx], ax			; Bytes to write
		HILO	v86r_ds, v86r_dx, esi		; Pointer to buffer where to write
		mov 	[v86r_ah], 40h			; Write handle function
		mov 	al, 21h				; Call INT 21h
		int 	33h
		jc	@@Error


		; ---------------------------------------------
		; Nb of bytes writen is less than bytes asked ?
		; ---------------------------------------------

		movzx	eax, [v86r_ax]			; Nb of bytes writen
		add	ebp, eax
		cmp	ax, [v86r_cx]			; Compare with nb of bytes asked
		jb	@@NoError			; Exit in this case, no more byte to write

		
		; --------------------
		; More data to write ?
		; --------------------

		add	edx, eax		
		sub	ecx, eax			; Update left nb bytes to write
		jnz	@@Loop
		jmp	@@NoError


		; --------
		; An error
		; --------
		
@@Error:		mov	ax, [v86r_ax]
		stc
		jmp	@@Exit


		; --------
		; No error
		; --------
		
@@NoError:	clr	ax				; No error
		mov	ebx, ebp			; Bytes writen
		clc


		; -----------------------------------------------------
		; Restore registers which are used and return to caller
		; -----------------------------------------------------
		
@@Exit:		pop	ebp esi edx ecx


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		ret

doWriteFile	endp


; ===============================================================
; = doSeekFile							=
; =								=
; = Low level function to move file pointer.			=
; =								=
; = I)	AL  : Method of moving (SEEK_SET, SEEK_CUR, SEEK_END)	=
; =	BX  : File handle					=
; =	ECX : Offset						=
; =								=
; = O)	AX : 0 if ok, else error code				=
; =	EBX : New file pointer location				=
; =								=
; = M) 	none							=
; =								=
; ===============================================================

		align	4
doSeekFile	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; Save registers which are used
		push	ecx

		; Call DOS function 42h (Open handle)
		mov	v86r_al, al			; Method of file moving
		mov	v86r_bx, bx			; File handle
		mov	v86r_dx, cx			; Low 16-bits in DX
		shr	ecx, 16				; Hi 16-bits
		mov	v86r_cx, cx			; in CX
		mov 	v86r_ah, 42h			; Move file pointer function
		mov 	al, 21h				; Call INT 21h
		int 	33h

		; Is there an error ?
		jnc	@@NoErr

		; An error
		mov	ax, v86r_ax
		jmp	@@Exit

		; No error
@@NoErr:	mov	bx, v86r_dx
		shl	ebx, 16
		mov	bx, v86r_ax			; In EBX, new r/w pointer location
		clr	ax				; No error
				
		; Restore registers which are used and return to caller
@@Exit:		pop	ecx


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		ret

doSeekFile	endp


; ===============================================================
; = doFilePos							=
; =								=
; = Low level function to return file pointer position.		=
; =								=
; = I)	BX  : File handle					=
; =								=
; = O)	AX  : 0 if ok, else error code				=
; =	EBX : Current file pointer location			=
; =								=
; = M) 	AX EBX							=
; =								=
; ===============================================================

		align	4
doFilePos	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; Call DOS function 42h
		mov	v86r_bx, bx			; File handle
		mov	v86r_dx, 0			; Low 16-bits
		mov	v86r_cx, 0			; Hi  1--bits
		mov 	v86r_ax, 4201h			; Move file pointer function
		mov 	al, 21h				; Call INT 21h
		int 	33h

		; Is there an error ?
		jnc	@@NoErr

		; An error
		mov	ax, v86r_ax
		jmp	@@Exit

		; No error
@@NoErr:	movzx	ebx, v86r_dx
		shl	ebx, 16
		mov	bx, v86r_ax			; In EBX, current r/w pointer location
		clr	ax				; No error
				

		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
@@Exit:		SUMMON	CTRLBREAK_REST


		; Restore registers which are used and return to caller
		ret

doFilePos	endp


; ===============================================================
; = doFileSize							=
; =								=
; = Low level function to return file size.			=
; =								=
; = I)	BX  : File handle					=
; =								=
; = O)	EAX : 0 if ok, else error code				=
; =	EBX : New file pointer location				=
; =								=
; = M)  EAX EBX							=
; =								=
; ===============================================================

		align	4
doFileSize	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; Save registers which are used
		push	cx dx esi

		; Ask DOS to know the current value of pointer
		mov	v86r_bx, bx			; File handle
		mov	v86r_dx, 0			; Low 16-bits
		mov	v86r_cx, 0			; Hi  1--bits
		mov 	v86r_ax, 4201h			; Move file pointer function
		mov 	al, 21h				; Call INT 21h
		int 	33h
		jc	@@Error
		mov	dx, v86r_ax			; Low
		mov	cx, v86r_dx			; Hi

		; Call DOS function 42h
		mov	v86r_bx, bx			; File handle
		mov	v86r_dx, 0			; Low 16-bits
		mov	v86r_cx, 0			; Hi  1--bits
		mov 	v86r_ax, 4202h			; Move file pointer function
		mov 	al, 21h				; Call INT 21h
		int 	33h

		; Is there an error ?
		jnc	@@NoErr

		; An error
@@Error:		mov	ax, v86r_ax
		jmp	@@Exit

		; No error
@@NoErr:		movzx	eax, v86r_dx
		shl	eax, 16
		mov	ax, v86r_ax			; In EBX, new r/w pointer location
		mov	esi, eax

		; Store pointer to initial position
		mov	v86r_bx, bx			; File handle
		mov	v86r_dx, dx			; Low 16-bits
		mov	v86r_cx, cx			; Hi  1--bits
		mov 	v86r_ax, 4200h			; Move file pointer function
		mov 	al, 21h				; Call INT 21h
		int 	33h
		jc	@@Error
		mov	ebx, esi
		clr	eax				; No error
				
		; Restore registers which are used and return to caller
@@Exit:		pop	esi dx cx


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		ret

doFileSize	endp


; ===============================================================
; = doCloseFile							=
; =								=
; = Low level function to close a file.				=
; =								=
; = I)	BX  : File handle					=
; =								=
; = O)	Carry set if error, else carry reset.			=
; = 	AX : 0 if ok, else error code				=
; =								=
; = M)	None.							=
; =								=
; ===============================================================

		align	4
doCloseFile	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; Call DOS function 42h (Open handle)
		mov	v86r_bx, bx			; File handle
		mov 	v86r_ah, 3Eh			; Close file function
		mov 	al, 21h				; Call INT 21h
		int 	33h

		; Is there an error ?
		jnc	@@NoErr

		; An error
		mov	ax, v86r_ax
		jmp	@@Exit

		; No error
@@NoErr:	clr	ax				; No error
				

		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
@@Exit:		SUMMON	CTRLBREAK_REST


		; Return to caller
		ret

doCloseFile	endp


; ===============================================================
; = doDeleteFile						=
; =								=
; = Low level function to delete a file.			=
; =								=
; = I)	EBX : pointer to filename.				=
; =								=
; = O)	AX : 0 if ok, else error code				=
; =								=
; = M) 								=
; =								=
; ===============================================================

		align	4
doDeleteFile	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; --------------------------------
		; Move filename string below 1Mb ?
		; --------------------------------

		ifdef	DOSREAL
		
		; Copy the string below 1 MB ?
		cmp	ebx, 10000h
		jb	@@Below1MB
		
		; Move string
		push	edx ecx ax
		mov	edx, [QuoteStrBuff]
		clr	ecx
@@MovStr1:	mov	al, [ebx + ecx]
		mov	[edx + ecx], al
		cmp	al, 0
		je	@@MovStr2
		inc	ecx
		jmp	@@MovStr1
@@MovStr2:	mov	ebx, edx
		pop	ax ecx edx
		
@@Below1MB:

		endif
		
		
		; Call DOS function
		HILO	v86r_ds, v86r_dx, ebx			; Pointer to filename
		mov 	v86r_ah, 41h				; Function
		mov 	al, 21h					; Call INT 21h
		int 	33h

		; Is there an error ?
		jnc	@@NoErr

		; An error
		mov	ax, v86r_ax
		ret						; jmp	@@Exit

		; No error
@@NoErr:		clr	ax
				

		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
@@Exit:		SUMMON	CTRLBREAK_REST


		; Return to caller
		ret

doDeleteFile	endp


; ===============================================================
; = doQFileExist						=
; =								=
; = Low level function to see if a file exist.			=
; =								=
; = I)	EBX : filename.						=
; =								=
; = O)	AX  : 0 if ok, else error code				=
; =	EBX : 1 if file exists, 0 else not.			=
; =								=
; = M) 								=
; =								=
; ===============================================================

		align	4
doQFileExist	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; --------------------------------
		; Move filename string below 1Mb ?
		; --------------------------------

		ifdef	DOSREAL
		
		; Copy the string below 1 MB ?
		cmp	ebx, 10000h
		jb	@@Below1MB
		
		; Move string
		push	edx ecx ax
		mov	edx, [QuoteStrBuff]
		clr	ecx
@@MovStr1:	mov	al, [ebx + ecx]
		mov	[edx + ecx], al
		cmp	al, 0
		je	@@MovStr2
		inc	ecx
		jmp	@@MovStr1
@@MovStr2:	mov	ebx, edx
		pop	ax ecx edx
		
@@Below1MB:

		endif
		
		
		; ---------------------
		; Call DOS function 43h
		; ---------------------

		HILO	v86r_ds, v86r_dx, ebx		; Filename pointer		
		mov 	v86r_ax, 4300h			; DOS function
		mov 	al, 21h				; Call INT 21h
		int 	33h

		; Is there an error ?
		jnc	@@Found

		; An error
@@Error:		mov	ax, v86r_ax
		cmp	ax, 2				; File not found ?
		je	@@NotFound
		cmp	ax, 3				; Path not found ?
		je	@@NotFound
		ret					; jmp	@@Exit

@@Found:		clr	ebx
		inc	ebx
		clr	ax				; No error
		ret					; jmp	@@Exit
		
@@NotFound:	clr	ebx
		clc					; Clear error status


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
@@Exit:		SUMMON	CTRLBREAK_REST


		; Return to caller
		ret

doQFileExist	endp

code32		ends
		end

