; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	DEBUG.ASM						!
; 									!
; Description:	Definitions for playing with debugging.			!
; 									!
; List of FROTH definitions defined in this module:			!
;	DOTS		.S						!
;	DUMP		DUMP						!
;	showVOC		.VOC						!
;	showNIA		.NIA						!
;	showCDA		.CDA						!
;	showDDIR	.DDIR						!
;	showCREF	.CREF						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc


ifdef TESTSTACK

; ===============================================================
; = MAKE_VSTACK							=
; =                                                         	=
; = Usage:							=
; =	Verify primary stack underflow (on definition entry).	=
; =                                                         	=
; = I) none							=
; = O) none							=
; = M) none							=
; =                                                         	=
; ===============================================================

MAKE_VSTACK	MACRO	name, nbargs

		align	4
name     	proc
	
		; Save used registers
		push	eax ecx

		; The number of words on the stack
		mov	eax, [CurrActThread]
		mov	eax, [PStackStart + eax * 4]

		; Verify stack frame (underflow ?)
		sub	eax, esi
		cmp	al, nbargs * @WordSize
		jae	@@Ok

		; Error: stack underflow
		add	esp, 4 + 4				; Clean stack
		mov	ah, ERR_STACKUNDERFLOW			; Stack underflow
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError

		; Return to caller
@@Ok:		pop	ecx eax
		ret

name     	endp

		endm

MAKE_VSTACK	VStack1,	1
MAKE_VSTACK	VStack2,	2
MAKE_VSTACK	VStack3,	3
MAKE_VSTACK	VStack4,	4
MAKE_VSTACK	VStack5,	5
MAKE_VSTACK	VStack6,	6
MAKE_VSTACK	VStack7,	7
MAKE_VSTACK	VStack8,	8


; ===============================================================
; = OSTACK							=
; =                                                         	=
; = Usage:							=
; =	Verify primary stack overflow (on definition exit).	=
; =                                                         	=
; = I) none							=
; = O) none							=
; = M) none							=
; ===============================================================

		align	4
OStack     	proc

		; Save EAX
		push	eax
	
		; The number of words on the stack
		mov	eax, [CurrActThread]
		mov	eax, [PStackStart + eax * 4]
		sub	eax, PSTACKSIZE

		; Stack overflow ?
		cmp	esi, eax
		jg	@@NotOverflow

		; Error: stack overflow
		mov	ah, ERR_STACKOVERFLOW			; Stack overflow
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

		; Not overflow error
@@NotOverflow:	pop	eax

		; Return to caller
		ret

E_OStack:	
OStack     	endp


; ===============================================================
; = VRStack                             			=
; =                                                         	=
; = Usage: Verify second stack frame.				=
; =                                                         	=
; = I) byte after CALL						=
; =                                                         	=
; = O) none							=
; =                                                         	=
; = M) none							=
; ===============================================================

		align	4
VRStack     	proc
	
		; Save used registers
		push	eax ebx ecx

		; The number of words on the 2nd stack
		mov	eax, [CurrActThread]
		mov	ebx, [RStackStart + eax * 4]
		mov	ecx, [RStackPtr   + eax * 4]
		sub	ebx, ecx
		mov	eax, ebx
		shr	eax, 2

		; Get the asked number of argumensts to test
		mov	ebx, esp
		add	ebx, 12					; 12 = 4 + 4 + 4
		mov	ecx, [ebx]

		; Verify stack frame
		cmp	al, [ecx]
		jae	@@Ok

		; Error: second stack underflow
		mov	esp, ebx
		mov	ah, ERR_RSTACKUNDERFLOW			; Second stack underflow
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError

		; Move the IP after the string
@@Ok:	      	inc	ecx					; 1 is the size of the parameter
		mov	[ebx], ecx

		; Return to caller
		pop	ecx ebx eax
		ret

VRStack     	endp

endif


; ===============================================================
; = PrStackDepth                        			=
; =                                                     	=
; = I) none                         				=
; = O) none                         				=
; = M) never                            			=
; ===============================================================

		align	4
PrStackDepth	proc

		; Save then set '?UpdateCursor'
		mov	eax, [edi + ddir_blkaddr + D_QUPDATECURSOR]
		mov	ebx, [eax]				; Get actual value
		push	ebx					; Save it
		mov	dword ptr [eax], TRUE			; ?UpdateCursor 1!
	
		; Just print ok ?
		GETFVAR	eax, D_QOK
		cmpz	eax, 0
		je	@@NotOk
		call	PDOTQ
		db	"ok ", 0
		jmp	@@next
@@NotOk:		

		; printf("%c", '(')
		call	PDOTQ
		db	"(", 0

		; The number of words on the stack
		mov	eax, [edi + ddir_blkaddr + D_BASE]
		push	dword ptr [eax]
		mov	dword ptr [eax], 10
		SUMMON	DEPTH
		SUMMON	DOT
		pop	ebx
		mov	eax, [edi + ddir_blkaddr + D_BASE]
		mov	[eax], ebx

ifdef NEVER
		mov	eax, [CurrActThread]
		mov	eax, [PStackStart + eax * @WordSize]
		sub	eax, esi
		shr	eax, 2
		clr	edx					; EDX:EAX / 100
		mov	ebx, 100
		idiv	ebx
		cmpz	al, 0
		je	@@LL1
		add	al, '0'
		call	COUT
@@LL1:		mov	eax, edx
		clr	edx					; EDX:EAX / 10
		mov	ebx, 10
		idiv	ebx
		cmpz	al, 0
		je	@@LL2
		add	al, '0'
		call	COUT
@@LL2:		mov	eax, edx
		add	al, '0'
		call	COUT
endif

		; printf(") ")
		call	PDOTQ
		db	') ', 0

		; Restore '?UpdateCursor'
@@next:		mov	eax, [edi + ddir_blkaddr + D_QUPDATECURSOR]
		pop	ebx
		mov	[eax], ebx

		; return to caller
		ret

PrStackDepth    endp


; ===============================================================
; = DOTS							=
; =                                                 		=
; = Code for the forth word:	.S				=
; =                     	                            	=
; = Stack:	( --- )						=
; ===============================================================

		align	4
DOTS		proc

		; --------------
		; Show the stack
		; --------------
		
		; Nb of integers on stack		
		mov	edx, [CurrActThread]
		mov	edx, [PStackStart + edx * @WordSize]
		sub	edx, esi
		shr	edx, 2
		cmpz	edx, 0
		jne	@@Full
		
		SUMMON	PDOTQ
;		db	13, 10
		db	"Stack: empty", 0
		jmp	@@Done

@@Full:		SUMMON	PDOTQ
;		db	13, 10
		db	"Stack: ", 0

		; Get value
@@Print:		dec	edx
		mov	eax, [esi + edx * @WordSize]
		sub	esi, @WordSize
		mov	[esi], eax
		push	edx
		SUMMON	DOT
		SUMMON	SPACE
		pop	edx
		
		; Next value, if any
		cmpz	edx, 0
		jne	@@Print
		
@@Done:


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_DOTS:
DOTS		endp


; ===============================================================
; = DUMP							=
; =                                                 		=
; = Code for the forth word:	DUMP				=
; =                     	                            	=
; = Stack:	( addr1 addr2 --- )				=
; =	      or ( addr n --- )					=
; ===============================================================

		align	4
DUMP		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; -------------
		; Show the dump
		; -------------

		; Start / End values
		mov	eax, PARAM1				; Get addr1
		mov	ebx, PARAM0				; Get addr2 or n
		cmp	ebx, eax
		jae	@@ok
		add	ebx, eax
		dec	ebx
@@ok:		mov	PARAM1, ebx
		mov	PARAM0, eax

		; Loop
		mov	al, 15

@@loop:		inc	al
		cmp	al, 8
		jne	@@NotPrSpc
		push	eax
		SUMMON	SPACE
		SUMMON	SPACE
		pop	eax
@@NotPrSpc:	cmp	al, 16
		jne	@@NotPrAddr
		SUMMON	CR
		SUMMON	_DUP
		SUMMON	HDOT
		IPUSH	':'
		SUMMON	EMIT
		SUMMON	SPACE		
		clr	al
@@NotPrAddr:	

		push	eax
		mov	ebx, PARAM0
		movzx	eax, byte ptr [ebx]
		IPUSH	eax
		SUMMON	Z2HDOT
		SUMMON	SPACE
		pop	eax

		inc	PARAM0
		mov	ebx, PARAM0
		cmp	ebx, PARAM1
		jbe	@@loop

		
		; ----------------
		; Return to caller
		; ----------------
		
		DISCARD	2
		ret

E_DUMP:
DUMP		endp


ifdef TESTSTACK

; ===============================================================
; = PrintMsg							=
; =                                                 		=
; = Run time function for the forth word ."			=
; =                     	                            	=
; = Regs modified:	BX ECX					=
; ===============================================================

		align	4
PrintMsg	proc

		push	ax
		push	ebx
		push	ecx

		; Store into DI the string address
		mov	ebx, esp
		add	ebx, 10					; 10 = 2 + 4 + 4
		mov	ecx, [ebx]

@@LL1:		mov	al, [ecx]
		cmp	al, 0
		jz	@@LL2
		call 	POUT
		inc	ecx
		jmp	@@LL1

		; Move the IP after the string
@@LL2:      	inc	ecx
		mov	[ebx], ecx

		; Return to caller
		pop	ecx
		pop	ebx
		pop	ax
		ret

PrintMsg	endp


; ===============================================================
; = PrintText							=
; =                                                 		=
; ===============================================================

		align	4
PrintText	proc

		push	ebx
		
		mov	ebx, eax
		mov	cl, 0

@@LL1:		mov	al, [ebx]
		cmp	al, 0
		jz	@@LL2
		call	POUT
		inc	ebx
		inc	cl
		jmp	@@LL1
@@LL2:		

		; Return to caller
		pop	ebx
		ret

PrintText	endp


; ===============================================================
; = PrintEAX							=
; =                                                 		=
; ===============================================================

		align	4
PrintEAX	proc

		; Save all registers
		pushad

		mov	ebx, eax
		shr	eax, 7*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL11
		add	al, '0'
		jmp	@@LL12
@@LL11:		add	al, 55
@@LL12:		call	POUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 6*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL21
		add	al, '0'
		jmp	@@LL22
@@LL21:		add	al, 55
@@LL22:		call	POUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 5*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL31
		add	al, '0'
		jmp	@@LL32
@@LL31:		add	al, 55
@@LL32:		call	POUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 4*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL41
		add	al, '0'
		jmp	@@LL42
@@LL41:		add	al, 55
@@LL42:		call	POUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 3*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL51
		add	al, '0'
		jmp	@@LL52
@@LL51:		add	al, 55
@@LL52:		call	POUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 2*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL61
		add	al, '0'
		jmp	@@LL62
@@LL61:		add	al, 55
@@LL62:		call	POUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 1*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL71
		add	al, '0'
		jmp	@@LL72
@@LL71:		add	al, 55
@@LL72:		call	POUT
		mov	eax, ebx
		
		mov	ebx, eax
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL81
		add	al, '0'
		jmp	@@LL82
@@LL81:		add	al, 55
@@LL82:		call	POUT

		mov	al, ' '
		call	POUT
		
		; Restore all registers
		popad

		ret

PrintEAX	endp


; ===============================================================
; = PrintAX							=
; =                                                 		=
; ===============================================================

		align	4
PrintAX		proc

		; Save all registers
		pushad

		mov	ebx, eax
		shr	eax, 3*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL51
		add	al, '0'
		jmp	@@LL52
@@LL51:		add	al, 55
@@LL52:		call	POUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 2*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL61
		add	al, '0'
		jmp	@@LL62
@@LL61:		add	al, 55
@@LL62:		call	POUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 1*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL71
		add	al, '0'
		jmp	@@LL72
@@LL71:		add	al, 55
@@LL72:		call	POUT
		mov	eax, ebx
		
		mov	ebx, eax
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL81
		add	al, '0'
		jmp	@@LL82
@@LL81:		add	al, 55
@@LL82:		call	POUT

		mov	al, ' '
		call	POUT
		
		; Restore all registers
		popad

		ret

PrintAX		endp


; ===============================================================
; = PrintAL							=
; =                                                 		=
; ===============================================================

		align	4
PrintAL		proc

		; Save all registers
		pushad

		mov	ebx, eax
		shr	eax, 1*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL71
		add	al, '0'
		jmp	@@LL72
@@LL71:		add	al, 55
@@LL72:		call	POUT
		mov	eax, ebx
		
		mov	ebx, eax
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL81
		add	al, '0'
		jmp	@@LL82
@@LL81:		add	al, 55
@@LL82:		call	POUT

		mov	al, ' '
		call	POUT
		
		; Restore all registers
		popad

		ret

PrintAL		endp


; ===============================================================
; = showVOC							=
; =                                                 		=
; = Code for the froth word:	.VOC				=
; =								=
; = Stack:	( --- )						=
; =								=
; = *** debug funtion ***					=
; ===============================================================

		align	4
showVOC		proc

		; Save all registers
		pushad

		; Restore all registers
		popad

		; Header
		call	PrintMsg
		db	0Fh
		db	0Dh, 0Ah
		db	'No       Name                             Father   Brother  LastWord WordsNb'
		db	13, 10, 0
		
		; Base address
		mov	edx, [VOC_START]

@@LL1:
		; No
		mov	eax, edx
		call	PrintEAX

		; Name
		mov	eax, edx
		add	eax, voc_name
		call	PrintText
@@NAME1:	mov	al, ' '
		call	POUT
		inc	cl
		cmp	cl, WORDMAXLEN+1
		jle	@@NAME1

		; Father ptr
		mov	eax, [edx.voc_fatherptr]
		call	PrintEAX

		; First word
		mov	eax, [edx.voc_lastword]
		call	PrintEAX

		; Words nb
		mov	eax, [edx.voc_wordsnb]
		call	PrintEAX

		; Next line
		call	PrintMsg
		db	0Dh, 0Ah, 0
		add	edx, size vocentry
		cmp	dword ptr [edx], -1
		jne	@@LL1

		; return to caller
		ret

E_showVOC:	
showVOC		endp


; ===============================================================
; = showNIA							=
; =                                                 		=
; = Code for the froth word:	.NIA				=
; =								=
; = Stack:	( --- )						=
; =								=
; = *** debug funtion ***					=
; ===============================================================

		align	4
showNIA		proc

		; Save all registers
		pushad

		; Header
		call	PrintMsg
		db	0Fh
		db	13, 10
		db	'Nb       No       Name                            Imm Syst VocPtr   PrevWord NextWord Code     Size Uses'
		db	0Dh, 0Ah, 0
		
		; Base address
		mov	edx, [NIA_START]
;	add	edx, 4780h		; !!!

@@LL1:
		; Nb
		mov	eax, edx
		sub	eax, [NIA_START]
		call	PrintEAX

		; No
		mov	eax, edx
		call	PrintEAX

		; Name
		mov	eax, edx
		add	eax, word_name
		call	PrintText
@@NAME1:	mov	al, ' '
		call	POUT
		inc	cl
		cmp	cl, WORDMAXLEN+1
		jle	@@NAME1

		; Immediate ?
		test	byte ptr [edx.word_flags], FIMMEDIATE
		jnz	@@IMM1
		mov	al, 'n'
		call	POUT
		jmp	@@IMM2
@@IMM1:		mov	al, 'Y'
		call	POUT
@@IMM2:		call	PrintMsg
		db	'   ', 0

		; System ?
		test	byte ptr [edx.word_flags], FSYSTEM
		cmp	al, 1
		jz	@@SYST1
		mov	al, 'n'
		call	POUT
		jmp	@@SYST2
@@SYST1:	mov	al, 'Y'
		call	POUT
@@SYST2:	call	PrintMsg
		db	'   ', 0

		; Voc ptr
		mov	eax, 0
		mov	ax, [edx.word_vocptr]
		call	PrintEAX
		
		; Previous word
		mov	eax, [edx.word_prevword]
		call	PrintEAX

		; Next word
		mov	eax, [edx.word_nextword]
		call	PrintEAX

		; Code ptr
		mov	eax, [edx.word_realaddr]
		call	PrintEAX
		
		; Code size
		mov	ax, [edx.word_CodeSize]
		call	PrintAX
		
		; Nb of uses
		mov	ax, [edx.word_NbUses]
		call	PrintAX

		; Next line
		call	PrintMsg
		db	0Dh, 0Ah, 0
		add	edx, size wordentry
		cmp	dword ptr [edx], -3
		jne	@@LL1

		; Restore all registers
		popad

		; return to caller
		ret

E_showNIA:
showNIA		endp


; ===============================================================
; = showCDA							=
; =                                                 		=
; = Code for the froth word:	.CDA				=
; =								=
; = Stack:	( --- )						=
; =								=
; = *** debug funtion ***					=
; ===============================================================

		align	4
showCDA		proc

		; Save all registers
		pushad

		; Address and size
		call	PrintMsg
		db	'StartAddr = ', 0
		mov	eax, [CDA_START]
		call	PrintEAX
		call	PrintMsg
		db	'  Size = ', 0
		mov	eax, [CDA_SIZE]
		call	PrintEAX

		; CR LF		
		call	PrintMsg
		db	0Dh, 0Ah, 0

		; Restore all registers
		popad

		; return to caller
		ret

E_showCDA:
showCDA		endp


; ===============================================================
; = showDDIR							=
; =                                                 		=
; = Code for the froth word:	.DDIR				=
; =								=
; = Stack:	( --- )						=
; =								=
; = *** debug funtion ***					=
; ===============================================================

		align	4
showDDIR	proc

		; Save all registers
		pushad

		; DDIR: Address and size
		call	PrintMsg
		db	0dh, 0Ah, 'DDIR:  StartAddr = ', 0
		mov	eax, [DDIR_START]
		call	PrintEAX
		call	PrintMsg
		db	'  Size = ', 0
		mov	eax, [DDIR_SIZE]
		call	PrintEAX
		call	PrintMsg
		db	'  First freed = ', 0
		mov	eax, [DDIR_FFREED]
		call	PrintEAX

		; BOUND: addresses
		call	PrintMsg
		db	"   Bounds: ", 0
		mov	eax, [DDA_BOUND]
		call	PrEAX
		mov	eax, [DDA_BOUND + @WordSize]
		call	PrEAX
		
		; DDA: Address and size
		call	PrintMsg
		db	0Dh, 0Ah
		db	' DDA:  StartAddr = ', 0
		mov	eax, [DDA_START]
		call	PrintEAX
		call	PrintMsg
		db	'  Size = ', 0
		mov	eax, [DDA_SIZE]
		call	PrintEAX
		call	PrintMsg
		db	'    Init size = ', 0
		mov	eax, [DDA_SIZE_INIT]
		call	PrintEAX
		call	PrintMsg
		db	'  Salvage = ', 0
		mov	eax, [DDIR_MEMSALV]
		call	PrintEAX
		
		; Header
		call	PrintMsg
		db	0dH, 0Ah
		db	'Nb       Addr     NxtFreed WordNb   Size     Addr     Mark? Deld? Thread Auto? Dyna?'
		db	0

		mov	ecx, [DDIR_START]
		add	ecx, [DDIR_SIZE]

		mov	ebx, [DDIR_START]
		add	ebx, [DDIR_SIZE_INIT]		; Comment this line to show also built-in variables
;							  -------------------------------------------------


@@LL1:		; Nb Adr
		call	PrintMsg
		db	0Dh, 0Ah, 0
		mov	eax, ebx
		sub	eax, [DDIR_START]
		call	PrintEAX
		mov	eax, ebx
		call	PrintEAX

		; Free
		mov	eax, [ebx.ddir_NextFreed]
		call	PrintEAX

		; Word nb
		movzx	eax, [ebx.ddir_wordnb]
		call	PrintEAX
		
		; Size
		mov	eax, [ebx.ddir_blksize]
		call	PrintEAX
		
		; Addr
		mov	eax, [ebx.ddir_blkaddr]
		call	PrintEAX
		call	PrintMsg
		db	"  ", 0
		
		; Marked as to delete ?
		mov	al, [ebx.ddir_flags]
		and	al, QDATA_MARKDEL
		call	PrintAL
		call	PrintMsg
		db	"    ", 0

		; Marked as deleted ?
		mov	al, [ebx.ddir_flags]
		and	al, QDATA_DELDONE
		call	PrintAL
		call	PrintMsg
		db	"    ", 0

		; Thread
		mov	al, [ebx.ddir_threadnb]
		call	PrintAL
		call	PrintMsg
		db	"    ", 0

		; Auto var ?
		mov	al, [ebx.ddir_flags]
		and	al, QDATA_AUTO
		call	PrintAL
		call	PrintMsg
		db	"  ", 0

		; Dynamic var ?
		mov	al, [ebx.ddir_flags]
		and	al, QDATA_DYNA
		call	PrintAL

		; PCR
		call	PrintMsg
		db	0

		; Next line, if any
		add	ebx, size ddirentry
		cmp	ebx, ecx
		jl	@@LL1

		; Restore all registers
		popad

		; return to caller
		ret

E_showDDIR:
showDDIR	endp


; ===============================================================
; = showCREF							=
; =                                                 		=
; = Code for the froth word:	.CREF				=
; =								=
; = Stack:	( --- )						=
; =								=
; = *** debug funtion ***					=
; ===============================================================

		align	4
showCREF	proc

		; Save all registers
		pushad
		
		; Print records
		call	PrintMsg
		db	0Fh		
		db	13, 10, 0
		call	PrintMsg
		db	"First freed record : ", 0
		mov	eax, [CrossRefFFreed]
		call	PrintEAX
		call	PrintMsg
		db	"   ", 0
		mov	eax, [CrossRefTotalNb]
		call	PrintEAX
		call	PrintMsg
		db	" total records   ", 0
		mov	eax, [CrossRefNb]
		call	PrintEAX
		call	PrintMsg
		db	"Records :"
		db	13, 10, 0
		clr	ecx					; Counter
		clr	bl
		mov	edx, [CrossRef]
@@Rec1:		call	PrintMsg
		db	"Record ", 0
		mov	al, cl
		call	PrintAL
		call	PrintMsg
		db	": ", 0
		mov	eax, [edx.wref_caller]
		call	PrintAX
		mov	eax, [edx.wref_called]
		call	PrintAX
		mov	ax, [edx.wref_calls]
		call	PrintAX
		call	PrintMsg
		db	"  ", 0
		inc	ecx
		cmp	ecx, [CrossRefTotalNb]
		jge	@@Rec3
		inc	bl
		cmp	bl, 4
		jne	@@Rec2
		call	PrintMsg
		db	13, 10, 0
		clr	bl
@@Rec2:		add	edx, size wrefentry
		jmp	@@Rec1
@@Rec3:
		
		; Print Index table on callers
		call	PrintMsg
		db	13, 10, 13, 10
		db	"Index table on (Callers, Calleds) :", 13, 10, 0
		clr	ecx
		clr	bl
		mov	edx, [CRefNdxRefRD]
@@Ndx1:		call	PrintMsg
		db	"Record ", 0
		mov	al, cl
		call	PrintAL
		call	PrintMsg
		db	": ", 0
		mov	eax, [edx]
		call	PrintAX
		call	PrintMsg
		db	"     ", 0
		inc	ecx
		cmp	ecx, [CrossRefTotalNb]
		jge	@@Ndx3
		inc	bl
		cmp	bl, 5
		jne	@@Ndx2
		call	PrintMsg
		db	13, 10, 0
		clr	bl
@@Ndx2:		add	edx, 4
		jmp	@@Ndx1
@@Ndx3:
		
		
		; Print Index table on calleds
		call	PrintMsg
		db	13, 10, 13, 10
		db	"Index table on (Calleds, Callers) :", 13, 10, 0
		clr	ecx
		clr	bl
		mov	edx, [CRefNdxRefDR]
@@Idx1:		
		call	PrintMsg
		db	"Record ", 0
		mov	al, cl
		call	PrintAL
		call	PrintMsg
		db	": ", 0
		mov	eax, [edx]
		call	PrintAX
		call	PrintMsg
		db	"     ", 0
		inc	ecx
		cmp	ecx, [CrossRefTotalNb]
		jge	@@Idx3
		inc	bl
		cmp	bl, 5
		jne	@@Idx2
		call	PrintMsg
		db	13, 10, 0
		clr	bl
@@Idx2:		add	edx, 4
		jmp	@@Idx1
@@Idx3:

		
		; Return to caller
		popad
		ret

E_showCREF:
showCREF	endp


		align	4
PrEAX		proc

		; Save all registers
		pushad

		mov	ebx, eax
		shr	eax, 7*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL11
		add	al, '0'
		jmp	@@LL12
@@LL11:		add	al, 55
@@LL12:		call	COUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 6*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL21
		add	al, '0'
		jmp	@@LL22
@@LL21:		add	al, 55
@@LL22:		call	COUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 5*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL31
		add	al, '0'
		jmp	@@LL32
@@LL31:		add	al, 55
@@LL32:		call	COUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 4*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL41
		add	al, '0'
		jmp	@@LL42
@@LL41:		add	al, 55
@@LL42:		call	COUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 3*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL51
		add	al, '0'
		jmp	@@LL52
@@LL51:		add	al, 55
@@LL52:		call	COUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 2*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL61
		add	al, '0'
		jmp	@@LL62
@@LL61:		add	al, 55
@@LL62:		call	COUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 1*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL71
		add	al, '0'
		jmp	@@LL72
@@LL71:		add	al, 55
@@LL72:		call	COUT
		mov	eax, ebx
		
		mov	ebx, eax
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL81
		add	al, '0'
		jmp	@@LL82
@@LL81:		add	al, 55
@@LL82:		call	COUT

		mov	al, ' '
		call	COUT
		
		; Restore all registers
		popad

		ret

PrEAX		endp


		align	4
PrAX		proc

		; Save all registers
		pushad

;		push	ax
;		mov	al, ' '
;		call	COUT
;		pop	ax

		and	eax, 0000FFFFh
		
		mov	ebx, eax
		shr	eax, 3*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL51
		add	al, '0'
		jmp	@@LL52
@@LL51:		add	al, 55
@@LL52:		call	COUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 2*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL61
		add	al, '0'
		jmp	@@LL62
@@LL61:		add	al, 55
@@LL62:		call	COUT
		mov	eax, ebx
		
		mov	ebx, eax
		shr	eax, 1*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL71
		add	al, '0'
		jmp	@@LL72
@@LL71:		add	al, 55
@@LL72:		call	COUT
		mov	eax, ebx
		
		mov	ebx, eax
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL81
		add	al, '0'
		jmp	@@LL82
@@LL81:		add	al, 55
@@LL82:		call	COUT

		mov	al, ' '
		call	COUT
		
		; Restore all registers
		popad

		ret

PrAX		endp


		align	4
PrAL		proc

		; Save all registers
		pushad

		and	eax, 000000FFh
		
		mov	ebx, eax
		shr	eax, 1*4
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL71
		add	al, '0'
		jmp	@@LL72
@@LL71:		add	al, 55
@@LL72:		call	COUT
		mov	eax, ebx
		
		mov	ebx, eax
		and	al, 0Fh
		cmp	al, 9
		jg	@@LL81
		add	al, '0'
		jmp	@@LL82
@@LL81:		add	al, 55
@@LL82:		call	COUT

		mov	al, ' '
		call	COUT
		
		; Restore all registers
		popad

		ret

PrAL		endp


endif

code32		ends
		end

