; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	DICTMNG2.ASM						!
; 									!
; Description:	Definitions for playing with the dictionnary.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	TICK		TICK						!
;	DTICK		DTICK						!
;	EXECUTE		EXECUTE						!
;	RECURSE		RECURSE						!
;	MYSELF		MYSELF						!
; 	DEFERRED	DEFERRED					!
; 	DEFER		DEFER						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions


; ===============================================================
; = TICK							=
; =                                                 		=
; = Code for the forth word:	TICK				=
; =                     	                            	=
; = 	   Stack:	( --- addr )				=
; = Input stream:	' word					=
; =                     	                            	=
; = Usage:	Push the code address of the given word		=
; =		(from input stream).				=
; ===============================================================

		align	4
TICK		proc


		; -----------------------------------------------
		; Parse next word. Must be defined, else an error
		; -----------------------------------------------

		call	IsolateWordEx				; NIA word in EDX

		; Get word code address
		mov	ebx, edx
		mov	eax, [edx.word_realaddr]
		IPUSH	eax

		cmp	[CompileMode], TRUE
		je	@@CompileState
		

		; -------------
		; Execute state
		; -------------

		jmp	@@done


		; -------------
		; Compile state
		; -------------

@@CompileState:	

		; Saved NIA address of called word
		push	ebx		

		; Put 'SUB ESI, 4'
		SUB_ESI_i8	@WordSize
		
		; Put 'MOV DWORD PTR [ESI], IMM32'
		MOV_xESIi8_i32
		mov		eax, PARAM0
		DISCARD		1
		call		CodeComma               	; Code,


		; Update word cross ref
		; ---------------------

		; Point to the last word defined in the current vocabulary: word currently being defined
		mov	ecx, [VOC_CURRENT]
		add	ecx, [VOC_START]
		mov	ecx, [ecx.voc_lastword]			; NIA pointer to caller word

		; Update word cross ref, only if it's not a system word
		pop	edx
		test	[edx.word_flags], FSYSTEM
		jnz	@@System2
		sub	edx, [NIA_START]
		call	CRef_add
@@System2:
		

		; ----------------
		; Return to caller
		; ----------------
		
@@done:		ret

E_TICK:
TICK		endp


; ===============================================================
; = DTICK							=
; =                                                 		=
; = Code for the forth word:	DTICK				=
; =                     	                            	=
; = 	   Stack:	( --- addr )				=
; = Input stream:	' word					=
; =                     	                            	=
; = Usage:	Push the code address of the given word		=
; =		(from input stream).				=
; ===============================================================

		align	4
DTICK		proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; ----------------
		; Return to caller
		; ----------------
		
@@done:		ret

E_DTICK:
DTICK		endp


; ===============================================================
; = EXECUTE							=
; =                                                 		=
; = Code for the forth word:	EXECUTE				=
; =                     	                            	=
; = Stack:	( addr --- )					=
; =                     	                            	=
; = Usage:	Execute the word whcih the address (code)	=
; =		is given.					=
; ===============================================================

		align	4
EXECUTE		proc


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -----------------------------
		; Execute code at given address
		; -----------------------------
		
		mov	eax, PARAM0			; get <addr>
		DISCARD	1				; drop
		jmp	eax

E_EXECUTE:
EXECUTE		endp


; ===============================================================
; = RECURSE							=
; =                                                 		=
; = Code for the forth word:	RECURSE				=
; =                     	                            	=
; = Stack:	( --- )						=
; =                     	                            	=
; = Usage:	Call current definition				=
; ===============================================================

		align	4
RECURSE		proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; Return to caller
		; ----------------
		
		ret

E_RECURSE:
RECURSE		endp


; ===============================================================
; = MYSELF							=
; =                                                 		=
; = Code for the forth word:	MYSELF				=
; =                     	                            	=
; = Stack:	( --- )						=
; =                     	                            	=
; = Usage:	Jump current definition				=
; ===============================================================

		align	4
MYSELF		proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; Return to caller
		; ----------------
		
		ret

E_MYSELF:
MYSELF		endp


; ===============================================================
; = DEFERRED							=
; =                                                 		=
; = Code for the forth word:	DEFERRED			=
; =                     	                            	=
; = 	   Stack:	compile-time:	( --- )			=
; =			    run-time:	( --- )			=
; = Input stream:	deferred word				=
; =                     	                            	=
; = Usage:	To defer to executer definitions.		=
; ===============================================================

		align	4
DEFERRED	proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------

		; 1 ALLOT DROP
		IPUSH	1
		SUMMON	RT_ALLOT
		DISCARD	1

		; DOES>		(this ends the compilation sequence)
		SUMMON	PDOES
		dd	R_DEFERRED

		
		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_DEFERRED:	

		; Return to caller
		ret

E_DEFERRED:
DEFERRED	endp


; ===============================================================
; = DEFER							=
; =                                                 		=
; = Code for the forth word:	DEFER				=
; =                     	                            	=
; = 	   Stack:	( --- )					=
; = Input stream:	defer word1 word2			=
; =                     	                            	=
; = Usage:	Set deferred <word2> to point to <word1>.	=
; ===============================================================

		align	4
DEFER		proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; Return to caller
		; ----------------
		
		ret

E_DEFER:
DEFER		endp


code32		ends
		end
