; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ED.ASM							!
; 									!
; Description:	Definitions for playing with FROSTED, built-in editor.	!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

__ED__		equ	1
include		Ed.inc					; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = EdInit							=
; =								=
; = Editor initializations.					=
; =								=
; = I)	none							=
; =								=
; = O)	CF = 0 if initialization is OK, else CF = 1		=
;	AH = error code if CF == 1				=
; =								=
; = M)	CF AX							=
; ===============================================================

		align	4
EdInit		proc

		; ------------------
		; Save all registers
		; ------------------
		
		push	edi ebp ecx
	

		; ----------------------
		; Initialize each buffer
		; ----------------------

		; First, initialize all with zero
		mov	edx, [EdBuffInfos.binf_BuffAddr]		; Pointer to start of all mem
		clr	al					; Value to write
		lea	edi, EdBuffInfos				; EDI points to each buffer
		mov	ecx, EDMAXNBBUFF * size buffedentry		; Size of the block to write
		cld						; Inc DI
		rep	stosb					; Write the block

		; Initialize all with zero
		clr	al					; Value to write
		mov	edi, edx					; EDI points to start address
		mov	ecx, [EdTotalBuffSize]			; Size of the block to write
		cld						; Inc DI
		rep	stosb					; Write the block
		mov	[EdCurrBuffSize], 0
		
		; Initialize loop
		lea	ebp, EdBuffInfos				; EBP points to each buffer
		mov	cx, EDMAXNBBUFF				; CX is a counter
		
@@Loop:		; Top of loop
		; -----------

		; FileName
		mov	byte ptr [ebp.binf_FileName], 0
		mov	byte ptr [ebp.binf_PrimFName], 0

		; Buffer address		
		mov	[ebp.binf_BuffAddr], edx
		add	[EdCurrBuffSize], INITSZBUFFED
		
		; Buffer current size, BEFORE calling 'EdPutCurLinBuff'
		mov	[ebp.binf_CurrSize], 3

		; Address of top screen line (1st line drawed on top)
		mov	[ebp.binf_1stLinAddr], edx

		; Current line address (in the edited text)
		mov	[ebp.binf_CurLinAddr], edx

		; Allocated size
		mov	[ebp.binf_BuffAllSz], INITSZBUFFED
		
		; Actual nb of lines
		mov	[ebp.binf_LinesNb], 1
		
		; Current position in the text
		mov	[ebp.binf_frow], 0			; First row # drawed (1st line of the screen)
		mov	[ebp.binf_fcol], 0			; Col # (real col, in expanded line, as show on screen)
		mov	[ebp.binf_trow], 0			; Row #
		mov	[ebp.binf_tcol], 0			; Col #
		
		; Modified and not saved ?
		mov	[ebp.binf_NotSaved], FALSE

		; No line tagged
		clr	ebx
@@NoTags:	mov	[ebp.binf_Tags_Row + ebx], -1
		mov	[ebp.binf_Tags_Col + ebx], -1
		inc	ebx
		cmp	bl, MAXNBTAGS
		jne	@@NoTags
		
		; Selected block
		mov	[ebp.binf_BlockBegin], -1
		
		; In memory (-> not to be reloaded)
		mov	[ebp.binf_InMemory], TRUE

		; Put a end-of-file mark at the end of text
		mov	byte ptr [edx],     0Dh
		mov	byte ptr [edx + 1], 0Ah
		mov	byte ptr [edx + 2], K_EOF


		; Loop
		; ----

		add	edx, INITSZBUFFED		; Update next buffer address		
		add	ebp, size buffedentry
		loopw	@@Loop


		; -----------------------
		; Current buffer # edited
		; -----------------------
	
		mov	[EdCurrBufferNb], 0

	
		; --------------------------
		; Mode insert or overwrite ?
		; --------------------------
	
		mov	[EdInsertMode], TRUE		; INS mode first


		; -------------------------
		; Allocate memory for scrap
		; -------------------------

		; Initialize all scrap with zero
		mov	edi, [EdScrap]			; EDI points to start address
		mov	ecx, [EdTotalScrapSize]		; Size of the block to write
		clr	al				; Value to write
		cld					; Inc DI
		rep	stosb				; Write the block
		
		; Scrap is empty
		mov	[EdScrapSize], 0


		; --------------------------
		; Allocate memory for macros
		; --------------------------

		; Initialize allocated region with zero
		mov	edi, [EdMacros]			; EDI points to start address
		mov	ecx, DEFSZ_EDMACROS		; Size of the block to write
		clr	al				; Value to write
		cld					; Inc DI
		rep	stosb				; Write the block

		; Macros region is empty
		mov	[EdMacrosSize], 0
		
		; No recording macro
		mov	[EdMacRecording], 0


		; --------------
		; Find / Replace
		; --------------

		mov	[EdFOpt_Word], FALSE		; Word On/Off ?
		mov	[EdFOpt_Case], FALSE		; Case sensitive On/Off ?
		mov	[EdFOpt_AllBuff], FALSE		; On all buffers ?
		mov	[EdPromptFind], 0		; Text to search


		; --------
		; No Error
		; --------
		
@@NoError:	clc					; CF = 0
		jmp	@@Exit
	

		; -----
		; Error
		; -----
		
@@Error:		stc					; CF = 1
	

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ecx ebp edi
		ret

EdInit		endp


; ===============================================================
; = EdNotAvailable						=
; =								=
; = Usage							=
; =	Print the msg "Function not still implemented."		=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdNotAvailable	proc

		; -----------------
		; Erase status line
		; -----------------
		
		call	EdClrStatusLine


		; -----------------
		; Print the message	(pointed by EDX)
		; -----------------

		mov	ah, [edc_sll]			; Attribute / Color
		clr	al				; Asking to compute msg length
		lea	edx, @@Msg			; Pointer on msg
		clr	bh				; Row #
		clr	bl				; Col #
		call	Ed_XYAS
		
		
		; ----------------
		; Wait for ESC key
		; ----------------
		
@@WaitEsc:	call	CharIn
		cmp	ax, K_ESC
		jne	@@WaitEsc
		

		; ----------------------
		; Redraw the status line
		; ----------------------
		
		call	EdPrStatusLine
		call	EdPosCur
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

		align	4		
@@Msg:		db	"This function is not still implemented... Press ESC "
		db	0
		
EdNotAvailable	endp

		
; ===============================================================
; = EdShowNotSavedF						=
; =								=
; = Usage:							=
; =	Show the current lig number on status line.		=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdShowNotSavedF	proc

		; Registers used
		push	ax ebx cx

		; Get adequate character: ' ' if saved, '*' if to save
		cmp	[ebp.binf_NotSaved], TRUE
		je	@@On
		mov	al, ' '
		jmp	@@Show
@@On:		mov	al, '*'

		; Show character
@@Show:		mov	ah, [edc_sli]				; Attribute / Color
		clr	bh					; Row # (status line)
		clr	bl					; Column #
		call	Ed_XYAC

		; Return to caller
		pop	cx ebx ax
		ret
		
EdShowNotSavedF	endp


; ===============================================================
; = EdShowLigNb							=
; =								=
; = Usage:							=
; =	Show the current lig number on status line.		=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdShowLigNb	proc

		; Save used registers
		push	eax
		push	bx
		push	cx

		; Print decimal number
		mov	cl, [edc_sli]					; Attr / Color
		mov	eax, [ebp.binf_trow]
		inc	eax
		clr	bh					; Row # (status line)
		mov	bl, 12					; Column #
		mov	ch, 5					; Field size
		call	EdPrNumber

		; Return to caller
		pop	cx
		pop	bx
		pop	eax
		ret

EdShowLigNb	endp


; ===============================================================
; = EdShowColNb							=
; =								=
; = Usage:							=
; =	Show the current col number on status line.		=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdShowColNb	proc

		; Save used registers
		push	eax
		push	bx
		push	cx

		; Print decimal number
		mov	cl, [edc_sli]				; Attr / Color
		mov	eax, [ebp.binf_tcol]
		inc	ax
		clr	bh					; Row # (status line)
		mov	bl, 24					; Column #
		mov	ch, 3					; Field size
		call	EdPrNumber

		; Return to caller
		pop	cx
		pop	bx
		pop	eax
		ret

EdShowColNb	endp


; ===============================================================
; = EdShowNbLines						=
; =								=
; = Usage:							=
; =	Show the current nb of lines				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdShowNbLines	proc

		; Save used registers
		push	eax bx cx

		; Print the size in lines nb
		mov	cl, [edc_sli]				; Attr / Color
		mov	eax, [ebp.binf_LinesNb]
		clr	bh					; Row # (status line)
		mov	bl, 35					; Column #
		mov	ch, 4					; Field size
		call	EdPrNumber

		; Return to caller
		pop	cx bx eax
		ret

EdShowNbLines	endp


; ===============================================================
; = EdShowCurrSize						=
; =								=
; = Usage:							=
; =	Show the current size of the edited text.		=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdShowCurrSize	proc

		; Save used registers
		push	eax bx cx

		; Print the size in lines nb
		mov	cl, [edc_sli]				; Attr / Color
		mov	eax, [ebp.binf_CurrSize]
		clr	bh					; Row # (status line)
		mov	bl, 46					; Column #
		mov	ch, 6					; Field size
		call	EdPrNumber

		; Return to caller
		pop	cx bx eax
		ret

EdShowCurrSize	endp


; ===============================================================
; = EdShowInsOvrMod						=
; =								=
; = Usage:							=
; =	Show the current mode for insert / overwrite.		=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdShowInsOvrMod	proc

		; Save used registers
		; -------------------

		push	ax bx edx


		; ---------------------		
		; Where is the cursor ?
		; ---------------------		
		
		cmp	[EdFastScreen], TRUE			; Fast screen function ?
		je	@@Fast1
		call	EdWhereXY
		push	ax
@@Fast1:


		; Which mode is active ?
		; ----------------------

		cmp	[EdInsertMode], FALSE
		je	@@OvrMode
		

		; Insertion mode
		; --------------
		
@@InsMode:	mov	ah, [edc_sli]				; Attr / Color
		mov	al, 3					; Nb of chars to print
		clr	bh					; Row #
		mov	bl, 58					; Col #
		lea	edx, EdMsgIns				; "INS"
		call	Ed_XYAS
		jmp	@@RestCurs


		; Overwrite mode
		; --------------
		
@@OvrMode:	mov	ah, [edc_sli]				; Attr / Color
		mov	al, 3					; Nb of chars to print
		clr	bh					; Row #
		mov	bl, 58					; Col #
		lea	edx, EdMsgOvr				; "INS"
		call	Ed_XYAS



		; Restore cursor position
		; -----------------------

@@RestCurs:	cmp	[EdFastScreen], TRUE			; Fast screen function ?
		je	@@Fast2
		pop	ax
		call	EdGotoXY
@@Fast2:


		; Return to caller
		; ----------------
		
@@Exit:		pop	edx bx ax
		ret

EdShowInsOvrMod	endp


; ===============================================================
; = EdShowFileName						=
; =								=
; = Usage:							=
; =	Show the filename of current buffer.			=
; =								=
; = I)	none							=
; = O)	none							=
; = M)	none							=
; ===============================================================

		align	4
EdShowFileName	proc

		; -------------------
		; Save used registers
		; -------------------
		
		pushad


		; -----------------
		; Draw the filename
		; -----------------

		clr	edx
@@Draw2:		mov	al, [ebp.binf_PrimFName + edx]
		cmpz	al, 0
		je	@@EndDraw

		mov	ah, [edc_sli]				; Attr / Color
		clr	bh					; Row # (status line)
		mov	bl, 68					; Column #
		add	bl, dl
		mov	cl, 1					; Only a character to print
		call	Ed_XYNAC
		
		inc	edx
		jmp	@@Draw2
		
@@EndDraw:


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		popad
		ret

EdShowFileName	endp


; ===============================================================
; = EdPrStatusLine						=
; =								=
; = Usage:							=
; = 	Redraw all the status line.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	None.							=
; ===============================================================

		align	4
EdPrStatusLine	proc

		; Save all registers
		pushad

		; Print the Status Line
		mov	ah, [edc_sll]				; Attr / Color
		mov	bx, [EdColsNb]				
		mov	al, bl					; Nb of chars to print (line width in AL)
		clr	bh					; Row #
		clr	bl					; Column #
		lea	edx, EdStatusLine				; The string to print
		call	Ed_XYAS
		
		; Print the buffer #
		mov	cl, [edc_sli]				; Attr / Color
		mov	eax, [EdCurrBufferNb]
		inc	eax					; Buffer #
		clr	bh					; Row # (status line)
		mov	bl, 2					; Column #
		mov	ch, 2					; Field size
		call	EdPrNumber

		; Print the flag 'Buffer not saved'
		call	EdShowNotSavedF

		; Print the current line #
		call	EdShowLigNb
				
		; Print the current column #
		call	EdShowColNb

		; Print the nb of lines
		call	EdShowNbLines
		
		; Print the current size (before 'EdPutCurLinBuff')
		call	EdShowCurrSize

		; Print the current mode (INS or OVR)
		call	EdShowInsOvrMod
		
		; Draw the filename
		call	EdShowFileName


		; ----------------
		; Return to caller
		; ----------------

		popad		
		ret

EdPrStatusLine	endp


; ===============================================================
; = EdInstallFrame						=
; =								=
; = Usage:							=
; = 	Erase all the screen, then install the editor frame.	=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; =								=
; = Note:							=
; =	Called initialy, when editor starts. And after the	=
; =	ESC-B command.						=
; ===============================================================

		align	4
EdInstallFrame	proc

		; -----------------------
		; Install the status line
		; -----------------------

		call	EdPrStatusLine
		

		; ----------------------
		; Clear the message line
		; ----------------------
		
		call	EdClrMsgLine


		; ------------------------
		; Erase all the text space
		; ------------------------

		mov	bx, [EdFirstRowNb]		
		mov	bh, bl					; Start with line #1
		mov	dx, [EdLastRowNb]			; Last line # to draw
		
@@EraseLine:	mov	ah, [edc_txn]				; Color / Attribute
		mov	al, ' '					; Character to print
		clr	bl					; We start at column #0
		mov	cx, [EdColsNb]				; The width of the screen (in CL)
		call	Ed_XYNAC
		
		inc	bh
		cmp	bh, dl
		jle	@@EraseLine


		; ----------------
		; Return to caller
		; ----------------
		
		ret

EdInstallFrame	endp


; ===============================================================
; = EdSelectBuffer						=
; =								=
; = Usage:							=
; =	Select a specified buffer.				=
; =								=
; = I)	EAX : buffer # (0, 1, ... )				=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdSelectBuffer	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff
		
		
		; ----------------------------------------------
		; Update EBP to point on new buffer informations      (Remember: EBP points always on current buffer informations)
		; ----------------------------------------------

		mov	edx, [EdCurrBufferNb]
		mov	[EdCurrBufferNb], eax
		imul	ebp, eax, size buffedentry
		add	ebp, offset EdBuffInfos
		
		
		; ----------------------
		; Redraw the status line
		; ----------------------
		
		call	EdPrStatusLine
		

		; ----------------------
		; Erase the message line
		; ----------------------
		
		call	EdClrMsgLine


		; ------------------------------------------------------
		; If buffer was unloaded to disk, re-read-it into memory
		; ------------------------------------------------------
		
		cmp	[ebp.binf_InMemory], TRUE
		je	@@InMemory
		
		; Ask room for the entire buffer
		mov	eax, [ebp.binf_CurrSize_S]		; Saved, when InMemory is FALSE
		call	EdRequestMem
		jnc	@@OK
@@OK:
		; Read file to buffer
		mov	eax, [EdCurrBufferNb]
		call	EdDiskReadBuffer
		
@@InMemory:


		; --------------------		
		; Redraw the text page
		; --------------------
		
		call	EdRedrawPage
		
		
		; -----------------------------------
		; Put screen cursor on right position
		; -----------------------------------
		
		call	EdPosCur
		
		
		; ----------------
		; Return to caller
		; ----------------
		
		ret

EdSelectBuffer	endp


; ===============================================================
; = EdHelp							=
; =								=
; = Usage:							=
; =	Draw a help page.					=
; =								=
; = I)	none.							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdHelp		proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff
		

		; --------------------------------------
		; Compute the Upper/Left corner position
		; --------------------------------------
		
		; Row in BH
		clr	ebx
		mov	bx, [EdLastRowNb]
		sub	bx, [EdFirstRowNb]
		inc	bx
		sub	bx, HELPSCR_YSIZE
		shr	bx, 1
		add	bx, [EdFirstRowNb]
		mov	bh, bl
		
		; Col in CL
		mov	cx, [EdColsNb]
		sub	cx, HELPSCR_XSIZE
		shr	cx, 1
		mov	bl, cl
		

		; ------------------
		; Draw the help page
		; ------------------
		
		; ESI : starting help msg
		lea	esi, EdHelpScreen
		
		; Current Attribute/Color
@@Draw0:	mov	ah, [edc_hln]
		
		; Draw line
@@Draw1:	mov	al, [esi]
		cmpz	al, 0
		je	@@Draw2
		
		; Translate characters ?
		cmp	al, 0AEh
		jne	@@tc2
		mov	al, 01Bh
		jmp	@@tc99
@@tc2:		cmp	al, 0AFh
		jne	@@tc3
		mov	al, 01Ah
		jmp	@@tc99
@@tc3:		cmp	al, 0C1h
		jne	@@tc4
		mov	al, 018h
		jmp	@@tc99
@@tc4:		cmp	al, 0C2h
		jne	@@tc5
		mov	al, 019h
		jmp	@@tc99
@@tc5:		cmp	al, 0B0h
		jne	@@tc99
		mov	al, 011h
@@tc99:

		; Change attribute/color ?
		cmp	al, 07Bh
		jne	@@cac2
		mov	ah, [edc_hlk]
		jmp	@@NextChar
@@cac2:		cmp	al, 07Dh
		jne	@@cac3
		mov	ah, [edc_hln]
		jmp	@@NextChar
@@cac3:		cmp	al, 05Bh
		jne	@@cac4
		mov	ah, [edc_hla]
		jmp	@@NextChar
@@cac4:		cmp	al, 05Dh
		jne	@@cac99
		mov	ah, [edc_hln]
		jmp	@@NextChar
@@cac99:
		
		; Draw character:   AL: char   AH: attr   BH: #row   BL: #col
		call	Ed_XYAC
		
		; Next char
		inc	bl					; Next col
@@NextChar:	inc	esi
		jmp	@@Draw1

@@Draw2:

		; End of line
		inc	esi
		cmp	byte ptr [esi], 0
		je	@@Draw3					; Finish

		; Next line
		inc	bh
		mov	bl, cl
		
		; Print next line
		jmp	@@Draw0
		
		; Finished
@@Draw3:


		; ----------------------
		; Wait ESC to be pressed
		; ----------------------
		
@@WaitEsc:	call	CharIn
		cmp	ax, K_ESC
		jne	@@WaitEsc		


		; ---------------
		; Redraw the page
		; ---------------
		
		; Redraw the text page
		call	EdRedrawPage
		
		; Put screen cursor on right position
		call	EdPosCur
		
		
		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdHelp		endp


; ===============================================================
; = IsKeyACommand						=
; =								=
; = Called by 'EdHeart' : see is the user key is a command.	=
; =								=
; = I)	AX  : key from keyboard					=
; = O)	FLAGS							=
; = M)	all registers.						=
; ===============================================================

		align	4
IsKeyACommand	proc

		; -----------------
		; Is-it a command ? 
		; -----------------
		
		lea	esi, EdFunctTable
@@Is2:		cmp	ax, [esi]
		je	@@Is3
		cmp	word ptr [esi + 2], 0
		je	@@Is4
		mov	bx, ax
		and	bl, 7Fh
		cmp	bx, [esi]
		je	@@Is3
@@Is4:		add	esi, 10					; 8: 2 + 2 + 4 + 2
		cmp	word ptr [esi], 0
		jne	@@Is2
		
		; End of table reached: AX not a command
		mov	[EdWasACmd], FALSE
		jmp	@@Exit
		
		; AX is a command
@@Is3:		movzx	eax, word ptr [esi + 8]			; Load parameter
		call	[esi + 4]				; Do specified command
		mov	[EdWasACmd], TRUE
				

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

IsKeyACommand	endp


; ===============================================================
; = EdTakeChar							=
; =								=
; = Usage:							=
; =	It's not a command key: take this char, and bring it	=
; =	into the current line edited.				=
; =								=
; = I)	AL  : char code						=
; = O)	none							=
; = M)	all registers.						=
; ===============================================================

		align	4
EdTakeChar	proc

		; ------------------------------
		; In OVR mode, just put the char
		; ------------------------------
		
		cmp	[EdInsertMode], TRUE
		je	@@InsMode

		
		; --------------
		; Overwrite mode
		; --------------
		
		; Store the char
@@StoreChar:	lea	ebx, EdCurrLineBuff				; Line buffer
		mov	ecx, [ebp.binf_tcol]				; Current col
		mov	[ebx + ecx], al					; Store char

		; Update the line on the screen
		jmp	@@UpdateLine
		
		
		; ----------------------------
		; In INS mode, insert the char
		; ----------------------------
		
@@InsMode:

		; Line too long ?
		push	eax
		call	EdCompCurLinSz
		pop	ebx
		xchg	eax, ebx
		inc	ebx
		cmp	bx, (EDMAXSZLINE - 1)
		je	@@Exit
		ifdef __NOHSCROLL__
		cmp	bx, [EdColsNb]
		je	@@Exit
		endif
	
		; Insert space
		lea	esi, (EdCurrLineBuff + EDMAXSZLINE - 2)		; Source address
		lea	edi, (EdCurrLineBuff + EDMAXSZLINE - 1)		; Destination address
		mov	ecx, (EDMAXSZLINE - 1)
		sub	ecx, [ebp.binf_tcol]				; Length of the block
		std							; Inc ESI and EDI
		rep	movsb  				               	; Move/Copy the block
		cld

		; Store character in right place
		jmp	@@StoreChar


		; -----------------
		; Horizontal scroll
		; -----------------

@@UpdateLine:	mov	eax, [ebp.binf_tcol]
		sub	eax, [ebp.binf_fcol]
		inc	eax
		cmp	ax, [EdColsNb]
		jne	@@NoHScroll
		ifdef __NOHSCROLL__
		jmp	@@Exit
		endif


		; Put current line in edited text
		call	EdPutCurLinBuff
		
		; Update the position of the first column shown (horizontal scrolling)
		mov	ebx, [ebp.binf_fcol]
		add	ebx, STEPMOVHSCROLL
		mov	[ebp.binf_fcol], ebx
		
		; Redraw the complete page
		call	EdRedrawPage
		jmp	@@NoHS2


		; -------------------------------------
		; Redraw the current line on the screen
		; -------------------------------------
		
@@NoHScroll:	; Just redraw the current line
		call	EdDrawCurrLine

		; Update cursor position
@@NoHS2:		inc	[ebp.binf_tcol]				; Next col
		
		; Print the current column #
		call	EdShowColNb

		; Update the cursor position on screen
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdTakeChar	endp


; ===============================================================
; = EdCurrBuffIsMod						=
; =								=
; = Usage:							=
; =	Mark current buffer as modified, and show it in		=
; =	status line.						=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdCurrBuffIsMod	proc

		; ------------------------------------------
		; Flag 'Modified and Not saved' already ON ?
		; ------------------------------------------
		
		cmp	[ebp.binf_NotSaved], FALSE
		jne	@@Exit
		
		
		; ------------------
		; Put the flag at ON
		; ------------------
		
		mov	[ebp.binf_NotSaved], TRUE
		
		
		; ----------------------
		; Update the status line
		; ----------------------

		call	EdShowNotSavedF


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdCurrBuffIsMod	endp


; ===============================================================
; = EdAbortBuffer						=
; =								=
; = Usage:							=
; =	Abort the current buffer.				=
; =								=
; = I)	none							=
; = O)	none							=
; = M)	all registers.						=
; ===============================================================

		align	4
EdAbortBuffer	proc

		; -----------------------------------
		; If buffer not saved, ask to confirm 
		; -----------------------------------
		
		cmp	[ebp.binf_NotSaved], FALSE
		je	@@Ok
		
		; Ask for confirmation
		lea	edx, EdMsgConfAbortBf
		call	EdConfirm
		jnc	@@Exit					; Abort or Not confirmation
		
@@Ok:

		; ----------------------------------------
		; Give back the memory used by this buffer
		; ----------------------------------------

		; Give back memory
		mov	eax, [ebp.binf_BuffAllSz]
		sub	eax, INITSZBUFFED
		jz	@@GiveBackDone
		call	EdGiveBackMem
@@GiveBackDone:


		; ------------------
		; Reset buffer infos
		; ------------------
		
		; FileName
		mov	byte ptr [ebp.binf_FileName], 0
		mov	byte ptr [ebp.binf_PrimFName], 0

		; Buffer current size, BEFORE calling 'EdPutCurLinBuff'
		mov	[ebp.binf_CurrSize], 3

		; Address of top screen line (1st line drawed on top)
		mov	edx, [ebp.binf_BuffAddr]
		mov	[ebp.binf_1stLinAddr], edx

		; Current line address (in the edited text)
		mov	[ebp.binf_CurLinAddr], edx

		; Actual nb of lines
		mov	[ebp.binf_LinesNb], 1
		
		; Current position in the text
		mov	[ebp.binf_frow], 0			; First row # drawed (1st line of the screen)
		mov	[ebp.binf_fcol], 0			; Col # (real col, in expanded line, as show on screen)
		mov	[ebp.binf_trow], 0			; Row #
		mov	[ebp.binf_tcol], 0			; Col #
		
		; Modified and not saved ?
		mov	[ebp.binf_NotSaved], FALSE

		; No line tagged
		clr	ebx
@@NoTags:	mov	[ebp.binf_Tags_Row + ebx], -1
		mov	[ebp.binf_Tags_Col + ebx], -1
		inc	ebx
		cmp	bl, MAXNBTAGS
		jne	@@NoTags
		
		; Selected block
		mov	[ebp.binf_BlockBegin], -1
		
		; In memory (-> not to be reloaded)
		mov	[ebp.binf_InMemory], TRUE

		; Put a end-of-file mark at the end of text
		mov	byte ptr [edx],     0Dh
		mov	byte ptr [edx + 1], 0Ah
		mov	byte ptr [edx + 2], K_EOF


		; ----------------------------
		; Now redraw the complete page
		; ----------------------------
		
		call	EdRedrawPage


		; ----------------------------------------
		; Bring 1st line of text to editing buffer
		; ----------------------------------------

		mov	ecx, [ebp.binf_CurLinAddr]
		call	EdGetCurLinBuff


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		

		; Redraw the entire status line
		call	EdPrStatusLine
		
		; Update the cursor position on screen
		call	EdPosCur

		; Return
		ret

EdAbortBuffer	endp


ifdef UNDERDEV
; !!!

; ===============================================================
; = EdShowInfos							=
; =								=
; = Usage							=
; =	On developpement version, print infos on bottom.	=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================


		align	4
edPDOTQ       	proc

		; Save used registers
		push    ax ecx

		; Store into DI the string address
		mov	ecx, [esp + 6]

@@LL1:      	mov	al, [ecx]
		cmpz	al, 0
		je	@@LL2
		call	COUT
		inc	ecx
		jmp	@@LL1
@@LL2:

		; Move the IP after the string
		inc	ecx
		mov	[esp + 6], ecx

		; Return to caller
		pop	ecx ax
		ret

edPDOTQ       	endp

		align	4
EdShowInfos	proc

		; ------------------
		; Save all registers
		; ------------------
		
		pushad


		; ----------------------------
		; Save current cursor position
		; ----------------------------
		
		call	EdWhereXY
		push	ax


		; --------------------------------------
		; Draw the 1st 15 char of current buffer
		; --------------------------------------

		mov	ax, [EdRowsNb]
		dec	al
		mov	ah, al
		clr	al
		call	EdGotoXY

		lea	esi, EdCurrLineBuff
		clr	ebx
@@Loop2:		mov	al, [esi + ebx]
		cmp	al, 020h
		jae	@@Loop2B
		mov	al, '.'
@@Loop2B:	call	COUT
		inc	ebx
		cmp	ebx, 11
		jne	@@Loop2


ifdef NEVER		
		; -----------------------------------
		; Draw the 1st 20 char of text buffer
		; -----------------------------------

		call	edPDOTQ
		db	" | ", 0

		mov	esi, [ebp.binf_BuffAddr]
		clr	ebx
@@Loop3:		mov	al, [esi + ebx]
		cmp	al, 020h
		jae	@@Loop3B
		mov	al, '.'
@@Loop3B:	call	COUT
		inc	ebx
		cmp	ebx, 20
		jne	@@Loop3
endif
		
		; ----------------
		; Show row and col
		; ----------------

ifdef NEVER
		call	edPDOTQ
		db	" | frow=", 0
		mov	eax, [ebp.binf_frow]
		call	PrAX

		call	edPDOTQ
		db	"trow=", 0
		mov	eax, [ebp.binf_trow]
		call	PrAX

		call	edPDOTQ
		db	"fcol=", 0
		mov	eax, [ebp.binf_fcol]
		call	PrAX

		call	edPDOTQ
		db	"tcol=", 0
		mov	eax, [ebp.binf_tcol]
		call	PrAX

endif


		; Buffer #0
		lea	edx, EdBuffInfos
		call	edPDOTQ
		db	"B0:", 0
		mov	eax, [edx.binf_BuffAddr]
		call	PrEAX
		mov	eax, [edx.binf_CurrSize]
		call	PrAX
		mov	eax, [edx.binf_BuffAllSz]
		call	PrAX

		; Buffer #1
		add	edx, size buffedentry
		call	edPDOTQ
		db	"  B1:", 0
		mov	eax, [edx.binf_BuffAddr]
		call	PrEAX
		mov	eax, [edx.binf_CurrSize]
		call	PrAX
		mov	eax, [edx.binf_BuffAllSz]
		call	PrAX

		; Buffer #2
		add	edx, size buffedentry
		call	edPDOTQ
		db	"  B2:", 0
		mov	eax, [edx.binf_BuffAddr]
		call	PrEAX
		mov	eax, [edx.binf_CurrSize]
		call	PrAX
		mov	eax, [edx.binf_BuffAllSz]
		call	PrAX

		; Buffer #3
		add	edx, size buffedentry
		call	edPDOTQ
		db	"  B3:", 0
		mov	eax, [edx.binf_BuffAddr]
		call	PrEAX
		mov	eax, [edx.binf_CurrSize]
		call	PrAX
		mov	eax, [edx.binf_BuffAllSz]
		call	PrAX

		; Buffer #4
		add	edx, size buffedentry
		call	edPDOTQ
		db	"  B4:", 0
		mov	eax, [edx.binf_BuffAddr]
		call	PrEAX
		mov	eax, [edx.binf_CurrSize]
		call	PrAX
		mov	eax, [edx.binf_BuffAllSz]
		call	PrAX


		; -------------------------------
		; Restore initial cursor position
		; -------------------------------
		
		pop	ax
		call	EdGotoXY
		
		
		; ----------------
		; Return to caller
		; ----------------
		
		popad
		ret

EdShowInfos	endp

endif


; ===============================================================
; = EdHeart							=
; =								=
; = Main loop of the editor.					=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; =								=
; = Usage:							=
; =	Called initialy, when editor starts.			=
; ===============================================================

		align	4
EdHeart		proc

@@WaitChar:

		; ------------------------------------------------------------------------------
		; In DEVELOPEMENT version, draw several informations on the bottom of the screen
		; ------------------------------------------------------------------------------
		
		ifdef UNDERDEV
		; !!!
;		call	EdShowInfos
		endif

		
		; -------------------
		; Read a keyboard key
		; -------------------

		call	CharIn
		

		; --------------------------------------------------------------------------------
		; If it is a command key, treat it as a command, then return with CF=1 (else CF=0)
		; 	If exit from editor, ZF=1, else ZF=0
		; 	If a command (except exit editor), CF=1, else CF=0
		; --------------------------------------------------------------------------------

		; Is-it a command key ?		
		call	IsKeyACommand
		
		; Exit from editor ?
		cmp	[EdExitEd], FALSE
		jne	@@Exit
		
		; It was a command ?
		cmp	[EdWasACmd], TRUE
		je	@@DoAgain
		
	
		; ----------------------------------
		; Cancel the block selection, if any
		; ----------------------------------
		
		call	EdInactiveBlock

		
		; --------------------------
		; Current buffer is modified
		; --------------------------
		
		call	EdCurrBuffIsMod


		; ------------------------------------------------------------------------------
		; It's not a command key: take this char and bring it to the current line edited
		; ------------------------------------------------------------------------------

		; Loose control characters
		cmp	ax, 020h
		jb	@@DoAgain
		cmp	ax, 100h
		jae	@@DoAgain

		; OK, take the caracter
		call	EdTakeChar
		
		
		; ------------------------------------------------
		; Do again and again: read next char from keyboard
		; ------------------------------------------------

@@DoAgain:	jmp	@@WaitChar


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdHeart		endp


; ===============================================================
; = Editor							=
; =								=
; = Enter in edit session.					=
; =								=
; = I)	none							=
; =								=
; = O)	CF = 1 means load is asked, else CF = 0			=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
Editor		proc

		; ----------------
		; Save EDI and ESI		(They are used by FROSTED !)
		; ----------------
		
		push	esi edi
	

		; ----------------------------------------------------------------------------------------
		; Is the editor active ?
		; This flag is used by some functions called outside editor session ('Update' for example)
		; ----------------------------------------------------------------------------------------
	
		mov	[EditorActive], TRUE			; Flag (0 | 1)


		; ----------------------
		; Get current video mode
		; ----------------------

		call	EdGetVideoMode
		mov	[EdPrevVideoMode], al			; Save value, for restore when exit editor


		; ---------------------
		; Set custom video mode
		; ---------------------

		SUMMON	ED_VIDEOMODE		
		mov	eax, [eax]
		cmp	eax, -1
		je	@@NoVideoMode
		cmpz	al, 0
		je	@@NoVideoMode
		call	EdSetVideoMode
@@NoVideoMode:


		; ------------------------------
		; Ask for the size of the screen
		; ------------------------------

		call	EdGetVideoMode
		shr	ax, 8
		mov	[EdColsNb], ax				; Nb of columns
		call	EdGetRowsNb
		mov	[EdRowsNb], ax				; Nb of rows

		; Update 1st and last line to drawing text
		mov	[EdFirstRowNb], 1
		dec	ax
		mov	[EdMsgRowNb], ax
		dec	ax
		mov	[EdLastRowNb], ax

		; Update empty line
		clr	eax
		mov	ax, [EdColsNb]
		mov	byte ptr [EdEmptyLine + eax], 0Dh
		mov	byte ptr [EdEmptyLine + eax + 1], 0Ah


		; -----------------------------------------------------------
		; Do we use fast screen video functions ? (direct video mode)
		; -----------------------------------------------------------
		
		SUMMON	ED_QDIRECTVIDEO
		mov	eax, [eax]
		mov	[EdFastScreen], al			; Update internal flag (for speed)
		cmp	al, FALSE
		je	@@NotFast
		call	EdGetScreenAddr				; Get each beginning line screen address
@@NotFast:


		; ----------------------------
		; Exit from editor ? Not yet !
		; ----------------------------

		mov	[EdExitEd], FALSE


		; ------------------------------------------------
		; Copy some FROTH variables here (to speed access)
		; ------------------------------------------------

		; Status line, label text		
		SUMMON	EDC_SLL
		mov	eax, [eax]
		mov	[edc_sll], al

		; Status line, infos text		
		SUMMON	EDC_SLI
		mov	eax, [eax]
		mov	[edc_sli], al

		; Text normal
		SUMMON	EDC_TXN
		mov	eax, [eax]
		mov	[edc_txn], al

		; Text selected
		SUMMON	EDC_TXS
		mov	eax, [eax]
		mov	[edc_txs], al

		; Help, normal
		SUMMON	EDC_HLN
		mov	eax, [eax]
		mov	[edc_hln], al

		; Help, alt
		SUMMON	EDC_HLA
		mov	eax, [eax]
		mov	[edc_hla], al

		; Help, keys
		SUMMON	EDC_HLK
		mov	eax, [eax]
		mov	[edc_hlk], al


		; ---------------------
		; Current buffer edited		(Remember: EBP points always on current buffer informations)
		; ---------------------
	
		mov	ebp, [EdCurrBufferNb]
		imul	ebp, size buffedentry
		add	ebp, offset EdBuffInfos
		
		
		; ----------------------------------
		; Words table needs to be computed ?
		; ----------------------------------

		mov	[EdWordsTableOk], FALSE


		; -----------------
		; Install the frame
		; -----------------
		
		call	EdInstallFrame


		; ------------------------------
		; Is a row# and col# specified ?
		; ------------------------------
		
		cmp	[ebp.binf_trow], 0
		je	@@NotRowCol
		
		mov	eax, [ebp.binf_trow]			; Line #
		mov	ebx, [ebp.binf_tcol]			; Colunn #
		call	EdDoGotoLine
		jmp	@@Heart


		; ----------------------------------------
		; Bring 1st line of text to editing buffer
		; ----------------------------------------
		
@@NotRowCol:	mov	ecx, [ebp.binf_CurLinAddr]


		; -----------
		; Redraw page
		; -----------
		
		call	EdRedrawPage


		; -----------------------------------
		; Put screen cursor on right position
		; -----------------------------------

		call	EdPosCur
		
		
		; ----------------------
		; Enter in the main loop
		; ----------------------
		
@@Heart:		call	EdHeart
		
		
		; ---------------------------
		; Restore previous video mode
		; ---------------------------

		mov	al, [EdPrevVideoMode]
		call	EdSetVideoMode


		; ----------------------------------------------------------------------------------------
		; Is the editor active ?
		; This flag is used by some functions called outside editor session ('Update' for example)
		; ----------------------------------------------------------------------------------------
	
		mov	[EditorActive], FALSE			; Flag (0 | 1)


		; -------------------
		; Restore EDI and ESI
		; -------------------

@@Exit:		pop	edi esi
	

		; ----------------
		; Return to caller
		; ----------------
		
		cmp	[EdExitEd], TRUE
		je	@@NoLoad
		stc						; Load
		jmp	@@Ret
@@NoLoad:	clc						; No load
@@Ret:		ret

Editor		endp


code32		ends
		end
