; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ED.ASM							!
; 									!
; Description:	Definitions for playing with FROSTED, built-in editor.	!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc			; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Ed.inc				; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = EdGetVideoMode						=
; =								=
; = Ask to BIOS the current video mode.				=
; =								=
; = I)	none							=
; =								=
; = O)	AL : video mode						=
; =	AH : nb of columns					=
; =	BH : active page					=
; =								=
; = M)	AX BH							=
; ===============================================================

		align	4
EdGetVideoMode	proc

		; ------------------
		; Call BIOS function
		; ------------------

		mov 	v86r_ah, 0Fh			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		mov	ax, v86r_ax			; Output
		mov	bh, v86r_bh

		
		; ----------------
		; Return to caller
		; ----------------
		
		ret

EdGetVideoMode	endp


; ===============================================================
; = EdSetVideoMode						=
; =								=
; = Set the current video mode.					=
; =								=
; = I)	AL : video mode						=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdSetVideoMode	proc

		; -------
		; Save AX
		; -------
		
		push	ax
		

		; ------------------
		; Call BIOS function
		; ------------------

		mov	v86r_al, al			; Input: video mode
		mov 	v86r_ah, 00h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ----------------
		; Return to caller
		; ----------------

		pop	ax		
		ret

EdSetVideoMode	endp


; ===============================================================
; = EdGetRowsNb							=
; =								=
; = Ask to BIOS the number of screen rows.			=
; =								=
; = I)	none							=
; =								=
; = O)	AX : nb of rows						=
; =								=
; = M)	AX							=
; ===============================================================

		align	4
EdGetRowsNb	proc

		; ------------------
		; Call BIOS function
		; ------------------

		mov	v86r_al, 30h			; ???
		mov 	v86r_ah, 11h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		clr	ah
		mov	al, v86r_dl			; Output
		inc	al				; For example: for 25 lines, BIOS give 18h (24)

		
		; ----------------
		; Return to caller
		; ----------------
		
		ret

EdGetRowsNb	endp


; ===============================================================
; = EdGetScreenAddr						=
; =								=
; = Ask to BIOS the number of screen rows.			=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	All registers.						=
; ===============================================================

		align	4
EdGetScreenAddr	proc

		; ----------------
		; In VGA mode only
		; ----------------
		
		mov	eax, 0B8000h
		lea	ebx, _code32a
		sub	eax, [ebx]


		; ---------------------------------------------
		; Compute the address of the start of each line
		; ---------------------------------------------
		
		clr	ecx					; Line counter
		lea	edx, EdScreenLinesAddr			; Where to address
		movzx	ebx, [EdColsNb]
		shl	ebx, 1					; Size of a line
@@Loop:		mov	[edx + ecx * 4], eax
		add	eax, ebx
		inc	ecx
		cmp	cx, [EdRowsNb]
		jne	@@Loop

		
		; ----------------
		; Return to caller
		; ----------------
		
		ret

EdGetScreenAddr	endp


; ===============================================================
; = Ed_C							=
; =								=
; = Print a character.						=
; =								=
; = I)	AL : the character to print				=
; =								=
; = O)	None							=
; =								=
; = M)	None							=
; ===============================================================

		align	4
Ed_C		proc

		; -------
		; Save AL
		; -------
		
		push	ax
		

		; ------------------
		; Call BIOS function
		; ------------------

		mov	v86r_al, al			; Input: char to print
		mov	v86r_cx, 1			; One character to print
		mov 	v86r_ah, 0Ah			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ----------------
		; Return to caller
		; ----------------

		pop	ax		
		ret

Ed_C		endp


; ===============================================================
; = Ed_AC							=
; =								=
; = Usage:							=
; =	Print one character with a given color / attribute at	=
; =	current cursor position. The cursor position is		=
; =	updated after the drawn.				=
; =								=
; = I)	AL : the character to print				=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
Ed_AC		proc

		; -------
		; Save AL
		; -------
		
		push	ax
		

		; ------------------
		; Call BIOS function
		; ------------------

		mov	v86r_bh, 0			; Page #
		mov	v86r_al, al			; Input: char to print
		mov	v86r_bl, ah			; Input: attribute / color
		mov 	v86r_ah, 0Eh			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ----------------
		; Return to caller
		; ----------------

		pop	ax		
		ret

Ed_AC		endp


; ===============================================================
; = Ed_XYAC_fast						=
; =								=
; = Print a character with a given color / attribute.		=
; =								=
; = I)	AL : the character to print				=
; =	AH : the attribute					=
; =	BH : row (from 0 to ..)					=
; =	BL : column (from 0 to ..)				=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
Ed_XYAC_fast	proc

		; --------------
		; Save used regs
		; --------------
		
		push	edx esi
		
		
		; ----------------------------
		; Fast write into video memory
		; ----------------------------
		
		movzx	edx, bh
		mov	edx, [EdScreenLinesAddr + edx * 4]		; Address of the beginning of the line
		movzx	esi, bl
		mov	[edx + 2 * esi], ax			; Store couple (char,attr) on screen


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	esi edx
		ret

Ed_XYAC_fast	endp


; ===============================================================
; = Ed_XYAC							=
; =								=
; = Print a character with a given color / attribute.		=
; =								=
; = I)	AL : the character to print				=
; =	AH : the attribute					=
; =	BH : row (from 0 to ..)					=
; =	BL : column (from 0 to ..)				=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
Ed_XYAC		proc

		; ----------------------
		; Fast screen function ?
		; ----------------------
		
		cmp	[EdFastScreen], TRUE
		je	Ed_XYAC_fast


		; -------
		; Save AL
		; -------
		
		push	ax
		

		; ---------------
		; Position cursor
		; ---------------

		push	ax
		mov	ah, bh				; Row #
		mov	al, bl				; Col #
		call	EdGotoXY
		pop	ax


		; ------------------
		; Call BIOS function
		; ------------------

		mov	v86r_bh, 0			; Page #
		mov	v86r_al, al			; Input: char to print
		mov	v86r_bl, ah			; Input: attribute / color
		mov	v86r_ch, 0
		mov	v86r_cl, 1			; Nb of characters to print
		mov 	v86r_ah, 09h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ----------------
		; Return to caller
		; ----------------

		pop	ax		
		ret

Ed_XYAC		endp


; ===============================================================
; = Ed_XYNAC_fast						=
; =								=
; = Print a character with a given color / attribute.		=
; =								=
; = I)	AL : the character to print				=
; =	AH : the attribute					=
; =	BH : row (from 0 to ..)					=
; =	BL : column (from 0 to ..)				=
; =	CL : nb of chars to print				=
; =								=
; = O)	none							=
; =								=
; = M)	CL							=
; ===============================================================

		align	4
Ed_XYNAC_fast	proc

		; --------------
		; Save used regs
		; --------------
		
		push	cx edx esi
		
		
		; ---------------
		; Initializations
		; ---------------
		
		; 1st screen address
		movzx	edx, bh
		mov	edx, [EdScreenLinesAddr + edx * 4]		; Address of the beginning of the line
		movzx	esi, bl
		add	edx, esi
		add	edx, esi

		
		; ----------------------------
		; Fast write into video memory
		; ----------------------------

@@Draw:		mov	[edx], ax				; Store couple (char,attr) on screen
		dec	cl
		jz	@@Exit
		add	edx, 2
		jmp	@@Draw


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	esi edx cx
		ret

Ed_XYNAC_fast	endp


; ===============================================================
; = Ed_XYNAC							=
; =								=
; = Print a character with a given color / attribute.		=
; =								=
; = I)	AL : the character to print				=
; =	AH : the attribute					=
; =	BH : row (from 0 to ..)					=
; =	BL : column (from 0 to ..)				=
; =	CL : nb of chars to print				=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
Ed_XYNAC		proc

		; ----------------------
		; Fast screen function ?
		; ----------------------
		
		cmp	[EdFastScreen], TRUE
		je	Ed_XYNAC_fast


		; -------
		; Save AL
		; -------
		
		push	ax
		

		; ---------------
		; Position cursor
		; ---------------

		push	ax
		mov	ah, bh				; Row #
		mov	al, bl				; Col #
		call	EdGotoXY
		pop	ax


		; ------------------
		; Call BIOS function
		; ------------------

		mov	v86r_bh, 0			; Page #
		mov	v86r_al, al			; Input: char to print
		mov	v86r_bl, ah			; Input: attribute / color
		mov	v86r_ch, 0
		mov	v86r_cl, cl			; Nb of characters to print
		mov 	v86r_ah, 09h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ----------------
		; Return to caller
		; ----------------

		pop	ax		
		ret

Ed_XYNAC		endp


; ===============================================================
; = Ed_XYAS_fast						=
; =								=
; = Print a String (not endded with '\0') with a given 		=
; = color / attribute.						=
; =								=
; = I)	AL  : nb of chars to print				=
; =	AH  : the attribute					=
; =	BH  : row (from 0 to ..)				=
; =	BL  : column (from 0 to ..)				=
; =	EDX : pointer to the string to print			=
; =								=
; = O)	AL : the length of the msg (if was 0).			=
; =								=
; = M)	None.							=
; =								=
; = Note:							=
; =	If AL == 0, then compute the length of the message	=
; = 	(pointed by EDX) ended by '\0'.				=
; =	The cursor position is NOT UPDATED after drawn.		=
; ===============================================================

		align	4
Ed_XYAS_fast	proc

		; --------------
		; Save used regs
		; --------------
		
		push	cx edx esi edi
		
		
		; ---------------
		; Initializations
		; ---------------
		
		; 1st screen address
		mov	edi, edx
		movzx	edx, bh
		mov	edx, [EdScreenLinesAddr + edx * 4]		; Address of the beginning of the line
		movzx	esi, bl
		add	edx, esi
		add	edx, esi

		
		; ----------------------------
		; Fast write into video memory
		; ----------------------------

		cmpz	al, 0
		je	@@Draw2

		; Case: length is given
		mov	cl, al
@@Draw:		cmpz	cl, 0
		je	@@Exit
		mov	al, [edi]
		mov	[edx], ax				; Store couple (char,attr) on screen
		add	edx, 2
		inc	edi
		dec	cl
		jmp	@@Draw

		; Case: length is not given, string endded by '\0'
@@Draw2:		clr	cl
@@Draw3:		mov	al, [edi]
		cmpz	al, 0
		je	@@Draw4
		mov	[edx], ax				; Store couple (char,attr) on screen
		add	edx, 2
		inc	edi
		inc	cl
		jmp	@@Draw3
@@Draw4:		mov	al, cl


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	edi esi edx cx
		ret

Ed_XYAS_fast	endp


; ===============================================================
; = Ed_XYAS							=
; =								=
; = Print a String (not endded with '\0') with a given 		=
; = color / attribute.						=
; =								=
; = I)	AL  : nb of chars to print				=
; =	AH  : the attribute					=
; =	BH  : row (from 0 to ..)				=
; =	BL  : column (from 0 to ..)				=
; =	EDX : pointer to the string to print			=
; =								=
; = O)	AL : the length of the msg (if was 0).			=
; =								=
; = M)	None.							=
; =								=
; = Note:							=
; =	If AL == 0, then compute the length of the message	=
; = 	(pointed by EDX) ended by '\0'.				=
; =	The cursor position is UPDATED after drawn.		=
; ===============================================================

		align	4
Ed_XYAS		proc

		; ----------------------
		; Fast screen function ?
		; ----------------------
		
		cmp	[EdFastScreen], TRUE
		je	Ed_XYAS_fast


		; -------
		; Save AL
		; -------
		
		push	ax


		; -----------------------------------------------
		; Asking to compute message length ? (if AL == 0)
		; -----------------------------------------------
		
		cmpz	al, 0
		jne	@@NoCompLen
		pop	ax
		push	edx
@@CompLen:	cmp	byte ptr [edx], 0
		je	@@EndCompLen
		inc	edx
		inc	al
		jmp	@@CompLen
@@EndCompLen:	pop	edx
		push	ax
@@NoCompLen:
		

		; ------------------
		; Call BIOS function
		; ------------------

		mov	v86r_bl, ah			; Input: attribute / color
		mov	v86r_bh, 0			; Input: Page #
		mov	v86r_ch, 0
		mov	v86r_cl, al			; Input: Length of the string
		mov	v86r_dh, bh			; Input: row #
		mov	v86r_dl, bl			; Input: col #
		mov	v86r_al, 1			; Mode 1: string contains only char (not attr), then update curs pos

		; Get <Seg,Ofs> from string address (in EDX)
		HILO	v86r_es, v86r_bp, edx
		
		mov 	v86r_ah, 13h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ----------------
		; Return to caller
		; ----------------

		pop	ax
		ret

Ed_XYAS		endp


; ===============================================================
; = EdGotoXY							=
; =								=
; = Set the cursor at (x, y) (x is the column, y the row).	=
; =								=
; = I)	AH  : row (from 0 to ..)				=
; =	AL  : column (from 0 to ..)				=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdGotoXY		proc

		; -------
		; Save AL
		; -------
		
		push	ax
		

		; ------------------
		; Call BIOS function
		; ------------------

		mov	v86r_bh, 0			; Input: page 0
		mov	v86r_dh, ah			; Input: row #
		mov	v86r_dl, al			; Input: col #
		mov 	v86r_ah, 02h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ----------------
		; Return to caller
		; ----------------

		pop	ax		
		ret

EdGotoXY		endp


; ===============================================================
; = EdWhereXY							=
; =								=
; = return where is the cursor (x is the column, y the row).	=
; =								=
; = I)	none							=
; =								=
; = O)	AH  : row (from 0 to ..)				=
; =	AL  : column (from 0 to ..)				=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdWhereXY	proc

		; ------------------
		; Call BIOS function
		; ------------------

		mov	v86r_bh, 0			; Input: page 0
		mov 	v86r_ah, 03h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		mov	ah, v86r_dh			; Output: row #
		mov	al, v86r_dl			; Output: col #

		
		; ----------------
		; Return to caller
		; ----------------
		
		ret

EdWhereXY	endp


; ===============================================================
; = EdScrollUp							=
; =								=
; = Usage:							=
; =	Scroll to up a portion of the screen.			=
; =								=
; = I)	al  : # of lines to scroll				=
; =	ah  : attrib/color of new lines created			=
; =	bh  : Upper row						=
; =	bl  : Upper left col					=
; =	ch  : Lower row						=
; =	cl  : Lower right col					=
; =								=
; = O)	None.							=
; =								=
; = M)	None							=
; ===============================================================

		align	4
EdScrollUp	proc

		; ------------------
		; Call BIOS function
		; ------------------

		mov	v86r_al, al			; Input: # of lines to scroll
		mov	v86r_bh, ah			; Input: new lines attrib/color
		mov	v86r_ch, bh			; Input: upper row
		mov	v86r_cl, bl			; Input: upper left col
		mov	v86r_dh, ch			; Input: Lower row
		mov	v86r_dl, cl			; Input: Lower right col
		mov 	v86r_ah, 06h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ----------------
		; Return to caller
		; ----------------
		
		ret

EdScrollUp	endp


; ===============================================================
; = EdScrollDown						=
; =								=
; = Usage:							=
; =	Scroll to down a portion of the screen.			=
; =								=
; = I)	al  : # of lines to scroll				=
; =	ah  : attrib/color of new lines created			=
; =	bh  : Upper row						=
; =	bl  : Upper left col					=
; =	ch  : Lower row						=
; =	cl  : Lower right col					=
; =								=
; = O)	None.							=
; =								=
; = M)	None							=
; ===============================================================

		align	4
EdScrollDown	proc

		; ------------------
		; Call BIOS function
		; ------------------

		mov	v86r_al, al			; Input: # of lines to scroll
		mov	v86r_bh, ah			; Input: new lines attrib/color
		mov	v86r_ch, bh			; Input: upper row
		mov	v86r_cl, bl			; Input: upper left col
		mov	v86r_dh, ch			; Input: Lower row
		mov	v86r_dl, cl			; Input: Lower right col
		mov 	v86r_ah, 07h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ----------------
		; Return to caller
		; ----------------
		
		ret

EdScrollDown	endp


code32		ends
		end
