; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ED_CMD4.ASM						!
; 									!
; Description:	Definitions for playing with FROSTED, built-in editor.	!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Ed.inc					; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = EdBuffSelection						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Show all buffers and selection				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdBuffSelection	proc

		; --------
		; Save EBP
		; --------
		
		push	ebp
		

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff
		

		; -------------
		; Install frame
		; -------------

		call	EdCls					; Clear screen		

		mov	ax, 0000h				; Row #0, Col #0
		call	EdGotoXY
		call	EdPDotQ
		db	"--------------------------------------------------------------------------------", 0
		
		mov	ax, 0100h				; Row #1, Col #0
		call	EdGotoXY
		call	EdPDotQ
		db	"Buffer Key     File                                                   Size   Alt", 0

		mov	ax, 0200h				; Row #2, Col #0
		call	EdGotoXY
		call	EdPDotQ
		db	"--------------------------------------------------------------------------------", 0
		
		mov	ax, [EdRowsNb]
		mov	ah, al
		sub	ah, 4
		clr	al
		call	EdGotoXY
		call	EdPDotQ
		db	"--------------------------------------------------------------------------------", 0

		inc	ah
		call	EdGotoXY		
		call	EdPDotQ
		db	"<ESC> to abort  <PgUp> for previous page (if any)  <PgDn> for next page (if any)", 0
		
		inc	ah
		call	EdGotoXY		
		call	EdPDotQ
		db	"------------------------------------------------------------------------------>", 0

	
		; --------------------------------
		; Show memory infos on bottom page
		; --------------------------------

		mov	ax, [EdRowsNb]
		mov	ah, al
		dec	ah					; # row
		clr	al					; #col
		call	EdGotoXY
		call	EdPDotQ
		db	"Memory used: ", 0

		mov	eax, [EdCurrBuffSize]			; Current size of buffers (in use)
		mov	bx, [EdRowsNb]
		mov	bh, bl
		dec	bh					; # row
		mov	bl, 13					; Column #
		mov	ch, 7					; Field size
		mov	cl, [edc_txn]				; Attr / Color
		call	EdPrNumber

		mov	ax, [EdRowsNb]
		mov	ah, al
		dec	ah					; # row
		mov	al, 20					; #col
		call	EdGotoXY
		call	EdPDotQ
		db	" / ", 0

		mov	eax, [EdTotalBuffSize]			; Total size allocated for all buffers
		mov	bx, [EdRowsNb]
		mov	bh, bl
		dec	bh					; # row
		mov	bl, 23					; Column #
		mov	ch, 7					; Field size
		mov	cl, [edc_txn]				; Attr / Color
		call	EdPrNumber

		mov	ax, [EdRowsNb]
		mov	ah, al
		dec	ah					; # row
		mov	al, 30					; #col
		call	EdGotoXY
		call	EdPDotQ
		db	" (", 0

		mov	eax, [EdCurrBuffSize]			; Current size of buffers (in use)
		imul	eax, 100
		clr	edx
		idiv	[EdTotalBuffSize]			; Result in EAX
		mov	bx, [EdRowsNb]
		mov	bh, bl
		dec	bh					; # row
		mov	bl, 32					; Column #
		mov	ch, 7					; Field size
		mov	cl, [edc_txn]				; Attr / Color
		call	EdPrNumber

		mov	ax, [EdRowsNb]
		mov	ah, al
		dec	ah					; # row
		mov	al, 39					; #col
		call	EdGotoXY
		call	EdPDotQ
		db	"%)", 0


		; ---------------
		; Initializations
		; ---------------
		
		clr	edx					; Buffer # (top, 1st drawn)
		
		
		; -----------
		; Draw a page
		; -----------
		
		; Initialize line (screen) counter
@@DrawPage:	clr	ecx
		

		; -------------
		; Draw the line
		; -------------

@@DrawLine:
		; Pointer on infos
		mov	eax, ecx
		add	eax, edx
		imul	eax, size buffedentry
		lea	ebp, EdBuffInfos + eax

		; In memory ?
		pushad
		cmp	[ebp.binf_InMemory], TRUE
		je	@@InMem
		mov	al, 'D'					; Char to draw
		jmp	@@NextMem
@@InMem:	mov	al, ' '					; Char to draw
@@NextMem:	mov	ah, [edc_txn]				; Attr / Color
		mov	bh, cl
		add	bh, 3					; Row #
		clr	bl					; Col #
		call	Ed_XYAC
		popad
		
		; Buffer #
		pushad
		mov	eax, edx
		add	eax, ecx
		inc	eax					; The nb to draw
		mov	bh, cl
		add	bh, 3					; Row #
		mov	bl, 2					; Column #
		mov	ch, 2					; Field size
		mov	cl, [edc_txn]				; Attr / Color
		call	EdPrNumber
		popad
		
		; Key name
		pushad
		mov	eax, ecx
		add	eax, edx
		lea	edx, @@KeyNames	+ eax * 8		; Pointer on msg
		mov	ah, [edc_txn]				; Attr / Color
		mov	al, 8					; String length
		mov	bh, cl
		add	bh, 3					; Row #
		mov	bl, 7					; Col #
		call	Ed_XYAS
		popad
		
		; File name
		pushad
		mov	edx, ebp
		add	edx, binf_FileName			; Pointer on msg
		mov	ah, [edc_txn]				; Attr / Color
		mov	al, 53					; String length
		mov	bh, cl
		add	bh, 3					; Row #
		mov	bl, 15					; Col #
		call	Ed_XYAS
		popad
		
		; Size
		pushad
		cmp	[ebp.binf_InMemory], FALSE
		je	@@NotInMem
		mov	eax, [ebp.binf_CurrSize]		; The nb to draw
		jmp	@@PrSize
@@NotInMem:	mov	eax, [ebp.binf_CurrSize_S]		; The nb to draw
@@PrSize:	mov	bh, cl
		add	bh, 3					; Row #
		mov	bl, 70					; Column #
		mov	ch, 6					; Field size
		mov	cl, [edc_txn]				; Attr / Color
		call	EdPrNumber
		popad
		
		; Alt ?
		pushad
		cmp	[ebp.binf_NotSaved], 0
		jne	@@Alt
		mov	al, ' '					; Char to draw
		jmp	@@Next
@@Alt:		mov	al, '*'					; Char to draw
@@Next:		mov	ah, [edc_txn]				; Attr / Color
		mov	bh, cl
		add	bh, 3					; Row #
		mov	bl, 78					; Col #
		call	Ed_XYAC
		popad
		

		; -----------------
		; Print next line ?
		; -----------------

		inc	ecx
		mov	eax, ecx
		add	eax, edx
		cmp	eax, EDMAXNBBUFF
		je	@@Stop
		movzx	eax, [EdRowsNb]
		sub	eax, 7
		cmp	ecx, eax
		jne	@@DrawLine
@@Stop:


		; ----------------
		; Wait user action
		; ----------------
		
@@WaitKey:	mov	ax, [EdRowsNb]
		mov	ah, al
		sub	ah, 2
		mov	al, 79
		call	EdGotoXY
		call	CharIn
		
		cmp	ax, K_ESC
		je	@@Abort
		
		cmp	ax, K_PGDN
		jne	@@NotPgDn
		movzx	edx, [EdRowsNb]
		sub	edx, 7
		sub	edx, EDMAXNBBUFF
		neg	edx
		jmp	@@DrawPage
@@NotPgDn:

		cmp	ax, K_PGUP
		jne	@@WaitKey
		
		clr	edx
		jmp	@@DrawPage
		
		
		; -----------------
		; Install the frame
		; -----------------
		
@@Abort:		call	EdInstallFrame


		; ---------------
		; Redraw the page
		; ---------------

		; Redraw the text page
		pop	ebp
		call	EdRedrawPage
		
		; Redraw the status line
		call	EdPrStatusLine

		; Put screen cursor on right position
		call	EdPosCur
		
		
		; ----------------
		; Return to caller
		; ----------------
		
		ret

@@KeyNames:	db	"     F1 "
		db	"     F2 "
		db	"     F3 "
		db	"     F4 "
		db	"     F5 "
		db	"     F6 "
		db	"     F7 "
		db	"     F8 "
		db	"     F9 "
		db	"    F10 "
		db	"Shft F1 "
		db	"Shft F2 "
		db	"Shft F3 "
		db	"Shft F4 "
		db	"Shft F5 "
		db	"Shft F6 "
		db	"Shft F7 "
		db	"Shft F8 "
		db	"Shft F9 "
		db	"Shft F10"
		db	"Ctrl F1 "
		db	"Ctrl F2 "
		db	"Ctrl F3 "
		db	"Ctrl F4 "
		db	"Ctrl F5 "
		db	"Ctrl F6 "
		db	"Ctrl F7 "
		db	"Ctrl F8 "
		db	"Ctrl F9 "
		db	"Ctrl F10"
		db	" Alt F1 "
		db	" Alt F2 "
		db	" Alt F3 "
		db	" Alt F4 "
		db	" Alt F5 "
		db	" Alt F6 "
		db	" Alt F7 "
		db	" Alt F8 "
		db	" Alt F9 "
		db	" Alt F10"

EdBuffSelection	endp


; ===============================================================
; = EdPrFindOptions						=
; =								=
; = Called by 'EdFindPrompt':					=
; = 	Print find options					=
; =								=
; = I)	CF = 1 if clear line, else 0				=
; =								=
; = O)	none							=
; =								=
; = M)	none.							=
; ===============================================================

		align	4
EdPrFindOptions	proc

		; ------------------
		; Save all registers
		; ------------------
	
		pushad


		; ------------------------
		; Save where is the cursor
		; ------------------------

		pushf		
		call	EdWhereXY
		popf


		; --------
		; Show msg
		; --------

		jnc	@@ClrDone		
		lea	edx, EdMsgFindOpt
		call	EdPrMsgMsg
@@ClrDone:	push	ax


		; -------------------------
		; Show 'Word search' option
		; -------------------------
		
		mov	bl, 16					; Col #
		cmp	[EdFOpt_Word], 0
		call	PrMsg


		; ----------------------------
		; Show 'Case sensitive' option
		; ----------------------------

		mov	bl, 41		
		cmp	[EdFOpt_Case], 0
		call	PrMsg
		

		; ----------------------------
		; Show 'On all buffers' option
		; ----------------------------

		mov	bl, 66
		cmp	[EdFOpt_AllBuff], 0
		call	PrMsg
		

		; ---------------------------
		; Restore where is the cursor
		; ---------------------------

		pop	ax		
		call	EdGotoXY


		; ---------------------
		; Restore all registers
		; ---------------------
	
		popad


		; ----------------
		; Return to caller
		; ----------------
		
		ret
		

		; ---------------------------
		; Print the message ON or OFF
		; ---------------------------
		
		align	4
PrMsg		proc

		jne	@@On
@@Off:		lea	edx, EdMsgOff
		jmp	@@Next
@@On:		lea	edx, EdMsgOn
@@Next:

		mov	ah, [edc_sli]				; Attr / Color
		mov	al, 3					; Msg length
		mov	cx, [EdMsgRowNb]
		mov	bh, cl					; Row #
								; Col # , in BL
		call	Ed_XYAS

		ret

PrMsg		endp


EdPrFindOptions	endp


; ===============================================================
; = EdFindPrompt						=
; =								=
; = Usage:							=
; =	Print a prompt msg, then ask for the parameter.		=
; =								=
; = I)	EDX : points on msg to print.				=
;	BL  : the max length for the text to input		=
; =								=
; = O)	CF = 1 if operation is aborted				=
; =	CF = 0 if operation is not aborted (CR)			=
; =	EDX  points on the answer.				=
; =	BH   is the length of the answer.			=
; =								=
; = M)	FLAGS							=
; =	All register.						=
; =								=
; = Note:							=
; =	The string pointed by EDX is endded with '\0'.		=
; ===============================================================

		align	4
EdFindPrompt	proc

		; ------------------------
		; Print msg on status line
		; ------------------------

		pushad
		call	EdPrMsgStatus
		popad
		
		
		; -------------------------
		; Print find options status
		; -------------------------

		pushad
		stc		
		call	EdPrFindOptions
		popad


		; ---------------
		; Read the answer
		; ---------------

		lea	edx, EdPromptFind			; Buffer to store the answer
		mov	ah, [edc_sll]				; Attr / Color
		clr	bh					; Nb of char read		

		; Get a new character
@@ReadChar:	call	CharIn

		; Ctrl-W : change 'Word search' option
		cmp	ax, K_CTRL_W
		jne	@@NotCtrlW
		xor	[EdFOpt_Word], 1			; Change state
		clc
		call	EdPrFindOptions
		jmp	@@ReadChar
@@NotCtrlW:

		; Ctrl-C : change 'Case sensitive' option
		cmp	ax, K_CTRL_C
		jne	@@NotCtrlC
		xor	[EdFOpt_Case], 1			; Change state
		clc
		call	EdPrFindOptions
		jmp	@@ReadChar
@@NotCtrlC:

		; Ctrl-A : change 'On all buffers' option
		cmp	ax, K_CTRL_A
		jne	@@NotCtrlA
		xor	[EdFOpt_AllBuff], 1			; Change state
		clc
		call	EdPrFindOptions
		jmp	@@ReadChar
@@NotCtrlA:

		cmp	ax, K_ESC
		je	@@Abort
		cmp	ax, K_CR
		jne	@@Rd2
		
		; CR, ok ?
		cmpz	bh, 0
		je	@@ReadChar				; Not accept CR if answer is empty...
		jmp	@@AnswerOk
		
@@Rd2:		cmp	ax, K_BS
		jne	@@NotBS


		; BS
		; --

		cmpz	bh, 0
		je	@@ReadChar				; No char !
		
		; Cursor left
		push	ax
		call	EdWhereXY
		dec	al
		call	EdGotoXY
		
		; Print a space
		mov	al, ' '
		call	Ed_C
		pop	ax
		
		; Update nb of char in the answer
		dec	bh

		; Store the character into the buffer
		dec	edx

		; Continue
		jmp	@@ReadChar
@@NotBS:		
		

		; Test char ok ?
		; --------------
		
		cmp	ax, 020h
		jb	@@ReadChar				; Control character not allowed
		cmp	ax, 07Fh
		jae	@@ReadChar				; Graphic character not allowed


@@RdTake:	
		
		; Max length reached ?
		cmp	bh, bl
		je	@@ReadChar
		
		; Char is OK
		call	Ed_C					; Print char

		; Update cursor position
		push	ax
		call	EdWhereXY
		inc	al
		call	EdGotoXY
		pop	ax
		
		; Update nb of char in the answer
		inc	bh

		; Store the character into the buffer
		mov	[edx], al
		inc	edx

		; Do again
		jmp	@@ReadChar
		

		; ---------------
		; Abort operation
		; ---------------
		
@@Abort:

		; Return with CF = 1
		stc
		jmp	@@Exit


		; ------------
		; Answer is OK
		; ------------
		
@@AnswerOk:

		; Return answer address
		mov	byte ptr [edx], 0				; Put '\0' at the end of the string
		lea	edx, EdPromptFind				; Buffer to store the answer

		; Return with CF = 0
		clc
		jmp	@@Exit


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pushad
		pushfd
		
		; -----------------
		; Erase status line
		; -----------------
		
		call	EdClrStatusLine


		popfd
		popad
		ret

EdFindPrompt	endp


; ===============================================================
; = _SearchText							=
; =								=
; = Called by 'EdFind'						=
; = 	Find promped text into a given line.			=
; =								=
; = I)	ESI : ptr to begin of line.				=
; =								=
; = O)	CF  : 1 if text found, else 0				=
; =      EDX : (if CF=1) relative position of found text.	=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
_SearchText	proc

		; --------------------
		; Saved used registers
		; --------------------
		
		push	esi eax ecx
		
		
		; ---------
		; Find text
		; ---------

		; Test 1st character of source character
@@NextChar:	mov	al, [esi]
		cmp	al, 0Dh
		je	@@StrNotFound

		; Given string starting this position found ?
		clr	ecx
		dec	ecx
@@ParseString:	inc	ecx
		mov	ah, [EdPromptFind + ecx]
		cmpz	ah, 0
		je	@@StrIsFound
		mov	al, [esi + ecx]
		cmp	al, 0Dh
		je	@@ForwardNextChr
		cmp	[EdFOpt_Case], TRUE
		je	@@ParseStr2
		TOUPPER	al
		TOUPPER	ah
@@ParseStr2:	cmp	al, ah
		je	@@ParseString
		
		; Next character from the source string
@@ForwardNextChr:	inc	esi
		inc	edx
		jmp	@@NextChar


		; -------------------------------
		; Given string is found into line
		; -------------------------------
		
@@StrNotFound:	stc
		jmp	@@Exit
		

		; --------------------------------
		; Given string not found into line
		; --------------------------------

@@StrIsFound:	clc
		

		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	ecx eax esi
		ret

_SearchText	endp


; ===============================================================
; = EdFind							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Find occurence.						=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdFind		proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff


		; ---------------------
		; Read the text to find
		; ---------------------

		lea	edx, EdMsgFind				; Text to print
		mov	bl, MAXSZFINDTXT			; Max size of text to find / replace
		call	EdFindPrompt
		jc 	@@Exit					; Abort
		

		; -----------------------------
		; Search given text into buffer
		; -----------------------------

doEdFind:	
		
		; Address of top screen line (1st line drawed on top)
		mov	esi, [ebp.binf_CurLinAddr]		; Current line address (in the edited text)
		add	esi, [ebp.binf_tcol]			; Current col # in text
		mov	ecx, [ebp.binf_trow]
		
		; Relative position of the text (if found)
		mov	edx, [ebp.binf_tcol]
		
@@NextLine:
		; Search text into this line
		call	_SearchText
		jnc	@@TextFound

		; Find next line address
		call	EdFindNextLine
		jc	@@EndOfText				; End of text reached ?

		; Update position in text		
		inc	ecx
		
		; Relative position of the text (if found)
		clr	edx
		
		; Next line
		jmp	@@NextLine
		
@@TextFound:	
		; Move current line to buffer
		push	ecx edx
		mov	[ebp.binf_CurLinAddr], esi
		mov	ecx, esi
		call	EdGetCurLinBuff
		pop	edx ecx

		; Goto position found		
		mov	eax, ecx				; Line #
		mov	ebx, edx				; Colunn #
		call	EdDoGotoLine

@@EndOfText:	


		; ----------------
		; Return to caller
		; ----------------

@@Exit:
		
		; Redraw the entire status line
		call	EdPrStatusLine
		
		; Clear the message line
		call	EdClrMsgLine

		; Update the cursor position on screen
		call	EdPosCur

		; Return
		ret

EdFind		endp


; ===============================================================
; = EdReplace							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Find and replace occurence.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdReplace	proc

		call	EdNotAvailable

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdReplace	endp


; ===============================================================
; = EdLastFindRepl						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Find and replace occurence.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdLastFindRepl	proc

		; --------------------------
		; A previous command exist ?
		; --------------------------
		
		cmp	[EdPromptFind], 0			; Text to search
		je	EdFind


		; ----------------------
		; Do next search command
		; ----------------------

		call	EdMoveRight
		jmp	doEdFind		

EdLastFindRepl	endp


; ===============================================================
; = EdMarkLine							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Mark the current line.					=
; =								=
; = I)	EAX : mark nb (0, 1, 2 or 3)				=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMarkLine	proc

		; -------------------------
		; Mark the current position
		; -------------------------

		; Current row		
		mov	ebx, [ebp.binf_trow]
		mov	[ebp.binf_Tags_Row + eax], ebx

		; Current col
		mov	ebx, [ebp.binf_tcol]
		mov	[ebp.binf_Tags_Col + eax], ebx


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdMarkLine	endp


; ===============================================================
; = EdGtoMarkedLine						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Mark the current line.					=
; =								=
; = I)	EAX : mark nb (0, 1, 2 or 3)				=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdGtoMarkedLine	proc

		; ----------------------------
		; Do nothing if tag is not set
		; ----------------------------

		; Column #
		mov	ebx, [ebp.binf_Tags_Col + eax]
		
		; Line #		
		mov	eax, [ebp.binf_Tags_Row + eax]
		cmp	eax, -1
		je	@@Exit
		
		
		; ------------------------------------
		; Goto on this line (row EAX, col EBX)
		; ------------------------------------
		
		call	EdDoGotoLine


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdGtoMarkedLine	endp


code32		ends
		end
