; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ED_GOTOW.ASM						!
; 									!
; Description:	Definitions for playing with FROSTED, built-in editor.	!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Ed.inc					; Constants, Structurs, Variables used by FROSTED


; ---> Local variables to this module ---> ---------------------------------------------------------------------------------------

		align	4
ColumnsNb	dd	?				; Nb of columns
RowsNb		dd	?				; Nb of rows
MaxWordsInPage	dd	?				; Max nb of words in a page (= ColumnsNb * RowsNb)

TopRecNb	dd	?				; Top record nb

; <--- Local variables to this module <--- ---------------------------------------------------------------------------------------


; ===============================================================
; = SearchWordPtr						=
; =								=
; = Usage:							=
; = 	Search a word reference.				=
; =								=
; = I) ECX : NIA pointer to caller word				=
; =								=
; = O) CF = 1, ref is found, so address is in EDX		=
; =    CF = 0, ref not found					=
; =								=
; = M) CF, EBX							=
; ===============================================================

		align	4
SearchWordPtr	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx esi edi
		
		
		; ---------------
		; Initializations
		; ---------------
		
		; Left limit
		clr	esi
		
		; Right limit
		mov	edi, [EdWordsNb]

		
		; ------------------
		; Search (dichotomy)
		; ------------------

		; Stop the search ?
@@DichoSearch:	cmp	esi, edi
		ja	@@NotFound
		cmp	edi, -1
		je	@@NotFound
		
		; Compute middle point
		mov	ebx, edi
		add	ebx, esi
		shr	ebx, 1					; Middle = (Left + Right) / 2
		
		cmp	ebx, [EdWordsNb]
		je	@@NotFound

		; Compare this point with the asked record
		cmp	[EdWordsTable + ebx * 4], ecx
		ja	@@GoLeft
		jb	@@GoRight		
		
		; Found
		mov	esi, ebx
		jmp	@@Found
		
		; Go left : search between [left, middle - 1]
@@GoLeft:	mov	edi, ebx
		dec	edi
		jmp	@@DichoSearch

		; Go right : search between [middle + 1, right]
@@GoRight:	mov	esi, ebx
		inc	esi
		jmp	@@DichoSearch
		
		
		; ----------------
		; Return to caller
		; ----------------
	
@@NotFound:	clc						; CF = 0
		jmp	@@Exit

@@Found:	stc						; CF = 1
		
@@Exit:		mov	edx, esi				; Output value
		pop	edi esi ebx eax
		ret

SearchWordPtr	endp


; ===============================================================
; = AddWordPtr							=
; =								=
; = Called by 'BuildWordsTable' : 				=
; = 	Add a new word pointer, if not already recorded.	=
; =								=
; = I) EBX : pointer to the NIA word.				=
; = O)	none							=
; = M)	none							=
; ===============================================================

		align	4
AddWordPtr	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ecx edx


		; -----------------
		; If table is empty
		; -----------------

		; Is table empty ?		
		cmp	[EdWordsNb], 0
		jne	@@NotEmpty
		
		; Create 1st index
		lea	edx, EdWordsTable
		mov	[edx], ebx
		
		; Update the nb of index
		mov	[EdWordsNb], 1

		; Done
		jmp	@@Done
		
@@NotEmpty:


		; ---------------------------
		; Already recorded in table ?
		; ---------------------------

		; in ECX, pointer to word
		mov	ecx, ebx
		call	SearchWordPtr
		jc	@@RecExists
		

		; --------------------------------
		; Record does not exist, create it
		; --------------------------------

		cmp	edx, [EdWordsNb]
		je	@@NotMoveNdx1
		push	ecx esi edi
		mov	ecx, [EdWordsNb]
		mov	edi, ecx
		shl	edi, 2
		add	edi, offset EdWordsTable		; Destination address
		mov	esi, edi
		sub	esi, 4					; Source address
		sub	ecx, edx
		std						; Dec ESI and EDI
		rep	movsd  		          	    	; Move the block
		cld
		pop	edi esi ecx
@@NotMoveNdx1:	mov	[EdWordsTable + edx * 4], ebx

		; Update the nb of records
		inc	[EdWordsNb]

		; Done
		jmp	@@Done		
		

		; ----------------------------------
		; Record already defined: do nothing
		; ----------------------------------
		
@@RecExists:	



		; ----------------
		; Return to caller
		; ----------------
		
@@Done:		pop	edx ecx eax
		ret

AddWordPtr	endp


; ===============================================================
; = BuildWordsTable						=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	Build the table of pointers on eacxh user word.		=
; =								=
; = I)	none							=
; = O)	none							=
; = M)	all regs.						=
; ===============================================================

		align	4
BuildWordsTable	proc

		; ---------------
		; Initializations
		; ---------------
		
		; Nb of user words
		mov	[EdWordsNb], 0

		; Pointer to cross ref table
		mov	edx, [CrossRef]
		
		; Ref counter
		clr	ecx


		; ------------------------
		; Scan each word reference
		; ------------------------

@@TopLoop:

		; All ref scanned ?
		cmp	ecx, [CrossRefTotalNb]
		je	@@Done

		; Record deleted ?
		cmp	[edx.wref_calls], -1
		je	@@LoopNext

		; Add the new word pointer (if not already recorded)
		mov	ebx, [edx.wref_caller]
		mov	eax, ebx
		add	eax, [NIA_START]
		cmp	[eax.word_LineNb], -1			; Infos 'line #' not available for this word ?
		je	@@NotAdd1
		call	AddWordPtr
@@NotAdd1:

		; Add the new word pointer (if not already recorded)
		mov	ebx, [edx.wref_called]
		mov	eax, ebx
		add	eax, [NIA_START]
		cmp	[eax.word_LineNb], -1			; Infos 'line #' not available for this word ?
		je	@@NotAdd2
		call	AddWordPtr
@@NotAdd2:

		; Next ref
@@LoopNext:	add	edx, size wrefentry
		inc	ecx
		jmp	@@TopLoop


		; ----------------
		; Return to caller
		; ----------------
		
@@Done:		ret

BuildWordsTable	endp


; ===============================================================
; = EdLineNbNotAvail						=
; =								=
; = Called by 'EdLoadFile'					=
; =	Line # informations is no longer available for the 	=
; =	words recorded into the current file.			=
; =								=
; = I)	none							=
; = O)	none							=
; = M)	all regs.						=
; ===============================================================

		align	4
EdLineNbNotAvail	proc

		; -------------------------------------------------------------------------------
		; Scan each word recorded to see if they belong to the current file (just loaded)
		; -------------------------------------------------------------------------------
		
		; Words counter
		clr	edx
		
		; Top loop
		; --------
		
@@TopLoop:	
		; End of table ?
		cmp	edx, [EdWordsNb]
		je	@@Done

		; Get word NIA address
		mov	esi, [EdWordsTable + edx * 4]
		add	esi, [NIA_START]
		
		; Does this word belong to the cuurent file ?
		clr	ecx					; Char counter
		dec	ecx
@@CompFileName:	inc	ecx
		mov	al, [esi.word_FileName + ecx]
		mov	ah, [ebp.binf_PrimFName + ecx]
		cmp	al, ah
		jne	@@NextWord
		or	al, ah
		jnz	@@CompFileName
		
		; Yes, discard the line # information
		mov	[esi.word_LineNb], -1
		jmp	@@NextWord
		
		; Goto next word, if any
@@NextWord:	inc	edx
		jmp	@@TopLoop

@@Done:
		
		
		; -----------------------------------------
		; Rebuild the table of pointer of all words
		; -----------------------------------------
		
		call	BuildWordsTable


		; ----------------
		; Return to caller
		; ----------------
		
		ret

EdLineNbNotAvail	endp


; ===============================================================
; = CompWordPos							=
; =								=
; = Called by 'DrwWords' : 					=
; = 	Compute the position (row, col) of a word.		=
; =								=
; = I)	ECX : sequential nb.					=
; =								=
; = O)	AH  : row						=
; =	AL  : col						=
; =								=
; = M)	none.							=
; ===============================================================

		align	4
CompWordPos	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	ecx edx
		
		
		; ---------------
		; Initializations
		; ---------------
		
		; Column counter
		clr	edx
		

		; ---------
		; Main loop
		; ---------
		
@@ColLoop:	mov	eax, [RowsNb]
		imul	eax, edx
		add	eax, [RowsNb]
		cmp	ecx, eax
		jge	@@NextCol
		
		; We are in the column EDX
		sub	ecx, eax
		add	ecx, [RowsNb]
		mov	eax, WORDMAXLEN + 1
		imul	eax, edx				; in AL: col
		mov	ah, cl					
		add	ah, 2					; in AH: row
		jmp	@@Exit
		
		; Try next column
@@NextCol:	inc	edx
		jmp	@@ColLoop


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	edx ecx
		ret

CompWordPos	endp


; ===============================================================
; = PrintWordName						=
; =								=
; = Usage:							=
; = 	Print a word name.					=
; =								=
; = I)  EDX : pointer to word (in 'WordsTable').		=
; =     ECX : sequential nb.					=
; =     CF  : 0 if normal video, 1 if inverse video.		= 
; =								=
; = O)	none							=
; =								=
; = M)	none.							=
; ===============================================================

		align	4
PrintWordName	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx edx
		pushf


		; -----------------------------
		; Move cursor to right position
		; -----------------------------
		
		call	CompWordPos				; O) ah = #row  al = #col
		call	EdGotoXY
		push	ax


		; ----------------------
		; Pointer to word record
		; ----------------------
		
		mov	edx, [EdWordsTable + edx * 4]
		add	edx, [NIA_START]
		add	edx, word_name
		
		
		; -----------------------------
		; Copy word name to temp buffer
		; -----------------------------
		
		; Initialize with blanks
		lea	ebx, @@WordName
		clr	eax
@@Zero:		mov	byte ptr [ebx + eax], ' '
		inc	eax
		cmp	al, WORDMAXLEN
		jne	@@Zero
		mov	byte ptr [ebx + eax], 0
		
		; Copy now the word name
		clr	ecx
@@Copy:		mov	al, [edx + ecx]
		cmpz	al, 0
		je	@@Copy2
		mov	[ebx + ecx], al
		inc	ecx
		jmp	@@Copy
@@Copy2:


		; --------------
		; Print the word
		; --------------

		; Position
		pop	bx					; BH = #row   BL = #col

		; Color / Attribute
		popf
		jc	@@Hight
		mov	ah, [edc_txn]
		jmp	@@HiLo
@@Hight:		mov	ah, [edc_txs]
@@HiLo:		

		; Nb of chars to print
		mov	al, WORDMAXLEN

		; String to print		
		lea	edx, @@WordName

		; Print now
		call	Ed_XYAS


		; ----------------
		; Return to caller
		; ----------------

		pop	edx ecx ebx eax		
		ret

		align	4
@@WordName	db	WORDMAXLEN + 1 dup (?)

PrintWordName	endp


; ===============================================================
; = PrintBlankName						=
; =								=
; = Usage:							=
; = 	Print a blank word name.				=
; =								=
; = I)  ECX : sequential nb.					=
; = O)  none							=
; = M)  none.							=
; ===============================================================

		align	4
PrintBlankName	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx edx

		
		; -----------------------------
		; Move cursor to right position
		; -----------------------------
		
		call	CompWordPos				; O) ah = #row  al = #col
		call	EdGotoXY


		; --------------
		; Print the word
		; --------------

		; Position
		mov	bx, ax					; BH = #row   BL = #col

		; Color / Attribute
		mov	ah, [edc_txn]

		; Nb of chars to print
		mov	al, WORDMAXLEN

		; String to print		
		lea	edx, @@WordName

		; Print now
		call	Ed_XYAS


		; ----------------
		; Return to caller
		; ----------------

		pop	edx ecx ebx eax		
		ret

		align	4
@@WordName	db	WORDMAXLEN dup (' ')
		db	0

PrintBlankName	endp


; ===============================================================
; = DrawWords							=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	List a page of user defined words.			=
; =								=
; = I)	none							=
; = O)	none							=
; = M)	none.							=
; ===============================================================

		align	4
DrawWords	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx edx


		; --------------------
		; Some initializations
		; --------------------
		
		; Counter of drawn words
		clr	ecx
		

		; --------------------
		; Main loop of drawing
		; --------------------

@@TopLoop:	

		; Page complete ?
		cmp	ecx, [MaxWordsInPage]
		je	@@Exit

		; End of words reached ?
		mov	eax, ecx
		add	eax, [TopRecNb]
		cmp	eax, [EdWordsNb]
		je	@@EndList

ifdef NEVER		
		; Move cursor to right position
		call	CompWordPos
		call	EdGotoXY
endif

		; Print the word name
		mov	edx, ecx
		add	edx, [TopRecNb]
		clc						; Normal video
		call	PrintWordName
		
		; Next word
		inc	ecx
		jmp	@@TopLoop


		; ------------------------------------------
		; End of list, complete the page with blanks
		; ------------------------------------------
		
@@EndList:

		; Page complete ?
		cmp	ecx, [MaxWordsInPage]
		je	@@Exit

ifdef NEVER		
		; Move cursor to right position
		call	CompWordPos
		call	EdGotoXY
endif

		; Print the word name
		call	PrintBlankName
		
		; Next word
		inc	ecx
		jmp	@@EndList


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	edx ecx ebx eax
		ret

DrawWords	endp


; ===============================================================
; = gotoNextWord						=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	Goto on next word, if possible.				=
; =								=
; = I)	EDX : current word (from 0).				=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
gotoNextWord	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx


		; ------------------------------
		; On the last word of the list ?
		; ------------------------------
		
		mov	eax, edx
		inc	eax
		add	eax, [TopRecNb]
		cmp	eax, [EdWordsNb]
		je	@@Exit


		; ------------------
		; On the same page ?
		; ------------------
		
		mov	eax, edx
		inc	eax
		cmp	eax, [MaxWordsInPage]
		je	@@NextPage

		; Skip to next word
		inc	edx
		jmp	@@Exit
		
@@NextPage:


		; --------------
		; Show next page
		; --------------
		
		; Next page
		mov	eax, [TopRecNb]
		add	eax, [MaxWordsInPage]
		mov	[TopRecNb], eax

		; Goto on the top of the page
		clr	edx

		; Draw a page of words
		call	DrawWords


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ecx ebx eax
		ret

gotoNextWord	endp


; ===============================================================
; = gotoPrevWord						=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	Goto on previous word, if possible.			=
; =								=
; = I)	EDX : current word (from 0).				=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
gotoPrevWord	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx


		; ------------
		; Not on top ?
		; ------------
		
		cmpz	ecx, 0
		je	@@Top

		; Skip to prev word
		dec	edx
		jmp	@@Exit
		
@@Top:

		
		; ----------------
		; On top of list ?
		; ----------------
		
		cmp	[TopRecNb], 0
		je	@@Exit
		
		
		; ------------------
		; Show previous page
		; ------------------
		
		; Prev page
		mov	eax, [TopRecNb]
		sub	eax, [MaxWordsInPage]
		cmpz	eax, 0
		jg	@@Ok
		clr	eax
@@Ok:		mov	[TopRecNb], 0

		; Goto on the bottom of the page
		mov	edx, [MaxWordsInPage]
		dec	edx

		; Draw a page of words
		call	DrawWords


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ecx ebx eax
		ret

gotoPrevWord	endp


; ===============================================================
; = gotoRightWord						=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	Goto on Right word, if possible.			=
; =								=
; = I)	EDX : current word (from 0).				=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
gotoRightWord	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx


		; ---------------------
		; End of list reached ?
		; ---------------------
		
		mov	eax, edx
		add	eax, [TopRecNb]
		add	eax, [RowsNb]
		cmp	eax, [EdWordsNb]
		jge	@@Exit

	
		; -------------
		; Next 'page' ?
		; -------------

		mov	eax, edx
		add	eax, [RowsNb]
		cmp	eax, [MaxWordsInPage]
		jge	@@NextPage

		; Goto right col		
		add	edx, [RowsNb]
		jmp	@@Exit
		
@@NextPage:

	
		; ----------------
		; Show next 'page'
		; ----------------
		
		; Next page
		mov	eax, [TopRecNb]
		add	eax, [RowsNb]
		mov	[TopRecNb], eax

		; Draw a page of words
		call	DrawWords


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ecx ebx eax
		ret

gotoRightWord	endp


; ===============================================================
; = gotoLeftWord						=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	Goto on Left word, if possible.				=
; =								=
; = I)	EDX : current word (from 0).				=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
gotoLeftWord	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx


		; -----------------
		; Previous 'page' ?
		; -----------------
		
		mov	eax, edx
		sub	eax, [RowsNb]
		cmpz	eax, 0
		jl	@@PrevPage

		; Prev column
		sub	edx, [RowsNb]
		jmp	@@Exit
		
@@PrevPage:


		; --------------------
		; Show previous 'page'
		; --------------------
		
		; No prev page ?
		mov	eax, [TopRecNb]
		cmp	eax, [RowsNb]
		jl	@@Exit

		; Prev page
		mov	eax, [TopRecNb]
		sub	eax, [RowsNb]
		mov	[TopRecNb], eax

		; Draw a page of words
		call	DrawWords


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ecx ebx eax
		ret

gotoLeftWord	endp


; ===============================================================
; = gotoTopList							=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	Goto top list.						=
; =								=
; = I)	none.							=
; =								=
; = O)	EDX : current position in the page.			=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
gotoTopList	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx


		; -------------
		; Goto top list
		; -------------
		
		; 1st page
		mov	[TopRecNb], 0

		; Update position in the page
		clr	edx

		; Draw a page of words
		call	DrawWords


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ecx ebx eax
		ret

gotoTopList	endp


; ===============================================================
; = gotoBottomList						=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	Goto bottom list.					=
; =								=
; = I)	none.							=
; =								=
; = O)	EDX : current position in the page.			=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
gotoBottomList	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx


		; ----------------
		; Goto bottom list
		; ----------------
		
		; Last page
		mov	eax, [EdWordsNb]
		sub	eax, [MaxWordsInPage]
		cmpz	eax, 0
		jge	@@Ok
		clr	eax
@@Ok:		mov	[TopRecNb], eax

		; Update position in the page
		mov	edx, [MaxWordsInPage]
		dec	edx
		mov	eax, edx
		add	eax, [TopRecNb]
		cmp	eax, [EdWordsNb]
		jl	@@Ok2
		mov	edx, [EdWordsNb]
		sub	edx, [TopRecNb]
		dec	edx
@@Ok2:		

		; Draw a page of words
		call	DrawWords


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ecx ebx eax
		ret

gotoBottomList	endp


; ===============================================================
; = gotoTopPage							=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	Goto top page.						=
; =								=
; = I)	none.							=
; =								=
; = O)	EDX : current position in the page.			=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
gotoTopPage	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx


		; -------------
		; Goto top page
		; -------------
		
		; Update position in the page
		clr	edx


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ecx ebx eax
		ret

gotoTopPage	endp


; ===============================================================
; = gotoBottomPage						=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	Goto bottom page.					=
; =								=
; = I)	none.							=
; =								=
; = O)	EDX : current position in the page.			=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
gotoBottomPage	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx


		; ----------------
		; Goto bottom page
		; ----------------
		
		; Update position in the page
		mov	edx, [MaxWordsInPage]
		dec	edx
		mov	eax, edx
		add	eax, [TopRecNb]
		cmp	eax, [EdWordsNb]
		jl	@@Ok
		mov	edx, [EdWordsNb]
		sub	edx, [TopRecNb]
		dec	edx
@@Ok:		


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ecx ebx eax
		ret

gotoBottomPage	endp


; ===============================================================
; = gotoNextPage						=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	Goto next page.						=
; =								=
; = I)	none.							=
; =								=
; = O)	EDX : current position in the page.			=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
gotoNextPage	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx


		; ----------------------
		; Goto next page, if any
		; ----------------------

		; End of list reached ?
		mov	eax, [TopRecNb]
		add	eax, [MaxWordsInPage]
		cmp	eax, [EdWordsNb]
		jge	@@Exit
		
		; Next page, if any
		mov	eax, [TopRecNb]
		add	eax, [MaxWordsInPage]
		mov	[TopRecNb], eax

		; Update position in the page
		clr	edx

		; Draw a page of words
		call	DrawWords


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ecx ebx eax
		ret

gotoNextPage	endp


; ===============================================================
; = gotoPrevPage						=
; =								=
; = Called by 'EdGotoWord' : 					=
; = 	Goto previous page.					=
; =								=
; = I)	none.							=
; =								=
; = O)	EDX : current position in the page.			=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
gotoPrevPage	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx


		; ------------------
		; Goto previous page
		; ------------------

		; Do nothing if top of list
		cmp	[TopRecNb], 0
		je	@@Exit
		
		; Previous page, if any
		mov	eax, [TopRecNb]
		sub	eax, [MaxWordsInPage]
		cmpz	eax, 0
		jge	@@Ok
		clr	eax
@@Ok:		mov	[TopRecNb], eax

		; Update position in the page
		clr	edx

		; Draw a page of words
		call	DrawWords


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ecx ebx eax
		ret

gotoPrevPage	endp


; ===============================================================
; = EdGotoWord							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	List all user defined words, then goto a word.		=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdGotoWord	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff
		

		; -------------
		; Install frame
		; -------------

		; Clear screen
		call	EdCls

		; Print the Status Line
		mov	ah, [edc_sll]				; Attr / Color
		mov	bx, [EdColsNb]				
		mov	al, bl					; Nb of chars to print (line width in AL)
		clr	bh					; Row #
		clr	bl					; Column #
		lea	edx, EdFrostedTittle			; The string to print
		call	Ed_XYAS

		; Print a user infos msg on msg line
		lea	edx, EdMsgGotoWord
		call	EdPrMsgMsg


		; --------------------
		; No reference words ?
		; --------------------
		
		cmp	[CrossRefTotalNb], 0
		jne	@@RefWordOk
		
		; Print a user infos msg on msg line
		lea	edx, EdMsgGotoWordNo			; "No word ref."
		call	EdPrMsgMsg
@@WaitEsc:	call	CharIn
		cmp	ax, EKEY_ESC
		jne	@@WaitEsc
		jmp	@@Abort

@@RefWordOk:

		; --------------------
		; Some initializations
		; --------------------

		; Build the table of pointer of all words
		cmp	[EdWordsTableOk], TRUE
		je	@@WordsTableOk
		call	BuildWordsTable
		mov	[EdWordsTableOk], TRUE
@@WordsTableOk:
		
		; 1st record shown (top record nb)
		mov	[TopRecNb], 0

		; Compute the nb of columns
		mov	ax, [EdColsNb]
		mov	bl, WORDMAXLEN + 1
		idiv	bl
		cbw
		cwde
		mov	[ColumnsNb], eax
		
		; Compute the nb of rows
		movzx	eax, [EdRowsNb]
		sub	eax, 4
		mov	[RowsNb], eax
		
		; Compute the max nb of words in a page
		imul	eax, [ColumnsNb]
		mov	[MaxWordsInPage], eax
		
		; Pointer to the highlighted word
		clr	edx
		

		; ---------
		; Main loop
		; ---------
		
		; Draw a page of words
		call	DrawWords

@@WaitKey:

		; Hightlight current word
ifdef NEVER
		mov	ecx, edx
		call	CompWordPos
		call	EdGotoXY
endif
		mov	ecx, edx
		push	edx
		add	edx, [TopRecNb]
		stc						; Reverse video
		call	PrintWordName
		pop	edx

		; Wait user action
		call	CharIn
		
		; ESC ? -> Abort operation
		cmp	ax, EKEY_ESC
		je	@@Abort

		; CR ? -> Search word
		cmp	ax, EKEY_CR
		je	@@SearchWord

		; Un-Hightlight current word
		push	eax
ifdef NEVER
		mov	ecx, edx
		call	CompWordPos
		call	EdGotoXY
endif
		mov	ecx, edx
		push	edx
		add	edx, [TopRecNb]
		clc						; Normal video
		call	PrintWordName
		pop	edx
		pop	eax

		; Next word ?
		cmp	ax, EKEY_DOWN
		je	@@Down
		cmp	ax, ' '
		je	@@Down
		cmp	ax, EKEY_DOWN + 80h
		jne	@@NotDown
@@Down:		call	gotoNextWord
		jmp	@@WaitKey
@@NotDown:
		
		; Previous word ?
		cmp	ax, EKEY_UP
		je	@@Up
		cmp	ax, EKEY_UP + 80h
		jne	@@NotUp
@@Up:		call	gotoPrevWord
		jmp	@@WaitKey
@@NotUp:

		; Right word ?
		cmp	ax, EKEY_RIGHT
		je	@@Right
		cmp	ax, EKEY_RIGHT + 80h
		jne	@@NotRight
@@Right:	call	gotoRightWord
		jmp	@@WaitKey
@@NotRight:

		; Left word ?
		cmp	ax, EKEY_LEFT
		je	@@Left
		cmp	ax, EKEY_LEFT + 80h
		jne	@@NotLeft
@@Left:		call	gotoLeftWord
		jmp	@@WaitKey
@@NotLeft:

		; Top of list ?
		cmp	ax, EKEY_CTRL_HOME
		je	@@CtrlHome
		cmp	ax, EKEY_CTRL_HOME + 80h
		jne	@@NoCtrlHome
@@CtrlHome:	call	gotoTopList
		jmp	@@WaitKey
@@NoCtrlHome:
		
		; Bottom of list ?
		cmp	ax, EKEY_CTRL_END
		je	@@CtrlEnd
		cmp	ax, EKEY_CTRL_END + 80h
		jne	@@NoCtrlEnd
@@CtrlEnd:	call	gotoBottomList
		jmp	@@WaitKey
@@NoCtrlEnd:
		
		; Top of page ?
		cmp	ax, EKEY_HOME
		je	@@Home
		cmp	ax, EKEY_HOME + 80h
		jne	@@NoHome
@@Home:		call	gotoTopPage
		jmp	@@WaitKey
@@NoHome:
		
		; Bottom of page ?
		cmp	ax, EKEY_END
		je	@@End
		cmp	ax, EKEY_END + 80h
		jne	@@NoEnd
@@End:		call	gotoBottomPage
		jmp	@@WaitKey
@@NoEnd:

		; Next page ?
		cmp	ax, EKEY_PGDN
		je	@@PgDn
		cmp	ax, EKEY_PGDN + 80h
		jne	@@NoPgDn
@@PgDn:		call	gotoNextPage
		jmp	@@WaitKey
@@NoPgDn:

		; Previous page ?
		cmp	ax, EKEY_PGUP
		je	@@PgUp
		cmp	ax, EKEY_PGUP + 80h
		jne	@@NoPgUp
@@PgUp:		call	gotoPrevPage
		jmp	@@WaitKey
@@NoPgUp:
		
		; Unknow action, readk key
		jmp	@@WaitKey


		; ---------------------
		; Search the given word
		; ---------------------
		
@@SearchWord:	

		; Pointer to NIA word
		add	edx, [TopRecNb]
		mov	edx, [EdWordsTable + edx * 4]

		; Search <filename> into all buffers
		add	edx, [NIA_START]
		clr	ecx					; Buffers counter
@@SrchFNBuff:	imul	ebx, ecx, size buffedentry
		add	ebx, offset EdBuffInfos
		cmp	byte ptr [ebx.binf_PrimFName], 0
		je	@@SrchFNBuff2

		push	ecx		
		clr	ecx
		dec	ecx
@@SrchFNBuff3:	inc	ecx
		mov	al, [ebx.binf_PrimFName + ecx]
		TOUPPER	al
		mov	ah, [edx.word_FileName + ecx]
		TOUPPER	ah
		cmp	al, ah
		jne	@@SrchFNBuff4
		or	al, ah
		jz	@@Found
		jmp	@@SrchFNBuff3
@@SrchFNBuff4:	pop	ecx

@@SrchFNBuff2:	inc	ecx
		cmp	cl, EDMAXNBBUFF
		jne	@@SrchFNBuff
		jmp	@@NotFound


		; -----------------------------------
		; The filename is found into a buffer
		; -----------------------------------
		
@@Found:	pop	ecx

		; Select buffer #ECX
		; Update EBP to point on new buffer informations      (Remember: EBP points always on current buffer informations)
		mov	ebp, ecx
		mov	[EdCurrBufferNb], ebp
		imul	ebp, size buffedentry
		add	ebp, offset EdBuffInfos

		; Re-install the editor frame
		push	edx
		call	EdInstallFrame
		pop	edx

		; Goto the given line
		mov	eax, [edx.word_LineNb]
		clr	ebx					; 1st column
		call	EdDoGotoLine

		; Done
		jmp	@@Exit
		

		; ---------------------------------------------
		; The filename is not found into an open buffer
		; ---------------------------------------------
		
@@NotFound:

		; Search a free buffer
		; --------------------
		
		clr	ecx
		dec	ecx
		lea	ebx, EdBuffInfos - size buffedentry
@@SrchFreeBuff:	inc	ecx
		add	ebx, size buffedentry
		cmp	byte ptr [ebx.binf_PrimFName], 0
		je	@@FreeBuffFound
		cmp	cl, EDMAXNBBUFF
		jne	@@SrchFreeBuff
		jmp	@@Abort						; All buffers in use !!!
@@FreeBuffFound:

		; Make this buffer the current buffer
		; -----------------------------------

		; Select buffer #ECX
		; Update EBP to point on new buffer informations      (Remember: EBP points always on current buffer infos)
		mov	[EdCurrBufferNb], ecx
		mov	ebp, ebx

		; Re-install the editor frame
		push	edx
		call	EdInstallFrame
		pop	edx

		; Copy the filename to temp buffer
		lea	esi, @@FileName			; Local buffer
		clr	ecx				; Char counter
		dec	ecx
@@CopyFileName:	inc	ecx
		mov	al, [edx.word_FileName + ecx]
		TOUPPER	al
		mov	[esi + ecx], al
		cmpz	al, 0
		jne	@@CopyFileName

		; Read the file into this buffer
		lea	eax, @@FileName
		call	EdLoadFile

		; Goto the given line
		mov	eax, [edx.word_LineNb]
		clr	ebx					; 1st column
		call	EdDoGotoLine

		; Done
		jmp	@@Exit
		

		
		; ---------------------------
		; Re-install the editor frame
		; ---------------------------
		
@@Abort:	call	EdInstallFrame


		; ---------------
		; Redraw the page
		; ---------------

		; Redraw the text page
		call	EdRedrawPage
		
		; Put screen cursor on right position
		call	EdPosCur
		
		
		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

		align	4
@@FileName	db	MAXSZFILENAME+1 dup (?)			; Local buffer to store modified filename

EdGotoWord	endp


code32		ends
		end
