; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	FROTH.ASM						!
; 									!
; Description:	Main module.						!
; 									!
; List of FROTH definitions defined in this module:			!
;	Quit		QUIT						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

		; Address of own code: this is the begin of moving (-> _EndAddr)
_StartAddr:					

__MAIN__	equ 1
include		equ.inc
include		vars.inc

include		dict.inc

include		ed.inc					; Definition about FROSTED, the integrated editor

public		_main

		; Address of own code: this is the begin of moving (-> _EndAddr)
_StartAddr2:					

; Defined in PMODE.ASM
extrn	_Excp_1:		dword


; ---> Local variables, *static* to this module ---> -----------------------------------------------------------------------------

		align	4
KbItActive	dd	FALSE					; Flag FALSE or TRUE: is the kbd handler installed ?

		align	4
KbdFileName	db	DEFKBD_FILENAME
		align	4
COLOR		sll,	LIGHTGRAY,	BLACK
		align	4
COLOR		sli,	LIGHTGRAY,	RED
		align	4
COLOR		txn,	BLUE,		WHITE			; Normal text
		align	4
COLOR		txs,	LIGHTGRAY,	BLUE			; Selected text
		align	4
COLOR		hln,	LIGHTGRAY,	BLACK			; Help, normal
		align	4
COLOR		hla,	LIGHTGRAY,	RED			; Help, alt
		align	4
COLOR		hlk,	LIGHTGRAY,	YELLOW			; Help, keys
		
		align	4
StartupFileName	db	STARTUP_FILENAME

		align	4
Argv		db	132 dup (?)				; String argument after switch (on cmd line)
	
; <--- end of local variables <--- -----------------------------------------------------------------------------------------------


_main:

		; --------------------------
		; Enable hardware interrupts
		; --------------------------

		sti


		; ---------------------------------------------
		; Analyse arguments on the command line, if any
		; ---------------------------------------------

		; Not available if running a turnkey exe
		cmp	[ExeFile], TRUE
		je	@@NoArg

		; Switch '?' 
		mov	al, '?'
		lea	edx, Argv
		call	GetArgc
		jc	@@NoSwitch_Q
@@PrArgMsg:	lea	edx, ArgMsg
@@PrArgMsg2:	mov	al, [edx]
		cmpz	al, 0
		je	_exit
		call	COUT
		inc	edx
		jmp	@@PrArgMsg2
@@NoSwitch_Q:

		; Switch 'e'
		mov	[EdTotalBuffSize], DEFSZ_EDBUFF		; Default size (if not given on cmd line)
		mov	al, 'e'
		lea	edx, Argv
		call	GetArgc
		jc	@@NoSwitch_e
		call	EdNumber				; I) EDX = stringz ptr    O) CF=1 if error, EAX=nb
		jc	@@PrArgMsg
		imul	eax, 1024				; Convert to nb of bytes
		mov	[EdTotalBuffSize], eax
@@NoSwitch_e:

		; Switch 's'
		mov	[EdTotalScrapSize], DEFSZ_SCRAPBUFF	; Default size (if not given on cmd line)
		mov	al, 's'
		lea	edx, Argv
		call	GetArgc
		jc	@@NoSwitch_s
		call	EdNumber				; I) EDX = stringz ptr    O) CF=1 if error, EAX=nb
		jc	@@PrArgMsg
		imul	eax, 1024				; Convert to nb of bytes
		mov	[EdTotalScrapSize], eax
@@NoSwitch_s:

		; Switch 't'
		mov	[TraceBufferMaxSz], DEFSZ_TRACEBUFFER	; Default size (if not given on cmd line)
		mov	al, 't'
		lea	edx, Argv
		call	GetArgc
		jc	@@NoSwitch_t
		call	EdNumber				; I) EDX = stringz ptr    O) CF=1 if error, EAX=nb
		jc	@@PrArgMsg
		imul	eax, 1024				; Convert to nb of bytes
		mov	[TraceBufferMaxSz], eax
@@NoSwitch_t:

@@NoArg:		


		; -------------------------------------------------------------------------------------------
		; READ/WRITE data greater than 64k or above 1Mb into MS/DOS files: allocate memory for buffer
		; -------------------------------------------------------------------------------------------
		
		mov	eax, FILEATOMSIZE
		call	asklomem				; LOMEM ONLY 
		dd	1
		mov	[FileIOBuff], eax
		
	
		; --------------------------------
		; Allocate memory for system stack
		; --------------------------------

		clr	ecx
		lea	edx, SStackStart
		lea	ebx, ThreadsESP
@@MemSStack:	mov	eax, SSTACKSIZE
		call	asklomem				; LOMEM ONLY
		dd	2
		add	eax, SSTACKSIZE
		mov	[edx], eax
		add	edx, @WordSize
		mov	[ebx], eax
		add	ebx, @WordSize
		inc	ecx
		cmp	cx, MAXNB_THREADS
		jne	@@MemSStack
		
		; Set stack pointer for thread #0
		mov	esp, [SStackStart]
		

		; -------------------------------------------
		; Allocate memory for buffers into low memory
		; -------------------------------------------
		
		mov	eax, size strentry + IBUFFMAXS + 1
		call	asklomem
		dd	3
		mov	[QuoteStrBuff], eax


		; -------------------------------------------
		; Allocate memory fo buffer for AskVideoInfos
		; -------------------------------------------

		mov	eax, 128
		call	asklomem
		dd	4
		mov	[AskVideoInfBuff], eax
		
	
		; --------------------------
		; Open and use the help file
		; --------------------------

		; No help if running a turnkey exe
		cmp	[ExeFile], TRUE
		je	@@NoHelp

		; Try to open		
		lea	ebx, HelpFileName			; Ptr to filename
		mov	ax, 10h                 		; 0 001 0000 -> (inherited, deny both, read)
		call	doOpenFile
		jnc	@@HlpOpenOk
		
		; "File '\FROTH\FROTH.HLP' not found."
@@HelpErr:	SUMMON	EdPDotQ
		db	13, 10
		db	"File "
		db	HELP_FILENAME
		db	" not found or invalid."
		db	13, 10, 0
		jmp	_exit
		
@@HlpOpenOk:

		; Save the file handle
		mov	[HelpFileHdl], bx
		
		; Ask for some memory to hold the 'help directory'
		mov	eax, (MAXNB_SYSTEMWORDS * size hlpentry)
		call	askmem
		dd	2
		mov	[HelpDir], eax
		
		; Read now the directory into memory
		mov	bx, [HelpFileHdl]				; File handle
		mov	ecx, (MAXNB_SYSTEMWORDS * size hlpentry) 	; Bytes to read
		mov	edx, eax					; Buffer address
		call	doReadFile					; in EBX: bytes read from the file
		jc	@@HelpErr
		
		; Ask for some memory to hold help data read for a given word
		mov	eax, MAXSZ_HELP_BUFFER				; Max size of the buffer for the help of a given word
		call	askmem
		dd	3
		mov	[HelpBuff], eax

@@NoHelp:	
		

		; -----------------------------------
		; Allocate memory for Vocabulary_Area
		; -----------------------------------
		
		mov	eax, VOC_MAXSIZE
		call	askmem
		dd	4
		mov	[VOC_START], eax
		
		; Move VOC area
		push	esi edi
		lea	esi, VOC_BEGIN             		; Source address
		mov	edi, [VOC_START]            		; Destination address
		mov	ecx, VOC_END - VOC_BEGIN   		; Size of the block to move
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi
		
		; Nb of vocabulary
		mov	eax, (VOC_END - 4 - VOC_BEGIN) / size vocentry
		mov	[VOC_NB], eax


		; -----------------------
		; Allocate memory for NIA	(NIA for Name Infos Area)
		; -----------------------

		; Ask for memory		
		mov	eax, NIA_WSIZE
		call	askmem
		dd	5
		mov	[NIA_START], eax

		; Move nucleus NIA
		push	esi
		lea	esi, NIA_BEGIN              		; Source address
		mov	edi, [NIA_START]            		; Destination address
		mov	ecx, NIA_END - NIA_BEGIN		; Size of the block to move
		cld                     			; Inc ESI and EDI
		rep	movsb                   		; Move the block
		pop	esi
		
		; Actual size
		mov	[NIA_SIZE], (NIA_END - NIA_BEGIN)
		

		; -----------------------------------------
		; Allocate memory for Cross Words Reference
		; -----------------------------------------

		; Not available if running a turnkey exe
		cmp	[ExeFile], TRUE
		je	@@NoCrossRef

		; Nb of words references defined
		mov	[CrossRefNb], 0
		
		; Nb total of records
		mov	[CrossRefTotalNb], 0

		; First record freed
		mov	[CrossRefFFreed], -1
		
		;  Table of all words references
		mov	eax, MAXNBWREF * size wrefentry
		call	askmem
		dd	66
		mov	[CrossRef], eax
		
		; Index on callers
		mov	eax, MAXNBWREF * @WordSize
		call	askmem
		dd	7
		mov	[CRefNdxRefRD], eax
		
		; Index on callers
		mov	eax, MAXNBWREF * @WordSize
		call	askmem
		dd	8
		mov	[CRefNdxRefDR], eax

@@NoCrossRef:	


		; -------------------------------------------------------
		; Count the nb of words for each vocabulary 'rom'-defined
		; -------------------------------------------------------
		
		; Not available if running a turnkey exe
		cmp	[ExeFile], TRUE
		je	@@NoCountWords

		; Init vocab loop
		mov	edx, [VOC_START]			; Pointer on 1st vocab
		clr	ecx					; Vocab counter

		; For each word of this vocab
@@CountWords:	mov	ebx, [edx.voc_lastword]
@@WordLoop:	cmp	ebx, -1					; Vocab empty ?
		je	@@NextVoc
		inc	[edx.voc_wordsnb]			; Update nb of words for this vocab
		add	ebx, [NIA_START]

		; Update the code size
		mov	eax, [ebx.word_EndCode]			; Code end
		sub	eax, [ebx.word_realaddr]		; Code start
		inc	eax
		mov	[ebx.word_CodeSize], ax			; Code size

		mov	ebx, [ebx.word_prevword]		; Previous word ptr, if any
		jmp	@@WordLoop
		
		; Next vocab, if any		
@@NextVoc:	add	edx, size vocentry
		inc	ecx
		cmp	ecx, [VOC_NB]
		jne	@@CountWords

@@NoCountWords:	
		

		; ------------------------------------
		; Allocate memory for parameters stack
		; ------------------------------------

		clr	ecx
		lea	edx, PStackStart
		lea	ebx, ThreadsFSP
@@MemPStack:	mov	eax, PSTACKSIZE
		call	askmem
		dd	9
		add	eax, PSTACKSIZE
		mov	[edx], eax
		mov	[ebx], eax
		add	edx, @WordSize
		add	ebx, @WordSize
		inc	ecx
		cmp	cx, MAXNB_THREADS
		jne	@@MemPStack

		; Set parameters stack pointer for thread #0
		mov	esi, [PStackStart]


		; -------------------------------------------
		; Allocate memory for second parameters stack
		; -------------------------------------------

		clr	ecx
		lea	edx, RStackStart
@@MemRStack:	mov	eax, RSTACKSIZE
		call	askmem
		dd	10
		add	eax, RSTACKSIZE
		mov	[edx], eax
		add	edx, @WordSize
		inc	ecx
		cmp	cx, MAXNB_THREADS
		jne	@@MemRStack


		; ---------------------------------------
		; Allocate memory for Loading Files Infos
		; ---------------------------------------
		
		mov	eax, MAXNBLOADFILES * size loadfileentry
		call	askmem
		dd	10
		mov	[LoadFilesInfos], eax


		; ----------------------------------------------
		; Allocate memory for Data Directory Area (DDIR)
		; ----------------------------------------------

		mov	eax, DDIR_WSIZE
		call	askmem
		dd	12
		mov	[DDIR_START], eax  	     		; Starting address
		mov	edi, eax               			; EDI is always the starting address of DDIR
		mov	[DDIR_SIZE], 0        			; Actual size

		; Memory which can be recovered with GARBAGE
		mov	[DDIR_MEMSALV], 0
		
		; 1st record freed : none
		mov	[DDIR_FFREED], -1
		
		
		; -----------------------------------------------
		; Allocate memory for the Local Dictionnary (LDA)
		; -----------------------------------------------
		
		; Ask for memory
		mov	eax, LDA_MAXNB
		imul	eax, size lwordentry
		call	askmem
		dd	13
		mov	[LDA_START], eax			; Starting address
		
		; Actual size
		mov	[LDA_SIZE], 0
		

		; ---------------------------------------------
		; Allocate memory for buffer to read load files
		; ---------------------------------------------

		; Not available if running a turnkey exe
		cmp	[ExeFile], TRUE
		je	@@NoLoadFiles

		mov	ebx, [LoadFilesInfos]
		clr	ecx
@@LL21:		mov	eax, LOADFILEBUFFSZ + IBUFFMAXS
		call	askmem
		dd	14
		mov	[ebx.loadfile_buff], eax
		add	ebx, size loadfileentry
		inc	ecx
		cmp	ecx, MAXNBLOADFILES
		jne	@@LL21

@@NoLoadFiles:	


		; ----------------------------------
		; Allocate memory for editor buffers
		; ----------------------------------

		; Not available if running a turnkey exe
		cmp	[ExeFile], TRUE
		je	@@NoEditor

		; For all buffers
		mov	eax, [EdTotalBuffSize]		
		call	askhimem
		dd	15
		mov	[EdBuffInfos.binf_BuffAddr], eax
		
		; For the scrap
		mov	eax, [EdTotalScrapSize]
		call	askmem
		dd	16
		mov	[EdScrap], eax

		; Ask for memory	
		mov	eax, DEFSZ_EDMACROS
		call	askmem
		dd	17
		mov	[EdMacros], eax

		; Internal buffer used by 'EdWritFileToBuf'
		mov	eax, EDWRITEBUFFSZ
		call	askmem
		dd	18
		mov	[EdWriteBuffAddr], eax
		
@@NoEditor:	
		

		; --------------------------------
		; Allocate memory for trace buffer
		; --------------------------------
		
		; Not available if running a turnkey exe
		cmp	[ExeFile], TRUE
		je	@@NoTrace

		mov	eax, [TraceBufferMaxSz]
		call	askmem
		dd	19
		mov	[TraceBuffer], eax

@@NoTrace:	


		; ----------
		; Call KB_US
		; ----------

		DISPOSE	1
		mov	PARAM0, offset KbdFileName		; " KB_FR.FKB"
		SUMMON	KBDFILE					; KbdFile		


		; ---------------------------------------------------------
		; Allocate memory for the Dictionnary (Code -> /// <- Data)
		; ---------------------------------------------------------

		call    _himemsize
		mov	ebx, eax
		call	askhimem				; HIMEM
		dd	21
		
		; Code Dict Area (CDA)
		mov	[CDA_START], eax	     		; Starting address
		mov	[CDA_SIZE], 0         			; Actual size
		
		; Data Dict Area (DDA)
		add	eax, ebx
		mov	[DDA_START], eax            		; Starting address
		mov	[DDA_SIZE], 0         			; Actual size
		

		; -----------------------------------------
		; Update each system word Start/End address
		; -----------------------------------------

		; Offset
		mov	eax, [CDA_START]
		sub	eax, offset _StartAddr
		mov	[OfsDictReloc], eax
		
		; Init vocab loop
		mov	edx, [VOC_START]			; Pointer on 1st vocab
		clr	ecx					; Vocab counter

		; For each word of this vocab
@@CountWords2:	mov	ebx, [edx.voc_lastword]
@@WordLoop2:	cmp	ebx, -1					; Vocab empty ?
		je	@@NextVoc2
		inc	[edx.voc_wordsnb]			; Update nb of words for this vocab
		add	ebx, [NIA_START]

		; Update the word Start/End address
		add	[ebx.word_realaddr], eax		; Code start
		add	[ebx.word_EndCode], eax			; Code end
		cmp	[ebx.word_runtaddr], 0
		je	@@WordLoop3
		add	[ebx.word_runtaddr], eax		; Code start
@@WordLoop3:	

		mov	ebx, [ebx.word_prevword]		; Previous word ptr, if any
		jmp	@@WordLoop2
		
		; Next vocab, if any		
@@NextVoc2:	add	edx, size vocentry
		inc	ecx
		cmp	ecx, [VOC_NB]
		jne	@@CountWords2


		; Address of functions used by PMODE
		mov	[_Excp_1], offset Excp1
		mov	eax, [OfsDictReloc]
		add	[_Excp_1], eax


		; --------------------------------
		; Initialize each nucleus variable
		; --------------------------------

		mov	ecx, [DDA_START]            		; ECX is Data Pointer       
		lea	ebx, DDIR_BEGIN

@@LL11:
		sub	ecx, [ebx.ddir_blksize]     		; Update Data pointer
		mov	[ebx.ddir_blkaddr], ecx         	; Save address in DDIR

		; Update Data Dict size     
		mov	eax, [ebx.ddir_blksize]
		add	[DDA_SIZE], eax

		; Update DDIR size
		add	[DDIR_SIZE], size ddirentry
				
		; Next variable, if any
		add	ebx, size ddirentry
		cmp	[ebx.ddir_NextFreed], 0
		jne	@@LL11

		; Memory allocated bounds
		mov	eax, [DDA_START]
		mov	[DDA_BOUND + @WordSize], eax
		sub	eax, [DDA_SIZE]
		mov	[DDA_BOUND], eax

		; Move Data Directory (DDIR)
		push	esi edi
		lea	esi, DDIR_BEGIN				; Source address
		mov	edi, [DDIR_START]			; Destination address
		mov	ecx, DDIR_END - DDIR_BEGIN		; Size of area to move
		cld						; Inc ESI and EDI
		rep	movsb					; Move the block
		pop	edi esi


		; -----------------------------------
		; Save initial size for Code and Data
		; -----------------------------------
		
		lea	eax, _EndAddr
		lea	ebx, _StartAddr
		lea	eax, [eax - ebx]
		mov	[CDA_SIZE], eax
		
		mov	eax, [CDA_SIZE]
		mov	[CDA_SIZE_INIT], eax

		mov	eax, [DDA_SIZE]
		mov	[DDA_SIZE_INIT], eax
		
		mov	eax, [DDIR_SIZE]
		mov	[DDIR_SIZE_INIT], eax
		
		
		; ----------------------------------------------------------
		; Store max nb of threads value into 'constant' MaxNbThreads
		; ----------------------------------------------------------
	
		SETFVAR	D_MAXNBTHREADS, MAXNB_THREADS


		; -------------------------------------------
		; Continue execution in the copy of FROTH.EXE
		; -------------------------------------------

		; Copy FROTH.EXE (rtl) to himem
		push	esi edi
		lea	esi, _StartAddr			; Source address
		mov	edi, [CDA_START]		; Destination address
		lea	ecx, _EndAddr
		sub	ecx, esi			; ECX = size of block to move
		cld                     		; Inc ESI and EDI
		rep	movsb  		        	; Move the block
		pop	edi esi


		; Go to copy		
		lea	eax, @@next
		add	eax, [OfsDictReloc]
		jmp	eax
@@next:		


		; --------------------------------------
		; Install the keyboard interrupt handler
		; --------------------------------------

		call	InstallKbIt

		; Flag FALSE or TRUE: is the kbd handler installed ?
		mov	[KbItActive], TRUE


		; ------------------------------
		; Select page #0 (VGA text mode)
		;	Bug fixed on Oct,10,95
		; ------------------------------

		IPUSH	0
		SUMMON	CHANGEPAGE			; 0 >page		


		; ----------------------------
		; Initialize things for editor
		; ----------------------------

		; Not available if running a turnkey exe
		cmp	[ExeFile], TRUE
		je	@@NoInitEd

		; Initialize default colors
		; -------------------------
		
		; Status line, label text
		sub	esi, @WordSize
		movzx	eax, [sll]
		mov	[esi], eax
		SUMMON	EDC_SLL
		SUMMON	STORE

		; Status line, infos text
		sub	esi, @WordSize
		movzx	eax, [sli]
		mov	[esi], eax
		SUMMON	EDC_SLI
		SUMMON	STORE

		; Normal text
		sub	esi, @WordSize
		movzx	eax, [txn]
		mov	[esi], eax
		SUMMON	EDC_TXN
		SUMMON	STORE

		; Selected text
		sub	esi, @WordSize
		movzx	eax, [txs]
		mov	[esi], eax
		SUMMON	EDC_TXS
		SUMMON	STORE

		; Help, normal
		sub	esi, @WordSize
		movzx	eax, [hln]
		mov	[esi], eax
		SUMMON	EDC_HLN
		SUMMON	STORE

		; Help, alt
		sub	esi, @WordSize
		movzx	eax, [hla]
		mov	[esi], eax
		SUMMON	EDC_HLA
		SUMMON	STORE

		; Help, keys
		sub	esi, @WordSize
		movzx	eax, [hlk]
		mov	[esi], eax
		SUMMON	EDC_HLK
		SUMMON	STORE

		; Default path for editor files
		mov	edx, [edi.ddir_blkaddr + D_ED_DIR]
		mov	[edx.str_currlen], 0			; Current length
		mov	[edx.str_currsize], MAXSZFILENAME	; Current size allocated
		mov	[edx.str_incsize], 0			; 0 (if fixed str)
		mov	[edx.str_memblknb], 0			; Memory block nb (ref to DDIR)
		mov	byte ptr [edx + size strentry], 0	; Null string

		; Null string
		mov	edx, [edi.ddir_blkaddr + D_NULLSTR]
		mov	[edx.str_currlen], 0			; Current length
		mov	[edx.str_currsize], 0			; Current size allocated
		mov	[edx.str_incsize], 0			; 0 (if fixed str)
		mov	[edx.str_memblknb], 0			; Memory block nb (ref to DDIR)
		mov	byte ptr [edx + size strentry], 0	; Null string

		; Call speed direct video functions ?
		SUMMON	ED_QDIRECTVIDEO
		SUMMON	ZEROSTORE

		; Using tab ?
		SUMMON	ED_QUSETAB
		SUMMON	ONESTORE

		; Column position for comment
		IPUSH	XCOMMENT_DFLT
		SUMMON	ED_COMMENTFIELD
		SUMMON	STORE

		; Line position on screen when goto a line
		IPUSH	LINEPOSADJUST
		SUMMON	ED_LINEPOSADJUST
		SUMMON	STORE

		; The key which calls FROSTED under INTERPRET
		IPUSH	-1					; -1
		SUMMON	ED_KEY					; ED_KEY
		SUMMON	STORE					; !
		
		; Editor video mode
		IPUSH	-1
		SUMMON	ED_VIDEOMODE
		SUMMON	STORE					; !
		

		; ------------------
		; Initialize FROSTED
		; ------------------
		
		mov	[EditorActive], TRUE
		call	EdInit
		mov	[EditorActive], FALSE

@@NoInitEd:	


		
		; ----------------------
		; Froth cold entry: COLD
		; ----------------------

		; Editor is not active at this time
		; ---------------------------------
		
		mov	[EditorActive], FALSE
		

		; Base for integer numeric I/O
		; ----------------------------

		SUMMON	DECIMAL					; DECIMAL


		; Set attribute for erase screen
		; ------------------------------

		SUMMON	VBLACK					; Background color
		SUMMON	VLIGHTGRAY				; Foreground color
		SUMMON	BACKFOREGROUND				; BackForeGround		
		SUMMON	SCREENBCKG				; ScreenBackground
		SUMMON	STORE					; !
		
		
		; Update cursor position after '!xyscreen'
		; ----------------------------------------
		
		SETFVAR	D_QUPDATECURSOR, TRUE			; ?UpdateCursor
		
		
		; Get the number of columns, rows and pages for the current video mode
		; --------------------------------------------------------------------

		call	doGetScreenInf
		

		; -------------------------------------
		; Which vocabulary: CONTEXT and CURRENT
		; -------------------------------------

		; FROTH
		mov	[VOC_CONTEXT], VOCN_FROTH
		
		; CURRENT
		mov	[VOC_CURRENT], VOCN_FROTH
		

		; ----------------------------
		; Threshold for automatic dust
		; ----------------------------

		SETFVAR	D_DUSTTHRESHOLD, -1		
		

		; ------------------
		; Set compiler flags
		; ------------------

		; Show (depth) (interpret prompt)
		SETFVAR	D_QOK, FALSE		

		; Safe system: default
		SETFVAR	D_QSAFE, TRUE		

		; Speed vs Space: default
		SETFVAR	D_QSPEED, TRUE

		; Detect integer arithmetic overflow
		SETFVAR	D_QIARITHOVERFLOW, TRUE
		
		; Chat while compiling ?
		SETFVAR	D_QCHAT, FALSE


		; ------------------------
		; Which language is used ?
		; ------------------------

		SETFVAR	D_LANGUAGE, 0				; 0 means English (but is there another language, uh?)


		; ------------------------------
		; Internally initialize graphics
		; ------------------------------

		call	InitGraphVars


		; ------------------
		; Start msec counter
		; ------------------

		mov 	[v86r_ah], 083h				; Function asked
		mov	[v86r_al], 0
		mov	[v86r_cx], 0FFFFh
		mov	[v86r_dx], 0FFFFh
		mov	[v86r_es], 040h				; BIOS segment,
		mov	[v86r_bx], 098h				; variable 'WAIT'
		mov 	al, 15h					; Call VIDEO_SERVICE
		int 	33h


		; --------------------------------
		; Give control to extern interpret
		; --------------------------------
		
		jmp	Abort
		

; ===============================================================
; = Turnkey							=
; =                                                         	=
; = Usage:							=
; =	Turnkey .EXE -> execute frozen application.		=
; ===============================================================

		align	4
Turnkey		proc

		; --------------------------
		; Load user's data directory
		; --------------------------

		; Open .EXE file
		mov	edx, [QuoteStrBuff]
		call	GetProgName
		mov	ebx, edx
		mov	eax, 2090h		; R/O  b13: no int24h   b7: don't pass handle to child   b4: no sharing
		clr	cl
		call	doWideOpenFile
		jc	@@Error
		mov	[@@FileHdl], bx

		; Seek into user's code area
		mov	al, SEEK_SET			; Move file pointer from beginning of file
;		mov	bx, [@@FileHdl]			; File handle
		mov	ecx, [ExeCdaBegin]		; Current position in the file
		call	doSeekFile
		jc	@@Error

		; Read user's code area into memory		
		mov	bx, [@@FileHdl]			; File handle
		mov	ecx, [ExeCdaSize]		; Bytes to read
		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]			; Buffer address
		call	doReadFile			; in EBX: bytes read from the file
		jc	@@Error

		; Read user's data area into memory		
		mov	bx, [@@FileHdl]			; File handle
		mov	ecx, [ExeDdaSize]		; Bytes to read
		mov	edx, [DDIR_START]
		add	edx, [DDIR_SIZE]		; Buffer address
		call	doReadFile			; in EBX: bytes read from the file
		jc	@@Error


		; -----------
		; Update DDIR
		; -----------

		mov	ecx, [DDA_START]            	; ECX is Data Pointer       
		sub	ecx, [DDA_SIZE]
		
		mov	ebx, [DDIR_START]
		add	ebx, [DDIR_SIZE_INIT]

		mov	edx, [DDIR_START]
		add	edx, [DDIR_SIZE]
		add	edx, [ExeDdaSize]

@@UpdtDDIR:	
		; End reached ?
		cmp	ebx, edx
		je	@@UpdtDDIRdone

		; Data address
		sub	ecx, [ebx.ddir_blksize]     		; Update Data pointer
		mov	[ebx.ddir_blkaddr], ecx         	; Save address in DDIR

		; Update Data Dict size     
		mov	eax, [ebx.ddir_blksize]
		add	[DDA_SIZE], eax

		; Update DDIR size
		add	[DDIR_SIZE], size ddirentry
				
		; Next variable, if any
		add	ebx, size ddirentry
		jmp	@@UpdtDDIR

		; Memory allocated bounds
@@UpdtDDIRdone:	mov	eax, [DDA_START]
		mov	[DDA_BOUND + @WordSize], eax
		sub	eax, [DDA_SIZE]
		mov	[DDA_BOUND], eax


		; ----------------------------------
		; Read DDA (Static Data Dictionnary)
		; ----------------------------------

@@ReadDDA:

		; Read DDIR reference
		mov	bx, [@@FileHdl]			; File handle
		mov	ecx, @WordSize			; Bytes to read
		lea	edx, @@nb			; Buffer address
		call	doReadFile			; in EBX: bytes read from the file
		jc	@@Error

		; End reached ?
		cmp	[@@nb], -1
		je	@@ReadDDAdone

		; Read data from .EXE file
		mov	ebx, [@@nb]
		add	ebx, [DDIR_START]
		mov	ecx, [ebx.ddir_blksize]		; Bytes to read
		mov	edx, [ebx.ddir_blkaddr]		; Buffer address
		mov	bx, [@@FileHdl]			; File handle
		call	doReadFile			; in EBX: bytes read from the file
		
		; Read next record, if any
		jmp	@@ReadDDA
		
@@ReadDDAdone:	

	
		; ---------------
		; Close .EXE file
		; ---------------

		mov	bx, [@@FileHdl]			; File handle
		call	doCloseFile


		; -----------------------
		; Now execute application 
		; -----------------------
		
		mov	[Exe2ndPass], TRUE		; 2nd pass (see Quit) ?
		mov	eax, [ExeWordAddr]
		add	eax, [CDA_START]
		call	eax
		jmp	BYE				; After application, return to host os
		

@@Error:		
		jmp	Abort
		
		align	4
@@FileHdl	dw	?
		align	4
@@nb		dd	?

Turnkey		endp
		
		

; ===============================================================
; = CtrlBreakRecov						=
; =                                                         	=
; = Usage:							=
; =	When Ctrl-break is held and '?CtrlBreak' is true,	=
; =	we jump here. All registers (even ESI and EDI)		=
; =	are in a unknow state...				=
; ===============================================================

		align	4
CtrlBreakRecov	proc

		; ---------------------------
		; Restore 'segment' registers
		; ---------------------------
		
		mov	ax, 10h
		mov	ds, ax
		mov	es, ax
		mov	fs, ax
		mov	gs, _selzero


		; ----------------------
		; Like any others errors
		; ----------------------

		mov	ah, ERR_CTRLBREAK		; "Ctrl-Break."
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		jmp	doError

CtrlBreakRecov	endp


; ===============================================================
; = Abort                           				=
; =                                                         	=
; = Give control to extern interpret                		=
; ===============================================================

		align	4
Abort       	proc

		; --------------------------
		; Enable hardware interrupts
		; --------------------------

		sti


Quit:

		; ---------------------------
		; Garbage collector
		; Discard all local variables
		; ---------------------------

		SUMMON	DUST
		call	DiscardLocalVars
		SUMMON	DUST


		; --------------------------------
		; Exit if turnkey app and 2nd time
		; --------------------------------
		
		cmp	[Exe2ndPass], TRUE			; 2nd pass (see Quit) ?
		je	BYE


		; ----------------------------
		; Restored saved STACK pointer
		; ----------------------------
		
		mov	ebx, [SStackStart]
		mov	esp, ebx

	
		; -----------------------------------------------------------
		; If not thread #0, copy this thread stack to thread #0 stack
		; -----------------------------------------------------------

		mov	eax, [CurrActThread]
		cmpz	eax, 0
		je	@@Thread0
		
		; Copy parameters stack
		push	esi edi
		mov	esi, [PStackStart + eax * 4]		; Source address
		dec	esi
		mov	edi, [PStackStart]			; Destination address
		dec	edi
		mov	ecx, PSTACKSIZE				; Nb of dwords to copy
		std                     			; Dec ESI and EDI
		rep	movsd  		            		; Move the block
		cld
		pop	edi esi
		
		; Update the pointer to the stack
		mov	ebx, [PStackStart + eax * 4]
		sub	ebx, esi
		mov	esi, [PStackStart]			; Stack for thread #0
		sub	esi, ebx
		
		jmp	@@ThreadNot0

@@Thread0:	; mov	esi, [PStackStart]			; Stack for thread #0

@@ThreadNot0:


		; -------------------------
		; Only one thread is active
		; -------------------------

		call	ResetThreads


		; --------------------------------------------------------------
		; Empty second parameters stack (reset parameters stack pointer)
		; --------------------------------------------------------------

		clr	ecx
		lea	edx, RStackStart
		lea	ebx, RStackPtr
@@EmptyRStack:	mov	eax, [edx]
		mov	[ebx], eax
		add	edx, @WordSize
		add	ebx, @WordSize
		inc	ecx
		cmp	ecx, MAXNB_THREADS
		jne	@@EmptyRStack


		; --------------------------------------------------
		; Put the system in mode immediate (not compilation)
		; --------------------------------------------------
		
		mov	[CompileMode], FALSE


		; ------------------------------------
		; Reset the system in mode not comment
		; ------------------------------------
		
		mov	[CDA_COMMENT], 0
		mov	[CDA_SLASHSLASH], FALSE			; Comment to end of line ?
		
		
		; --------------------------------------------------------------
		; Reset the system in mode: not store directly code op into word
		; --------------------------------------------------------------
		
		mov	[CDA_BRACKET_MODE], FALSE
		
		
		; -----------------------------------------------------------
		; Reset the system in mode: not Infixed Expression Evaluation
		; -----------------------------------------------------------
		
		mov	[InfixExprEvalQ], FALSE
		

		; -----------------------------------------
		; Flags for 'BUILD> ... DOES>' construction
		; -----------------------------------------

		mov	[CDA_BUILD_CALLED], FALSE		; Is 'BUILD>' called ?		
		mov	[CDA_DOES_CALLED], FALSE		; Is 'DOES>'  called ?
		mov	[CDA_ALLOT_CALLED], FALSE		; Is 'ALLOT'  called ?
		

		; --------------------------------
		; Flag for using dynamic variables
		; --------------------------------
		
		mov	[DynamicFlag], FALSE		; DYNAMIC word called ?
		mov	[DynVarAddr], 0			; Address of beginning for dynamic variable
		mov	[AllocFlag], FALSE		; ALLOC word called ?
		mov	[AllocFound], FALSE		; ALLOC word found while compiling ?
		mov	[FreeFlag], FALSE		; FREE  word called ?
		mov	[FreeFound], FALSE		; FREE  word found while compiling ?


		; ------------------------------------------
		; Nb of nested loops (reset loop stacked nb)
		; ------------------------------------------

		clr	ecx		
@@ZLoopStack:	mov	[LoopStackNb + ecx * @WordSize], 0
		inc	ecx
		cmp	cx, MAXNB_THREADS
		jne	@@ZLoopStack
		
		
		; -----------------------
		; Control structurs stack
		; -----------------------

		lea	ecx, ControlStack		
		mov	[ControlStackPtr], ecx
		
		
		; --------------------------
		; Nb of nested case..endcase
		; --------------------------
		
		clr	ecx		
@@ZCaseStack:	mov	[CaseStackNb + ecx * @WordSize], 0
		inc	ecx
		cmp	cx, MAXNB_THREADS
		jne	@@ZCaseStack


		; ---------------------------
		; Does 'STARTUP.FTH' loaded ?
		; ---------------------------
		
		cmp	[StartupLoaded], FALSE
		jne	@@StartupDone
		cmp	[ExeFile], TRUE				; Are-we executing a saved file.exe ?
		je	@@StartupDone
		
		; Now 'STARTUP.FTH' is loaded
		mov	[StartupLoaded], TRUE
		
		; Load the file 'STARTUP.FTH'
		DISPOSE	1
		mov	PARAM0, offset StartupFileName
		SUMMON	QUOTELOAD				; "LOAD
		
@@StartupDone:


		; -----------------
		; Allows Ctrl-Break
		; -----------------
		
		mov	[NoCtrlBreak], 0
		mov	[CtrlBreakEvent], FALSE


		; ----------------------
		; Is-it a turnkey copy ?
		; ----------------------
		
		cmp	[ExeFile], TRUE
		je	Turnkey


		; -------------------------------------------
		; Enter at last in the extern froth interpret
		; -------------------------------------------
	
		jmp	INTERPRET

E_Quit:		
Abort       	endp


; ===============================================================
; = Excp1							=
; =                                                         	=
; = Usage:							=
; =	Function called when 'irq 1-15' interrupt occurs.	=
; =                                                     	=
; = I) AL = int #						=
; ===============================================================

		align	4
Excp1		proc

		mov	ah, ERR_EXCP_0			; Error #
		add	ah, al
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		jmp	doError

Excp1		endp


; ===============================================================
; = PrWordName                          			=
; =                                                         	=
; = Usage:							=
; =	Print the word name, the address is given.		=
; =                                                     	=
; = I) EDX : word code address (from begin to end)		=
; =                                                         	=
; = O) none							=
; =                                                         	=
; = M) none							=
; ===============================================================

		align	4
PrWordName     	proc

		; Save used registers
		pushad
		
		; Scan all words
		clr	ebx
		mov	ecx, [NIA_START]
@@Scan:		cmp	byte ptr [ecx.word_name], 0
		je	@@WordEmpty
		
		; Word address is this one ?
		cmp	edx, [ecx.word_realaddr]
		jb	@@NotThisWord
		cmp	edx, [ecx.word_EndCode]
		jbe	@@ThisWord
		
@@NotThisWord:
@@WordEmpty:	add	ecx, size wordentry
		add	ebx, size wordentry
		cmp	ebx, [NIA_SIZE]				; !!!
		jl	@@Scan
	mov	eax, edx
	call	PrEAX
		jmp	@@Exit		

		; Word is found
@@ThisWord:	add	ecx, word_name
		sub	esi, @WordSize
		mov	[esi], ecx
		SUMMON	_TYPE					; TYPE
		SUMMON	SPACE					; SPACE

		; Return to caller
@@Exit:		popad
		ret

PrWordName	endp


; ===============================================================
; = doError                           				=
; =                                                         	=
; = Froth error procedure                   			=
; =                                                     	=
; = I) AH  : error number (0, 1, ...)                		=
; =	     ( if -1, print only the given msg in EDX).		=
; =    EDX : pointer on msg, or 0 if not msg			=
; =    ECX : line nb, or 0 if none				=
; ===============================================================

		align	4
doError       	proc

		; ---------------------
		; Save enter parameters
		; ---------------------
		
		mov	[@@SavedAH], ah
		mov	[@@SavedECX], ecx
		mov	[@@SavedEDX], edx


		; -----------------------------------------------
		; Restore EDI and ESI (perhaps updated by caller)
		; -----------------------------------------------
		
		mov	[NoCtrlBreak], 1			; CTRL-BREAK not allowed
		mov	[CtrlBreakEvent], FALSE
		mov	edi, [DDIR_START]       		; Data table: DDIR
		mov	[@@SavedESI], esi
		mov	esi, [PStackStart]			; Stack pointer for thread #0


		; ------------------------------
		; Show the nb of running threads
		; ------------------------------

		SUMMON	CR

		; Count the nb of running threads
		clr	ecx
		clr	edx
@@CountRT:	cmp	[ThreadsActive + ecx], FALSE
		je	@@CountRT2
		inc	edx
@@CountRT2:	inc	ecx
		cmp	cx, MAXNB_THREADS
		jne	@@CountRT
		cmp	dx, 1
		je	@@NoPrintRT

		SUMMON	PDOTQ
		db	"Threads running: ", 0
		clr	ecx
@@ShowRT:	cmp	[ThreadsActive + ecx], FALSE
		je	@@ShowRT2
		IPUSH	ecx
		SUMMON	DOT					; .
		SUMMON	SPACE
@@ShowRT2:	inc	ecx
		cmp	cx, MAXNB_THREADS
		jne	@@ShowRT
		SUMMON	CR

@@NoPrintRT:	


		; ------------------------------------
		; Show the thread nb, if not thread #0
		; ------------------------------------
		
		SUMMON	DECIMAL					; DECIMAL
		cmp	[CurrActThread], 0
		je	@@ThreadDone

		SUMMON	PDOTQ					; rt_."
		db	"Active thread was thread #", 0
		mov	eax, [CurrActThread]
		IPUSH	eax
		SUMMON	DOT					; .
		SUMMON	CR
@@ThreadDone:


		; --------------
		; Show the stack
		; --------------

		; Save ESI
		push	esi
		
		; Restore initial ESI, it it seems to be OK
		mov	eax, [CurrActThread]			; Current active thread #
		mov	eax, [PStackStart + eax * 4]
		cmp	[@@SavedESI], eax
		ja	@@ESIbad
		sub	eax, PSTACKSIZE
		cmp	[@@SavedESI], eax
		jb	@@ESIbad
		mov	esi, [@@SavedESI]
@@ESIbad:	SUMMON	DOTS					; .S

		pop	esi
		SUMMON	CR


		; ------------------------
		; Show definitions callers
		; ------------------------

		; Is ESP ok ?
		mov	eax, [CurrActThread]			; Current active thread
		mov	ebx, [SStackStart + eax*4]
		cmp	esp, ebx
		ja	@@ShowDone
		sub	ebx, SSTACKSIZE
		cmp	esp, ebx
		jb	@@ShowDone

		; Pointer into FROTH code region ?
		lea	ebx, _StartAddr2
		add	ebx, [OfsDictReloc]			; Offset of code dict relocation
		cmp	[esp], ebx
		jb	@@ShowDone				; Too low

		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		cmp	[esp], ebx
		ja	@@ShowDone				; Too hight

		SUMMON	PDOTQ					; rt_."
		db	"Calls: ", 0
		mov	ebx, esp
@@Show:		mov	edx, [ebx]
		cmp	edx, offset Execute
		je	@@Stop
		call	PrWordName
		add	ebx, @WordSize
		cmp	ebx, [SStackStart + eax * @WordSize]
		jb	@@Show
@@Stop:		SUMMON	CR					; CR
@@ShowDone:	


		; ----------------------------
		; Restored saved STACK pointer
		; ----------------------------
		
		mov	ebx, [SStackStart]
		mov	esp, ebx


		; -----------------------------------
		; Show the thread #, if not thread #0
		; -----------------------------------
		
		mov	ebx, [CurrActThread]			; Current active thread #
		cmpz	bx, 0
		je	@@ThreadShown
		SUMMON	PDOTQ					; ."
		db	"Thread #", 0
		IPUSH	ebx
		SUMMON	DOT					; .
		SUMMON	CR
@@ThreadShown:

		; Reset to thread #0
		call	ResetThreads
		

		; --------------------------
		; Are we loading from file ?
		; --------------------------

		cmp	[LoadFilesNb], 0
		je	@@NoOpenFiles
		
		; Close all the files open
		mov	edx, [LoadFilesInfos]
		clr	ecx					; Counter
@@CloseLoop:	mov	bx, [edx.loadfile_hdl]			; Get file handle (if 0, file is not opened)
		cmpz	bx, 0
		je	@@NextFile
		call	doCloseFile				; Close this file
		mov	[edx.loadfile_hdl], 0			; File is closed
@@NextFile:	add	edx, size loadfileentry
		inc	ecx
		cmp	cl, MAXNBLOADFILES
		jne	@@CloseLoop		

		; Show the filename
		PRSTR	"File: "
		mov	eax, [LoadFilesNb]
		dec	eax
		imul	eax, size loadfileentry
		add	eax, [LoadFilesInfos]
		add	eax, loadfile_name
		sub	esi, @WordSize
		mov	PARAM0, eax
		SUMMON	_TYPE					; TYPE
		
		; Show the line #
		PRSTR	"  Line: "
		mov	eax, [LoadFilesNb]
		dec	eax
		imul	eax, size loadfileentry
		add	eax, [LoadFilesInfos]
		mov	ebx, [eax.loadfile_linenb]
		sub	esi, @WordSize
		mov	PARAM0, ebx
		SUMMON	DOT					; .

		; Show the col #
		PRSTR	"  Col: "
		movzx	ebx, [eax.loadfile_colnb]
		sub	esi, @WordSize
		mov	PARAM0, ebx
		SUMMON	DOT					; .
		SUMMON	CR					; CR

@@NoOpenFiles:


		; ------------------------------
		; Show msg, if any (addr in EDX)
		; ------------------------------

		; A given msg to print only (case of: "ERROR) ?
		cmp	[@@SavedAH], -1					; AX = error #
		jne	@@Msg2C
		mov	edx, [@@SavedEDX]				; EDX = ptr to msg (or NULL if none)		
		IPUSH	edx
		SUMMON	_TYPE
		SUMMON	SPACE
		jmp	@@MsgDone
@@Msg2C:

		mov	edx, [@@SavedEDX]				; EDX = ptr to msg (or NULL if none)		
		cmpz	edx, 0
		je	@@MsgDone
		IPUSH	edx
		SUMMON	_TYPE
		cmp	[@@SavedECX], 0
		jne	@@Msg2
		SUMMON	PDOTQ
		db	" : ", 0
		jmp	@@MsgDone
@@Msg2:		SUMMON	PDOTQ
		db	" ", 0
@@MsgDone:

		; ----------------------------
		; Show line #, of any (in ECX)
		; ----------------------------

		mov	ecx, [@@SavedECX]				; ECX = line # (or 0if none)		
		cmpz	ecx, 0
		je	@@LineNbDone
		SUMMON	PDOTQ
		db	"(line #", 0
		IPUSH	ecx
		SUMMON	DOT
		SUMMON	PDOTQ
		db	") : ", 0
@@LineNbDone:


		; -----------------
		; Found msg address
		; -----------------

		cmp	[@@SavedAH], -1
		je	@@LL4						; Case of: "ERROR
		mov	ah, [@@SavedAH]					; AX = error #		
		lea	ebx, ERRMSG
@@LL1:      	cmp	ah, [ebx]
		je 	@@LL3
@@LL2:      	inc	ebx
		cmp	byte ptr [ebx], 0
		jne	@@LL2
		inc	ebx
		jmp	@@LL1
@@LL3:      	inc	ebx
		mov	al, [ebx]
		cmpz	al, 0
		je 	@@LL4
		call	COUT
		jmp	@@LL3
@@LL4:

		; --------------------------------------------------------------
		; If compile state, then forget the word currently being defined	
		; --------------------------------------------------------------

		cmp	[CompileMode], FALSE
		je	@@ExecState

		; Forget word currently being defined
		mov	edx, [VOC_CURRENT]
		add	edx, [VOC_START]
		mov	edx, [edx.voc_lastword]
		add	edx, [NIA_START]
		mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		sub	eax, [edx.word_realaddr]
		mov	[edx.word_CodeSize], ax
		call	doForgetWord

@@ExecState:


		; ---------------------------
		; Garbage collector
		; Discard all local variables
		; ---------------------------

		SUMMON	DUST
		call	DiscardLocalVars
		SUMMON	DUST


		; ------------------------------------------------------------------------
		; If error from a loaded file, is this file loaded into a FROSTED buffer ?
		; ------------------------------------------------------------------------

		mov	ecx, [LoadFilesNb]
		mov	[LoadFilesNb], 0			; Reset now the 'depth' of files loading
		cmpz	ecx, 0
		je	@@abort
		
		; Current filename (where error occurs)
		dec	ecx
		imul	ecx, size loadfileentry
		add	ecx, [LoadFilesInfos]

		lea	edx, EdBuffInfos
		mov	[@@BuffNb], 0				; Buffers counter

@@FilePresent:	; Test filename
		clr	ebx					; Chars counter
@@FP2:		mov	al, [ecx.loadfile_name + ebx]
		TOUPPER	al
		mov	ah, [edx.binf_FileName + ebx]
		TOUPPER	ah
		cmp	al, ah
		jne	@@FP3
		or	al, ah
		je	@@FP4
		inc	ebx
		jmp	@@FP2

		; Next buffer, if any
@@FP3:		add	edx, size buffedentry
		inc	[@@BuffNb]
		cmp	[@@BuffNb], EDMAXNBBUFF
		jne	@@FilePresent
		jmp	@@abort

		; Filename found
@@FP4:		SUMMON	PDOTQ
		db	CMSG_ERR_RETED			; "Press <CR> to edit file or <ESC> to return to interpret "
@@FP5:		SUMMON	CharIn
		cmp	ax, EKEY_ESC
		je	@@abort
		cmp	ax, EKEY_CR
		jne	@@FP5

		; Select the current buffer to edit
		movzx	eax, [@@BuffNb]
		mov	[EdCurrBufferNb], eax
		
		; Select the current line where to go
		movzx	eax, [ecx.loadfile_colnb]
		dec	eax
		mov	[edx.binf_tcol], eax			; col #
		mov	eax, [ecx.loadfile_linenb]
		dec	eax
		mov	[edx.binf_trow], eax

		; Enter into editor
		SUMMON	ED					; ED
		

		; --------------------------------
		; Give control to extern interpret
		; --------------------------------
		
@@abort:	jmp	Abort
		
		align	4
@@BuffNb	db	?
		align	4
@@SavedESI	dd	?

@@SavedAH	db	?
		align	4
@@SavedECX	dd	?
@@SavedEDX	dd	?

doError       	endp


; ===============================================================
; = askmem							=
; =                                                         	=
; = Usage:							=
; =	Ask for memory.						=
; ===============================================================

		align	4
askmem		proc

		; Ask to pmode
		push	eax
		call	_getmem
		jnc	@@Exit
		
		; Not enought memory
		clr	bl
		jmp	MemError
		
		; Normal return
@@Exit:		add	esp, @WordSize
		add	dword ptr [esp], @WordSize			; Skip next cell
		ret

askmem		endp


; ===============================================================
; = asklomem							=
; =                                                         	=
; = Usage:							=
; =	Ask for low memory.					=
; ===============================================================

		align	4
asklomem	proc

		; Ask to pmode
		push	eax
		call	_getlomem
		jnc	@@Exit
		
		; Not enought memory
		mov	bl, 1
		jmp	MemError
		
		; Normal return
@@Exit:		add	esp, @WordSize
		add	dword ptr [esp], @WordSize			; Skip next cell
		ret

asklomem	endp


; ===============================================================
; = askhimem							=
; =                                                         	=
; = Usage:							=
; =	Ask for high memory.					=
; ===============================================================

		align	4
askhimem	proc

		; Ask to pmode
		push	eax
		call	_gethimem
		jnc	@@Exit
		
		; Not enought memory
		mov	bl, 2
		jmp	MemError
		
		; Normal return
@@Exit:		add	esp, @WordSize
		add	dword ptr [esp], @WordSize			; Skip next cell
		ret

askhimem	endp


; ===============================================================
; = MemError							=
; =                                                         	=
; = Usage:							=
; =	Test is memory asked with '_getlomem', '_gethimem' or	=
; =	'_getmem' is given.					=
; =                                                     	=
; = I) eax : block addr, or -1 if not enought memory.		=
; =                                                     	=
; = O) none                                                    	=
; =                                                     	=
; = M) none                                                    	=
; ===============================================================

		align	4
MemError	proc

		; Not enought memory !
		cmpz	bl, 0
		jne	@@1
		lea	edx, @@Msg1
		jmp	@@Print1
@@1:		cmp	bl, 2
		jne	@@2
		lea	edx, @@MsgA
		jmp	@@Print1
@@2:		lea	edx, @@MsgB
@@Print1:	mov	al, [edx]
		cmpz	al, 0
		je	@@Print1Done
		call	COUT
		inc	edx
		jmp	@@Print1
@@Print1Done:

		; Print the amount of memory asked
		pop	eax
		call	PrEAX
		
		; Print the mark
		lea	edx, @@Msg2
@@Print2:	mov	al, [edx]
		cmpz	al, 0
		je	@@Print2Done
		call	COUT
		inc	edx
		jmp	@@Print2
@@Print2Done:
		mov	eax, [esp]
		mov	eax, [eax]
		call	PrEAX
		
		; Print the amount of lo mem free
		lea	edx, @@Msg3
@@Print3:	mov	al, [edx]
		cmpz	al, 0
		je	@@Print3Done
		call	COUT
		inc	edx
		jmp	@@Print3
@@Print3Done:
		call	_lomemsize
		call	PrEAX

		; Print the amount of hi mem free
		lea	edx, @@Msg4
@@Print4:	mov	al, [edx]
		cmpz	al, 0
		je	@@Print4Done
		call	COUT
		inc	edx
		jmp	@@Print4
@@Print4Done:
		call	_himemsize
		call	PrEAX

		; CR
		mov	al, 13
		call	COUT
		mov	al, 10
		call	COUT

		; Desinstall the keyboard interrupt handler ?
		cmp	[KbItActive], FALSE
		je	@@KbIt
		call	DesinstallKbIt
@@KbIt:

		; Return to host operating system
		jmp	_exit
		
		; Return to caller
@@Exit:		ret

@@Msg1		db	13, 10
		db	"Not enought memory.", 13, 10
		db	"Asked is : "
		db	0

@@MsgA		db	13, 10
		db	"Not low enought memory.", 13, 10
		db	"Asked is : "
		db	0

@@MsgB		db	13, 10
		db	"Not hight enought memory.", 13, 10
		db	"Asked is : "
		db	0

@@Msg2		db	13, 10
		db	" Mark is : "
		db	0

@@Msg3		db	13, 10
		db	" lo mem free is : "
		db	0

@@Msg4		db	13, 10
		db	" hi mem free is : "
		db	0

MemError	endp


code32		ends
		end
