; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	HGR.ASM							!
; 									!
; Description:	Definitions for playing with graphics.			!
; 									!
; List of FROTH definitions defined in this module:			!
;	GNBROWS		#grows						!
;	GNBCOLS		#gcols						!
;	GNBCOLORS	#gcolors					!
;	GNBPAGES	#gpages						!
;	M320x200x16	320x200x16					!
;	M640x200x16	640x200x16					!
;	M640x350x4	640x350x4					!
;	M640x350x16	640x350x16					!
;	M640x480x2	640x480x2					!
;	M640x480x16	640x460x16					!
;	M320x200x256	320x200x256					!
;	M800x600x16	800x600x16					!
;	GNOTFILLED	NOTFILLED					!
;	GFILLED		FILLED						!
;	CLEAR		CLEAR						!
;	SELECTGPAGE	SelectGPage					!
;	CLIP		CLIP						!
;	GMODE_COPY	gmode_copy					!
;	GMODE_XOR	gmode_xor					!
;	GMODE_OR	gmode_or					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

__HGR__ equ 1
include		hgr.inc


; ===============================================================
; = _GraphMoodeNotSet                        			=
; =								=
; = Called if graphics mode not set.				=
; ===============================================================

		align	4
_GraphModeNotSet	proc

		; "Graphics mode not set."
		mov	ah, ERR_GMODENOTSET			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError

_GraphModeNotSet	endp


; ===============================================================
; = _NotImplemented						=
; =								=
; = Called if graphics function is not still implemented.	=
; ===============================================================

		align	4
_NotImplemented	proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError

_NotImplemented	endp


; ===============================================================
; = GNBROWS                            				=
; =								=
; = Code for the forth word:    #grows               		=
; =								=
; = Stack:  ( --- n )	                 			=
; ===============================================================

		align	4
GNBROWS		proc

		; Read data
		GETFVAR	eax, D_GNBROWS
		sub	esi, @WordSize
		mov	[esi], eax

		; Verify stack frame overflow, then return to caller
		OVSTACK
		
		; Return to caller
		ret

E_GNBROWS:
GNBROWS		endp


; ===============================================================
; = GNBCOLS                            				=
; =								=
; = Code for the forth word:    #gcols               		=
; =								=
; = Stack:  ( --- n ) 	                			=
; ===============================================================

		align	4
GNBCOLS		proc

		; Read data
		GETFVAR	eax, D_GNBCOLS
		sub	esi, @WordSize
		mov	[esi], eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_GNBCOLS:
GNBCOLS		endp


; ===============================================================
; = GNBCOLORS                            			=
; =								=
; = Code for the forth word:    #gcolors             		=
; =								=
; = Stack:  ( --- n ) 	                			=
; ===============================================================

		align	4
GNBCOLORS	proc

		; Read data
		GETFVAR	eax, D_GNBCOLORS
		sub	esi, @WordSize
		mov	[esi], eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_GNBCOLORS:
GNBCOLORS	endp


; ===============================================================
; = GNBPAGES                            			=
; =								=
; = Code for the forth word:    #gpages             		=
; =								=
; = Stack:  ( --- n ) 	                			=
; ===============================================================

		align	4
GNBPAGES	proc

		; Read data
		GETFVAR	eax, D_GNBPAGES
		sub	esi, @WordSize
		mov	PARAM0, eax

		; Verify stack frame overflow, then return to caller
		OVSTACK
		ret

E_GNBPAGES:
GNBPAGES	endp


_ClipAllScreen	proc

		; x0 y0 x1 y1
		mov	[_clip_x0], 0
		mov	[_clip_y0], 0
		mov	ax, VSP.vsp_usSX
		dec	ax
		mov	[_clip_x1], ax
		mov	ax, VSP.vsp_usSY
		dec	ax
		mov	[_clip_y1], ax

		; Return to caller
		ret

_ClipAllScreen	endp


; ===============================================================
; = InitGraphVars						=
; =								=
; = Usage							=	
; =   Internally initialize graphics.				=
; =								=
; = Called on cold start only.					=
; ===============================================================

		align	4
InitGraphVars	proc

		; Pointer to draw functions
		lea	eax, _NotImplemented
		add	eax, [OfsDictReloc]		; Offset of code dict relocation
		mov	[_clear_], eax
		mov	[_setpix_], eax
		mov	[_getpix_], eax
		mov	[_line_], eax
		mov	[_getvideo_], eax
		mov	[_putvideo_], eax
		
		; Return to caller
		ret

InitGraphVars	endp

		
; ===============================================================
; = M320x200x16							=
; =								=
; = Code for the froth word:	320x200x16			=
; =								=
; = Stack:  ( --- )						=
; ===============================================================

		align	4
M320x200x16   	proc

		; ---------------------
		; Initialize video mode
		; ---------------------
		
		SUMMON	GETVIDEOMODE
		mov	eax, PARAM0
		DISCARD	1
		cmp	eax, 00Dh
		je	@@Init1
		IPUSH	00Dh
		SUMMON	SETVIDEOMODE
@@Init1:		


		; -----------------
		; Initialize things
		; -----------------
		
		; Copy specific values
		push	esi edi
		lea	esi, VSP_320x200x16            		; Source address
		lea	edi, VSP		          	; Destination address
		mov	ecx, size VIDEOSTATEPARAMS        	; Size of the block to move
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update current mode informations
		SETFVAR	D_GNBCOLS, 320
		SETFVAR	D_GNBROWS, 200
		SETFVAR	D_GNBCOLORS, 16

		; Clip to all screen
		call	_ClipAllScreen
		
		; Pointer to draw functions
		mov	[_clear_], offset FillVBuff_10
		mov	[_setpix_], offset SetPixel_10
		mov	[_getpix_], offset GetPixel_10
		mov	[_line_], offset Line_10
		mov	[_getvideo_], offset _BLTTORAM_10
		mov	[_putvideo_], offset _BLTTOVBUFF_10

		; Initializes a timeout value for detecting the start of the vertical nondisplay interval.
		call	INITVNWAIT
		

		; ----------------
		; Return to caller
		; ----------------

		ret
			
E_M320x200x16:
M320x200x16	endp


; ===============================================================
; = M640x200x16							=
; =								=
; = Code for the froth word:	640x200x16			=
; =								=
; = Stack:  ( --- )						=
; ===============================================================

		align	4
M640x200x16   	proc

		; ---------------------
		; Initialize video mode
		; ---------------------
		
		SUMMON	GETVIDEOMODE
		mov	eax, [esi]
		add	esi, @WordSize
		cmp	eax, 00Eh
		je	@@Init1
		IPUSH	00Eh
		SUMMON	SETVIDEOMODE
@@Init1:		


		; -----------------
		; Initialize things
		; -----------------
		
		; Copy specific values
		push	esi edi
		lea	esi, VSP_640x200x16            		; Source address
		lea	edi, VSP		          	; Destination address
		mov	ecx, size VIDEOSTATEPARAMS        	; Size of the block to move
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update current mode informations
		SETFVAR	D_GNBCOLS, 640
		SETFVAR	D_GNBROWS, 200
		SETFVAR	D_GNBCOLORS, 16
		
		; Clip to all screen
		call	_ClipAllScreen
		
		; Pointer to draw functions
		mov	[_clear_], offset FillVBuff_10
		mov	[_setpix_], offset SetPixel_10
		mov	[_getpix_], offset GetPixel_10
		mov	[_line_], offset Line_10
		mov	[_getvideo_], offset _BLTTORAM_10
		mov	[_putvideo_], offset _BLTTOVBUFF_10
		
		; Initializes a timeout value for detecting the start of the vertical nondisplay interval.
		call	INITVNWAIT
		

		; ----------------
		; Return to caller
		; ----------------

		ret
			
E_M640x200x16:
M640x200x16	endp


; ===============================================================
; = M640x350x4							=
; =								=
; = Code for the froth word:	640x350x4			=
; =								=
; = Stack:  ( --- )						=
; ===============================================================

		align	4
M640x350x4   	proc

		; ---------------------
		; Initialize video mode
		; ---------------------
		
		SUMMON	GETVIDEOMODE
		mov	eax, [esi]
		add	esi, @WordSize
		cmp	eax, 00Fh
		je	@@Init1
		IPUSH	00Fh
		SUMMON	SETVIDEOMODE
@@Init1:		


		; -----------------
		; Initialize things
		; -----------------
		
		; Copy specific values
		push	esi edi
		lea	esi, VSP_640x350x4            		; Source address
		lea	edi, VSP		          	; Destination address
		mov	ecx, size VIDEOSTATEPARAMS        	; Size of the block to move
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update current mode informations
		SETFVAR	D_GNBCOLS, 640
		SETFVAR	D_GNBROWS, 350
		SETFVAR	D_GNBCOLORS, 4
		
		; Clip to all screen
		call	_ClipAllScreen
		
		; Pointer to draw functions
		mov	[_clear_], offset FillVBuff_10
		mov	[_setpix_], offset SetPixel_10
		mov	[_getpix_], offset GetPixel_10
		mov	[_line_], offset Line_10
		mov	[_getvideo_], offset _BLTTORAM_10
		mov	[_putvideo_], offset _BLTTOVBUFF_10
		
		; Initializes a timeout value for detecting the start of the vertical nondisplay interval.
		call	INITVNWAIT
		

		; ----------------
		; Return to caller
		; ----------------

		ret
			
E_M640x350x4:
M640x350x4	endp


; ===============================================================
; = M640x350x16							=
; =								=
; = Code for the froth word:	640x350x16			=
; =								=
; = Stack:  ( --- )						=
; ===============================================================

		align	4
M640x350x16   	proc

		; ---------------------
		; Initialize video mode
		; ---------------------
		
		SUMMON	GETVIDEOMODE
		mov	eax, [esi]
		add	esi, @WordSize
		cmp	eax, 010h
		je	@@Init1
		IPUSH	010h
		SUMMON	SETVIDEOMODE
@@Init1:		


		; -----------------
		; Initialize things
		; -----------------
		
		; Copy specific values
		push	esi edi
		lea	esi, VSP_640x350x16            		; Source address
		lea	edi, VSP		          	; Destination address
		mov	ecx, size VIDEOSTATEPARAMS        	; Size of the block to move
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update starting graphics page address
		GETFVAR	ebx, D_PAGENB				; PAGE#
		IPUSH	ebx
		SUMMON	SELECTGPAGE

		; Update current mode informations
		SETFVAR	D_GNBCOLS, 640
		SETFVAR	D_GNBROWS, 350
		SETFVAR	D_GNBCOLORS, 16
		
		; Clip to all screen
		call	_ClipAllScreen
		
		; Pointer to draw functions
		mov	[_clear_], offset FillVBuff_10
		mov	[_setpix_], offset SetPixel_10
		mov	[_getpix_], offset GetPixel_10
		mov	[_line_], offset Line_10
		mov	[_getvideo_], offset _BLTTORAM_10
		mov	[_putvideo_], offset _BLTTOVBUFF_10
		
		; Initializes a timeout value for detecting the start of the vertical nondisplay interval.
		call	INITVNWAIT
		

		; ----------------
		; Return to caller
		; ----------------

		ret
			
E_M640x350x16:
M640x350x16	endp


; ===============================================================
; = M640x480x2							=
; =								=
; = Code for the froth word:	640x480x2			=
; =								=
; = Stack:  ( --- )						=
; ===============================================================

		align	4
M640x480x2   	proc

		; ---------------------
		; Initialize video mode
		; ---------------------
		
		SUMMON	GETVIDEOMODE
		mov	eax, [esi]
		add	esi, @WordSize
		cmp	eax, 011h
		je	@@Init1
		IPUSH	011h
		SUMMON	SETVIDEOMODE
@@Init1:		


		; -----------------
		; Initialize things
		; -----------------
		
		; Copy specific values
		push	esi edi
		lea	esi, VSP_640x480x2            		; Source address
		lea	edi, VSP		          	; Destination address
		mov	ecx, size VIDEOSTATEPARAMS        	; Size of the block to move
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update current mode informations
		SETFVAR	D_GNBCOLS, 640
		SETFVAR	D_GNBROWS, 480
		SETFVAR	D_GNBCOLORS, 2
		
		; Clip to all screen
		call	_ClipAllScreen
		
		; Pointer to draw functions
		mov	[_clear_], offset FillVBuff_10
		mov	[_setpix_], offset SetPixel_10
		mov	[_getpix_], offset GetPixel_10
		mov	[_line_], offset Line_10
		mov	[_getvideo_], offset _BLTTORAM_10
		mov	[_putvideo_], offset _BLTTOVBUFF_10
		
		; Initializes a timeout value for detecting the start of the vertical nondisplay interval.
		call	INITVNWAIT
		

		; ----------------
		; Return to caller
		; ----------------

		ret
			
E_M640x480x2:
M640x480x2	endp


; ===============================================================
; = M640x480x16							=
; =								=
; = Code for the froth word:	640x480x16			=
; =								=
; = Stack:  ( --- )						=
; ===============================================================

		align	4
M640x480x16   	proc

		; ---------------------
		; Initialize video mode
		; ---------------------
		
		SUMMON	GETVIDEOMODE
		mov	eax, [esi]
		add	esi, @WordSize
		cmp	eax, 012h
		je	@@Init1
		IPUSH	012h
		SUMMON	SETVIDEOMODE
@@Init1:		


		; -----------------
		; Initialize things
		; -----------------
		
		; Copy specific values
		push	esi edi
		lea	esi, VSP_640x480x16            		; Source address
		lea	edi, VSP		          	; Destination address
		mov	ecx, size VIDEOSTATEPARAMS        	; Size of the block to move
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update current mode informations
		SETFVAR	D_GNBCOLS, 640
		SETFVAR	D_GNBROWS, 480
		SETFVAR	D_GNBCOLORS, 16
		
		; Clip to all screen
		call	_ClipAllScreen
		
		; Pointer to draw functions
		mov	[_clear_], offset FillVBuff_10
		mov	[_setpix_], offset SetPixel_10
		mov	[_getpix_], offset GetPixel_10
		mov	[_line_], offset Line_10
		mov	[_getvideo_], offset _BLTTORAM_10
		mov	[_putvideo_], offset _BLTTOVBUFF_10
		
		; Initializes a timeout value for detecting the start of the vertical nondisplay interval.
		call	INITVNWAIT
		

		; ----------------
		; Return to caller
		; ----------------

		ret
			
E_M640x480x16:
M640x480x16	endp


; ===============================================================
; = M320x200x256						=
; =								=
; = Code for the froth word:	320x200x256			=
; =								=
; = Stack:  ( --- )						=
; ===============================================================

		align	4
M320x200x256   	proc

		; ---------------------
		; Initialize video mode
		; ---------------------
		
		SUMMON	GETVIDEOMODE
		mov	eax, [esi]
		add	esi, @WordSize
		cmp	eax, 013h
		je	@@Init1
		IPUSH	013h
		SUMMON	SETVIDEOMODE
@@Init1:		


		; -----------------
		; Initialize things
		; -----------------
		
		; Copy specific values
		push	esi edi
		lea	esi, VSP_320x200x256            	; Source address
		lea	edi, VSP		          	; Destination address
		mov	ecx, size VIDEOSTATEPARAMS        	; Size of the block to move
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update current mode informations
		SETFVAR	D_GNBCOLS, 320
		SETFVAR	D_GNBROWS, 200
		SETFVAR	D_GNBCOLORS, 256
		
		; Clip to all screen
		call	_ClipAllScreen
		
		; Pointer to draw functions
		mov	[_clear_], offset FillVBuff_13
		mov	[_setpix_], offset SetPixel_13
		mov	[_getpix_], offset GetPixel_13
		mov	[_line_], offset Line_13
		mov	[_getvideo_], offset _NotImplemented
		mov	[_putvideo_], offset _NotImplemented
		
		; Initializes a timeout value for detecting the start of the vertical nondisplay interval.
		call	INITVNWAIT
		

		; ----------------
		; Return to caller
		; ----------------

		ret
			
E_M320x200x256:
M320x200x256	endp


; ===============================================================
; = M800x600x16							=
; =								=
; = Code for the froth word:	800x600x16			=
; =								=
; = Stack:  ( n --- )						=
; = n	is the BIOS video mode.					=
; ===============================================================

		align	4
M800x600x16   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ---------------------
		; Initialize video mode
		; ---------------------
		
		SUMMON	GETVIDEOMODE
		mov	eax, [esi]
		add	esi, @WordSize
		cmp	eax, [esi]
		je	@@Init1
		SUMMON	SETVIDEOMODE
		jmp	@@Init2
@@Init1:		add	esi, @WordSize
@@Init2:


		; -----------------
		; Initialize things
		; -----------------
		
		; Copy specific values
		push	esi edi
		lea	esi, VSP_800x600x16            		; Source address
		lea	edi, VSP		          	; Destination address
		mov	ecx, size VIDEOSTATEPARAMS        	; Size of the block to move
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update current mode informations
		SETFVAR	D_GNBCOLS, 800
		SETFVAR	D_GNBROWS, 600
		SETFVAR	D_GNBCOLORS, 16
		
		; Clip to all screen
		call	_ClipAllScreen
		
		; Pointer to draw functions
		mov	[_clear_], offset FillVBuff_10
		mov	[_setpix_], offset SetPixel_10
		mov	[_getpix_], offset GetPixel_10
		mov	[_line_], offset Line_10
		mov	[_getvideo_], offset _BLTTORAM_10
		mov	[_putvideo_], offset _BLTTOVBUFF_10
		
		; Initializes a timeout value for detecting the start of the vertical nondisplay interval.
		call	INITVNWAIT
		

		; ----------------
		; Return to caller
		; ----------------

		ret
			
E_M800x600x16:
M800x600x16	endp


; ===============================================================
; = GNOTFILLED                            			=
; =								=
; = Code for the forth word:    NOTFILLED           		=
; =								=
; = Stack:  ( --- n ) 	                			=
; ===============================================================

		align	4
GNOTFILLED	proc

		; Push immediate value onto stack
		IPUSH	FALSE

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_GNOTFILLED:
GNOTFILLED	endp


; ===============================================================
; = GFILLED                            				=
; =								=
; = Code for the forth word:    FILLED           		=
; =								=
; = Stack:  ( --- n ) 	                			=
; ===============================================================

		align	4
GFILLED		proc

		; Push immediate value onto stack
		IPUSH	TRUE

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_GFILLED:
GFILLED		endp



ucPixel	=	byte ptr [ebp + 8]

FillVBuff_10	proc

		enter	0, 0

		; load the specified pixel value into the Set/Reset register
		mov	ah, ucPixel
		clr	al					; AL = 0 (Set/Reset reg #)
		mov	dx, 3CEh				; Graphics Controller I/O addr
		out	dx, ax

		mov	ax, 0F01h				; AH = 1111b (Enable Set/Reset mask)
								; AL = 1 (Enable Set/Reset reg #)
		out	dx, ax

		; fill all 4 bit planes
		mov	ebx, [VSP.vsp_VBuf]			; EBX -> Start of video buffer
		clr	ecx
@@Zero:		mov	dword ptr fs:[ebx + ecx*4], 0
		inc	ecx
		cmp	ecx, 10000h/4
		jne	@@Zero

		; restore the default Enable Set/Reset value
		mov	ax, 1
		out	dx, ax

		leave
		ret	2

FillVBuff_10	endp


FillVBuff_13	proc

		enter	0, 0

		mov	ah, ucPixel
		mov	al, ah
		shl	eax, 16
		mov	ah, ucPixel
		mov	al, ah

		mov	ebx, [VSP.vsp_VBuf]			; EBX -> Start of video buffer
		clr	ecx
@@Zero:		mov	dword ptr fs:[ebx + ecx*4], eax
		inc	ecx
		cmp	ecx, 10000h/4
		jne	@@Zero

		leave
		ret	2

FillVBuff_13	endp


; ===============================================================
; = CLEAR							=
; =								=
; = Code for the froth word:	Clear				=
; =								=
; = Stack:  ( o --- )						=
; ===============================================================

		align	4
CLEAR	   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -------------------------------
		; Get parameters and do operation
		; -------------------------------

		inc	[NoCtrlBreak]			; CTRL-BREAK not allowed
		push	fs
		mov	fs, [_selzero]
		mov	eax, PARAM0
		DISCARD	1				; DROP
		push	ax
		call	[_clear_]
		pop	fs
		
		
		; -----------------------------
		; Set cursor position in (0, 0)
		; -----------------------------

		call	doGetVideoPage			; PAGE#
		mov	[v86r_dh], 0			; Input: row #
		mov	[v86r_dl], 0			; Input: col #
		mov	[v86r_ah], 02h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ----------------
		; Return to caller
		; ----------------

		SUMMON	CTRLBREAK_REST			; CTRL-BREAK allowed
		ret
			
E_CLEAR:
CLEAR		endp


; ===============================================================
; = SELECTGPAGE							=
; =								=
; = Code for the froth word:	SELECTGPAGE			=
; =								=
; = Stack: ( n --- )						=
; ===============================================================

		align	4
SELECTGPAGE	proc

		; -------------------------------------
		; Update starting graphics page address
		; -------------------------------------
	
		mov	ebx, PARAM0
		DISCARD	1
		movzx	eax, [VSP.vsp_PageSz]
		mul	ebx
		add	eax, 0A0000h
		mov	[VSP.vsp_VBuf], eax


		; ----------------
		; Return to caller
		; ----------------

		ret

E_SELECTGPAGE:			
SELECTGPAGE	endp


; ===============================================================
; = CLIP							=
; =								=
; = Code for the froth word:	Clip				=
; =								=
; = Stack:  ( x0 y0 x1 y1  --- )				=
; ===============================================================

		align	4
CLIP	   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	4


		; -------------------------------
		; Get parameters and do operation
		; -------------------------------

		mov	eax, PARAM3				; x0
		mov	[_clip_x0], ax
		
		mov	eax, PARAM2				; y0
		mov	[_clip_y0], ax
		
		mov	eax, PARAM1				; x1
		mov	[_clip_x1], ax
		
		mov	eax, PARAM0				; y1
		mov	[_clip_y1], ax
		
		DISCARD	4
		
		
		; ----------------
		; Return to caller
		; ----------------

		ret
			
E_CLIP:
CLIP		endp


; ===============================================================
; = GMODE_COPY                            			=
; =								=
; = Code for the forth word:    GMODE_COPY           		=
; =								=
; = Stack:  ( --- ) 	                			=
; ===============================================================

		align	4
GMODE_COPY	proc

		; Select given rasted operation
		mov	[VSP.vsp_cRMW], RMW_REPLACE

		; Return to caller
		ret

E_GMODE_COPY:
GMODE_COPY	endp


; ===============================================================
; = GMODE_XOR                            			=
; =								=
; = Code for the forth word:    GMODE_XOR			=
; =								=
; = Stack:  ( --- ) 	                			=
; ===============================================================

		align	4
GMODE_XOR	proc

		; Select given rasted operation
		mov	[VSP.vsp_cRMW], RMW_XOR

		; Return to caller
		ret

E_GMODE_XOR:
GMODE_XOR	endp


; ===============================================================
; = GMODE_AND                            			=
; =								=
; = Code for the forth word:    GMODE_AND			=
; =								=
; = Stack:  ( --- ) 	                			=
; ===============================================================

		align	4
GMODE_AND	proc

		; Select given rasted operation
		mov	[VSP.vsp_cRMW], RMW_AND

		; Return to caller
		ret

E_GMODE_AND:
GMODE_AND	endp


; ===============================================================
; = GMODE_OR                            			=
; =								=
; = Code for the forth word:    GMODE_OR			=
; =								=
; = Stack:  ( --- ) 	                			=
; ===============================================================

		align	4
GMODE_OR	proc

		; Select given rasted operation
		mov	[VSP.vsp_cRMW], RMW_OR

		; Return to caller
		ret

E_GMODE_OR:
GMODE_OR	endp


code32		ends
		end
