; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	HLINE.ASM						!
; 									!
; Description:	Definitions for playing with graphics.			!
; 		(drawing line).						!
; 									!
; List of FROTH definitions defined in this module:			!
;	LINE		LINE						!
;	HLINE		HLINE						!
;	VLINE		VLINE						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		hgr.inc


; Parameters for all SetPixel functions
; -------------------------------------

ucPixel	equ	byte ptr [ebp + 8]
y2	equ	[ebp + 10]
x2	equ	[ebp + 12]
y1	equ	[ebp + 14]
x1	equ	[ebp + 16]


; Temporary variables
; -------------------

LineFn	equ	dword ptr [ebp - 4]
VertIncr equ	word ptr [ebp - 6]
Incr2	equ	word ptr [ebp - 8]
Incr1	equ	word ptr [ebp - 10]			; (used only in Line256)
IncrY	equ	word ptr [ebp - 12]			; (used only in Line256)


include	h_lin_10.asm
include	h_lin_13.asm


; ===============================================================
; = LINE							=
; =								=
; = Code for the froth word:	line				=
; =								=
; = Stack:  ( x1 y1 x2 y2 o --- )				=
; ===============================================================

		align	4
LINE	   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	5


		; -------------
		; Draw the line
		; -------------

		push	fs
		mov	fs, [_selzero]
		mov	eax, [PARAM4]			; Get x1
		mov	ebx, [PARAM3]			; Get y1
		push	ax bx
		mov	eax, [PARAM2]			; Get x2
		mov	ebx, [PARAM1]			; Get y2
		push	ax bx
		mov	eax, [PARAM0]			; Get color
		push	ax
		call	[_line_]
		pop	fs


		; ----------------
		; Return to caller
		; ----------------

@@Exit:		DISCARD	5
		ret

E_LINE:
LINE		endp


; ===============================================================
; = HLINE							=
; =								=
; = Code for the froth word:	hline				=
; =								=
; = Stack:  ( y x1 x2 o --- )					=
; ===============================================================

		align	4
HLINE	   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	4


		; -------------
		; Draw the line
		; -------------

		push	fs
		mov	fs, [_selzero]
		mov	eax, [PARAM2]			; Get x1
		mov	ebx, [PARAM3]			; Get y
		push	ax bx
		mov	eax, [PARAM1]			; Get x2
		push	ax bx
		mov	eax, [PARAM0]			; Get color
		push	ax
		call	[_line_]
		pop	fs


		; ----------------
		; Return to caller
		; ----------------

@@Exit:		DISCARD	4
		ret

E_HLINE:
HLINE		endp


; ===============================================================
; = VLINE							=
; =								=
; = Code for the froth word:	vline				=
; =								=
; = Stack:  ( x y1 y2 o --- )					=
; ===============================================================

		align	4
VLINE	   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	4


		; -------------
		; Draw the line
		; -------------

		push	fs
		mov	fs, [_selzero]
		mov	eax, [PARAM3]			; Get x
		mov	ebx, [PARAM2]			; Get y1
		push	ax bx
		mov	ebx, [PARAM1]			; Get y2
		push	ax bx
		mov	eax, [PARAM0]			; Get color
		push	ax
		call	[_line_]
		pop	fs


		; ----------------
		; Return to caller
		; ----------------

@@Exit:		DISCARD	4
		ret

E_VLINE:
VLINE		endp


code32		ends
		end
