; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	HSPRITE.ASM						!
; 									!
; Description:	Definitions for playing with graphics.			!
; 		(drawing sprite).					!
; 									!
; List of FROTH definitions defined in this module:			!
;	SPRITE			SPRITE					!
;	DEFCOLOR_SPRITE		>color_sprite				!
; 	DEF_SPRITE		>sprite					!
; 	INIT_SPRITE		init_sprite				!
; 	DRAW_SPRITE		draw_sprite				!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		hgr.inc


; ===============================================================
; = SPRITE							=
; =								=
; = Code for the froth word:	SPRITE				=
; =								=
; = Stack:  ( x y --- )						=
; ===============================================================

		align	4
SPRITE	   	proc

		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------

		; Verify stack frame
		VSTACK	2

		; Compute the size of the video buffer
		clr	eax
		push	eax						; Address to buffer to store region
		inc	eax
		push	ax ax						; x y
		mov	eax, PARAM1					; Get width
		push	ax
		mov	eax, PARAM0					; Get height
		push	ax
		call	[_getvideo_]
		mov	[t_videosize], eax

		; Allocate space
		SUMMON	DDUP						; DDUP
		SUMMON	STAR						; *
		add	PARAM0, size sprite
		mov	eax, PARAM0
		mov	[t_videoofs], eax
		mov	eax, [t_videosize]
		add	PARAM0, eax
		SUMMON	RT_BALLOT
		mov	edx, PARAM0
		DISCARD	1

		; Initialize sprite structure
		mov	[edx.sprite_sign], 'SprT'			; Signature
		mov	[edx.sprite_initdone], FALSE			; Sprite not initialized
		mov	eax, PARAM1
		cmp	eax, 1
		jl	@@InvParams
		cmp	eax, SPRITE_XMAX
		jg	@@InvParams
		mov	[edx.sprite_dx], ax				; X size (width)
		mov	eax, PARAM0
		cmp	eax, 1
		jl	@@InvParams
		cmp	eax, SPRITE_YMAX
		jg	@@InvParams
		mov	[edx.sprite_dy], ax				; Y size (height)
		clr	ecx
@@zero:		mov	[edx.sprite_chars + ecx], 0			; Chars defining colors
		inc	ecx
		cmp	cx, SPRITE_COLORSMAX
		jne	@@zero
		mov	eax, [t_videoofs]
		mov	[edx.sprite_videoofs], eax
		mov	eax, [t_videosize]
		mov	[edx.sprite_videosize], eax
		

		; DOES>		(this ends the compilation sequence)
		DISCARD	2
		SUMMON	PDOES
		dd	R_SPRITE

		
		; "Invalid parameters."
@@InvParams:	mov	ah, ERR_INVPARAMS			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError


		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_SPRITE:

		; Return to caller
		ret

		align	4
t_videoofs	dd	?
t_videosize	dd	?

E_SPRITE:	
SPRITE		endp


; ===============================================================
; = DEFCOLOR_SPRITE						=
; =								=
; = Code for the froth word:	DEFCOLOR_SPRITE			=
; =								=
; = Stack:  ( c n1 n2 --- )					=
; ===============================================================

		align	4
DEFCOLOR_SPRITE	proc
		
		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3
		
		
		; -----------------------------
		; Get sprite address and verify
		; -----------------------------
		
		mov	edx, PARAM0
		bound	edx, qword ptr DDA_BOUND
		cmp	[edx.sprite_sign], 'SprT'
		je	@@SpriteOK
		
		; "Not a sprite."
		mov	ah, ERR_SPRITENOTA		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError

@@SpriteOK:	
		

		; -----------------
		; Store given color
		; -----------------
		
		clr	ecx
		dec	ecx
@@FindFree:	inc	ecx
		cmp	[edx.sprite_chars + ecx], 0
		je	@@FoundFree
		cmp	ecx, SPRITE_COLORSMAX
		jl	@@FindFree

		; "Invalid parameters."
		mov	ah, ERR_INVPARAMS		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError

@@FoundFree:	mov	eax, PARAM2			; Get character
		mov	[edx.sprite_chars + ecx], al
		mov	eax, PARAM1			; Get color
		mov	[edx.sprite_colors + ecx], al

		
		; ----------------
		; Return to caller
		; ----------------

		DISCARD	3		
		ret

E_DEFCOLOR_SPRITE:	
DEFCOLOR_SPRITE	endp


; ===============================================================
; = DEF_SPRITE							=
; =								=
; = Code for the froth word:	>sprite				=
; =								=
; = Stack:  ( n addrz addr --- )				=
; ===============================================================

		align	4
DEF_SPRITE	proc

		; --------------
		; Save used regs
		; --------------
		
		push	edi ebp

		
		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3
		
		
		; -----------------------------
		; Get sprite address and verify
		; -----------------------------
		
		mov	ebp, PARAM0
		bound	ebp, qword ptr DDA_BOUND
		cmp	[ebp.sprite_sign], 'SprT'
		je	@@SpriteOK
		
		; "Not a sprite."
		pop	ebp edi				; Clean stack
		mov	ah, ERR_SPRITENOTA		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError

@@SpriteOK:	

		; -------------------
		; Get the line number
		; -------------------
		
		mov	edx, PARAM2
		cmp	edx, 0
		jl	@@InvParam
		cmp	dx, [ebp.sprite_dy]
		jl	@@LineNbOK

		; "Invalid parameters."
@@InvParam:	pop	ebp edi				; Clean stack
		mov	ah, ERR_INVPARAMS		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError

@@LineNbOK:	movzx	eax, [ebp.sprite_dx]
		imul	edx, eax
		lea	edx, [edx + ebp + size sprite]


		; --------------------------
		; Get and test string length
		; --------------------------
		
		mov	edi, PARAM1
		clr	ecx
		dec	ecx
@@TestStrLen:	inc	ecx
		cmp	byte ptr [edi + ecx], 0
		jne	@@TestStrLen
		cmp	cx, [ebp.sprite_dx]
		je	@@StrLenOK
		
		; "Invalid defining string length (not equal to sprite width)."
		pop	ebp edi				; Clean stack
		mov	ah, ERR_SPRITEINVSTRLEN		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError

@@StrLenOK:	

		; ----------------------------
		; Scan and use defining string
		; ----------------------------
		
		clr	ecx
		dec	ecx
@@UseStr:	inc	ecx
		mov	al, [edi + ecx]
		cmpz	al, 0
		je	@@finish
		
		; Search this character into the sprite table
		clr	ebx
		dec	ebx
@@SearchChar:	inc	ebx
		mov	ah, [ebp.sprite_chars + ebx]
		cmpz	ah, 0
		je	@@CharNotDef
		cmp	ah, al
		jne	@@SearchChar
		
		; Get and Store color
		mov	al, [ebp.sprite_colors + ebx]
		mov	[edx], al
		inc	edx
		
		; Continue until end of string is reached
		jmp	@@UseStr
		

		; "Character not defined."
@@CharNotDef:	pop	ebp edi				; Clean stack
		mov	ah, ERR_SPRITEINVCHAR		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
@@finish:	pop	ebp edi
		DISCARD	3
		ret

E_DEF_SPRITE:	
DEF_SPRITE	endp


; ===============================================================
; = INIT_SPRITE							=
; =								=
; = Code for the froth word:	init_sprite			=
; =								=
; = Stack:  ( addr --- )					=
; ===============================================================

		align	4
INIT_SPRITE	proc

t_cRMW		equ	dword ptr [ebp - 4]
t_VideoRegion	equ	dword ptr [ebp - 8]

		; ------------------
		; Verify stack frame
		; ------------------

		VSTACK	1
		enter	8, 0
		

		; -----------------------------
		; Get sprite address and verify
		; -----------------------------
		
		mov	edx, PARAM0
		bound	edx, qword ptr DDA_BOUND
		cmp	[edx.sprite_sign], 'SprT'
		je	@@SpriteOK
		
		; "Not a sprite."
		mov	ah, ERR_SPRITENOTA		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError

@@SpriteOK:	

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; -------------------------------------------------------
		; Save video region which will be used to draw the sprite
		; -------------------------------------------------------

		; Sprite is now initialized
		mov	[edx.sprite_initdone], TRUE

		; Ask for some temp memory
		mov	eax, [edx.sprite_videosize]
		mov	ecx, [CDA_START]
		add	ecx, [CDA_SIZE]
		mov	edx, [DDA_START]
		sub	edx, [DDA_SIZE]
		sub	edx, eax
		cmp	edx, ecx
		jge	@@MemOk

		; Out of memory !
		mov	ah, ERR_NOMEMORY			; Out of memory.
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError
		
@@MemOk:

		; Store x size
		mov	ecx, PARAM0
		movzx	eax, [ecx.sprite_dx]		
		mov	[edx.vr_psBMP.biWidth], eax			; X size (width)

		; Store y size
		movzx	ebx, [ecx.sprite_dy]		
		mov	[edx.vr_psBMP.biHeight], ebx			; Y size (height)
		
		; Save region
		mov	[t_VideoRegion], edx
		push	edx						; Address to buffer to store region
		clr	cx
		push	cx cx						; x y
		push	ax bx						; width height
		call	[_getvideo_]
		
		
		; -------------------
		; Draw now the sprite
		; -------------------

		; Select given rasted operation
		movzx	eax, [VSP.vsp_cRMW]
		mov	[t_cRMW], eax					; Save previous value
		mov	[VSP.vsp_cRMW], RMW_REPLACE

		; Draw sprite pixels
		push	fs esi edi
		mov	fs, [_selzero]
		mov	esi, PARAM0
		lea	edi, [esi + size sprite]
		clr	eax						; y
		dec	eax
@@NextRow:	inc	eax
		cmp	ax, [esi.sprite_dy]
		je	@@done
		clr	ebx
		dec	ebx
@@NextCol:	inc	ebx
		cmp	bx, [esi.sprite_dx]
		je	@@NextRow
		movzx	ecx, byte ptr [edi]
		push	ax bx
		push	bx ax cx					; x y color
		call	[_setpix_]					; draw the pixel
		pop	bx ax
		inc	edi
		jmp	@@NextCol
@@done:		pop	edi esi fs

		; Save the sprite video region
		mov	edx, PARAM0
		add	edx, [edx.sprite_videoofs]			
		push	edx						; Address to buffer to store region;
		clr	cx
		push	cx cx						; x y
		mov	edx, PARAM0
		mov	ax, [edx.sprite_dx]				; dx
		mov	bx, [edx.sprite_dy]				; dy
		push	ax bx						; dx dy
		call	[_getvideo_]


		; ---------------------------------
		; Restore the previous video region
		; ---------------------------------

		; Restore previous rasted operation
		mov	eax, [t_cRMW]
		mov	[VSP.vsp_cRMW], al

		; Restore region		
		clr	ax
		push	ax ax						; x y
		mov	edx, [t_VideoRegion]
		push	edx						; Address to buffer to store region;
		clr	ax
		push	ax						; flag
		call	[_putvideo_]


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST
		

		; ----------------
		; Return to caller
		; ----------------

		DISCARD	1
		leave
		ret

E_INIT_SPRITE:	
INIT_SPRITE	endp


; ===============================================================
; = DRAW_SPRITE							=
; =								=
; = Code for the froth word:	draw_sprite			=
; =								=
; = Stack:  ( x y addr --- )					=
; ===============================================================

		align	4
DRAW_SPRITE	proc
		
		; ------------------
		; Verify stack frame
		; ------------------

		VSTACK	3
		
		
		; -----------------------------
		; Get sprite address and verify
		; -----------------------------
		
		mov	edx, PARAM0
		bound	edx, qword ptr DDA_BOUND
		cmp	[edx.sprite_sign], 'SprT'
		je	@@SpriteOK
		
		; "Not a sprite."
		mov	ah, ERR_SPRITENOTA		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError

@@SpriteOK:	


		; -----------------------
		; Is sprite initialized ?
		; -----------------------
		
		cmp	[edx.sprite_initdone], TRUE
		je	@@SpriteInit
		
		; "SPRITE: Not initialized."
		mov	ah, ERR_SPRITENOTINIT		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError

@@SpriteInit:	
		

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ---------------
		; Draw the sprite
		; ---------------

		mov	eax, PARAM2				; x
		mov	ebx, PARAM1				; y
		push	ax bx					; x y
		add	edx, [edx.sprite_videoofs]			
		push	edx					; Address to buffer to store region;
		clr	ax
		push	ax					; flag
		call	[_putvideo_]


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST
		

		; ----------------
		; Return to caller
		; ----------------

		DISCARD	3
		ret

E_DRAW_SPRITE:	
DRAW_SPRITE	endp


code32		ends
		end
