; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	HPIXEL.ASM						!
; 									!
; Description:	Definitions for playing with graphics			!
; 		(video functions).					!
; 									!
; List of FROTH definitions defined in this module:			!
;	VIDEOREGION	VideoRegion					!
;	GETVIDEO	GetVideo					!
;	PUTVIDEO	PutVideo					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		hgr.inc


; ===============================================================
; = VIDEOREGION							=
; =								=
; = Code for the froth word:	VIDEOREGION			=
; =								=
; = Stack:  ( x y --- )						=
; ================================================================

		align	4
VIDEOREGION   	proc

		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------
		
		; Verify stack frame
		VSTACK	2

		; Compute space
		clr	eax
		push	eax					; Address to buffer to store region
		inc	eax
		push	ax					; x
		push	ax					; y
		mov	eax, PARAM1				; Get width
		push	ax
		mov	eax, PARAM0				; Get height
		push	ax
		call	[_getvideo_]

		; Allocate space
		sub	esi, @WordSize
		mov	PARAM0, eax
		SUMMON	RT_BALLOT
		mov	edx, PARAM0
		DISCARD	1

		; Store x size
		mov	eax, PARAM1					; x-size
		cmp	eax, 1
		jl	@@InvParams
		GETFVAR	ebx, D_GNBCOLS
		cmp	eax, ebx
		jg	@@InvParams
		mov	[edx.vr_psBMP.biWidth], eax			; X size (width)

		; Store y size
		mov	eax, PARAM0
		cmp	eax, 1
		jl	@@InvParams
		GETFVAR	ebx, D_GNBROWS
		cmp	eax, ebx
		jg	@@InvParams
		mov	[edx.vr_psBMP.biHeight], eax			; Y size (height)

		; DOES>		(this ends the compilation sequence)
		DISCARD	2
		SUMMON	PDOES
		dd	R_VIDEOREGION

		
		; "Invalid parameters."
@@InvParams:	mov	ah, ERR_INVPARAMS		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_VIDEOREGION:

		; Return to caller
		ret

E_VIDEOREGION:	
VIDEOREGION	endp


; ===============================================================
; = GETVIDEO							=
; =								=
; = Code for the froth word:	GETVIDEO			=
; =								=
; = Stack:  ( x y addr --- )					=
; ===============================================================

		align	4
GETVIDEO   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; -------------
		; Call function
		; -------------
		
		mov	edx, PARAM0				; Address to buffer to store region
		push	edx
		mov	eax, PARAM2				; Pop x
		mov	ebx, PARAM1				; Pop y
		push	ax bx
		mov	eax, [edx.vr_psBMP.biWidth]		; Get width
		mov	ebx, [edx.vr_psBMP.biHeight]		; Get height
		push	ax bx			
		call	[_getvideo_]


		; ----------------
		; Return to caller
		; ----------------

		DISCARD	3		
		ret

E_GETVIDEO:	
GETVIDEO		endp


; ===============================================================
; = PUTVIDEO							=
; =								=
; = Code for the froth word:	PUTVIDEO			=
; =								=
; = Stack:  ( x y addr --- )					=
; ===============================================================

		align	4
PUTVIDEO   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; -------------
		; Call function
		; -------------
		
		mov	eax, PARAM2				; Pop x
		push	ax
		mov	eax, PARAM1				; Pop y
		push	ax
		mov	edx, PARAM0				; Address to buffer to store region
		push	edx
		clr	ax
		push	ax					; flag
		call	[_putvideo_]


		; ----------------
		; Return to caller
		; ----------------

		DISCARD	3		
		ret

E_PUTVIDEO:	
PUTVIDEO		endp



; PackPixels10
;
; At entry: BL = bits from plane 0
;	    BH = bits from plane 1
;	    DL = bits from plane 2
;	    DH = bits from plane 3
;	    CL = number of pixels to pack (1-4)
; Leaves:   AX = packed pixels
; Destroys: CH

PackPixels10	macro
		LOCAL	M000

		clr	ax
		mov	ch, cl

M000:		rol	dh, 1			; cf = bit from plane 3
		rcl	ax, 1
		rol	dl, 1			; cf = bit from plane 2
		rcl	ax, 1
		rol	bh, 1			; cf = bit from plane 1
		rcl	ax, 1
		rol	bl, 1			; cf = bit from plane 0
		rcl	ax, 1
		dec	cl
		jnz	M000

		; rotate packed pixels into position in AX
		mov	cl, ch
		add	cl, 2
		and	cl, 3
		shl	cl, 2
		ror	ax, cl

		endm


; UnpackPixels10
;
; At entry: ESI -> packed 4-bit pixels
; Leaves:   ESI -> next 8 pixels
;	    BL = bits for plane 0
;	    BH = bits for plane 1
;	    DL = bits for plane 2
;	    DH = bits for plane 3
; Destroys: AX, CX

UnpackPixels10	macro
		LOCAL	M100, M101

		mov	ch, 2	    		; outer-loop limit

M100:		mov	cl, 4	    		; inner-loop limit
		lodsw
		xchg	ah, al	    		; AX = 4 pixels

M101:		shl	ax, 1
		rcl	dh, 1	    		; update bits for plane 3
		shl	ax, 1
		rcl	dl, 1	    		; update bits for plane 2
		shl	ax, 1
		rcl	bh, 1	    		; update bits for plane 1
		shl	ax, 1
		rcl	bl, 1	    		; update bits for plane 0
		dec	cl
		jnz	M101

		dec	ch
		jnz	M100

		endm


; parameters for all BltToRAM functions
h	equ	word ptr [ebp + 8]
w	equ	word ptr [ebp + 10]
y	equ	word ptr [ebp + 12]
x	equ	word ptr [ebp + 14]
pBMP	equ	dword ptr [ebp + 16]

; BltToRAM_10
;
;   ULONG VS2API
;   BltToRAM10( LPVOID pBMP, USHORT x, USHORT y, USHORT w, USHORT h );
;
; Notes:
;   Parameters x,y specify the upper left corner of the bitmap;
;   w,h specify the width and height in pixels.
;
;   No pixel data is copied if the specified buffer pointer is null.
;
;   Returns the actual size of the bitmap.

ofsRow	equ	dword ptr [ebp - 4]
RowIncr	equ	word ptr  [ebp - 6]
PixRot	equ	byte ptr  [ebp - 7]
nPad	equ	byte ptr  [ebp - 8]
cbLow	equ	word ptr  [ebp - 10]
cbHigh	equ	word ptr  [ebp - 12]

		align	4
_BLTTORAM_10	proc

		enter	12, 0					; Room for local variables (10 bytes)
		push	esi edi fs
		mov	fs, [_selzero]

		; compute the size of the pixel data in each row
		mov	ax, w	    				; AX = pixels per row
		inc	ax
		shr	ax, 1	    				; AX = bytes of pixel data per row
		mov	si, ax	    				; SI = bytes of pixel data per row
		add	ax, 3
		and	al, 0FCh    				; round up to nearest multiple of 4

		mov	bx, ax
		sub	bx, si	    				; BX = # of bytes of padding
		shr	bx, 1	    				; BX = # of words of padding
		mov	nPad, bl

		; compute the size of the bitmap
		mul	h	    				; DX:AX = # bytes of bitmap data
		mov	bx, ax
		mov	cx, dx	    				; CX:BX = # bytes of bitmap data
	
		add	ax, size BITMAPINFOHEADER
		adc	dx, 0
		add	ax, 16*4    				; size of 16-color table
		adc	dx, 0
	
		mov	cbLow, ax
		mov	cbHigh, dx

		; verify that the specified buffer pointer is not null
		mov	edi, pBMP    				; EDI -> BMP in RAM
		cmpz	edi, 0
		je	L090

		; initialize the BITMAPINFOHEADER data structure
L000:		cld
		mov	ax, size BITMAPINFOHEADER
		stosw
		clr	ax
		stosw		    			; biSize
	
		mov	ax, w
		stosw	
		clr	ax
		stosw		    			; biWidth
	
		mov	ax, h
		stosw
		clr	ax
		stosw			  		; biHeight
	
		mov	ax, 1	    			; biPlanes (always 1 according to Microsoft's documentation)
		stosw
	
		mov	ax, 4	    			; biBitCount
		stosw

		mov	ax, BI_RGB  			; (no data compression)
		stosw
		clr	ax
		stosw		    			; biCompression
	
		mov	ax, bx
		stosw
		mov	ax, cx
		stosw	    				; biSizeImage
	
		mov	ecx, 8
		clr	ax
		rep	stosw				; biXPelsPerMeter, biYPelsPerMeter,
				    			;  biClrUsed, biClrImportant

		; build bit masks for converting palette-register values to DAC indexes
		mov	al, 14h     				; Pixel Padding register number
		call	_GetATCReg   				; AL = current Pixel Padding register value
		mov	bl, al
		shl	bl, 4	    				; move bits 2-3 into bits 6-7
		mov	bh, 0Fh     				; BH = AND mask for 4-bit palette-reg values
	
		mov	al, 10h     				; Mode Control register number
		call	_GetATCReg   				; AL = current Mode Control reg value
		test	al, 80h     				; test bit 7
		jnz	L010	    				; jump if 4-bit palette-register values

		; come here if using 6-bit palette-register values
		mov	bh, 3Fh     				; BH = AND mask for 6-bit palette-reg values
		and	bl, 0C0h    				; BL = OR mask for 6-bit palette-reg values

		; copy the current color values from the video DAC
L010:		mov	cx, 16	    				; CX = # of palette registers

		call	VNWAIT	    				; wait for vertical nondisplay interval

L011:		mov	al, cl
		neg	al
		and	al, 0Fh     				; AL = register number (00h-0Fh)
		call	_GetATCReg   				; AL = palette register value
		and	al, bh	    				; isolate low-order bits
		or	al, bl	    				; set high-order bits
	
		call	_GetDACRGB   				; store RGB quad at ES:[DI]
		add	edi, 4	    				; point to next RGB quad in color table
		loopw	L011

		; verify that the bitmap is not null
		mov	ax, h
		mul	w	    				; DX:AX = w*h
		or	ax, dx
		jnz	L015
		jmp	L090	    				; exit if empty bitmap

		; compute the number of bits to shift the pixel data
L015:		mov	cx, x
		and	cl, 7	    				; CL = x MOD 8
		mov	PixRot, cl

		; save the number of bytes per row in the video buffer
		mov	ax, VSP.vsp_cbRow
		mov	RowIncr, ax

		; point to the first pixel
		push	edi
		push	x
		mov	ax, y
		add	ax, h
		dec	ax	    				; AX = y-coordinate of bottommost row
		push	ax
		call	PixelAddr_10
		mov	esi, edi					; ESI -> lower-left pixel
		pop	edi
		mov	ofsRow, esi

		; come here at the start of each row of pixels
L020:		push	w

		; get pixels from the next byte
L025:		mov	cl, PixRot  				; CL = # of bits to rotate left
	
		mov	dx, 3CEh    				; DX = Graphics Controller I/O addr
		mov	ax, 304h    				; select bit plane 3 for CPU reads
		out	dx, ax
	
		mov	bx, fs:[esi]    				; BX = bits from bit plane 3
		rol	bx, cl
		mov	ch, bl	    				; CH = rotated bits from plane 3
	
		dec	ah	    				; select bit plane 2
		out	dx, ax

		mov	bx, fs:[esi]    				; BX = bits from bit plane 2
		rol	bx, cl	   				; BL = rotated bits from plane 2
		mov	bh, ch	   				; BH = rotated bits from plane 3
		push	bx
	
		dec	ah	   				; select bit plane 1
		out	dx, ax
	
		mov	bx, fs:[esi]    				; BX = bits from bit plane 1
		rol	bx, cl
		mov	ch, bl	   				; CH = rotated bits from plane 1
	
		dec	ah	   				; select bit plane 0
		out	dx, ax
	
		mov	bx, fs:[esi]    				; BX = bits from bit plane 0
		rol	bx, cl	   				; BL = rotated bits from plane 0
		mov	bh, ch	   				; BH = rotated bits from plane 1
		pop	dx	   				; DL = rotated bits from plane 2
				   				; DH = rotated bits from plane 3
	
		inc	esi	   				; ESI -> next byte

		; do while 8 or more pixels remain in the row
		pop	cx
		sub	cx, 8
		js	L030
		push	cx

		; serialize the 8 pixels
		mov	cl, 4
		PackPixels10
		stosw
		mov	cl, 4
		PackPixels10
		stosw
		jmp	L025

		; come here when fewer than 8 pixels remain
L030:		add	cx, 8
		cmp	cx, 4
		jb	L031	    				; jump if fewer than 4 pixels remain
	
		push	cx
		mov	cl, 4
		PackPixels10
		stosw

		pop	cx
		sub	cx, 4

		; come here when fewer than 4 pixels remain
L031:		jcxz	L032	    				; jump if no pixels remain

		PackPixels10
		stosw

L032:		cmp	nPad, 0
		jz	L033	    				; jump if no padding needed

		clr	ax
		stosw		    				; add a word of padding

		; come here when an entire row of pixels has been copied
L033:		dec	h
		jz	L090	    				; exit if no more rows to copy

		; point to the first pixel in the next row
		mov	esi, ofsRow
		movzx	eax, RowIncr
		sub	esi, eax
		mov	ofsRow, esi
		jmp	L020

		; restore registers and exit
L090:		pop	fs edi esi
		movzx	eax, cbHigh
		shl	eax, 16
		mov	ax, cbLow
		leave
		ret	12	   				; sizeof(LPVOID)+4*sizeof(USHORT)

_BLTTORAM_10	endp


; GetATCReg
;
; Caller:   AL = register number
; Returns:  AL = current value of register

		align	4
_GetATCReg	proc

		push	dx
	
		mov	ah, al	   				; AH = register number
		mov	dx, 3DAh    				; CRT Status register I/O addr
		cli
		in	al, dx	   				; reset the Attribute Controller flip-flop
		mov	dl, 0C0h    				; DX = 3C0h (Attribute Controller I/O addr)
		mov	al, ah
		out	dx, al
		inc	dx	   				; DX = 3C1h
		in	al, dx	   				; AL = requested register value
		sti
	
		push	ax
		mov	dl, 0DAh
		in	al, dx	   				; reset the flip-flop again
		mov	dl, 0C0h
		mov	al, 20h
		out	dx, al	   				; set bit 5 (PAS bit) to keep
		pop	ax	   				;  attribute decoding enabled
	
		pop	dx
		ret
	
_GetATCReg	endp

; GetDACRGB
;
; Caller:   AL = DAC register number
;	    ES:DI -> buffer

		align	4
_GetDACRGB	proc

		push	dx
	
		mov	ah, al	   				; AH = DAC register number
		mov	dx, 3C7h    				; DX = 3C7h (DAC read I/O address)
		mov	al, ah
		out	dx, al
	
		mov	dl, 0C9h    				; DX = 3C9h (DAC data register I/O addr)
		in	al, dx	   				; AL = red value
		clr	ah	   				; AH = 0
		shl	ax, 2	   				; convert 6-bit values to 8 bits
		mov	[edi + 2], ax				; store red value and null
		in	al, dx
		mov	ah, al	   				; AH = green value
		in	al, dx	   				; AL = blue value
		shl	ax, 2	   				; convert 6-bit values to 8 bits
		mov	[edi], ax
	
		pop	dx
		ret
	
_GetDACRGB	endp



; parameters for all BltToVBuf functions
flag	equ	word ptr  [ebp + 8]
pBMP	equ	dword ptr [ebp + 10]
y	equ	word ptr  [ebp + 14]
x	equ	word ptr  [ebp + 16]

; BltToVBuf10
;
;   void VS2API
;   BltToVBuf10( USHORT x, USHORT y, LPVOID pBMP, USHORT flag );
;
; Notes:
;   If the flag is nonzero, this routine updates the Attribute
;   Controller palette registers with the color-table values and zeroes
;   Mode Control register bit 7.  (Bits 2-3 of the Color Select register
;   are not modified.)

ofsRow	equ	dword ptr [ebp - 4]
RowIncr	equ	word ptr  [ebp - 6]
PixShr	equ	word ptr  [ebp - 8]
w	equ	word ptr  [ebp - 10]
h	equ	word ptr  [ebp - 12]
LMask	equ	word ptr  [ebp - 14]
RMask	equ	word ptr  [ebp - 16]
Bits3	equ	byte ptr  [ebp - 17]
Bits2	equ	byte ptr  [ebp - 18]
Bits1	equ	byte ptr  [ebp - 19]
Bits0	equ	byte ptr  [ebp - 20]

_BLTTOVBUFF_10	proc

		enter	20, 0
		push	esi edi fs
		mov	fs, [_selzero]

		; configure the Graphics Controller for write mode 0, read mode 1
		mov	dx, 3CEh    			; DX = Graphics Controller I/O addr
		mov	ax, 805h    			; AH = 8; AL = Mode register number
		out	dx, ax
	
		mov	ah, VSP.vsp_cRMW			; AH = Read-Modify-Write bits
		mov	al, 3				; AL = Data Rotate/Func Select reg #
		out	dx, ax
	
		mov	ax, 7				; AH = 0; AL = Color Don't Care reg number
		out	dx, ax

		; save the number of bytes per row
		mov	ax, VSP.vsp_cbRow
		mov	RowIncr, ax

		; compute the address of the first pixel
		push	x
		push	y
		call	PixelAddr_10			; EDI -> upper-left pixel

		; get the width, height, and number of colors in the color table
		mov	esi, pBMP
		mov	cx, WORD PTR (BITMAPINFOHEADER PTR [esi]).biWidth
		mov	w, cx
		mov	ax, WORD PTR (BITMAPINFOHEADER PTR [esi]).biHeight
		mov	h, ax
		mov	cx, WORD PTR (BITMAPINFOHEADER PTR [esi]).biClrUsed

		; point to the last row of the destination
		dec	ax	    			; height - 1
		mul	RowIncr
		movzx	eax, ax
		add	edi, eax				; EDI -> lower-left pixel
		mov	[ofsRow], edi

		; verify the number of entries in the color table
		or	cx, cx
		jnz	L300

		; use the number of bits per pixel to compute the number of colors used
		mov	cx, (BITMAPINFOHEADER PTR [esi]).biBitCount
		mov	ax, 1
		shl	ax, cl
		mov	cx, ax	    			; CX = number of colors

		; point past the BITMAPINFOHEADER to the color table
L300:		add	esi, size BITMAPINFOHEADER

		; if the specified flag is zero, don't update the palette registers
		cmp	flag, 0
		jnz	L301

		; come here to skip over the color table
		shl	cx, 2	    			; CX = size of color table
		movzx	eax, cx
		add	esi, eax	    			; ESI -> bitmap data
		jmp	L310

		; come here to update palette registers
L301:		mov	al, 10h     			; AL = 10h (Mode Control reg number)
		call	_GetATCReg   			; AL = current value of Mode Control reg
		mov	ah, al
		and	ah, 7Fh     			; AH = new Mode Control reg value
				   			; (bit 7 = 0 for 6-bit palette values)

		; set palette registers 0-0Fh with 6-bit values 0-0Fh
		mov	dx, 3C0h    			; DX = 3C0h (Attribute Controller I/O addr)
		mov	al, 0Fh
		call	VNWAIT	   			; wait for vertical nondisplay interval

L303:		cli
		out	dx, al	    			; select a palette register
		out	dx, al	    			; update the register
		sti
		dec	al
		jns	L303

		; update the Mode Control register
		mov	al, 30h     			; AL bits 0-4 = 10h (Mode Control reg #)
		cli		   			; AL bit 5 = 1 (Palette Address Source bit)
		out	dx, al
		mov	al, ah	   			; AL = value for Mode Control register
		out	dx, al
		sti

		; set DAC registers 0-15 with the values from the BMP color table
		mov	dl, 0C8h    			; DX = 3C8h (DAC write I/O address)
		clr	al
	
		cli
		out	dx, al
		inc	dx	   			; DX = 3C9h (DAC data register I/O addr)

L304:		lodsw
		mov	bx, ax	    			; BL = blue; BH = green
		lodsw		    			; AL = red
		shr	al, 2	    			; convert 8 bits to 6 bits
		out	dx, al
		mov	al, bh	    			; AL = green
		shr	al, 2	    			; convert 8 bits to 6 bits
		out	dx, al
		mov	al, bl	    			; AL = blue
		shr	al, 2	    			; convert 8 bits to 6 bits
		out	dx, al
		loopw	L304
		sti

		; compute a bit mask for the rightmost pixels
L310:		mov	ax, 0FFh
		mov	cx, x
		add	cx, w
		dec	cx	    			; CL = (x+w-1)
		and	cx, 7	    			; CL = (x+w-1) MOD 8
		inc	cx	    			; CL = (x+w) MOD 8, with 0 mapped to 8
		ror	ax, cl	    			; AH = mask
		mov	al, 8	    			; AL = Bit Mask register number
		mov	RMask, ax

		; compute the number of bits to shift pixels rightward
		movzx	ecx, x
		and	cx, 7	    			; CX = x MOD 8
		mov	PixShr, cx

		; compute a bit mask for the leftmost pixels
		mov	ah, 0FFh
		shr	ah, cl	   			; AH = mask
		mov	LMask, ax   			; store mask and Bit Map register number
	
		add	cx, w
		cmp	cx, 8
		ja	L330	   			; jump if at least 2 bytes of pixels per row

		; come here if only one byte of pixels per row
		and	RMask, ax   			; RMask = RMask & LMask

L320:		mov	ax, RMask
		mov	dx, 3CEh    			; DX = Graphics Controller I/O addr
		out	dx, ax

		UnpackPixels10
		call	_WritePixels_10

		dec	h
		jnz	L321
		jmp	L390	   			; exit if no more rows

		; advance to the next row
L321:		mov	edi, ofsRow
		movzx	eax, RowIncr
		sub	edi, eax
		mov	ofsRow, edi
		jmp	L320

		; come here if 2 or more bytes of pixels per row
L330:		push	w	   			; push w (number of pixels per row)

		; set the leftmost pixels
		mov	ax, LMask
		mov	dx, 3CEh    			; DX = Graphics Controller I/O addr
		out	dx, ax
	
		UnpackPixels10
		call	_WritePixels_10

		; compute the number of pixels remaining in the row
		pop	cx
		add	cx, PixShr
		sub	cx, 8	   			; CX = w-(8-PixShr) = # of pixels remaining
	
		; enable all 8 pixel positions
		mov	dx, 3CEh    			; DX = Graphics Controller I/O addr
		mov	ax, 0FF08h  			; AH = 0FFh; AL = Bit Mask register number
		out	dx, ax

		; set all but the rightmost pixels
L331:		cmp	cx, 8
		jb	L332	   			; jump if fewer than 8 pixels remain

		push	cx
		UnpackPixels10
		call	_WritePixels_10
		pop	cx
	
		sub	cx, 8
		jmp	L331

		; set the rightmost pixels
L332:		jcxz	L334	    			; jump if no pixels remain

		mov	ax, RMask
		mov	dx, 3CEh				; DX = Graphics Controller I/O addr
		out	dx, ax
	
		cmp	cx, PixShr
		jbe	L333				; jump if all pixels are already unpacked

		UnpackPixels10

L333:		call	_WritePixels_10

L334:		dec	h
		jz	L390	    			; exit if no more rows

		; advance to the next row
		mov	edi, ofsRow
		movzx	eax, RowIncr
		sub	edi, eax
		mov	ofsRow, edi
		jmp	L330

		; restore the default Sequencer and Graphics Controller states
L390:		mov	dx, 3C4h
		mov	ax, 0F02h   			; Map Mask
		out	dx, ax
	
		mov	dl, 0CEh    			; DX = 3CEh
		mov	ax, 3	   			; Data Rotate/Function Select
		out	dx, ax
	
		mov	al, 5	   			; Mode register
		out	dx, ax
	
		mov	ax, 0FF08h  			; Bit Mask
		out	dx, ax

		; restore registers and exit
		pop	fs edi esi
		leave
		ret	10	   			; 3*sizeof(USHORT)+sizeof(LPVOID)

_BLTTOVBUFF_10	endp



; WritePixels10
;
; At entry:  ESI -> 8 packed 4-bit pixels
;	    EDI -> video buffer
; Leaves:    ESI -> next 8 packed pixels
;	    EDI -> next byte in video buffer
; Destroys: AX, CX

_WritePixels_10	proc

		mov	cx, PixShr  			; CL = # of bits to shift right
	
		push	dx
		mov	dx, 3C4h    			; DX = Sequencer I/O addr
		mov	ax, 802h    			; AH = mask for bit plane 3
		out	dx, ax
		pop	dx
	
		mov	ah, Bits3
		mov	al, dh
		mov	Bits3, dh   			; save new bits for plane 3
		shr	ax, cl			
		and	fs:[edi], al 			; update bit plane 3

		push	dx
		mov	dx, 3C4h
		mov	ax, 402h    			; AH = mask for bit plane 2
		out	dx, ax
		pop	dx
	
		mov	ah, Bits2
		mov	al, dl
		mov	Bits2, dl   			; save new bits for plane 2
		shr	ax, cl
		and	fs:[edi], al 			; update bit plane 2
	
		mov	dx, 3C4h
		mov	ax, 202h    			; AH = mask for bit plane 1
		out	dx, ax
	
		mov	ah, Bits1
		mov	al, bh
		mov	Bits1, bh   			; save new bits for plane 1
		shr	ax, cl
		and	fs:[edi], al 			; update bit plane 1
	
		mov	ax, 102h    			; AH = mask for bit plane 0
		out	dx, ax
	
		mov	ah, Bits0
		mov	al, bl
		mov	Bits0, bl   			; save new bits for plane 0
		shr	ax, cl
		and	fs:[edi], al 			; update bit plane 0

		inc	edi	   			; point to the next byte in video buffer
		ret

_WritePixels_10	endp


code32		ends
		end
