; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	MEMORY.ASM						!
; 									!
; Description:	Definitions for playing with the memory.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	MEMFREE		MEMFREE						!
;	[@]		RAT						!
;	(@)		PAT						!
;	@		AT						!
;	[C@]		RCAT						!
;	(C@)		PCAT						!
;	C8@		CAT						!
;	[W@]		RWAT						!
;	(W@)		PWAT						!
;	W@		WAT						!
;	!		STORE						!
;	(!)		PSTORE						!
;	C!		CSTORE						!
;	(C!)		PCSTORE						!
;	W!		WSTORE						!
;	(W!)		PWSTORE						!
;	@!		ATSTORE						!
;	C@SWAP		CATWAP						!
;	W@SWAP		WATSWAP						!
;	@SWAP		ATSWAP						!
;	@==		ATEQU						!
;	@!=		ATNEQU						!
;	@0==		ATZEROEQU					!
;	@0!=		ATZERONEQU					!
;	0!		ZEROSTORE					!
;	1!		ONESTORE					!
;	2!		TWOSTORE					!
;	-1!		MINUSONESTORE					!
;	-2!		MINUSTWOSTORE					!
;	+!		PLUSSTORE					!
;	1+!		ONEPLUSSTORE					!
;	2+!		TWOPLUSSTORE					!
;	-!		MINUSSTORE					!
;	1-!		ONEMINUSSTORE					!
;	2-!		TWOMINUSSTORE					!
;	*!		STARSTORE					!
;	2*!		TWOSTARSTORE					!
;	/!		SLASHSTORE					!
;	2/!		TWOSLASHSTORE					!
;	?		QUEST						!
;	C?		CQUEST						!
;	W?		WQUEST						!
;	DYNAMIC		DYNAMIC						!
;	ALLOC		ALLOCATE					!
;	FREE		FREE						!
;	?AllotFail	QALLOTFAIL					!
;	ALLOT		ALLOT						!
;	AllotMore	ALLOTMORE					!
;	RT_DOES		PDOES						!
;	DOES>		DOES						!
;	DataSalvage	DATASALVAGE					!
;	DUST		DUST						!
;	DUSTTHRESHOLD	DUSTTHRESHOLD					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions

		
		; Used in: ALLOT and DOES
		extrn	InterpretAddr:	proc		; See INTERPRET


; ===============================================================
; = MEMFREE							=
; =                                                 		=
; = Code for the forth word:	MemFree				=
; =                     	                            	=
; = Stack:	( --- n )					=
; ===============================================================

		align	4
MEMFREE		proc

		; Compute the free memory available into (code + data) dictionnary
		mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		mov	ebx, [DDA_START]
		sub	ebx, [DDA_SIZE]
		sub	ebx, eax
		IPUSH	ebx

		; Verify stack frame overflow, then return to caller
		OVSTACK
		
		; Return to caller
		ret

E_MEMFREE:
MEMFREE		endp


; ===============================================================
; = RAT								=
; =                                                 		=
; = Code for the forth word:	[@]				=
; =								=
; = Stack:	( addr --- n )					=
; ===============================================================

		align	4
RAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push
		mov	eax, PARAM0
		push	fs
		mov	fs, [_selzero]
		mov	eax, fs:[eax]
		pop	fs
		mov	PARAM0, eax

		; Return to caller
		ret
E_RAT:
RAT		endp


; ===============================================================
; = PAT								=
; =                                                 		=
; = Code for the forth word:	(@)				=
; =								=
; = Stack:	( addr --- n )					=
; ===============================================================

		align	4
PAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push
		mov	eax, PARAM0
		mov	eax, [eax]
		mov	PARAM0, eax

		; Return to caller
		ret
E_PAT:
PAT		endp


; ===============================================================
; = AT								=
; =                                                 		=
; = Code for the forth word:	@				=
; =								=
; = Stack:	( addr --- n )					=
; ===============================================================

		align	4
AT		proc

		; Verify stack frame
		VSTACK	1

		; Get addr
		mov	eax, PARAM0
		
		; Check addr against bound
		bound	eax, qword ptr DDA_BOUND
		
		; Read value then push
		mov	eax, [eax]
		mov	PARAM0, eax

		; Return to caller
		ret

E_AT:
AT		endp


; ===============================================================
; = RCAT							=
; =                                                 		=
; = Code for the forth word:	[C@]				=
; =								=
; = Stack:	( adr --- b )					=
; ===============================================================

		align	4
RCAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push		
		mov	eax, PARAM0
		push	fs
		mov	fs, [_selzero]
		movzx	eax, byte ptr fs:[eax]
		pop	fs
		mov	PARAM0, eax

		; Return to caller
		ret

E_RCAT:
RCAT		endp


; ===============================================================
; = PCAT							=
; =                                                 		=
; = Code for the forth word:	(C@)				=
; =								=
; = Stack:	( adr --- b )					=
; ===============================================================

		align	4
PCAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push		
		mov	eax, PARAM0
		movzx	eax, byte ptr [eax]
		mov	PARAM0, eax

		; Return to caller
		ret

E_PCAT:
PCAT		endp


; ===============================================================
; = CAT								=
; =                                                 		=
; = Code for the forth word:	C@				=
; =								=
; = Stack:	( adr --- b )					=
; ===============================================================

		align	4
CAT		proc

		; Verify stack frame
		VSTACK	1

		; Read addr
		mov	eax, PARAM0

		; Check addr against bound
		bound	eax, qword ptr DDA_BOUND

		; Read value then push		
		movzx	eax, byte ptr [eax]
		mov	PARAM0, eax

		; Return to caller
		ret

E_CAT:
CAT		endp


; ===============================================================
; = RWAT							=
; =                                                 		=
; = Code for the forth word:	[W@]				=
; =								=
; = Stack:	( adr --- w )					=
; ===============================================================

		align	4
RWAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push
		mov	eax, PARAM0
		push	fs
		mov	fs, [_selzero]
		movzx	eax, word ptr fs:[eax]
		pop	fs
		mov	PARAM0, eax

		; Return to caller
		ret

E_RWAT:
RWAT		endp


; ===============================================================
; = PWAT							=
; =                                                 		=
; = Code for the forth word:	(W@)				=
; =								=
; = Stack:	( adr --- w )					=
; ===============================================================

		align	4
PWAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push
		mov	eax, PARAM0
		movzx	eax, word ptr [eax]
		mov	PARAM0, eax

		; Return to caller
		ret

E_PWAT:
PWAT		endp


; ===============================================================
; = WAT								=
; =                                                 		=
; = Code for the forth word:	W@				=
; =								=
; = Stack:	( adr --- w )					=
; ===============================================================

		align	4
WAT		proc

		; Verify stack frame
		VSTACK	1

		; Read addr
		mov	eax, PARAM0
		
		; Check addr against bound
		bound	eax, qword ptr DDA_BOUND
		
		; Read value then push
		movzx	eax, word ptr [eax]
		mov	PARAM0, eax

		; Return to caller
		ret

E_WAT:
WAT		endp


; ===============================================================
; = STORE							=
; =                                                 		=
; = Code for the forth word:	!				=
; =								=
; = Stack:	( n adr --- )					=
; =								=
; = ALL REGISTERS ARE PRESERVED.				=
; ===============================================================

		align	4
STORE		proc

		; Verify stack frame
		VSTACK	2

		; Get addr
		mov	eax, PARAM0				; Pop adr

		; Check addr against bound
		bound	eax, qword ptr DDA_BOUND

		; Store value
		mov	ebx, PARAM1				; Pop n
		mov	[eax], ebx
		DISCARD	2

		; Return to caller
		ret

E_STORE:
STORE		endp


; ===============================================================
; = PSTORE							=
; =                                                 		=
; = Code for the forth word:	(!)				=
; =								=
; = Stack:	( n adr --- )					=
; ===============================================================

		align	4
PSTORE		proc

		; Verify stack frame
		VSTACK	2

		; Store value
		mov	eax, PARAM0				; Pop adr
		mov	ebx, PARAM1				; Pop n
		mov	[eax], ebx
		DISCARD	2

		; Return to caller
		ret

E_PSTORE:
PSTORE		endp


; ===============================================================
; = PCSTORE							=
; =                                                 		=
; = Code for the forth word:	(C!)				=
; =								=
; = Stack:	( b adr --- )					=
; ===============================================================

		align	4
PCSTORE		proc

		; Verify stack frame
		VSTACK	2

		; Store value
		mov	eax, PARAM0				; Pop adr
		mov	ebx, PARAM1				; Pop n
		mov	[eax], bl
		DISCARD	2

		; Return to caller
		ret

E_PCSTORE:
PCSTORE		endp


; ===============================================================
; = CSTORE							=
; =                                                 		=
; = Code for the forth word:	C!				=
; =								=
; = Stack:	( b adr --- )					=
; ===============================================================

		align	4
CSTORE		proc

		; Verify stack frame
		VSTACK	2

		; Get addr
		mov	eax, PARAM0				; Pop adr

		; Check addr against bound
		bound	eax, qword ptr DDA_BOUND

		; Store value
		mov	ebx, PARAM1				; Pop n
		mov	[eax], bl
		DISCARD	2

		; Return to caller
		ret

E_CSTORE:
CSTORE		endp


; ===============================================================
; = WSTORE							=
; =                                                 		=
; = Code for the forth word:	W!				=
; =								=
; = Stack:	( w adr --- )					=
; ===============================================================

		align	4
WSTORE		proc

		; Verify stack frame
		VSTACK	2

		; Get addr
		mov	eax, PARAM0				; Pop adr

		; Check addr against bound
		bound	eax, qword ptr DDA_BOUND

		; Store value
		mov	ebx, PARAM1				; Pop w
		mov	[eax], bx
		DISCARD	2

		; Return to caller
		ret

E_WSTORE:
WSTORE		endp


; ===============================================================
; = PWSTORE							=
; =                                                 		=
; = Code for the forth word:	(W!)				=
; =								=
; = Stack:	( w adr --- )					=
; ===============================================================

		align	4
PWSTORE		proc

		; Verify stack frame
		VSTACK	2

		; Store value
		mov	eax, PARAM0				; Pop adr
		mov	ebx, PARAM1				; Pop w
		mov	[eax], bx
		DISCARD	2

		; Return to caller
		ret

E_PWSTORE:
PWSTORE		endp


; ===============================================================
; = ATSTORE							=
; =                                                 		=
; = Code for the forth word:	@!				=
; =								=
; = Stack:	( addr1 addr2 --- )				=
; ===============================================================

		align	4
ATSTORE		proc

		; Verify stack frame
		VSTACK	2

		; Get addr1 (source)
		mov	eax, PARAM1				; Pop addr1
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	eax, [eax]

		; Store value
		mov	ebx, PARAM0				; Pop addr2
		bound	ebx, qword ptr DDA_BOUND		; Check addr against bound
		mov	[ebx], eax
		DISCARD	2

		; Return to caller
		ret

E_ATSTORE:
ATSTORE		endp


; ===============================================================
; = CATSWAP							=
; =                                                 		=
; = Code for the forth word:	C@swap				=
; =								=
; = Stack:	( addr1 addr2 --- )				=
; ===============================================================

		align	4
CATSWAP		proc

		; Verify stack frame
		VSTACK	2

		; Swap values
		mov	eax, PARAM1				; Addr1
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, PARAM0				; Addr2
		bound	ebx, qword ptr DDA_BOUND		; Check addr against bound
		mov	cl, [eax]
		mov	dl, [ebx]
		mov	[eax], dl
		mov	[ebx], cl
		
		; Return to caller
		DISCARD	2					; drop drop
		ret

E_CATSWAP:
CATSWAP		endp



; ===============================================================
; = WATSWAP							=
; =                                                 		=
; = Code for the forth word:	W@swap				=
; =								=
; = Stack:	( addr1 addr2 --- )				=
; ===============================================================

		align	4
WATSWAP		proc

		; Verify stack frame
		VSTACK	2
		
		; Swap values
		mov	eax, PARAM1				; Addr1
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, PARAM0				; Addr2
		bound	ebx, qword ptr DDA_BOUND		; Check addr against bound
		mov	cx, [eax]
		mov	dx, [ebx]
		mov	[eax], dx
		mov	[ebx], cx
		
		; Return to caller
		DISCARD	2					; drop drop
		ret

E_WATSWAP:
WATSWAP		endp



; ===============================================================
; = ATSWAP							=
; =                                                 		=
; = Code for the forth word:	@swap				=
; =								=
; = Stack:	( addr1 addr2 --- )				=
; ===============================================================

		align	4
ATSWAP		proc

		; Verify stack frame
		VSTACK	2
		
		; Swap values
		mov	eax, PARAM1				; Addr1
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, PARAM0				; Addr2
		bound	ebx, qword ptr DDA_BOUND		; Check addr against bound
		mov	ecx, [eax]
		mov	edx, [ebx]
		mov	[eax], edx
		mov	[ebx], ecx
		
		; Return to caller
		DISCARD	2					; drop drop
		ret

E_ATSWAP:
ATSWAP		endp



; ===============================================================
; = ATEQU							=
; =								=
; = Code for the forth word:	@==				=
; =								=
; = Stack:	( n addr --- f )				=
; ===============================================================

		align	4
ATEQU		proc

		; Verify stack frame
		VSTACK	2

		; Read [adr]
		mov	ebx, PARAM1				; pop <n>
		DISCARD	1
		mov	eax, PARAM0				; pop <addr>
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		cmp	dword ptr [eax], ebx
		je	@@True
		
@@False:		mov	PARAM0, FALSE
		ret						; jmp	@@Exit

@@True:		mov	PARAM0, TRUE

		; Return to caller
@@Exit:		ret

E_ATEQU:
ATEQU		endp


; ===============================================================
; = ATNEQU							=
; =								=
; = Code for the forth word:	@!=				=
; =								=
; = Stack:	( n addr --- f )				=
; ===============================================================

		align	4
ATNEQU		proc

		; Verify stack frame
		VSTACK	2

		; Read [adr]
		mov	ebx, PARAM1				; pop <n>
		DISCARD	1
		mov	eax, PARAM0				; pop <addr>
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		cmp	dword ptr [eax], ebx
		jne	@@True
		
@@False:		mov	PARAM0, FALSE
		ret						; jmp	@@Exit

@@True:		mov	PARAM0, TRUE

		; Return to caller
@@Exit:		ret

E_ATNEQU:
ATNEQU	endp


; ===============================================================
; = ATZEROEQU							=
; =                                                 		=
; = Code for the forth word:	@0==				=
; =								=
; = Stack:	( adr --- f )					=
; ===============================================================

		align	4
ATZEROEQU	proc

		; Verify stack frame
		VSTACK	1

		; Read [adr]
		mov	eax, PARAM0
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		cmp	dword ptr [eax], 0
		je	@@True
		
@@False:	mov	PARAM0, FALSE
		ret						; jmp	@@Exit

@@True:		mov	PARAM0, TRUE

		; Return to caller
@@Exit:		ret

E_ATZEROEQU:
ATZEROEQU	endp


; ===============================================================
; = ATZERONEQU							=
; =                                                 		=
; = Code for the forth word:	@0!=				=
; =								=
; = Stack:	( adr --- f )					=
; ===============================================================

		align	4
ATZERONEQU	proc

		; Verify stack frame
		VSTACK	1

		; Read [adr]
		mov	eax, PARAM0
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		cmp	dword ptr [eax], 0
		je	@@False
		
@@True:		mov	PARAM0, TRUE
		ret						; jmp	@@Exit

@@False:	mov	PARAM0, FALSE

		; Return to caller
@@Exit:		ret

E_ATZERONEQU:
ATZERONEQU	endp


; ===============================================================
; = ZEROSTORE							=
; =                                                 		=
; = Code for the forth word:	0!				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
ZEROSTORE	proc

		; Verify stack frame
		VSTACK	1

		; Store value 0 at address
		mov	eax, PARAM0				; Pop adr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	dword ptr [eax], 0
		DISCARD	1

		; Return to caller
		ret

E_ZEROSTORE:
ZEROSTORE	endp


; ===============================================================
; = ONESTORE							=
; =                                                 		=
; = Code for the forth word:	1!				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
ONESTORE	proc

		; Verify stack frame
		VSTACK	1

		; Store value 1 at address
		mov	eax, PARAM0				; Pop adr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	dword ptr [eax], 1
		DISCARD	1

		; Return to caller
		ret

E_ONESTORE:
ONESTORE	endp


; ===============================================================
; = TWOSTORE							=
; =                                                 		=
; = Code for the forth word:	2!				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
TWOSTORE	proc

		; Verify stack frame
		VSTACK	1

		; Store value 2 at address
		mov	eax, PARAM0				; Pop adr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	dword ptr [eax], 2
		DISCARD	1

		; Return to caller
		ret

E_TWOSTORE:
TWOSTORE		endp


; ===============================================================
; = MINUSONESTORE						=
; =                                                 		=
; = Code for the forth word:	-1!				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
MINUSONESTORE	proc

		; Verify stack frame
		VSTACK	1

		; Store value -1 at address
		mov	eax, PARAM0				; Pop adr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	dword ptr [eax], -1
		DISCARD	1

		; Return to caller
		ret

E_MINUSONESTORE:
MINUSONESTORE	endp


; ===============================================================
; = MINUSTWOSTORE						=
; =                                                 		=
; = Code for the forth word:	-2!				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
MINUSTWOSTORE	proc

		; Verify stack frame
		VSTACK	1

		; Store value -2 at address
		mov	eax, PARAM0				; Pop adr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	dword ptr [eax], -2
		DISCARD	1

		; Return to caller
		ret

E_MINUSTWOSTORE:
MINUSTWOSTORE	endp


; ===============================================================
; = PLUSSTORE							=
; =                                                 		=
; = Code for the forth word:	+!				=
; =								=
; = Stack:	( n adr --- )					=
; ===============================================================

		align	4
PLUSSTORE	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ------------
		; Do operation
		; ------------
		
		mov	eax, PARAM0				; Addr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, PARAM1				; Pop n
		DISCARD	2
		add	[eax], ebx


		; ----------------
		; Return to caller
		; ----------------
		
		into						; Integer arithmetic overflow test
		ret

E_PLUSSTORE:
PLUSSTORE	endp


; ===============================================================
; = ONEPLUSSTORE						=
; =                                                 		=
; = Code for the forth word:	1+!				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
ONEPLUSSTORE	proc

		; Verify stack frame
		VSTACK	1

		; Do operation
		mov	eax, PARAM0				; Addr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		DISCARD	1
		inc	dword ptr [eax]

		; Return to caller
		into						; Integer arithmetic overflow test
		ret

E_ONEPLUSSTORE:
ONEPLUSSTORE	endp


; ===============================================================
; = TWOPLUSSTORE						=
; =                                                 		=
; = Code for the forth word:	2+!				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
TWOPLUSSTORE	proc

		; Verify stack frame
		VSTACK	1

		; Do operation
		mov	eax, PARAM0				; Addr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		DISCARD	1
		add	dword ptr [eax], 2

		; Return to caller
		into						; Integer arithmetic overflow test
		ret

E_TWOPLUSSTORE:
TWOPLUSSTORE	endp


; ===============================================================
; = MINUSSTORE							=
; =                                                 		=
; = Code for the forth word:	-!				=
; =								=
; = Stack:	( n adr --- )					=
; ===============================================================

		align	4
MINUSSTORE	proc

		; Verify stack frame
		VSTACK	2

		; Do operation
		mov	eax, PARAM0				; Addr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, PARAM1				; Pop n
		DISCARD	2
		sub	[eax], ebx

		; Return to caller
		into						; Integer arithmetic overflow test
		ret

E_MINUSSTORE:
MINUSSTORE	endp


; ===============================================================
; = ONEMINUSSTORE						=
; =                                                 		=
; = Code for the forth word:	1-!				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
ONEMINUSSTORE	proc

		; Verify stack frame
		VSTACK	1

		; Do operation
		mov	eax, PARAM0				; Addr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		DISCARD	1
		dec	dword ptr [eax]

		; Return to caller
		into						; Integer arithmetic overflow test
		ret

E_ONEMINUSSTORE:
ONEMINUSSTORE	endp


; ===============================================================
; = TWOMINUSSTORE						=
; =                                                 		=
; = Code for the forth word:	2-!				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
TWOMINUSSTORE	proc

		; Verify stack frame
		VSTACK	1

		; Do operation
		mov	eax, PARAM0				; Addr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		DISCARD	1
		sub	dword ptr [eax], 2

		; Return to caller
		into						; Integer arithmetic overflow test
		ret

E_TWOMINUSSTORE:
TWOMINUSSTORE	endp


; ===============================================================
; = STARSTORE							=
; =                                                 		=
; = Code for the forth word:	*!				=
; =								=
; = Stack:	( n adr --- )					=
; ===============================================================

		align	4
STARSTORE	proc

		; Verify stack frame
		VSTACK	2

		; Do operation
		mov	eax, PARAM1				; Pop n
		mov	ebx, PARAM0				; Addr
		bound	ebx, qword ptr DDA_BOUND		; Check addr against bound
		DISCARD	2
		imul	dword ptr [ebx]				; EAX *= [adr]
		mov	[ebx], eax

		; Return to caller
		into						; Integer arithmetic overflow test
		ret

E_STARSTORE:
STARSTORE	endp


; ===============================================================
; = TWOSTARSTORE						=
; =                                                 		=
; = Code for the forth word:	2*!				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
TWOSTARSTORE	proc

		; Verify stack frame
		VSTACK	1

		; Do operation
		mov	eax, PARAM0				; Addr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		DISCARD	1
		sal	dword ptr [eax], 1

		; Return to caller
		into						; Integer arithmetic overflow test
		ret

E_TWOSTARSTORE:
TWOSTARSTORE	endp


; ===============================================================
; = SLASHSTORE							=
; =                                                 		=
; = Code for the forth word:	/!				=
; =								=
; = Stack:	( n adr --- )					=
; ===============================================================

		align	4
SLASHSTORE	proc

		; Verify stack frame
		VSTACK	2

		; Do operation
		mov	eax, PARAM1				; Pop n
		mov	ebx, PARAM0				; Addr
		bound	ebx, qword ptr DDA_BOUND		; Check addr against bound
		clr	edx
		idiv	dword ptr [ebx]				; EAX *= [adr]
		mov	[ebx], eax
		DISCARD	2					; drop drop

		; Return to caller
		ret

E_SLASHSTORE:
SLASHSTORE	endp


; ===============================================================
; = TWOSLASHSTORE						=
; =                                                 		=
; = Code for the forth word:	2/!				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
TWOSLASHSTORE	proc

		; Verify stack frame
		VSTACK	1

		; Do operation
		mov	eax, PARAM0				; Addr
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		DISCARD	1
		sar	dword ptr [eax], 1

		; Return to caller
		ret

E_TWOSLASHSTORE:
TWOSLASHSTORE	endp


; ===============================================================
; = QUEST							=
; =                                                 		=
; = Code for the forth word:	?				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
QUEST		proc

		; Verify stack frame
		VSTACK	1

		; Do action
		mov	eax, PARAM0
		mov	eax, [eax]
		mov	PARAM0, eax				; @		
		SUMMON	DOT					; .

		; Return to caller
		ret

E_QUEST:
QUEST		endp


; ===============================================================
; = CQUEST							=
; =                                                 		=
; = Code for the forth word:	C?				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
CQUEST		proc

		; Verify stack frame
		VSTACK	1

		; Do action
		mov	eax, PARAM0
		movzx	eax, byte ptr [eax]
		mov	PARAM0, eax				; C@
		SUMMON	DOT					; .

		; Return to caller
		ret

E_CQUEST:
CQUEST		endp


; ===============================================================
; = WQUEST							=
; =                                                 		=
; = Code for the forth word:	W?				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
WQUEST		proc

		; Verify stack frame
		VSTACK	1

		; Do action
		mov	eax, PARAM0
		movzx	eax, word ptr [eax]
		mov	PARAM0, eax				; W@
		SUMMON	DOT					; .

		; Return to caller
		ret

E_WQUEST:
WQUEST		endp


; ===============================================================
; = DYNAMIC							=
; =                                                 		=
; = Code for the forth word:	DYNAMIC				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
DYNAMIC		proc

		; ----------------------------------------
		; Verify if the system is is execute state
		; ----------------------------------------

		SUMMON	QEXECUTE
		
		
		; ------------------------------------
		; Allready called and still not used ?
		; ------------------------------------
		
		cmp	[DynamicFlag], FALSE
		je	@@Ok
		
		; "DYNAMIC word already called"
		mov	ah, ERR_DYNAALCALL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError
		
@@Ok:

		; -----------
		; Update flag
		; -----------
		
		inc	[DynamicFlag]


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_DYNAMIC:
DYNAMIC		endp


; ===============================================================
; = ALLOCATE							=
; =                                                 		=
; = Code for the forth word:	ALLOCATE			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
ALLOCATE		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; ----------------------------------------
		; Store operation: 'MOV [AllocFlag], TRUE'
		; ----------------------------------------

		; Code operation		
		MOV_xMEM_i8
		
		; Memory relative address (rel32)
		lea	eax, AllocFlag
		call	CodeComma			; Code,
		
		; Value to store
		mov	al, TRUE
		call	CodeCComma			; CodeC,


		; ----------------------------------------
		; The word ALLOC was found while compiling
		; ----------------------------------------
		
		mov	[AllocFound], TRUE


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_ALLOCATE:
ALLOCATE		endp


; ===============================================================
; = FREE							=
; =                                                 		=
; = Code for the forth word:	FREE				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
FREE		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; -------------------------------------
		; Store operation: MOV [FreeFlag], TRUE
		; -------------------------------------

		; Code operation		
		MOV_xMEM_i8
		
		; Memory relative address (rel32)
		lea	eax, FreeFlag
		call	CodeComma			; Code,
		
		; Value to store
		mov	al, TRUE
		call	CodeCComma			; CodeC,


		; ---------------------------------------
		; The word FREE was found while compiling
		; ---------------------------------------
		
		mov	[FreeFound], TRUE


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_FREE:
FREE		endp


; ===============================================================
; = CodeComma							=
; =                                     	            	=
; = Code,							=
; =	Store a dword into CDA (Code Dictionnary Area)		=
; =	(it's a compiler action)				=
; =                       		                       	=
; = I) 	EAX value ro store					=
; = O)	none 							=
; = M) 	none							=
; ===============================================================

		align	4
CodeComma	proc

		; Dynamic allocation ?
		; --------------------
		
		cmp	[DynVarAddr], 0
		jne	@@DynaAlloc


		; NOT dynamic allocation
		; ----------------------

		; Store value
		push	ebx
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		mov	[ebx], eax
		pop	ebx
		
		; Adjust CDA size
		add	[CDA_SIZE], @WordSize
		
		; Done
		ret					; jmp	@@Exit
		
	
		; Dynamic allocation
		; ------------------
		
@@DynaAlloc:
		
		; Store value
		push	ebx
		mov	ebx, [DynVarAddr]
		mov	[ebx], eax
		pop	ebx
		
		; Ajust new address
		add	[DynVarAddr], @WordSize
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_CodeComma:
CodeComma	endp


; ===============================================================
; = CodeWComma							=
; =                                     	            	=
; = Code,							=
; =	Store a word into CDA (Code Dictionnary Area)		=
; =	(it's a compiler action)				=
; =                       		                       	=
; = I) 	AX value ro store					=
; = O)	none 							=
; = M) 	none							=
; ===============================================================

		align	4
CodeWComma	proc

		; Dynamic allocation ?
		; --------------------
		
		cmp	[DynVarAddr], 0
		jne	@@DynaAlloc


		; NOT dynamic allocation
		; ----------------------

		; Store value
		push	ebx
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		mov	[ebx], ax
		pop	ebx
		
		; Adjust CDA size
		add	[CDA_SIZE], 2
		
		; Done
		ret					; jmp	@@Exit
		
	
		; Dynamic allocation
		; ------------------
		
@@DynaAlloc:
		
		; Store value
		push	ebx
		mov	ebx, [DynVarAddr]
		mov	[ebx], ax
		pop	ebx
		
		; Ajust new address
		add	[DynVarAddr], 2
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_CodeWComma:
CodeWComma	endp


; ===============================================================
; = CodeCComma							=
; =                                     	            	=
; = Code,							=
; =	Store a byte into CDA (Code Dictionnary Area)		=
; =	(it's a compiler action)				=
; =                       		                       	=
; = I) 	AL value ro store					=
; = O)	none 							=
; = M) 	none							=
; ===============================================================

		align	4
CodeCComma	proc

		; Dynamic allocation ?
		; --------------------
		
		cmp	[DynVarAddr], 0
		jne	@@DynaAlloc


		; NOT dynamic allocation
		; ----------------------

		; Store value
		push	ebx
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		mov	[ebx], al
		pop	ebx
		
		; Adjust CDA size
		inc	[CDA_SIZE]
		
		; Done
		ret					; jmp	@@Exit
		
	
		; Dynamic allocation
		; ------------------
		
@@DynaAlloc:
		
		; Store value
		push	ebx
		mov	ebx, [DynVarAddr]
		mov	[ebx], al
		pop	ebx
		
		; Ajust new address
		inc	[DynVarAddr]
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_CodeCComma:
CodeCComma	endp


; ===============================================================
; = QALLOTFAIL                      				=
; =                                                         	=
; = Code for the forth word:    ?AllotFail			=
; =                                                         	=
; = Stack:  ( --- addr ) 	                		=
; ===============================================================

QALLOTFAIL	proc

		; Integer run-time
		DOINT	D_QALLOTFAIL
		
		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_QALLOTFAIL:
QALLOTFAIL	endp


; ===============================================================
; = Malloc							=
; =                                     	            	=
; = Usage:	Allocate memory into the Data Dictionnary,	=
; =		via DDIR indirection (Data Directory Area).	=
; =                       		                       	=
; = I)		EAX	memory amount asked (in bytes)		=
; =		EBX	0: static memory			=
; =			1: local memory				=
; =			2: dynamic memory			=
; =                                 	                	=
; = O)		EAX	the data block reference		=
; =		EBX	the real data address			=
; =                                         	        	=
; = M)		ECX EDX						=
; ===============================================================

		align	4
Malloc		proc

		; Is there enought memory ?
		; -------------------------

		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, eax
		cmp	ecx, edx
		jge	@@MemOk

		; Out of memory !
		mov	ah, ERR_NOMEMORY		; Out of memory.
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		jmp	doError
		
@@MemOk:

		; Add a new record into DDIR
		; --------------------------

		push	ebx
		mov	ebx, [DDIR_FFREED]
		cmp	ebx, -1
		je	@@Rec2

		; A record is freed, take it
		add	ebx, [DDIR_START]
		mov	ecx, [ebx.ddir_NextFreed]
		mov	[DDIR_FFREED], ecx
		jmp	@@RecDone
		
		; No record freed, add one (if possible...)
@@Rec2:		mov	ebx, [DDIR_SIZE]
		cmp	ebx, DDIR_WSIZE
		jl	@@RecNbOk

		; DDIR table full !
		pop	ebx				; Clean stack
		mov	ah, ERR_DDIRFULL		; "Data allocation table full."
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		jmp	doError
		
@@RecNbOk:	add	ebx, [DDIR_START]		; EBX points to the new record
		
		; Update DDIR size
		add	[DDIR_SIZE], size ddirentry

@@RecDone:

		
		; Get asked memory
		; ----------------

		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, eax

		; Update memory pointers
		add	[DDA_SIZE], eax

		; Update memory allocated bounds
		mov	edx, [DDA_START]
		sub	edx, [DDA_SIZE]
		mov	[DDA_BOUND], edx


		; Fill the DDIR entry
		; -------------------
		
		; The record is not free
		mov	[ebx.ddir_NextFreed], -1
		
		; Block size
		mov	[ebx.ddir_blksize], eax
		
		; Memory address
		mov	[ebx.ddir_blkaddr], ecx

		; Reference to the word which owns this data
		mov	eax, [VOC_CURRENT]
		add	eax, [VOC_START]
		mov	eax, [eax.voc_lastword]
		push	ecx
		mov	ecx, size wordentry
		clr	edx
		idiv	ecx					; eax /= size wordentry
		pop	ecx
		mov	[ebx.ddir_wordnb], ax

		; Thread nb
		mov	eax, [CurrActThread]
		mov	[ebx.ddir_threadnb], al
		
		; Thread nb, local? dynamic?
		clr	al					; Initialize flags
		pop	edx
		cmp	dl, 1
		jne	@@Not1
		or	al, QDATA_AUTO				; Automatic variable
		jmp	@@12
@@Not1:		cmp	dl, 2
		jne	@@12
		or	al, QDATA_DYNA				; Dynamic variable
@@12:		mov	[ebx.ddir_flags], al			; Set flags
		

		; Return to caller
		; ----------------

		; EAX = data block reference
		sub	ebx, [DDIR_START]
		mov	eax, ebx
		
		; EBX = real data address
		mov	ebx, ecx

		; Done				
		ret

Malloc		endp


; ===============================================================
; = ReAlloc							=
; =                                     	            	=
; = Usage:	Re-Allocate memory into the Data Dictionnary	=
; =		via DDIR indirection (Data Directory Area).	=
; =                       		                       	=
; = I)		EAX	memory amount asked (in bytes)		=
; =		EDX	data block nb				=
; =                                 	                	=
; = O)		EAX	the data block reference		=
; =		EBX	the real data address			=
; =                                         	        	=
; = M)		ECX EDX						=
; ===============================================================

		align	4
ReAlloc		proc

		; Free the data used
		; ------------------

		push	edx					; Save data block nb (for future use)
		call	UnAllot
		

		; Is there enought memory ?
		; -------------------------

		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, eax
		cmp	ecx, edx
		jge	@@MemOk

		; Out of memory !
		add	esp, 1 * @WordSize			; Clean stack
		mov	ah, ERR_NOMEMORY			; Out of memory.
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError
		
@@MemOk:


		; Add a new record into DDIR
		; --------------------------

		mov	ebx, [DDIR_FFREED]
		cmp	ebx, -1
		je	@@Rec2

		; A record is freed, take it
		add	ebx, [DDIR_START]
		mov	ecx, [ebx.ddir_NextFreed]
		mov	[DDIR_FFREED], ecx
		jmp	@@RecDone
		
		; No record freed, add one (if possible...)
@@Rec2:		mov	ebx, [DDIR_SIZE]
		cmp	ebx, DDIR_WSIZE
		jl	@@RecNbOk

		; DDIR table full !
		add	esp, 1 * @WordSize		; Clean stack
		mov	ah, ERR_DDIRFULL		; "Data allocation table full."
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		jmp	doError
		
@@RecNbOk:	add	ebx, [DDIR_START]		; EBX points to the new record
		
		; Update DDIR size
		add	[DDIR_SIZE], size ddirentry

@@RecDone:
		
		
		; Get asked memory
		; ----------------

		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, eax

		; Update memory pointers
		add	[DDA_SIZE], eax

		; Update memory allocated bounds
		mov	edx, [DDA_START]
		sub	edx, [DDA_SIZE]
		mov	[DDA_BOUND], edx


		; Fill the DDIR entry
		; -------------------
		
		; The record is not free
		mov	[ebx.ddir_NextFreed], -1
		
		; Block size
		mov	[ebx.ddir_blksize], eax
		
		; Memory address
		mov	[ebx.ddir_blkaddr], ecx

		; Reference to the word which owns this data
		pop	edx					; Get saved previous record used
		mov	ax, [edi + edx.ddir_wordnb]
		mov	[ebx.ddir_wordnb], ax
		
		; Thread nb
		mov	eax, [CurrActThread]
		mov	[ebx.ddir_threadnb], al
		
		; Thread nb, local? dynamic?
		movzx	eax, [edi + edx.ddir_flags]		; Get previous flags
		and	al, (not QDATA_MARKDEL)			; Record not marked as deleted
		mov	[ebx.ddir_flags], al
		

		; Return to caller
		; ----------------

		; EAX = data block reference
		sub	ebx, [DDIR_START]
		mov	eax, ebx
		
		; EBX = real data address
		mov	ebx, ecx

		; Done				
		ret

ReAlloc		endp


; ===============================================================
; = BWD_ALLOT							=
; =                                                 		=
; = Code for the forth word:	RT_(BW)ALLOT			=
; =								=
; = Stack:	( n --- addr )					=
; =								=
; = Usage:	Allocate (n * [1,2,4]) bytes in the Data	=
; =		Dictionnary.					=
; = O) EAX : memory block nb (DDIR table).			=
; ===============================================================

BWD_ALLOT	MACRO	sz

		align	4

		if	sz eq 'B'
RT_BALLOT	proc
		elseif	sz eq 'W'
RT_WALLOT	proc
		else
RT_ALLOT	proc
		endif

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------

		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ---------------------------------
		; Dynamic allocation ? (with ALLOC)
		; ---------------------------------
		
		cmp	[AllocFlag], TRUE
		je	@@StaticAll


		; ----------------------------------------------------------------
		; Automatic allocation ? No if Compile Actions called by INTERPRET
		; ----------------------------------------------------------------

		lea	eax, InterpretAddr
		add	eax, [OfsDictReloc]
		cmp	dword ptr [esp + 4], eax
		je	@@StaticAll

		; Amount of memory asked
		mov	eax, PARAM0
		if	sz eq 'W'
		shl	eax, 1					; * 2
		elseif	sz eq 'D'
		shl	eax, 2					; * 4
		endif

		; Is memory already allocated for this auto variable ? If yes, release this memory before
		mov	edx, [esp + 4]
		mov	edx, [edx]
		cmp	edx, -1
		je	@@FirstAlloc


		; Re-allocation of memory
		; -----------------------

		call	ReAlloc
		jmp	@@MemDone
		

		; Get memory asked (first allocation)
		; -----------------------------------
		
@@FirstAlloc:	mov	bl, 1					; 1 mean Local variable
		call	Malloc					; Get memory


		; Store the data block address
@@MemDone:	mov	PARAM0, ebx 				; Push block address on stack
		mov	ecx, [esp + 4]				; Addr of cell where we have to store DataBlockAddr
		mov	[ecx], eax

		; Done for dynamic allocation
		jmp	@@Exit					; ret

		
		; -------------------------------------------
		; Static allocation (not inside a definition)
		; -------------------------------------------
		
@@StaticAll:

		; Compile: CALL $
		CALL_rel32
		clr	eax
		call	CodeComma               		; Code,

		; Compile: ADD DWORD PTR [ESP], 4+6+3+2+1
		ADD_xESPi8_i8	0, 4+6+3+2+1

		; Get memory asked (returned: <n>, block memory nb)
		; then compile: MOV EAX, [EDI.ddir_blkaddr + <n>]
		MOV_EAX_xEDIi32
		mov	eax, PARAM0
		if	sz eq 'W'
		shl	eax, 1					; * 2
		elseif	sz eq 'D'
		shl	eax, 2					; * 4
		endif
		cmp	[AllocFlag], TRUE
		je	@@DynaVar
		clr	bl					; 0 means static  variable
		jmp	@@malloc
@@DynaVar:	mov	bl, 2					; 2 means dynamic variable
@@malloc:	call	Malloc					; Get memory
		mov	PARAM0, ebx 				; Push block address on stack
		add	eax, ddir_blkaddr
		call	CodeComma				; Code,
		
		; Compile: SUB ESI, 4
		SUB_ESI_i8	@WordSize
		
		; Compile: MOV [ESI], eax
		MOV_xESIi8_EAX
		
		; Compile: RET
		_RET_

		; Record the addr of the block, which can be used by ALLOTMORE (if static allocation)
		cmp	[AllocFlag], TRUE
		mov	[REF_BLK_ALLCTED], eax

		; On output, EAX is block nb
		sub	eax, ddir_blkaddr
		

		; ------------------------------------------------------
		; Return to caller, with real block address on top stack
		; ------------------------------------------------------
		
@@Exit:		
		; CTRL-BREAK allowed
		SUMMON	CTRLBREAK_REST

		ret

		if	sz eq 'B'
E_RT_BALLOT:
		elseif	sz eq 'W'
E_RT_WALLOT:
		else
E_RT_ALLOT:
		endif
		endp


; ===============================================================
; = ALLOT							=
; =                                                 		=
; = Code for the forth word:	ALLOT				=
; =								=
; = Stack:	( n --- addr )					=
; =								=
; = Usage:	Allocate (n * [1, 2, 4]) bytes in the Data 	=
; =		Dictionnary.					=
; ===============================================================

		align	4

		if	sz eq 'B'
BALLOT		proc
		elseif	sz eq 'W'
WALLOT		proc
		else
ALLOT		proc
		endif

		; 'BUILD>' not called before ?
		cmp	[CDA_BUILD_CALLED], FALSE
		jne	@@Ok

		; 'BUILD>' not called beforehand.
		mov	ah, ERR_BUILDNOTCALLED			; "'BUILD>' not called beforehand."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok:

		; DOES> already called ?
		cmp	[CDA_DOES_CALLED], FALSE
		je	@@Ok2
		
		; 'DOES>' already called
		mov	ah, ERR_DOESALCALLED			; "'DOES>' already called."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok2:		

		; ALLOT already called ?
		cmp	[CDA_ALLOT_CALLED], FALSE
		je	@@Ok3
		
		; 'ALLOT' already called
		mov	ah, ERR_ALLOTALCALLED			; "'ALLOT' already called."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok3:		mov	[CDA_ALLOT_CALLED], TRUE

		; Put 'CALL_rel32'
		if	sz eq 'B'
			compCALL_rel32	RT_BALLOT
		elseif	sz eq 'W'
			compCALL_rel32	RT_WALLOT
		else
			compCALL_rel32	RT_ALLOT
		endif
		
		; Return to caller
		ret

		if	sz eq 'B'
E_BALLOT:
		elseif	sz eq 'W'
E_WALLOT:
		else
E_ALLOT:
		endif
		endp

		endm


BWD_ALLOT	'B'
BWD_ALLOT	'W'
BWD_ALLOT	'D'


; ===============================================================
; = BWD_ALLOTMORE						=
; =                                                 		=
; = Code for the forth words:	ALLOTMORE			=
; =								=
; = Stack:	( n --- addr )					=
; =								=
; = Usage:	Re-Allocate (n * [1,2,4]) bytes in the Data 	=
; =		Dictionnary.					=
; ===============================================================

BWD_ALLOTMORE	MACRO	sz

		align	4

		if	sz eq 'B'
RT_BALLOTMORE	proc
		elseif	sz eq 'W'
RT_WALLOTMORE	proc
		else
RT_ALLOTMORE	proc
		endif

		; CTRL-BREAK not allowed
		; ----------------------
		inc	[NoCtrlBreak]


		; Get last block allocated by ALLOT
		cmp	[AllocFlag], TRUE
		je	@@NOK
		mov	ebx, [REF_BLK_ALLCTED]
		cmp	ebx, -1
		jne	@@OK

		; "'ALLOTMORE' can be used only after ALLOT (non dynamic)."
@@NOK:		mov	ah, ERR_ALLOTMOREBAD			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@OK:		sub	ebx, ddir_blkaddr
		add	ebx, [DDIR_START]

		; Get size to add
		mov	eax, PARAM0
		if	sz eq 'W'
		shl	eax, 1					; * 2
		elseif	sz eq 'D'
		shl	eax, 2					; * 4
		endif


		; Get asked memory
		; ----------------

		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, eax
		cmp	ecx, edx
		jge	@@MemOk

		; Out of memory !
		mov	ah, ERR_NOMEMORY			; Out of memory.
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		
@@MemOk:

		; Update memory pointers
		add	[DDA_SIZE], eax

		; Update memory allocated bounds
		mov	edx, [DDA_START]
		sub	edx, [DDA_SIZE]
		mov	[DDA_BOUND], edx

		; Move data
		push	esi edi
		mov	edi, PARAM0				; Size of mem asked
		if	sz eq 'W'
		shl	edi, 1					; * 2
		elseif	sz eq 'D'
		shl	edi, 2					; * 4
		endif
		mov	esi, [ebx.ddir_blkaddr]			; Source address
		neg	edi
		add	edi, esi
		mov	ecx, [ebx.ddir_blksize]			; Size of block to move
		cld                    				; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		
		; Update the DDIR entry
		; ---------------------
		
		sub	[ebx.ddir_blkaddr], eax
		add	[ebx.ddir_blksize], eax

		
		; Push on stack begin of new block allocated
		; ------------------------------------------
		
		mov	ecx, [ebx.ddir_blkaddr]
		add	ecx, [ebx.ddir_blksize]
		sub	ecx, eax
		mov	PARAM0, ecx
	

		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST
		

		; Return to caller
		; ----------------
		
		ret

		if	sz eq 'B'
E_RT_BALLOTMORE:
		elseif	sz eq 'W'
E_RT_WALLOTMORE:
		else
E_RT_ALLOTMORE:
		endif
		endp


; ===============================================================
; = BWD_ALLOTMORE						=
; =                                                 		=
; = Code for the forth word:	(BW)ALLOTMORE			=
; =								=
; = Stack:	( n --- addr )					=
; =								=
; = Usage:	Re-Allocate (n * [1,2,4]) bytes in the Data 	=
; =		Dictionnary.					=
; ===============================================================

		align	4

		if	sz eq 'B'
BALLOTMORE	proc
		elseif	sz eq 'W'
WALLOTMORE	proc
		else
ALLOTMORE	proc
		endif


		; Verify if the system is in compile state
		call	QCOMPILE


		; Put 'CALL_rel32'
		if	sz eq 'B'
			compCALL_rel32	RT_BALLOTMORE
		elseif	sz eq 'W'
			compCALL_rel32	RT_WALLOTMORE
		else
			compCALL_rel32	RT_ALLOTMORE
		endif
		
		
		; Return to caller
		ret

		if	sz eq 'B'
E_BALLOTMORE:
		elseif	sz eq 'W'
E_WALLOTMORE:
		else
E_ALLOTMORE:
		endif
		endp


		endm


BWD_ALLOTMORE	'B'
BWD_ALLOTMORE	'W'
BWD_ALLOTMORE	'D'


; ===============================================================
; = PDOES							=
; =								=
; = Code for the forth word:	RT_DOES>			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
PDOES		proc

		; ------------------------------------
		; Dynamic allocation (with ALLOCATE) ?
		; ------------------------------------
		
		cmp	[AllocFlag], FALSE
		je	@@NotDynaAlloc

		; Reset flag
		mov	[AllocFlag], FALSE
		mov	[DynVarAddr], 0

		; Done
		jmp	@@Exit
		
@@NotDynaAlloc:	


		; ------------------------------------------------------------
		; Local allocation ? No if Compile Actions called by INTERPRET
		; ------------------------------------------------------------

		lea	eax, InterpretAddr
		add	eax, [OfsDictReloc]
		cmp	dword ptr [esp + 4], eax
		je	@@StaticAll
		
		; Skip a level return
		add	esp, @WordSize
		mov	eax, [esp]
		add	eax, @WordSize				; Skip cell where is stored DataBlockAddr
		mov	[esp], eax
		ret
		
@@StaticAll:

		; -------------------------------------
		; Put 'JMP rel32' to run-time procedure
		; -------------------------------------

		; Store codop
		CALL_rel32		
		mov	eax, [esp]				; Cell where is stored where to jump
		mov	eax, [eax]
		add	eax, [NIA_START]
		mov	eax, [eax]
		sub	eax, [CDA_START]
		sub	eax, [CDA_SIZE]
		sub	eax, 4					; 4 for size REL32
		call	CodeComma				; Code,

		; ret
		_RET_

		; align 4
@@Align4:	mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		and	al, 3
		jz	@@Align4Done
		_NOP_
		jmp	@@Align4
@@Align4Done:


		; -----------------------------------------------------
		; Update the end code address of the word being defined
		; -----------------------------------------------------

		; Pointer on the word NIA address		
		mov	eax, [VOC_CURRENT]
		add	eax, [VOC_START]
		mov	eax, [eax.voc_lastword]
		add	eax, [NIA_START]

		; Compute the size of the code
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		sub	ebx, [eax.word_realaddr]
		
		; Store the code size
		mov	[eax.word_CodeSize], bx
		
		; Update the end of the code
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		dec	ebx
		mov	[eax.word_EndCode], ebx


		; -----------------------------------------------
		; Terminate caller precedure (and not this one !)
		; -----------------------------------------------

@@Exit:		add	esp, 4					; Skip a level return !
		ret

E_PDOES:
PDOES		endp


; ===============================================================
; = DOES							=
; =                                                 		=
; = Code for the forth word:	DOES>				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
DOES		proc

		; 'BUILD>' not called before ?
		cmp	[CDA_BUILD_CALLED], FALSE
		jne	@@Ok

		; 'BUILD>' not called beforehand.
		mov	ah, ERR_BUILDNOTCALLED			; "'BUILD>' not called beforehand."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok:

		; DOES> already called ?
		cmp	[CDA_DOES_CALLED], FALSE
		je	@@Ok2
		
		; 'DOES>' already called
		mov	ah, ERR_DOESALCALLED			; "'DOES>' allready called."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok2:		mov	[CDA_DOES_CALLED], TRUE


		; Put 'CALL_rel32 rt_does'
		compCALL_rel32	PDOES
		
		; Store run-time call nb
		mov	ebx, [LastWordDefWColon]		; Last word (NIA) defined with : (colon)
		add	ebx, [NIA_START]
		mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		add	eax, @WordSize
		mov	[ebx.word_runtaddr], eax
		
		; Store run-time call nb
		mov	eax, ebx
		sub	eax, [NIA_START]
		add	eax, word_runtaddr
		call	CodeComma				; Code,
		
		; Return to caller
		ret

E_DOES:
DOES		endp


; ===============================================================
; = LocalVar							=
; =                                                 		=
; = Code for the forth word:	LocalVar			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
LocalVar		proc

		; Save used registers
		push	eax ebx edx

		; Store into DI the cell address
		mov	ebx, esp
		add	ebx, 4 * 3				; 12 = 4 (eax) + 4 (ebx) + 4 (edx)
		mov	edx, [ebx]

		; Get data reference (in EBX)
		mov	eax, [edx]
		mov	eax, [edx + eax]

		; Move the IP after the data reference
		add	edx, 4
		mov	[ebx], edx				; 12 = 4 (eax) + 4 (ebx) + 4 (edx)

		; Push the data address
		add	eax, edi
		mov	eax, [eax.ddir_blkaddr]
		IPUSH	eax

		; Return to caller
		pop	edx ebx eax
		ret

E_LocalVar:
LocalVar		endp


; ===============================================================
; = DATASALVAGE							=
; =                                                 		=
; = Code for the forth word:	DataSalvage			=
; =								=
; = Stack:	( --- n )					=
; =								=
; = Usage							=
; =	Memory which can recovered by GARBAGE collector.	=
; ===============================================================

		align	4
DATASALVAGE	proc

		; Push value on stack
		mov	eax, [DDIR_MEMSALV]
		IPUSH	eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_DATASALVAGE:
DATASALVAGE	endp


; ===============================================================
; = DustData							=
; =                                                 		=
; = I) EDX : pointer to record to delete.			=
; ===============================================================

		align	4
DustData	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	esi edi eax ebx ecx
		
		
		; -----------------------
		; Is-it the last record ?
		; -----------------------
		
		mov	eax, [DDIR_START]
		add	eax, [DDIR_SIZE]
		sub	eax, size ddirentry
		cmp	edx, eax
		jne	@@NotLastRec
		
		; Update the memory used
		mov	eax, [edx.ddir_blksize]
		sub	[DDA_SIZE], eax
		
		; Update the amount of memory which can be salvaged by GARBAGE collector
		sub	[DDIR_MEMSALV], eax

		; Update the size of the table
		sub	[DDIR_SIZE], size ddirentry
		
		; Update memory allocated bounds
		mov	ebx, [DDA_START]
		sub	ebx, [DDA_SIZE]
		mov	[DDA_BOUND], ebx

		; Done
		jmp	@@Exit
		
@@NotLastRec:

		; ------------------------
		; It's not the last record
		; ------------------------
		
		; Move memory
		; -----------
		
		; Source address
		mov	esi, [edx.ddir_blkaddr]
		dec	esi
		
		; Destination address
		mov	edi, esi
		add	edi, [edx.ddir_blksize]

		; Size to move
		lea	ecx, [esi + 1]
		add	ecx, [DDA_SIZE]
		sub	ecx, [DDA_START]
		js	@@DontMove
		jz	@@DontMove

		; Move block
		std                     				; Dec ESI and EDI
		rep	movsb  		            	    		; Move the block
		cld

		
		; Update each block addr 
		; ----------------------
		
		mov	ecx, [DDIR_START]
		mov	ebx, ecx
		add	ecx, [DDIR_SIZE]
		add	ebx, [DDIR_SIZE_INIT]
		mov	eax, [edx.ddir_blksize]				; Size of the block deleted
		mov	esi, [edx.ddir_blkaddr]				; Addr block deleted
		sub	ebx, size ddirentry
@@UpdtBlkAddr:	

		add	ebx, size ddirentry
		cmp	ebx, ecx
		je	@@UpdtBA2

		cmp	[ebx.ddir_blkaddr], esi
		jae	@@UpdtBlkAddr

		add	[ebx.ddir_blkaddr], eax
		jmp	@@UpdtBlkAddr
@@UpdtBA2:

@@DontMove:	
		mov	eax, [edx.ddir_blksize]				; Size of the block deleted


		; Mark the record as reusable
		; ---------------------------
		
		mov	eax, [DDIR_FFREED]
		mov	ebx, edx
		sub	ebx, [DDIR_START]
		mov	[edx.ddir_NextFreed], eax
		mov	[DDIR_FFREED], ebx


		; Update the amount of memory which can be salvaged by GARBAGE collector
		; ----------------------------------------------------------------------
		
		mov	eax, [edx.ddir_blksize]
		sub	[DDIR_MEMSALV], eax

		
		; Block is no more marked to delete
		; ---------------------------------

		and	[edx.ddir_flags], (not QDATA_MARKDEL)
		or	[edx.ddir_flags], QDATA_DELDONE
		
		
		; Update the amount of memory used
		; --------------------------------
		
		sub	[DDA_SIZE], eax


		; Update memory allocated bounds
		; ------------------------------
		
		mov	ebx, [DDA_START]
		sub	ebx, [DDA_SIZE]
		mov	[DDA_BOUND], ebx


		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	ecx ebx eax edi esi
		ret

DustData	endp

	
; ===============================================================
; = DUST							=
; =                                                 		=
; = Code for the forth word:	DUST				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
DUST		proc
	
		; ---------
		; Dust data
		; ---------
		
		cmp	[DDIR_MEMSALV], 0			; Memory which can be recovered with GARBAGE (see: DataSalvage)
		je	@@DataDone
		
		; Scan each record
		mov	edx, [DDIR_START]
		add	edx, [DDIR_SIZE]
@@Data1:	
		mov	eax, [DDIR_START]
		add	eax, [DDIR_SIZE_INIT]
		cmp	edx, eax
		je	@@DataDone
		sub	edx, size ddirentry
		test	[edx.ddir_flags], QDATA_MARKDEL
		jz	@@Data1
		call	DustData
		jne	@@Data1
@@DataDone:


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_DUST:
DUST		endp


; ===============================================================
; = DUSTTHRESHOLD                     				=
; =                                                         	=
; = Code for the forth word:    DustThreshold			=
; =                                                         	=
; = Stack:  ( --- adr )                 			=
; ===============================================================

		align	4
DUSTTHRESHOLD	proc

		; Integer run-time
		DOVAR  	D_DUSTTHRESHOLD

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_DUSTTHRESHOLD:
DUSTTHRESHOLD	endp


; ===============================================================
; = UnAllot							=
; =                                                 		=
; = Usage							=
; =	Free memory allocated for a word with ALLOT.		=
; =								=
; = I) EDX : memory block nb					=
; =								=
; = M) none							=
; ===============================================================

		align	4
UnAllot		proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax


		; ----------------------------------------
		; Is-it the last block ? If yes, delete it
		; ----------------------------------------

		mov	eax, [DDIR_SIZE]
		sub	eax, size ddirentry
		cmp	eax, edx
		jne	@@NotLast
		
		; Update the memory used
		mov	eax, [edi + edx.ddir_blksize]
		sub	[DDA_SIZE], eax
		
		; Update the size of the table
		sub	[DDIR_SIZE], size ddirentry
		
		; Update memory allocated bounds
		mov	eax, [DDA_START]
		sub	eax, [DDA_SIZE]
		mov	[DDA_BOUND], eax

		; Done
		jmp	@@Exit
		
@@NotLast:


		; ----------------------------------------------------------------------
		; Update the amount of memory which can be salvaged by GARBAGE collector
		; ----------------------------------------------------------------------
		
		test	[edi + edx.ddir_flags], QDATA_MARKDEL
		jnz	@@UpdtSalvag
		mov	eax, [edi + edx.ddir_blksize]
		add	[DDIR_MEMSALV], eax
@@UpdtSalvag:	
		

		; ---------------------------------------------------------------------
		; Mark this block as free, but not deleted (it will be deleted by DUST)
		; ---------------------------------------------------------------------

		; Mark the block as to be deleted
		or	[edi + edx.ddir_flags], QDATA_MARKDEL

		
		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	eax
		ret

E_UnAllot:
UnAllot		endp


code32		ends
		end
