; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	NUMIO.ASM						!
; 									!
; Description:	Definitions for playing with integer I/O.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	BASE		BASE						!
;	DECIMAL		DECIMAL						!
;	HEX		HEXADECIMAL					!
;	<#		BSHARP						!
;	#		SHARP						!
;	U#		USHARP						!
;	HOLD		HOLD						!
;	SIGN		SIGN						!
;	#S		SHARPS						!
;	U#S		USHARPS						!
;	#>		ESHARP						!
;	<#S>		BSHARPSE					!
;	NUMBER		NUMBER						!
;	.		DOT						!
;	.R		DOTR						!
;	.L		DOTL						!
;	U.		UDOT						!
;	H.		HDOT						!
;	02H.		Z2HDOT						!
; 									!
; ----------------------------------------------------------------------+


		.386p
		jumps
		locals

code32		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc


; ===============================================================
; = BASE                            				=
; =                                                         	=
; = Code for the forth word:    BASE                		=
; =                                                         	=
; = Stack:  ( --- addr )                 			=
; =                                                         	=
; ===============================================================

		align	4
_BASE		proc

		; Get system var address
		DOVAR	D_BASE
		
		; Verify stack frame overflow, then return to caller
		OVSTACK
		
		; Return to caller
		ret

E__BASE:
_BASE		endp


; ===============================================================
; = DECIMAL							=
; =                                                 		=
; = Code for the forth word:	DECIMAL				=
; =								=
; = Stack:	( --- )						=
; =								=
; = REGISTERS ARE PRESERVED.					=
; ===============================================================

		align	4
DECIMAL		proc

		; Base for integer numeric I/O
		push	eax
		SETFVAR	D_BASE, 10
		pop	eax
		
		; Return to caller
		ret

E_DECIMAL:
DECIMAL		endp


; ===============================================================
; = HEXADECIMAL							=
; =                                                 		=
; = Code for the forth word:	HEXADECIMAL			=
; =								=
; = Stack:	( --- )						=
; =								=
; = REGISTERS ARE PRESERVED.					=
; ===============================================================

		align	4
HEXADECIMAL	proc

		; Base for integer numeric I/O
		push	eax
		SETFVAR	D_BASE, 16
		pop	eax
		
		; Return to caller
		ret

E_HEXADECIMAL:
HEXADECIMAL		endp


; ===============================================================
; = BSHARP							=
; =                                                 		=
; = Code for the forth word:	<#				=
; =								=
; = Stack:	( n --- n )					=
; ===============================================================

		align	4
BSHARP		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ----------------------------------------------
		; Reset pointer into Numeric Output Buffer (NOB)
		; ----------------------------------------------
		
		lea	eax, NOB - 1
		mov	byte ptr [eax], 0			; Set end zero
		dec	eax
		mov	[NOBPTR], eax

		
		; --------------------------------
		; Is the initial number negative ?
		; --------------------------------
		
		cmp	PARAM0, 0
		jl	@@Neg
		mov	[NegFlag], FALSE
		jmp	@@Pos
@@Neg:		mov	[NegFlag], TRUE
@@Pos:


		; ----------------------
		; Verify that BASE is ok
		; ----------------------

		GETFVAR	eax, D_BASE				; get base
		cmp	eax, 2
		jl	@@NOK
		cmp	eax, 36
		jle	@@OK
@@NOK:		SUMMON	DECIMAL
@@OK:		


		; ----------------
		; Return to caller
		; ----------------

		ret

E_BSHARP:
BSHARP		endp


; ===============================================================
; = SHARP							=
; =                                                 		=
; = Code for the forth word:	#				=
; =								=
; = Stack:	( n1 --- n2 )					=
; ===============================================================

		align	4
SHARP		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -----------------------
		; Diviseur (current base)
		; -----------------------

		GETFVAR	ebx, D_BASE				; Get base

		; Dividende
		mov	eax, [esi]
		cdq
		
		; Make division (signed)
		idiv	ebx

		; Store new quotient
		mov	[esi], eax
		
		; Store rest
		cmp	dl, 80h
		jb	@@Positive
		neg	dl
@@Positive:	mov	al, dl
		cmp	al, 9
		jg	@@LL1
		add	al, '0'
		jmp	@@LL2
@@LL1:		add	al, 55
@@LL2:		mov	ebx, [NOBPTR]
		mov	[ebx], al
		dec	ebx
		mov	[NOBPTR], ebx


		; ----------------
		; Return to caller
		; ----------------
		
		ret
E_SHARP:
SHARP		endp


; ===============================================================
; = USHARP							=
; =                                                 		=
; = Code for the forth word:	U#				=
; =								=
; = Stack:	( u1 --- u2 )					=
; =								=
; ===============================================================

		align	4
USHARP		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -----------------------
		; Diviseur (current base)
		; -----------------------
	
		GETFVAR	ebx, D_BASE				; Get base

		; Dividend
		mov	eax, [esi]
		clr	edx
		
		; Make division (unsigned)
		div	ebx

		; Store new quotient
		mov	[esi], eax
		
		; Store rest
		mov	al, dl
		cmp	al, 9
		jg	@@LL1
		add	al, '0'
		jmp	@@LL2
@@LL1:		add	al, 55
@@LL2:		mov	ebx, [NOBPTR]
		mov	[ebx], al
		dec	ebx
		mov	[NOBPTR], ebx


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_USHARP:
USHARP		endp


; ===============================================================
; = HOLD							=
; =                                                 		=
; = Code for the forth word:	HOLD				=
; =								=
; = Stack:	( n c --- n )					=
; =								=
; ===============================================================

		align	4
HOLD		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; -----------
		; Insert char
		; -----------

		mov	eax, PARAM0
		DISCARD	1				; drop
		mov	ebx, [NOBPTR]
		mov	[ebx], al
		dec	ebx
		mov	[NOBPTR], ebx


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_HOLD:
HOLD		endp


; ===============================================================
; = SIGN							=
; =                                                 		=
; = Code for the forth word:	SIGN				=
; =								=
; = Stack:	( n --- n )					=
; =								=
; ===============================================================

		align	4
SIGN		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ---------------------------------
		; Insert '-' if number was negative
		; ---------------------------------
		
		cmp	[NegFlag], FALSE
		je	@@Exit
		
		mov	ebx, [NOBPTR]
		mov	byte ptr [ebx], '-'
		dec	ebx
		mov	[NOBPTR], ebx


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_SIGN:
SIGN		endp


; ===============================================================
; = SHARPS							=
; =                                                 		=
; = Code for the forth word:	#S				=
; =								=
; = Stack:	( n --- 0 )					=
; =								=
; ===============================================================

		align	4
SHARPS		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; #
@@LL1:		SUMMON	SHARP					; #
		
		; 0 = 
		cmp	dword ptr [esi], 0
		jne	@@LL1

		; Insert '-' sign if negative
		SUMMON	SIGN					; SIGN


		; ----------------
		; Return to caller
		; ----------------
		
		ret
		
E_SHARPS:
SHARPS		endp


; ===============================================================
; = USHARPS							=
; =                                                 		=
; = Code for the forth word:	U#S				=
; =								=
; = Stack:	( un --- 0 )					=
; ===============================================================

		align	4
USHARPS		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; U#
@@LL1:		SUMMON	USHARP					; U#
		
		; 0 = 
		cmp	dword ptr [esi], 0
		jne	@@LL1


		; ----------------
		; Return to caller
		; ----------------
		
		ret
		
E_USHARPS:
USHARPS		endp


; ===============================================================
; = ESHARP							=
; =                                                 		=
; = Code for the forth word:	#>				=
; =								=
; = Stack:	( n --- addrz )					=
; =								=
; ===============================================================

		align	4
ESHARP		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -------------------------------
		; Create a (pseudo) strinh header 
		; -------------------------------

		mov	edx, [NOBPTR]
		sub	edx, (size strentry - 1)
		lea	eax, NOB
		sub	eax, edx
		sub	eax, (size strentry + 1)
		mov	[edx.str_currlen], eax 		; Current length
		inc	eax
		mov	[edx.str_currsize], eax		; Current size allocated
		mov	[edx.str_incsize], 0		; Increment size (if dynamic string), else 0 (if fixed string)
		mov	[edx.str_memblknb], -1		; NO Memory block nb (ref to DDIR)


		; --------------------------------------------
		; Put pointer into Numeric Output Buffer (NOB)
		; --------------------------------------------
		
		add	edx, size strentry
		mov	PARAM0, edx


		; ----------------
		; Return to caller
		; ----------------
		
		ret
		
E_ESHARP:
ESHARP		endp


; ===============================================================
; = BSHARPSE							=
; =                                                 		=
; = Code for the forth word:	<#S>				=
; =								=
; = Stack:	( n --- addr )					=
; =								=
; ===============================================================

		align	4
BSHARPSE	proc

		; <# #S #>
		SUMMON	BSHARP			; <#
		SUMMON	SHARPS			; #S
		SUMMON	ESHARP			; #>
		
		; Return to caller
		ret
		
E_BSHARPSE:
BSHARPSE	endp


; ===============================================================
; = NUMBER							=
; =                                                 		=
; = Code for the forth word:	NUMBER				=
; =								=
; = Stack:	( addrz --- n 1)	if conversion is OK	=
; =		( addrz --- 0)	if conversion is NOK		=
; =								=
; ===============================================================

		align	4
NUMBER		proc

		; Save used registers
		push	eax ebx ecx edx edi ebp

		; Get the base in ECX
		GETFVAR	ecx, D_BASE				; Get base
		mov	dl, cl					; Save current base in DL
		sub	ecx, 2
		mov	ecx, [POWERTAB + ecx * 4]

		; Is the 1st character '-' ?
		clr	dh					; Flag: dh=1 if '-'
		mov	ebp, PARAM0
		cmp	byte ptr [ebp], '-'
		jne	@@NotMinus
		inc	ebp
		inc	dh
@@NotMinus:

		; ESI: address of first char of the number
		; EDI: address of last  char of the number
		mov	edi, ebp
@@LL1:		cmp	byte ptr [edi], 0
		je	@@LL2
		inc	edi
		jmp	@@LL1
@@LL2:		dec	edi

		; Init number being converted
		mov	PARAM0, 0

		; Convert next digit
@@LL3:		clr	eax
		mov	al, [edi]

		; If lowercase, convert to uppercase
		cmp	al, 'a'
		jl	@@UL1
		cmp	al, 'z'
		jg	@@UL1
		sub	al, 020h
@@UL1:		
		
		; Is digit correct in the base ?
		cmp	al, '0'
		jl	@@NOK

		; Is base <= 10 ?
		cmp	dl, 10
		jg	@@BG10
		
		; Base is <= 10
		sub	al, '0'
		cmp	al, dl
		jge	@@NOK
		jmp	@@DP

		; Base is >= 10
@@BG10:		cmp	al, '9'
		jg	@@BG11
		sub	al, '0'
		jmp	@@DP
		
@@BG11:		sub	al, 55
		cmp	al, dl		
		jge	@@NOK
		cmp	al, 10
		jl	@@NOK

@@DP:
		; digit * power
		mov	ebx, [ecx]
		imul	eax, ebx
		add	PARAM0, eax

		; Test overflow error
		jno	@@NotOverflow

		; "Constant integer out of range."		
		add	esp, 6 * @WordSize			; Clean stack
		mov	ah, ERR_CONSTINTOUTOFRANGE		; Error msg #
		mov	edx, ebp				; Additionnal msg
		clr	ecx					; No line #
		call	doError

@@NotOverflow:	
		; Next digit
		cmp	edi, ebp
		je	@@OK
		dec	edi
		add	ecx, 4
		jmp	@@LL3

		; Not a valid number
@@NOK:		mov	PARAM0, FALSE				; ( adr --- ff )
		jmp	@@LL5

		; A valid number
@@OK:		cmpz	dh, FALSE
		jz	@@NotNeg
		neg	PARAM0
@@NotNeg:
		IPUSH	TRUE					; ( n --- n tf )


		; End of proc
		; -----------

@@LL5:
		; Restore ESI and EDI which are used localy
		pop	ebp edi edx ecx ebx eax
		cld

		; Return to caller
		ret

E_NUMBER:
NUMBER		endp


; ===============================================================
; = DOT								=
; =                                                 		=
; = Code for the forth word:	.				=
; =								=
; = Stack:	( n --- )					=
; =								=
; = ALL REGISTERS ARE PRESERVED					=
; ===============================================================

		align	4
DOT		proc

		; Verify stack frame
		VSTACK	1

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; Save all registers
		push	eax ebx ecx edx

		; <# #S >#
		SUMMON	BSHARP					; <#	( n --- n )
		SUMMON	SHARPS					; #S	( n --- 0 )
		SUMMON	ESHARP					; #>	( 0 --- addr )
		
		; TYPE
		SUMMON	_TYPE					; TYPE

		; Restore regs
		pop	edx ecx ebx eax

		; CTRL-BREAK allowed
		SUMMON	CTRLBREAK_REST

		; Return to caller
		ret

E_DOT:
DOT		endp


; ===============================================================
; = DOTR							=
; =                                                 		=
; = Code for the forth word:	.R				=
; =								=
; = Stack:	( n1 n2 --- )					=
; ===============================================================

		align	4
DOTR		proc

		; Verify stack frame
		VSTACK	1

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; Save all registers
		push	eax ebx ecx edx 

		; <# #S >#
		SUMMON	BSHARP					; <#	( n --- n )
		SUMMON	SHARPS					; #S	( n --- 0 )
		SUMMON	ESHARP					; #>	( 0 --- addrz )
		
		; TYPE
		SUMMON	DTYPE					; -TYPE

		; Restore regs
		pop	edx ecx ebx eax

		; CTRL-BREAK allowed
		SUMMON	CTRLBREAK_REST

		; Return to caller
		ret

E_DOTR:
DOTR		endp


; ===============================================================
; = DOTL							=
; =                                                 		=
; = Code for the forth word:	.L				=
; =								=
; = Stack:	( n1 n2 --- )					=
; ===============================================================

		align	4
DOTL		proc

		; Verify stack frame
		VSTACK	1

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; Save all registers
		push	eax ebx ecx edx

		; <# #S >#
		SUMMON	BSHARP					; <#	( n --- n )
		SUMMON	SHARPS					; #S	( n --- 0 )
		SUMMON	ESHARP					; #>	( 0 --- adr )
		
		; TYPE
		SUMMON	TYPED					; TYPE-

		; Restore regs
		pop	edx ecx ebx eax

		; CTRL-BREAK allowed
		SUMMON	CTRLBREAK_REST

		; Return to caller
		ret

E_DOTL:
DOTL		endp


; ===============================================================
; = UDOT							=
; =                                                 		=
; = Code for the forth word:	U.				=
; =								=
; = Stack:	( un --- )					=
; =								=
; = ALL REGISTERS ARE PRESERVED					=
; ===============================================================

		align	4
UDOT		proc

		; Verify stack frame
		VSTACK	1

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; Save all registers
		push	eax ebx ecx edx

		; <# #S >#
		SUMMON	BSHARP					; <#	( n --- n )
		SUMMON	USHARPS					; U#S	( n --- 0 )
		SUMMON	ESHARP					; #>	( 0 --- adr )
		
		; TYPE
		SUMMON	_TYPE					; TYPE

		; Restore regs
		pop	edx ecx ebx eax

		; CTRL-BREAK allowed
		SUMMON	CTRLBREAK_REST

		; Return to caller
		ret

E_UDOT:
UDOT		endp


; ===============================================================
; = HDOT							=
; =                                                 		=
; = Code for the forth word:	H.				=
; =								=
; = Stack:	( n --- )					=
; =								=
; = ALL REGISTERS ARE PRESERVED					=
; ===============================================================

		align	4
HDOT		proc

		; Verify stack frame
		VSTACK	1

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; Save all registers
		push	eax ebx ecx edx

		; Do H.		
		SUMMON	_BASE					; BASE
		SUMMON	AT					; @
		SUMMON	HEXADECIMAL				; HEXADECIMAL
		SUMMON	SWAP					; SWAP
		SUMMON	UDOT					; U.
		SUMMON	_BASE					; BASE
		SUMMON	STORE					; !

		; Restore regs
		pop	edx ecx ebx eax

		; CTRL-BREAK allowed
		SUMMON	CTRLBREAK_REST

		; Return to caller
		ret

E_HDOT:
HDOT		endp


; ===============================================================
; = Z2HDOT							=
; =                                                 		=
; = Code for the forth word:	02H.				=
; =								=
; = Stack:	( c --- )					=
; =								=
; = ALL REGISTERS ARE PRESERVED					=
; ===============================================================

		align	4
Z2HDOT		proc

		; Verify stack frame
		VSTACK	1

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; Save all registers
		push	eax ebx ecx edx

		; Do H.		
		SUMMON	_BASE					; BASE
		SUMMON	AT					; @
		SUMMON	HEXADECIMAL				; HEXADECIMAL
		SUMMON	SWAP					; SWAP
		SUMMON	BSHARP					; <#
		SUMMON	SHARP					; #
		SUMMON	SHARP					; #
		SUMMON	ESHARP					; #>
		SUMMON	_TYPE					; TYPE
		SUMMON	_BASE					; BASE
		SUMMON	STORE					; !

		; Restore regs
		pop	edx ecx ebx eax

		; CTRL-BREAK allowed
		SUMMON	CTRLBREAK_REST

		; Return to caller
		ret

E_Z2HDOT:
Z2HDOT		endp


code32		ends
		end

