; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	PORTIO.ASM						!
; 									!
; Description:	Definitions for playing with port I/O.			!
; 									!
; List of FROTH definitions defined in this module:			!
;	INPORT		IN						!
;	CINPORT		CIN						!
;	WINPORT		WIN						!
;	OUTPORT		OUT						!
;	COUTPORT	COUT						!
;	WOUTPORT	WOUT						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc


; ===============================================================
; = INPORT							=
; =								=
; = Code for the forth word:	IN				=
; =								=
; = Stack:	( n1 --- n2 )					=
; ===============================================================

		align	4
INPORT		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------
		; Do operation
		; ------------
		
		mov	edx, PARAM0				; Get <n1>
		in	eax, dx
		mov	PARAM0, eax
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_INPORT:
INPORT		endp


; ===============================================================
; = CINPORT							=
; =								=
; = Code for the forth word:	CIN				=
; =								=
; = Stack:	( n1 --- n2 )					=
; ===============================================================

		align	4
CINPORT		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------
		; Do operation
		; ------------
		
		mov	edx, PARAM0				; Get <n1>
		clr	eax
		in	al, dx
;		cbw						; AX = AL
;		cwde						; EAX = AX
		mov	PARAM0, eax
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_CINPORT:
CINPORT		endp


; ===============================================================
; = WINPORT							=
; =								=
; = Code for the forth word:	WIN				=
; =								=
; = Stack:	( n1 --- n2 )					=
; ===============================================================

		align	4
WINPORT		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------
		; Do operation
		; ------------
		
		mov	edx, PARAM0				; Get <n1>
		clr	eax
		in	ax, dx
;		cwde						; EAX = AX
		mov	PARAM0, eax
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_WINPORT:
WINPORT		endp


; ===============================================================
; = OUTPORT							=
; =								=
; = Code for the forth word:	OUT				=
; =								=
; = Stack:	( n1 n2 --- )					=
; ===============================================================

		align	4
OUTPORT		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------
		; Do operation
		; ------------

		mov	eax, PARAM1					; Get <n1>
		mov	edx, PARAM0					; Get <n2>
		out	dx, eax
		DISCARD	2						; drop drop
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_OUTPORT:
OUTPORT		endp


; ===============================================================
; = COUTPORT							=
; =								=
; = Code for the forth word:	COUT				=
; =								=
; = Stack:	( n1 n2 --- )					=
; ===============================================================

		align	4
COUTPORT	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------
		; Do operation
		; ------------

		mov	eax, PARAM1					; Get <n1>
		mov	edx, PARAM0					; Get <n2>
		out	dx, al
		DISCARD	2						; drop drop
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_COUTPORT:
COUTPORT	endp


; ===============================================================
; = WOUTPORT							=
; =								=
; = Code for the forth word:	WOUT				=
; =								=
; = Stack:	( n1 n2 --- )					=
; ===============================================================

		align	4
WOUTPORT	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------
		; Do operation
		; ------------

		mov	eax, PARAM1					; Get <n1>
		mov	edx, PARAM0					; Get <n2>
		out	dx, ax
		DISCARD	2						; drop drop
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_WOUTPORT:
WOUTPORT	endp


code32		ends
		end
