; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	SAVEEXE.ASM						!
; 									!
; Description:	Definitions for playing with building turkey		!
;		application.						!
; 									!
; List of FROTH definitions defined in this module:			!
;	SAVE_EXE	SAVE.EXE					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc					; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc


; ===============================================================
; = SAVE_EXE							=
; =								=
; = Code for the forth word:	SAVE_EXE			=
; =								=
; = Stack:	( addrz --- n )					=
; ===============================================================

		align	4
SAVE_EXE	proc

		; -----------------------------------------
		; Save input parameters into Temp variables
		; -----------------------------------------

t_FileHdl	equ	word ptr [ebp - 4]
t_FrothExeHdl	equ	word ptr [ebp - 8]
t_FrothExeSize	equ	[ebp - 12]
t_BuffAddr	equ	[ebp - 16]
t_nb		equ	dword ptr [ebp - 20]
		enter	5 * @WordSize, 0


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1

		
		; -----------------
		; Nothing to save ?
		; -----------------
		
		mov	eax, [CDA_SIZE]
		cmp	eax, [CDA_SIZE_INIT]
		jne	@@Saveable

		; "Nothing to save !"
		mov	ah, ERR_FROTHEXENOUWORDS		; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Saveable:	

		; -----------------------------
		; Garbage collector (NO hole !)
		; -----------------------------
		
		SUMMON	DUST


		; --------------------
		; Try to open the file
		; --------------------
		
		; Pointer to filename FROTH.EXE
		mov	ebx, PARAM0		; Get filename
		
		; Access mode
		mov	eax, 2091h		; W/O  b13: no int24h   b7: don't pass handle to child   b4: no sharing

		; Authorize new file creation ? (update ZF flag)
		mov	cl, 1
		
		; Call BIOS
		call	doWideOpenFile
		jnc	@@OpenOk
		
		mov	PARAM0, -1		; Error
		jmp	@@CloseFile
		
@@OpenOk:	
		; Save file handle
		mov	[t_FileHdl], bx
		

		; ------------------------
		; Read FROTH.EXE in memory
		; ------------------------

		; Open FROTH.EXE file
		mov	edx, [QuoteStrBuff]
		call	GetProgName
		mov	ebx, edx
		mov	eax, 2090h		; R/O  b13: no int24h   b7: don't pass handle to child   b4: no sharing
		clr	cl
		call	doWideOpenFile
		jc	@@CloseFile
		mov	[t_FrothExeHdl], bx

		; Size of FROTH.EXE
		call	doFileSize
		mov	[t_FrothExeSize], ebx

		; Try to find memory for a buffer enought big to hold FROTH.EXE
		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, [t_FrothExeSize]
		cmp	ecx, edx
		jl	@@CloseFile
		mov	[t_BuffAddr], ecx

		; Read entire FROTH.EXE file
		mov	bx, [t_FrothExeHdl]			; File handle
		mov	edx, [t_BuffAddr]			; Buffer address
		mov	ecx, [t_FrothExeSize]			; Bytes to read
		call	doReadFile				; in EBX: bytes read from the file

		; Close FROTH.EXE file
		mov	bx, [t_FrothExeHdl]
		call	doCloseFile


		; ----------------------------------
		; Find mark into buffer, then update
		; ----------------------------------

		; Find mark
		mov	ebx, [t_BuffAddr]
		clr	edx
@@FindMark:	cmp	dword ptr [ebx + edx], 'iSlO'
		jne	@@FindMark2
		cmp	dword ptr [ebx + edx + 4], 'CHPC'
		je	@@FindMark3
@@FindMark2:	inc	edx
		jmp	@@FindMark
@@FindMark3:	

		; Now modify variables
		lea	ebx, [ebx + edx + 8]
		mov	byte ptr [ebx], TRUE			; ExeFile
		
		; Word to execute relative address
		mov	eax, [NIA_START]
		add	eax, [NIA_SIZE]
		sub	eax, size wordentry
		mov	eax, [eax.word_realaddr]
		sub	eax, [CDA_START]
		mov	dword ptr [ebx + 1*4], eax		; ExeWordAddr

		; Adress and Size of user's code into .EXE file
		mov	eax, [t_FrothExeSize]
		mov	dword ptr [ebx + 2*4], eax		; ExeCdaBegin
		mov	eax, [CDA_SIZE]
		sub	eax, [CDA_SIZE_INIT]			; Bytes to write
		mov	dword ptr [ebx + 3*4], eax		; ExeCdaSize

		; Adress of user's data into .EXE file
		add	eax, [t_FrothExeSize]			
		mov	dword ptr [ebx + 4*4], eax		; ExeDdaBegin
		mov	eax, [DDIR_SIZE]
		sub	eax, [DDIR_SIZE_INIT]			; Bytes to write
		mov	dword ptr [ebx + 5*4], eax		; ExeDdaSize


		; -----------------------------		
		; Save FROTH.EXE into .EXE file
		; -----------------------------		
		
		mov	bx, [t_FileHdl]				; File handle
		mov	ecx, [t_FrothExeSize]			; Bytes to write
		mov	edx, [t_BuffAddr]			; Buffer address
		call	doWriteFile
		jc	@@CloseFile


		; ------------------------------
		; Save user defined words (code)
		; ------------------------------

		mov	bx, [t_FileHdl]				; File handle
		mov	ecx, [CDA_SIZE]
		sub	ecx, [CDA_SIZE_INIT]			; Bytes to write
		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE_INIT]			; Buffer address
		call	doWriteFile
		jc	@@CloseFile


		; ------------------------------
		; Save user defined words (data)
		; ------------------------------

		mov	bx, [t_FileHdl]				; File handle
		mov	ecx, [DDIR_SIZE]
		sub	ecx, [DDIR_SIZE_INIT]			; Bytes to write
		mov	edx, [DDIR_START]
		add	edx, [DDIR_SIZE_INIT]			; Buffer address
		call	doWriteFile
		jc	@@CloseFile


		; ----------------
		; Save static data
		; ----------------

		mov	ebx, [DDIR_START]
		add	ebx, [DDIR_SIZE_INIT]

		mov	edx, [DDIR_START]
		add	edx, [DDIR_SIZE]

@@SaveDDA:	
		; End reached ?
		cmp	ebx, edx
		je	@@SaveDDA3

		; Save only static variables (not auto, not dynamic)
		test	[ebx.ddir_flags], (QDATA_AUTO or QDATA_DYNA)		; Automatic or Dynamic variable ?
		jnz	@@SaveDDA2

		; Save DDIR reference
		pushad
		mov	eax, ebx
		sub	eax, [DDIR_START]
		mov	[t_nb], eax
		mov	bx, [t_FileHdl]				; File handle
		mov	ecx, @WordSize				; Bytes to write
		lea	edx, t_nb				; Buffer address
		call	doWriteFile
		popad

		; Save DDA
		pushad
		mov	ecx, [ebx.ddir_blksize]			; Bytes to write
		mov	edx, [ebx.ddir_blkaddr]			; Buffer address
		mov	bx, [t_FileHdl]				; File handle
		call	doWriteFile
		popad
		jmp	@@SaveDDA2

		; Next variable, if any
@@SaveDDA2:	add	ebx, size ddirentry
		jmp	@@SaveDDA

@@SaveDDA3:	

		; Write final -1
		mov	[t_nb], -1
		mov	bx, [t_FileHdl]				; File handle
		mov	ecx, @WordSize				; Bytes to write
		lea	edx, t_nb				; Buffer address
		call	doWriteFile


		; ---------------------------------
		; Push size of .EXE file onto stack
		; ---------------------------------

		mov	bx, [t_FileHdl]				; File handle
		call	doFilePos
		mov	PARAM0, ebx				; Store file size


		; --------------
		; Close the file
		; --------------
		
@@CloseFile:	mov	bx, [t_FileHdl]
		call	doCloseFile

		
		; ----------------
		; Return to caller
		; ----------------

		leave		
		ret
		
E_SAVE_EXE:
SAVE_EXE	endp


code32		ends
		end
