; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	SCREEN.ASM						!
; 									!
; Description:	Definitions for playing with the screen (text mode).	!
; 									!
; List of FROTH definitions defined in this module:			!
;	NBPAGES			#PAGES					!
;	PAGENB			PAGE#					!
;	NBROWS			#ROWS					!
;	NBCOLS			#COLS					!
;	SCREENBCK		SCREENBACKGROUND			!
;	QUPDATECURSOR		?UpdateCursor				!
;	SETVIDEOMODE		SetVideoMode				!
;	GetVideoMode		GetVideoMode				!
;	SELECTPAGE		PAGE>					!
;	CHANGEPAGE		>PAGE					!
;	CURSOR			CURSOR					!
;	GOTOXY			GOTOXY					!
;	GOTOX			GOTOX					!
;	GOTOY			GOTOY					!
;	WHEREX			WHEREX					!
;	WHEREY			WHEREY					!
;	BACKFOREGROUND		BackForeGround				!
;	ATXYCHAR		@XYCHAR					!
;	ATXYACHAR		@XYACHAR				!
;	STOREXYCHAR		!XYCHAR					!
;	STOREXYACHAR		!XYACHAR				!
;	STOREXYNCHAR		!XYNCHAR				!
;	STOREXYNACHAR		!XYNACHAR				!
;	ATTRCLS			AttrCls					!
;	CLS			CLS					!
;	STORESTRING		!STRING					!
;	STOREASTRING		!ASTRING				!
;	STOREXYSTRING		!XYSTRING				!
;	STOREXYASTRING		!XYASTRING				!
;	TOXLSTRING_CHAR		>xlstring_char				!
;	TOXLSTRING_ATTR		>xlstring_attr				!
;	STOREXYLSTRING		!XYLSTRING				!
;	ACLRLINE		ACLRLINE				!
;	CLRLINE			CLRLINE					!
;	ACLREOL			ACLREOL					!
;	CLREOL			CLREOL					!
;	LOADCHARSET_8x8		LoadCharset_8x8				!
;	LOADCHARSET_8x14	LoadCharset_8x14			!
;	LOADCHARSET_8x16	LoadCharset_8x16			!
;	BLACK			BLACK					!
;	BLUE			BLUE					!
;	GREEN			GREEN					!
;	CYAN			CYAN					!
;	RED			RED					!
;	MAGENTA			MAGENTA					!
;	BROWN			BROWN					!
;	LIGHTGRAY		LIGHTGRAY				!
;	DARKGRAY		DARKGRAY				!
;	LIGHTBLUE		LIGHTBLUE				!
;	LIGHTGREEN		LIGHTGREEN				!
;	LIGHTCYAN		LIGHTCYAN				!
;	LIGHTRED		LIGHTRED				!
;	LIGHTMAGENTA		LIGHTMAGENTA				!
;	YELLOW			YELLOW					!
;	WHITE			WHITE					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

		
; ---> Local variables ---> ------------------------------------------------------------------------------------------------------

		align	4
_xylstring_char	db	'~'

		align	4
_xylstring_attr	db	8 dup (0)

		align	4
_msgbuff	db	MAXSZ_SCREENSTR+1

; ---> End of local variables ---> -----------------------------------------------------------------------------------------------


; ===============================================================
; = isUpdateCursor						=
; =                                                         	=
; = Usage							=
; =	Local function to see if FROTH variable			=
; =	'?UpdateCursor'						=
; =                                                         	=
; = I) none							=
; = O) CF = 1 if true, else false				=
; = M) CF only							=
; ===============================================================

		align	4
isUpdateCursor	proc

		; Test flag
		push	eax
		mov	eax, [edi + ddir_blkaddr + D_QUPDATECURSOR]
		cmp	dword ptr [eax], FALSE
		jne	@@On
		
@@Off:		pop	eax
		clc						; CF = 0
		ret						; jmp	@@Exit

@@On:		pop	eax
		stc						; CF = 1

		; Return to caller
@@Exit:		ret

isUpdateCursor	endp

		
; ===============================================================
; = NBPAGES                            				=
; =								=
; = Code for the forth word:    #Pages                		=
; =								=
; = Stack:  ( --- n )	                 			=
; =								=
; = Usage:							=
; =	Return the number of pages for the current video mode.	=
; ===============================================================

		align	4
NBPAGES		proc

		; Read data
		GETFVAR	eax, D_NBPAGES
		sub	esi, @WordSize
		mov	[esi], eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_NBPAGES:
NBPAGES		endp


; ===============================================================
; = PAGENB                            				=
; =								=
; = Code for the forth word:    PAGE#                		=
; =								=
; = Stack:  ( --- addr )                 			=
; ===============================================================

		align	4
PAGENB		proc

		; Get system integer var address
		DOVAR	D_PAGENB

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_PAGENB:
PAGENB		endp


; ===============================================================
; = NBROWS                            				=
; =								=
; = Code for the forth word:    #Rows                		=
; =								=
; = Stack:  ( --- n )	                 			=
; =								=
; = Usage:							=
; =	Return the number of rows for the current video mode.	=
; ===============================================================

		align	4
NBROWS		proc

		; Read data
		GETFVAR	eax, D_NBROWS
		sub	esi, @WordSize
		mov	[esi], eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_NBROWS:
NBROWS		endp


; ===============================================================
; = NBCOLS                            				=
; =								=
; = Code for the forth word:    #Cols                		=
; =								=
; = Stack:  ( --- n ) 	                			=
; ===============================================================

		align	4
NBCOLS		proc

		; Read data
		GETFVAR	eax, D_NBCOLS
		sub	esi, @WordSize
		mov	[esi], eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_NBCOLS:
NBCOLS		endp


; ===============================================================
; = getScreenBckg						=
; =								=
; = Usage:							=
; =	Read the froth integer variable 'ScreenBackground'	=
; =								=
; = I) none							=
; = O) AL = attribute						=
; = M) none, else AL						=
; ===============================================================

		align	4
getScreenBckg	proc

		; Save used registers
		push	ebx

		; Compute the data address (in EAX)
		mov	ebx, [edi + ddir_blkaddr + D_SCREENBCKG]
		
		; Get value
		mov	al, [ebx]

		; Return to caller
		pop	ebx
		ret

getScreenBckg	endp

; ===============================================================
; = SCREENBCKG                         				=
; =								=
; = Code for the forth word:    ScreenBackground       		=
; =								=
; = Stack:  ( --- adr ) 	                		=
; ===============================================================

		align	4
SCREENBCKG	proc

		; Get system integer var address
		DOVAR	D_SCREENBCKG
		
		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_SCREENBCKG:
SCREENBCKG	endp


; ===============================================================
; = QUPDATECURSOR                      				=
; =								=
; = Code for the forth word:    ?UpdateCursor			=
; =								=
; = Stack:  ( --- addr ) 	                		=
; ===============================================================

		align	4
QUPDATECURSOR	proc

		; Get system integer var address
		DOVAR	D_QUPDATECURSOR
		
		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_QUPDATECURSOR:
QUPDATECURSOR	endp


; ===============================================================
; = AskVideoInfos						=
; =								=
; = Usage:							=
; =	Ask informations on current video mode.			=
; =								=
; = I) none							=
; = O) EDX : pointer to buffer.					=
; = M) none							=
; ===============================================================

		align	4
AskVideoInfos	proc

		; Save used registers
		push	ax

		; Call BIOS function
		mov 	v86r_ah, 1Bh				; Function asked
		mov	v86r_bx, 0				; BX = 0
		mov	edx, [AskVideoInfBuff]			; Buffer address
		HILO	v86r_es, v86r_di, edx			; Pointer to buffer
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h

		; Return to caller
		pop	ax
		ret

AskVideoInfos	endp


; ===============================================================
; = doGetScreenInf						=
; =								=
; = Usage:							=
; =	Update the pseudo-constants: #rows #cols #pages		=
; =								=
; = Called by:							=
; =	SETVIDEOMODE  COLD					=
; ===============================================================

		align	4
doGetScreenInf	proc

		; -------------------------
		; Get the number of columns
		; -------------------------

		; Call BIOS GetVideoMode function to have the nb of cols
		mov 	v86r_ah, 0Fh			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		movzx	eax, v86r_ah			; Get nb of cols
		
		; Store value into 'constant'
		mov	ebx, [edi + ddir_blkaddr + D_NBCOLS]
		mov	[ebx], eax


		; ----------------------
		; Get the number of rows
		; ----------------------
		
		; Call BIOS GetRowsNb function to have the nb of rows
		mov	v86r_al, 30h			; ???
		mov 	v86r_ah, 11h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		movzx	eax, v86r_dl			; Get nb of rows
		inc	eax				; For example: for 25 lines, BIOS give 18h (24)
		
		; Store value into 'constant'
		mov	ebx, [edi + ddir_blkaddr + D_NBROWS]
		mov	[ebx], eax


		; -----------------------------
		; Get the number of video pages
		; -----------------------------
		
		call	AskVideoInfos
		movzx	eax, byte ptr [edx + 29h]		; # of video pages
		dec	eax
		and	eax, 7					; No more than 8 video pages
		inc	eax
		mov	ebx, [edi + ddir_blkaddr + D_NBPAGES]
		mov	[ebx], eax				; screen #pages !
		mov	ebx, [edi + ddir_blkaddr + D_GNBPAGES]
		mov	[ebx], eax				; graphics #gpages !


		; ----------------
		; Return to caller
		; ----------------
		
		ret

doGetScreenInf	endp


; ===============================================================
; = SETVIDEOMODE						=
; =								=
; = Code for the forth word:	SetVideoMode			=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
SETVIDEOMODE	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------------
		; Call BIOS function
		; ------------------

		mov	eax, [esi]			; Pop video mode asked
		add	esi, @WordSize
		mov	v86r_al, al			; Input: video mode
		mov 	v86r_ah, 00h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; -----------------------------------------
		; Get the number of columns, rows and pages
		; -----------------------------------------

		call	doGetScreenInf
		

		; -------------------------
		; Select the 1st video page
		; -------------------------
		
		mov	eax, [edi.ddir_blkaddr + D_PAGENB]	; PAGE#
		mov	dword ptr [eax], 0			; 0!


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SETVIDEOMODE:
SETVIDEOMODE	endp


; ===============================================================
; = GETVIDEOMODE						=
; =								=
; = Code for the forth word:	GetVideoMode			=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

		align	4
GETVIDEOMODE	proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; ------------------
		; Call BIOS function
		; ------------------

		mov 	v86r_ah, 0Fh			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ---------------------------
		; Push the current video mode
		; ---------------------------

		movzx	eax, v86r_al
		sub	esi, @WordSize
		mov	[esi], eax


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; --------------------------------------------------
		; Verify stack frame overflow, then return to caller
		; --------------------------------------------------
		
		OVSTACK


		; Return to caller
		ret
		
E_GETVIDEOMODE:
GETVIDEOMODE	endp


; ===============================================================
; = SELECTPAGE							=
; =								=
; = Code for the forth word:	PAGE>				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
SELECTPAGE	proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; Get current page nb
		GETFVAR	eax, D_PAGENB			; PAGE# @

		; Call BIOS function
		mov 	v86r_ah, 05h			; Function asked
		mov	v86r_al, al			; Page number
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		; Update graphics page address
		GETFVAR	eax, D_PAGENB			; PAGE# @
		sub	esi, @WordSize
		mov	PARAM0, eax
		SUMMON	SELECTGPAGE


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; Return to caller
		ret

E_SELECTPAGE:
SELECTPAGE	endp


; ===============================================================
; = CHANGEPAGE							=
; =								=
; = Code for the forth word:	>PAGE				=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
CHANGEPAGE	proc

		; Verify stack frame
		VSTACK	1

		; Put current page nb
		DOINT  	D_PAGENB			; PAGE#
		SUMMON	STORE				; !

		; Do function
		SUMMON	SELECTPAGE			; PAGE>

		; Return to caller
		ret

E_CHANGEPAGE:
CHANGEPAGE	endp


; ===============================================================
; = CURSOR							=
; =								=
; = Code for the forth word:	CURSOR				=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
CURSOR		proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -----------------------
		; Get the number of lines
		; -----------------------
		
		call	AskVideoInfos
		movzx	eax, byte ptr [edx + 23h]		; # of lines

		
		; --------------
		; 0 : cursor off
		; --------------
		
		cmp	dword ptr [esi], 0
		jne	@@Not0

		inc	al
		mov	v86r_ch, al			; Start line
		mov	v86r_cl, al			; End line
		jmp	@@DoCurs			; Call BIOS

@@Not0:


		; ---------------------
		; 1 : cursor under char
		; ---------------------
		
		cmp	dword ptr [esi], 1
		jne	@@Not1

		sub	al, 2
		mov	v86r_ch, al			; Start line
		inc	al
		mov	v86r_cl, al			; End line
		jmp	@@DoCurs			; Call BIOS

@@Not1:


		; ----------------
		; 2 : plain cursor
		; ----------------
		
		cmp	dword ptr [esi], 2
		jne	@@Exit

		dec	al
		mov	v86r_cl, al			; End line
		clr	al
		mov	v86r_ch, al			; Start line
		jmp	@@DoCurs			; Call BIOS


		; ---------
		; Call BIOS
		; ---------

@@DoCurs:	mov 	v86r_ah, 01h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ----------------
		; Return to caller
		; ----------------

@@Exit:		add	esi, @WordSize			; drop


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST

		ret

E_CURSOR:
CURSOR		endp


; ===============================================================
; = GOTOXY							=
; =								=
; = Code for the forth word:	GotoXY				=
; =								=
; = Stack:	( x y --- )					=
; ===============================================================

		align	4
GOTOXY		proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ------------------
		; Call BIOS function
		; ------------------

		call	doGetVideoPage			; PAGE#
		mov	eax, [esi]			; Get row
		mov	v86r_dh, al			; Input: row #
		mov	eax, [esi + @WordSize]		; Get column
		add	esi, 2 * @WordSize		; pop pop
		mov	v86r_dl, al			; Input: col #
		mov 	v86r_ah, 02h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_GOTOXY:
GOTOXY		endp


; ===============================================================
; = GOTOX							=
; =								=
; = Code for the forth word:	GotoX				=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
GOTOX		proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ---------------------------------
		; What is the actual row position ?
		; ---------------------------------

		call	doGetVideoPage			; PAGE#
		mov 	v86r_ah, 03h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		clr	eax


		; ------------------
		; Call BIOS function
		; ------------------

		call	doGetVideoPage			; PAGE#
;		mov	v86r_dh, al			; Input: row #
		mov	eax, [esi]			; Get column
		add	esi, @WordSize			; pop
		mov	v86r_dl, al			; Input: col #
		mov 	v86r_ah, 02h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_GOTOX:
GOTOX		endp


; ===============================================================
; = GOTOY							=
; =								=
; = Code for the forth word:	GotoY				=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
GOTOY		proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -------------------------------------
		; What is the current column position ?
		; -------------------------------------

		call	doGetVideoPage			; PAGE#
		mov 	v86r_ah, 03h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ------------------
		; Call BIOS function
		; ------------------

		call	doGetVideoPage			; PAGE#
		mov	eax, [esi]			; Get row
		mov	v86r_dh, al			; Input: row #
		add	esi, @WordSize			; pop
;		mov	v86r_dl, al			; Input: col #
		mov 	v86r_ah, 02h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_GOTOY:
GOTOY		endp


; ===============================================================
; = WHEREXY							=
; =								=
; = Code for the forth word:	WhereXY				=
; =								=
; = Stack:	( --- x y )					=
; ===============================================================

		align	4
WHEREXY		proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; ------------------
		; Call BIOS function
		; ------------------

		call	doGetVideoPage				; PAGE#
		mov 	v86r_ah, 03h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		movzx	eax, v86r_dl				; Output: col #
		mov	[esi - @WordSize], eax
		mov	al, v86r_dh				; Output: row #
		sub	esi, 2 * @WordSize			; push push
		mov	[esi], eax


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; --------------------------------------------------
		; Verify stack frame overflow, then return to caller
		; --------------------------------------------------
		
		OVSTACK


		; Return to caller
		ret
		
E_WHEREXY:
WHEREXY		endp


; ===============================================================
; = WHEREX							=
; =								=
; = Code for the forth word:	WhereX				=
; =								=
; = Stack:	( --- x )					=
; ===============================================================

		align	4
WHEREX		proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; ------------------
		; Call BIOS function
		; ------------------

		call	doGetVideoPage				; PAGE#
		mov 	v86r_ah, 03h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		movzx	eax, v86r_dl				; Output: col #
		sub	esi, @WordSize
		mov	[esi], eax


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; --------------------------------------------------
		; Verify stack frame overflow, then return to caller
		; --------------------------------------------------
		
		OVSTACK


		; Return to caller
		ret
		
E_WHEREX:
WHEREX		endp


; ===============================================================
; = WHEREY							=
; =								=
; = Code for the forth word:	WhereY				=
; =								=
; = Stack:	( --- y )					=
; ===============================================================

		align	4
WHEREY		proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; ------------------
		; Call BIOS function
		; ------------------

		call	doGetVideoPage				; PAGE#
		mov 	v86r_ah, 03h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		movzx	eax, v86r_dh				; Output: row #
		sub	esi, @WordSize
		mov	[esi], eax


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; --------------------------------------------------
		; Verify stack frame overflow, then return to caller
		; --------------------------------------------------
		
		OVSTACK


		; Return to caller
		ret
		
E_WHEREY:
WHEREY		endp


; ===============================================================
; = BackForeGround						=
; =								=
; = Code for the forth word:	BackForeGround			=
; =								=
; = Stack:	( n1 n2 --- n3 )				=
; ===============================================================

		align	4
BACKFOREGROUND	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2

		
		; -----------------
		; Compute attribute
		; -----------------

		mov	eax, [esi]				; Get foreground
		add	esi, @WordSize
		mov	ebx, [esi]				; Get background
		shl	ebx, 4					; background << 4
		add	eax, ebx
		mov	[esi], eax
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_BACKFOREGROUND:
BACKFOREGROUND	endp


; ===============================================================
; = ATXYCHAR							=
; =								=
; = Code for the forth word:	@XYCHAR				=
; =								=
; = Stack:	( x y --- attr char )				=
; ===============================================================

		align	4
ATXYCHAR		proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; -------------------
		; Set cursor position
		; -------------------

		call	doGetVideoPage				; PAGE#
		mov	eax, [esi]				; Get row
		mov	v86r_dh, al				; Input: row #
		mov	eax, [esi + @WordSize]			; Get column
		mov	v86r_dl, al				; Input: col #
		mov 	v86r_ah, 02h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h


		; ------------------
		; Call BIOS function
		; ------------------

;		call	doGetVideoPage				; PAGE#
		mov 	v86r_ah, 08h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		
		; Push char
		movzx	eax, v86r_al				; Get character
		add	esi, @WordSize				; drop
		mov	[esi], eax
		

		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_ATXYCHAR:
ATXYCHAR		endp


; ===============================================================
; = ATXYACHAR							=
; =								=
; = Code for the forth word:	@XYACHAR			=
; =								=
; = Stack:	( x y --- attr char )				=
; ===============================================================

		align	4
ATXYACHAR	proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; -------------------
		; Set cursor position
		; -------------------

		call	doGetVideoPage				; PAGE#
		mov	eax, [esi]				; Get row
		mov	v86r_dh, al				; Input: row #
		mov	eax, [esi + @WordSize]			; Get column
		mov	v86r_dl, al				; Input: col #
		mov 	v86r_ah, 02h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h


		; ------------------
		; Call BIOS function
		; ------------------

;		call	doGetVideoPage				; PAGE#
		mov 	v86r_ah, 08h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		
		; Push attribute
		movzx	eax, v86r_ah				; Get attribute
		mov	[esi + @WordSize], eax
		
		; Push char
		mov	al, v86r_al				; Get character
		mov	[esi], eax
		

		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_ATXYACHAR:
ATXYACHAR	endp


; ===============================================================
; = doGetVideoPage						=
; =								=
; = Usage:							=
; =	Return in AL the value of the user variable PAGE#	=
; =								=
; = I) none							=
; = O) v86r_bh : page#						=
; = M) EAX							=
; ===============================================================

		align	4
doGetVideoPage	proc

		; Used registers
		push	eax

		; The data address (EAX)
		mov	eax, [DDIR_START]			; Don't use EDI
		mov	eax, [eax.ddir_blkaddr + D_PAGENB]
		
		; The user variable PAGE# value
		mov	al, [eax]
		mov	v86r_bh, al				; Input: page #

		; Return to caller
		pop	eax
		ret

doGetVideoPage	endp


; ===============================================================
; = STOREXYCHAR							=
; =								=
; = Code for the forth word:	!xychar				=
; =								=
; = Stack:	( x y char --- )				=
; ===============================================================

		align	4
STOREXYCHAR	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; --------------------------------
		; Call BIOS function for 'WhereXY'
		; --------------------------------

		call	isUpdateCursor				; ?UpdateCusor @
		jc	@@UpdtCurs
		call	doGetVideoPage				; PAGE#
		mov 	v86r_ah, 03h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		mov	dx, v86r_dx				; row in DH, col in DL
@@UpdtCurs:	


		; -------------------
		; Set cursor position
		; -------------------

		call	doGetVideoPage				; PAGE#
		mov	eax, [esi + 1 * @WordSize]		; Get row
		mov	v86r_dh, al				; Input: row #
		mov	eax, [esi + 2 * @WordSize]		; Get column
		mov	v86r_dl, al				; Input: col #
		mov 	v86r_ah, 02h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h


		; --------------------------------
		; Call BIOS function to print char
		; --------------------------------

;		call	doGetVideoPage				; PAGE#
		mov	eax, [esi]				; pop char
		mov	v86r_al, al				; Input: char to print
		add	esi, 3 * @WordSize			; drop drop drop
		call	getScreenBckg				; Get current attribute
		mov	v86r_bl, al				; Input: attribute / color
		mov	v86r_cx, 1				; Only one character to print
		mov 	v86r_ah, 09h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h

	
		; ----------------------------------
		; Restore previous cursor position ?
		; ----------------------------------

		call	isUpdateCursor				; ?UpdateCusor @
		jc	@@UpdtCurs2
		call	doGetVideoPage				; PAGE#
		mov	v86r_dx, dx				; Row in DH, Col in DL
		mov 	v86r_ah, 02h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
@@UpdtCurs2:	

		
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_STOREXYCHAR:
STOREXYCHAR	endp


; ===============================================================
; = STOREXYACHAR						=
; =								=
; = Code for the forth word:	!xyachar			=
; =								=
; = Stack:	( x y attr char --- )				=
; ===============================================================

		align	4
STOREXYACHAR	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	4


		; --------------------------------
		; Call BIOS function for 'WhereXY'
		; --------------------------------

		call	isUpdateCursor				; ?UpdateCusor @
		jc	@@UpdtCurs
		call	doGetVideoPage				; PAGE#
		mov 	v86r_ah, 03h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		mov	dx, v86r_dx				; row in DH, col in DL
@@UpdtCurs:	


		; -------------------
		; Set cursor position
		; -------------------

		call	doGetVideoPage				; PAGE#
		mov	eax, [esi + 2 * @WordSize]			; Get row
		mov	v86r_dh, al				; Input: row #
		mov	eax, [esi + 3 * @WordSize]			; Get column
		mov	v86r_dl, al				; Input: col #
		mov 	v86r_ah, 02h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h


		; --------------------------------
		; Call BIOS function to print char
		; --------------------------------

;		call	doGetVideoPage				; PAGE#
		mov	eax, [esi]				; pop char
		mov	v86r_al, al				; Input: char to print
		mov	eax, [esi + @WordSize]			; pop attribute
		add	esi, 4 * @WordSize			; drop drop drop drop
		mov	v86r_bl, al				; Input: attribute / color
		mov	v86r_cx, 1				; Only one character to print
		mov	v86r_ah, 09h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h

		
		; ----------------------------------
		; Restore previous cursor position ?
		; ----------------------------------

		call	isUpdateCursor				; ?UpdateCusor @
		jc	@@UpdtCurs2
		call	doGetVideoPage				; PAGE#
		mov	v86r_dx, dx				; Row in DH, Col in DL
		mov 	v86r_ah, 02h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
@@UpdtCurs2:	

		
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_STOREXYACHAR:
STOREXYACHAR	endp


; ===============================================================
; = STOREXYNCHAR						=
; =								=
; = Code for the forth word:	!xynchar			=
; =								=
; = Stack:	( x y n char --- )				=
; ===============================================================

		align	4
STOREXYNCHAR	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	4


		; --------------------------------
		; Call BIOS function for 'WhereXY'
		; --------------------------------

		call	isUpdateCursor				; ?UpdateCusor @
		jc	@@UpdtCurs
		call	doGetVideoPage				; PAGE#
		mov 	v86r_ah, 03h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		mov	dx, v86r_dx				; row in DH, col in DL
@@UpdtCurs:	


		; -------------------
		; Set cursor position
		; -------------------

		call	doGetVideoPage				; PAGE#
		mov	eax, PARAM2				; Get row
		mov	v86r_dh, al				; Input: row #
		mov	eax, PARAM3				; Get column
		mov	v86r_dl, al				; Input: col #
		mov 	v86r_ah, 02h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h


		; --------------------------------
		; Call BIOS function to print char
		; --------------------------------

;		call	doGetVideoPage				; PAGE#
		mov	eax, PARAM0				; pop char
		mov	v86r_al, al				; Input: char to print
		call	getScreenBckg				; Get current attribute
		mov	v86r_bl, al				; Input: attribute / color
		mov	eax, PARAM1				; pop number
		DISCARD	4					; drop drop drop drop
		mov	v86r_cx, ax				; Input: nb chars to print
		mov 	v86r_ah, 09h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h

		
		; ----------------------------------
		; Restore previous cursor position ?
		; ----------------------------------

		call	isUpdateCursor				; ?UpdateCusor @
		jc	@@UpdtCurs2
		call	doGetVideoPage				; PAGE#
		mov	v86r_dx, dx				; Row in DH, Col in DL
		mov 	v86r_ah, 02h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
@@UpdtCurs2:	

		
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_STOREXYNCHAR:
STOREXYNCHAR	endp


; ===============================================================
; = STOREXYNACHAR						=
; =								=
; = Code for the forth word:	!xynachar			=
; =								=
; = Stack:	( x y n attr char --- )				=
; ===============================================================

		align	4
STOREXYNACHAR	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	5


		; --------------------------------
		; Call BIOS function for 'WhereXY'
		; --------------------------------

		call	isUpdateCursor				; ?UpdateCusor @
		jc	@@UpdtCurs
		call	doGetVideoPage				; PAGE#
		mov 	v86r_ah, 03h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		mov	dx, v86r_dx				; row in DH, col in DL
@@UpdtCurs:	


		; -------------------
		; Set cursor position
		; -------------------

		call	doGetVideoPage				; PAGE#
		mov	eax, PARAM3				; Get row
		mov	v86r_dh, al				; Input: row #
		mov	eax, PARAM4				; Get column
		mov	v86r_dl, al				; Input: col #
		mov 	v86r_ah, 02h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h


		; --------------------------------
		; Call BIOS function to print char
		; --------------------------------

;		call	doGetVideoPage				; PAGE#
		mov	eax, PARAM0				; pop char
		mov	v86r_al, al				; Input: char to print
		mov	eax, PARAM1				; pop attribute
		mov	v86r_bl, al				; Input: attribute / color
		mov	eax, PARAM2				; pop number
		DISCARD	5					; drop drop drop drop
		mov	v86r_cx, ax				; Input: nb chars to print
		mov 	v86r_ah, 09h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h

		
		; ----------------------------------
		; Restore previous cursor position ?
		; ----------------------------------

		call	isUpdateCursor				; ?UpdateCusor @
		jc	@@UpdtCurs2
		call	doGetVideoPage				; PAGE#
		mov	v86r_dx, dx				; Row in DH, Col in DL
		mov 	v86r_ah, 02h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
@@UpdtCurs2:	

		
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_STOREXYNACHAR:
STOREXYNACHAR	endp


; ===============================================================
; = ATTRCLS							=
; =								=
; = Code for the forth word:	AttrCls				=
; =								=
; = Stack:	( attr  --- )					=
; ===============================================================

		align	4
ATTRCLS		proc

		; CTRL-BREAK not allowed
		inc	[NoCtrlBreak]

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -------------------------
		; Get the number of columns
		; -------------------------

		; Call BIOS GetVideoMode function to have the nb of cols
		mov 	v86r_ah, 0Fh			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		movzx	ecx, v86r_ah			; Get nb of cols
		

		; ----------------------
		; Get the number of rows
		; ----------------------
		
		; Call BIOS GetRowsNb function to have the nb of rows
		mov 	v86r_ah, 11h			; Function asked
		mov	v86r_al, 30h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		movzx	edx, v86r_dl			; Get nb of rows (-1)
		inc	dl				; For example: for 25 lines, BIOS give 18h (24)

		
		; -----------------------------
		; Set cursor position in (0, 0)
		; -----------------------------

		call	doGetVideoPage			; PAGE#
		mov	v86r_dh, 0			; Input: row #
		mov	v86r_dl, 0			; Input: col #
		mov	v86r_ah, 02h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ---------------------------------
		; Call BIOS function to print chars
		; ---------------------------------

;		call	doGetVideoPage			; PAGE#
		mov	v86r_al, ' '			; Input: char to print
		mov	eax, [esi]			; pop attribute
		add	esi, @WordSize			; pop
		mov	v86r_bl, al			; Input: attribute / color
		mov	ax, cx
		imul	ax, dx				; EAX = ecx * edx
		mov	v86r_cx, ax			; Input: nb chars to print
		mov 	v86r_ah, 09h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_ATTRCLS:
ATTRCLS		endp


; ===============================================================
; = CLS								=
; =								=
; = Code for the forth word:	Cls				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
CLS		proc

		; -------------
		; Get attribute
		; -------------
		
		SUMMON	SCREENBCKG				; ScreenBackground
		SUMMON	AT					; @
		
		
		; ------------
		; Clear screen
		; ------------
		
		SUMMON	ATTRCLS					; AttrCls

		
		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_CLS:
CLS		endp


; ===============================================================
; = STORESTRING							=
; =								=
; = Code for the forth word:	!string				=
; =								=
; = Stack:	( addrz  --- )					=
; =								=
; = REGISTERS ARE PRESERVED.					=
; ===============================================================

		align	4
STORESTRING	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1

	
		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx


		; --------------------------------
		; Call BIOS function for 'WhereXY'
		; --------------------------------

		call	doGetVideoPage			; PAGE#
		mov 	v86r_ah, 03h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		mov	al, v86r_dl			; Output: col #
		mov	v86r_dl, al			; col #
		mov	al, v86r_dh			; Output: row #
		mov	v86r_dh, al			; row #


		; -------------------------------------
		; Copy string address to below memory ?
		; -------------------------------------
		
		ifdef	DOSREAL

		; Copy the string below 1 MB ?
		cmp	dword ptr [esi], 10000h
		jb	@@Below1MB
		
		; Move string
		mov	ebx, [esi]
		push	edx ecx ax
		lea	edx, _msgbuff
		clr	ecx
@@MovStr1:	cmp	cx, MAXSZ_SCREENSTR
		je	@@MovStr3
		mov	al, [ebx + ecx]
		mov	[edx + ecx], al
		cmpz	al, 0
		je	@@MovStr2
		inc	ecx
		jmp	@@MovStr1
@@MovStr3:	mov	byte ptr [edx + ecx], 0
@@MovStr2:	mov	[esi], edx
		pop	ax ecx edx
		
@@Below1MB:

		endif


		; ------------------
		; Call BIOS function
		; ------------------

		call	getScreenBckg			; Get current attribute
		mov	v86r_bl, al			; Input: attribute / color
		
		call	doGetVideoPage			; PAGE#

		mov	ebx, [esi]			; string addr
		clr	eax				; str length
@@Len:		inc	eax
		cmp	byte ptr [ebx + eax], 0
		jne	@@Len
		mov	v86r_ch, ah
		mov	v86r_cl, al			; Input: Length of the string
		
		; Select mode 0 (no cursor update) or mode 1 (cursor update)
		call	isUpdateCursor			; ?UpdateCusor @
		jc	@@Mode1
		mov	v86r_al, 0			; Mode 0: string contains only char (not attr), then don't update curs pos
		jmp	@@ModeDone
@@Mode1:	mov	v86r_al, 1			; Mode 1: string contains only char (not attr), then update curs pos
@@ModeDone:
		
		; Get <Seg,Ofs> from string address (in EDX)
		mov	ebx, [esi]			; pop string address
		HILO	v86r_es, v86r_bp, ebx		; String address

		mov 	v86r_ah, 13h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ----------------
		; Return to caller
		; ----------------
		
		add	esi, 1 * @WordSize		; drop
		pop	ebx eax


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		ret

E_STORESTRING:
STORESTRING	endp


; ===============================================================
; = STOREASTRING						=
; =								=
; = Code for the forth word:	!astring			=
; =								=
; = Stack:	( attr addrz  --- )				=
; ===============================================================

		align	4
STOREASTRING	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; --------------------------------
		; Call BIOS function for 'WhereXY'
		; --------------------------------

		call	doGetVideoPage			; PAGE#
		mov 	v86r_ah, 03h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		mov	al, v86r_dl			; Output: col #
		mov	v86r_dl, al			; col #
		mov	al, v86r_dh			; Output: row #
		mov	v86r_dh, al			; row #


		; -------------------------------------
		; Copy string address to below memory ?
		; -------------------------------------
		
		ifdef	DOSREAL

		; Copy the string below 1 MB ?
		cmp	dword ptr [esi], 10000h
		jb	@@Below1MB
		
		; Move string
		mov	ebx, [esi]
		lea	edx, _msgbuff
		clr	ecx
@@MovStr1:	cmp	cx, MAXSZ_SCREENSTR
		je	@@MovStr3
		mov	al, [ebx + ecx]
		mov	[edx + ecx], al
		cmpz	al, 0
		je	@@MovStr2
		inc	ecx
		jmp	@@MovStr1
@@MovStr3:	mov	byte ptr [edx + ecx], 0
@@MovStr2:	mov	[esi], edx
		
@@Below1MB:

		endif


		; ------------------
		; Call BIOS function
		; ------------------

		mov	eax, [esi + @WordSize]		; pop attribute
		mov	v86r_bl, al			; Input: attribute / color
		
		call	doGetVideoPage			; PAGE#

		mov	ebx, [esi]			; string addr
		clr	eax				; str length
@@Len:		inc	eax
		cmp	byte ptr [ebx + eax], 0
		jne	@@Len
		mov	v86r_ch, ah
		mov	v86r_cl, al			; Input: Length of the string
		
		; Select mode 0 (no cursor update) or mode 1 (cursor update)
		call	isUpdateCursor			; ?UpdateCusor @
		jc	@@Mode1
		mov	v86r_al, 0			; Mode 0: string contains only char (not attr), then don't update curs pos
		jmp	@@ModeDone
@@Mode1:	mov	v86r_al, 1			; Mode 1: string contains only char (not attr), then update curs pos
@@ModeDone:
		
		; Get <Seg,Ofs> from string address (in EDX)
		mov	edx, [esi]			; pop string address
		HILO	v86r_es, v86r_bp, edx		; String address

		mov 	v86r_ah, 13h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		add	esi, 2 * @WordSize		; drop drop
		ret

E_STOREASTRING:
STOREASTRING	endp


; ===============================================================
; = STOREXYSTRING						=
; =								=
; = Code for the forth word:	!xystring			=
; =								=
; = Stack:	( x y addrz --- )				=
; ===============================================================

		align	4
STOREXYSTRING	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; -------------------------------------
		; Copy string address to below memory ?
		; -------------------------------------
		
		ifdef	DOSREAL

		; Copy the string below 1 MB ?
		cmp	dword ptr [esi], 10000h
		jb	@@Below1MB
		
		; Move string
		mov	ebx, [esi]
		lea	edx, _msgbuff
		clr	ecx
@@MovStr1:	cmp	cx, MAXSZ_SCREENSTR
		je	@@MovStr3
		mov	al, [ebx + ecx]
		mov	[edx + ecx], al
		cmpz	al, 0
		je	@@MovStr2
		inc	ecx
		jmp	@@MovStr1
@@MovStr3:	mov	byte ptr [edx + ecx], 0
@@MovStr2:	mov	[esi], edx
		
@@Below1MB:

		endif


		; ------------------
		; Call BIOS function
		; ------------------

		call	getScreenBckg			; Get current attribute
		mov	v86r_bl, al			; Input: attribute / color

		call	doGetVideoPage			; PAGE#

		mov	ebx, [esi]			; string addr
		clr	eax				; str length
@@Len:		inc	eax
		cmp	byte ptr [ebx + eax], 0
		jne	@@Len
		mov	v86r_ch, ah
		mov	v86r_cl, al			; Input: Length of the string
		
		mov	eax, [esi + 2 * @WordSize]	; pop col #
		mov	v86r_dl, al			; Input: column #
		
		mov	eax, [esi + @WordSize]		; pop col #
		mov	v86r_dh, al			; Input: row #

		; Select mode 0 (no cursor update) or mode 1 (cursor update)
		call	isUpdateCursor			; ?UpdateCusor @
		jc	@@Mode1
		mov	v86r_al, 0			; Mode 0: string contains only char (not attr), then don't update curs pos
		jmp	@@ModeDone
@@Mode1:	mov	v86r_al, 1			; Mode 1: string contains only char (not attr), then update curs pos
@@ModeDone:
		
		; Get <Seg,Ofs> from string address (in EDX)
		mov	ebx, [esi]			; pop string address
		HILO	v86r_es, v86r_bp, ebx

		mov 	v86r_ah, 13h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		add	esi, 3 * @WordSize		; drop drop drop
		ret

E_STOREXYSTRING:
STOREXYSTRING	endp


; ===============================================================
; = STOREXYASTRING						=
; =								=
; = Code for the forth word:	!xyastring			=
; =								=
; = Stack:	( x y attr addrz  --- )				=
; ===============================================================

		align	4
STOREXYASTRING	proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	4


		; -------------------------------------
		; Copy string address to below memory ?
		; -------------------------------------
		
		ifdef	DOSREAL

		; Copy the string below 1 MB ?
		cmp	dword ptr [esi], 10000h
		jb	@@Below1MB
		
		; Move string
		mov	ebx, [esi]
		lea	edx, _msgbuff
		clr	ecx
@@MovStr1:	cmp	cx, MAXSZ_SCREENSTR
		je	@@MovStr3
		mov	al, [ebx + ecx]
		mov	[edx + ecx], al
		cmpz	al, 0
		je	@@MovStr2
		inc	ecx
		jmp	@@MovStr1
@@MovStr3:	mov	byte ptr [edx + ecx], 0
@@MovStr2:	mov	[esi], edx
		
@@Below1MB:

		endif


		; ------------------
		; Call BIOS function
		; ------------------

		mov	eax, [esi + @WordSize]		; pop attribute
		mov	v86r_bl, al			; Input: attribute / color

		call	doGetVideoPage			; PAGE#

		mov	ebx, [esi]			; string addr
		clr	eax				; str length
@@Len:		inc	eax
		cmp	byte ptr [ebx + eax], 0
		jne	@@Len
		mov	v86r_ch, ah
		mov	v86r_cl, al			; Input: Length of the string
		
		mov	eax, [esi + 12]			; pop col #
		mov	v86r_dl, al			; Input: column #
		
		mov	eax, [esi + 8]			; pop col #
		mov	v86r_dh, al			; Input: row #

		; Select mode 0 (no cursor update) or mode 1 (cursor update)
		call	isUpdateCursor			; ?UpdateCusor @
		jc	@@Mode1
		mov	v86r_al, 0			; Mode 0: string contains only char (not attr), then don't update curs pos
		jmp	@@ModeDone
@@Mode1:	mov	v86r_al, 1			; Mode 1: string contains only char (not attr), then update curs pos
@@ModeDone:
		
		; Get <Seg,Ofs> from string address (in EDX)
		mov	ebx, [esi]			; pop string address
		HILO	v86r_es, v86r_bp, ebx		; String address

		mov 	v86r_ah, 13h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		DISCARD	4
		ret

E_STOREXYASTRING:
STOREXYASTRING	endp


; ===============================================================
; = TOXYLSTRING_CHAR						=
; =								=
; = Code for the forth word:	>xylstring_char			=
; =								=
; = Stack:	( c --- )					=
; ===============================================================

		align	4
TOXYLSTRING_CHAR	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ----------
		; Save value
		; ----------
		
		mov	eax, PARAM0
		mov	[_xylstring_char], al


		; ----------------
		; Return to caller
		; ----------------
		
		DISCARD	1
		ret

E_TOXYLSTRING_CHAR:
TOXYLSTRING_CHAR	endp


; ===============================================================
; = TOXYLSTRING_ATTR						=
; =								=
; = Code for the forth word:	>xylstring_attr			=
; =								=
; = Stack:	( n1 n2 n3 --- )				=
; ===============================================================

		align	4
TOXYLSTRING_ATTR	proc

		; ------------------
		; Verify stack frame
		; ------------------
		

		VSTACK	3


		; -----------
		; Save values
		; -----------

		mov	eax, PARAM2
		cmp	eax, 0
		jl	@@error
		cmp	eax, 8
		jge	@@error

		mov	ebx, PARAM1
		mov	ecx, PARAM0
		shl	bl, 4
		and	cl, 0Fh
		or	bl, cl
		mov	[_xylstring_attr + eax], bl


		; ----------------
		; Return to caller
		; ----------------
		
@@error:		DISCARD	3
		ret

E_TOXYLSTRING_ATTR:
TOXYLSTRING_ATTR	endp


; ===============================================================
; = STOREXYLSTRING						=
; =								=
; = Code for the forth word:	!xylstring			=
; =								=
; = Stack:	( x y addrz  --- )				=
; ===============================================================

		align	4
STOREXYLSTRING	proc

		; --------------
		; Temp variables
		; --------------

		enter	4+4, 0
t_attrib	equ	byte ptr [ebp - 4]
t_prevposcur	equ	word ptr [ebp - 8]
		
		; Initialize current attribute
		mov	al, [_xylstring_attr]
		mov	[t_attrib], al


		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; --------------------------------
		; Call BIOS function for 'WhereXY'
		; --------------------------------

		call	isUpdateCursor				; ?UpdateCusor @
		jc	@@UpdtCurs
		call	doGetVideoPage				; PAGE#
		mov 	[v86r_ah], 03h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		mov	dx, [v86r_dx]				; row in DH, col in DL
		mov	[t_prevposcur], dx
@@UpdtCurs:	


		; -------------------------------------
		; Copy string address to below memory ?
		; -------------------------------------
		
		ifdef	DOSREAL

		; Copy the string below 1 MB ?
		cmp	PARAM0, 10000h
		jb	@@Below1MB
		
		; Move string
		mov	ebx, PARAM0
		lea	edx, _msgbuff
		clr	ecx
@@MovStr1:	cmp	cx, MAXSZ_SCREENSTR
		je	@@MovStr3
		mov	al, [ebx + ecx]
		mov	[edx + ecx], al
		cmpz	al, 0
		je	@@MovStr2
		inc	ecx
		jmp	@@MovStr1
@@MovStr3:	mov	byte ptr [edx + ecx], 0
@@MovStr2:	mov	PARAM0, edx
		
@@Below1MB:

		endif

	
		; ------------------------
		; Starting cursor position
		; ------------------------
		
		mov	eax, PARAM2			; x
		mov	ebx, PARAM1			; y
		mov	dh, bl
		mov	dl, al


		; --------------------------
		; Interpret and Print string
		; --------------------------

		mov	ebx, PARAM0
		mov	ecx, ebx
@@SrchChar:	mov	al, [ecx]
		cmpz	al, 0
		je	@@EndStr
		cmp	al, [_xylstring_char]
		je	@@FoundChar
		inc	ecx
		jmp	@@SrchChar

@@FoundChar:	
		; Print string
		cmp	ecx, ebx
		je	@@PrintDone
		mov	al, [t_attrib]			; get current attribute
		mov	[v86r_bl], al			; Input: attribute / color
		call	doGetVideoPage			; PAGE#
		mov	eax, ecx
		sub	eax, ebx			; String length
		mov	[v86r_cx], ax			; Input: Length of the string
		mov	[v86r_dx], dx			; Input: row in DH, column in DL
		add	dl, al				; Update column position (for next string)
		mov	[v86r_al], 1			; Mode 1: string contains only char (not attr), then update curs pos
		HILO	v86r_es, v86r_bp, ebx		; String address
		mov 	[v86r_ah], 13h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
@@PrintDone:	
		movzx	eax, byte ptr [ecx + 2]
		mov	al, [_xylstring_attr + eax - '0']
		mov	[t_attrib], al
		lea	ebx, [ecx + 4]
		mov	ecx, ebx
		jmp	@@SrchChar

@@EndStr:	cmp	ecx, ebx
		je	@@finish

		mov	al, [t_attrib]			; get current attribute
		mov	[v86r_bl], al			; Input: attribute / color
		call	doGetVideoPage			; PAGE#
		mov	eax, ecx
		sub	eax, ebx			; String length
		mov	[v86r_cx], ax			; Input: Length of the string
		mov	[v86r_dx], dx			; Input: row in DH, column in DL
		mov	[v86r_al], 1			; Mode 1: string contains only char (not attr), then update curs pos
		HILO	v86r_es, v86r_bp, ebx		; String address
		mov 	[v86r_ah], 13h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

@@finish:	

		; ----------------------------------
		; Restore previous cursor position ?
		; ----------------------------------

		call	isUpdateCursor			; ?UpdateCusor @
		jc	@@UpdtCurs2
		call	doGetVideoPage			; PAGE#
		mov	dx, [t_prevposcur]		; Restore previous cursor position
		mov	[v86r_dx], dx			; Row in DH, Col in DL
		mov 	[v86r_ah], 02h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
@@UpdtCurs2:	

		
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		DISCARD	3
		leave
		ret

E_STOREXYLSTRING:
STOREXYLSTRING	endp


; ===============================================================
; = ACLRLINE							=
; =								=
; = Code for the forth word:	aClrLine			=
; =								=
; = Stack:	( y attr   --- )				=
; ===============================================================

		align	4
ACLRLINE		proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; -------------------
		; Set cursor position
		; -------------------

		call	doGetVideoPage			; PAGE#
		mov	eax, [esi + @WordSize]		; Get row
		mov	v86r_dh, al			; Input: row #
		clr	al				; Get column
		mov	v86r_dl, al			; Input: col #
		mov 	v86r_ah, 02h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ---------------------------
		; Get the width of the screen
		; ---------------------------
		
		SUMMON	NBCOLS
		mov	ecx, [esi]
		add	esi, @WordSize


		; --------------------------------
		; Call BIOS function to print char
		; --------------------------------

;		call	doGetVideoPage			; PAGE#
		mov	al, ' '				; pop char
		mov	v86r_al, al			; Input: char to print
		mov	eax, [esi]			; pop attribute
		mov	v86r_bl, al			; Input: attribute / color
		add	esi, 2 * @WordSize		; pop pop
		mov	v86r_cx, cx			; Input: nb chars to print
		mov 	v86r_ah, 09h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_ACLRLINE:
ACLRLINE		endp


; ===============================================================
; = CLRLINE							=
; =								=
; = Code for the forth word:	ClrLine				=
; =								=
; = Stack:	( y --- )					=
; ===============================================================

		align	4
CLRLINE		proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -------------------
		; Set cursor position
		; -------------------

		call	doGetVideoPage			; PAGE#
		mov	eax, [esi]			; Get row
		mov	v86r_dh, al			; Input: row #
		clr	al				; Get column
		mov	v86r_dl, al			; Input: col #
		mov 	v86r_ah, 02h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h


		; ---------------------------
		; Get the width of the screen
		; ---------------------------
		
		SUMMON	NBCOLS
		mov	ecx, [esi]
		add	esi, @WordSize


		; --------------------------------
		; Call BIOS function to print char
		; --------------------------------

;		call	doGetVideoPage			; PAGE#
		mov	al, ' '				; pop char
		mov	v86r_al, al			; Input: char to print
		mov	eax, [esi]			; pop attribute
		call	getScreenBckg			; Get current attribute
		mov	v86r_bl, al			; Input: attribute / color
		add	esi, @WordSize			; drop
		mov	v86r_cx, cx			; Input: nb chars to print
		mov 	v86r_ah, 09h			; Function asked
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h

		
		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_CLRLINE:
CLRLINE		endp


; ===============================================================
; = ACLREOL							=
; =								=
; = Code for the forth word:	aClrEOL				=
; =								=
; = Stack:	( attr  --- )					=
; ===============================================================

		align	4
ACLREOL		proc

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------

		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ---------------------------
		; Get the width of the screen
		; ---------------------------
		
		SUMMON	NBCOLS
		mov	ecx, [esi]
		add	esi, @WordSize


		; ----------------------------------------------------
		; Call BIOS function to see where is the cursor (#col)
		; ----------------------------------------------------

		call	doGetVideoPage				; PAGE#
		mov 	v86r_ah, 03h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h
		movzx	eax, v86r_dl				; Output: col #
		sub	ecx, eax					; Is the nb of char to print
		dec	ecx
		cmpz	cx, 0
		je	@@Exit


		; --------------------------------
		; Call BIOS function to print char
		; --------------------------------

;		call	doGetVideoPage				; PAGE#
		mov	al, ' '					; pop char
		mov	v86r_al, al				; Input: char to print
		mov	eax, [esi]				; pop attribute
		mov	v86r_bl, al				; Input: attribute / color
		mov	v86r_cx, cx				; Input: nb chars to print
		mov 	v86r_ah, 09h				; Function asked
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h

		
		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		add	esi, @WordSize				; drop


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		SUMMON	CTRLBREAK_REST

		ret

E_ACLREOL:
ACLREOL		endp


; ===============================================================
; = ClrEOL							=
; =								=
; = Code for the forth word:	ClrEOL				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
CLREOL		proc

		; -------------
		; Get attribute
		; -------------
		
		SUMMON	SCREENBCKG				; ScreenBackground
		SUMMON	AT					; @
		
		
		; -------
		; Clr EOL
		; -------
		
		SUMMON	ACLREOL					; aClrEOL
		

		; ----------------
		; Return to caller
		; ----------------
	
		ret

E_CLREOL:
CLREOL		endp



; ===============================================================
; = LOADCHARSET_8x8						=
; =								=
; = Code for the forth word:	LoadCharset_8x8			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
LOADCHARSET_8x8	proc

		; Call BIOS 
		mov 	[v86r_ah], 011h				; Function asked
		mov	[v86r_al], 012h				; Function
		mov	[v86r_bl], 0
		mov 	al, 10h					; Call VIDEO_SERVICE
		int 	33h

		; Update the number of rows		
		call	doGetScreenInf

		; Return to caller
		ret

E_LOADCHARSET_8x8:
LOADCHARSET_8x8	endp


; ===============================================================
; = LOADCHARSET_8x14						=
; =								=
; = Code for the forth word:	LoadCharset_8x14		=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
LOADCHARSET_8x14	proc

		; Call BIOS 
		mov 	[v86r_ah], 011h			; Function asked
		mov	[v86r_al], 011h			; Function
		mov	[v86r_bl], 0
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		
		; Update the number of rows		
		call	doGetScreenInf

		; Return to caller
		ret

E_LOADCHARSET_8x14:
LOADCHARSET_8x14	endp


; ===============================================================
; = LOADCHARSET_8x16						=
; =								=
; = Code for the forth word:	LoadCharset_8x16		=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
LOADCHARSET_8x16	proc

		; Call BIOS 
		mov 	[v86r_ah], 011h			; Function asked
		mov	[v86r_al], 014h			; Function
		mov	[v86r_bl], 0
		mov 	al, 10h				; Call VIDEO_SERVICE
		int 	33h
		
		; Update the number of rows		
		call	doGetScreenInf

		; Return to caller
		ret

E_LOADCHARSET_8x16:
LOADCHARSET_8x16	endp


; ===============================================================
; = V<color>							=
; =								=
; = Code for the forth word:	V<color>			=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

ICOL		macro	color

		align	4
V&color		proc

		; Push value
		IPUSH	color

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_V&color:
V&color		endp

		endm

		ICOL	BLACK
		ICOL	BLUE
		ICOL	GREEN
		ICOL	CYAN
		ICOL	RED
		ICOL	MAGENTA
		ICOL	BROWN
		ICOL	LIGHTGRAY
		ICOL	DARKGRAY
		ICOL	LIGHTBLUE
		ICOL	LIGHTGREEN
		ICOL	LIGHTCYAN
		ICOL	LIGHTRED
		ICOL	LIGHTMAGENTA
		ICOL	YELLOW
		ICOL	WHITE

		purge	ICOL


code32		ends
		end
