; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	MEMORY.ASM						!
; 									!
; Description:	Definitions for playing with the memory.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	TYPE		TYPE						!
;	DTYPE		-TYPE						!
;	TYPED		TYPE-						!
;	CR		CR						!
;	SPACE		SPACE						!
;	SPACES		SPACES						!
;	PDOTQ		rt_."						!
;	PQUOTE		rt."						!
;	QUOTE		"						!
;	DOTQUOTE	."						!
;	CHARUPCASE	CharUppercase					!
;	CHARLOCASE	CharLowerCase					!
;	FSTRING		FSTRING						!
;	DSTRING		DSTRING						!
;	CSTRING		CSTRING						!
;	NULLSTR		NULLSTR						!
;	STR_LEN		$LEN						!
;	STR_MLEN	$MLEN						!
;	STR_EQU		$==						!
;	STR_LESS	$<						!
;	STR_LESSEQ	$<=						!
;	STR_GREAT	$>						!
;	STR_GREAT	$>=						!
;	STR_STORE	$!						!
;	STR_ADD		$+						!
;	ADD_STR		+$						!
;	STR_REPL	$REPLACE		$REPL			!
;	STR_SHRINK	$SHRINK						!
;	STR_FIND	$FIND						!
;	STR_COPY	$COPY						!
;	STR_COPY_TOC	$CopyToChar					!
;	STR_CONVNB	$ConvNb						!
;	STR_SRCCHCHR	$SearchChr					!
;	STR_SRCCHSTR	$SearchStr					!
;	INPUT		INPUT						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions


; ===============================================================
; = TstStr							=
; =                                                 		=
; = Usage:	Test if a valid string (strinh header ok ?)	=
; =								=
; = I) reg : point to string (NOT string buffer)		=
; = REGISTERS ARE PRESERVED.					=
; ===============================================================

TSTSIGN	MACRO	reg
	ENDM

		align	4
_StringNotOk	proc

		; "Not a string (invalid string header)"
		mov	ah, ERR_NOTASTRING
		clr	edx					; No msg
		clr	ecx					; No line #
		call	doError

_StringNotOk	endp


; ===============================================================
; = TYPE							=
; =                                                 		=
; = Code for the froth word:	TYPE				=
; =								=
; = Stack:	( adr --- )					=
; =								=
; = REGISTERS ARE PRESERVED.					=
; ===============================================================

		align	4
_TYPE		proc

		; Verify stack frame
		VSTACK	1

		; Save used regs
		push	ax ebx

		; String address
		mov	ebx, [esi]
		add	esi, @WordSize

		; Loop until end of string		
@@LL1:		mov	al, [ebx]
		cmpz	al, 0
		je	@@LL2
		call	COUT
		inc	ebx
		jmp	@@LL1

@@LL2:
		; Return to caller
		pop	ebx ax
		ret

E__TYPE:
_TYPE		endp


; ===============================================================
; = DTYPE							=
; =                                                 		=
; = Code for the froth word:	-TYPE				=
; =								=
; = Stack:	( n addrz --- )					=
; =								=
; = REGISTERS ARE PRESERVED.					=
; ===============================================================

		align	4
DTYPE		proc

		; Verify stack frame
		VSTACK	2

		; Save used regs
		push	eax ebx ecx

		; String address
		mov	ebx, PARAM0
		clr	eax
@@StrLen:	cmp	byte ptr [ebx + eax], 0
		je	@@StrLen2
		inc	eax
		jmp	@@StrLen
@@StrLen2:	

		; Compute nb of spaces
		mov	ecx, PARAM1				; Get <n>
		DISCARD	2
		jbe	@@Spaces
		
		DISPOSE	1
		sub	ecx, eax
		mov	PARAM0, ecx
		SUMMON	SPACES
		
@@Spaces:

		; Loop until end of string		
@@LL1:		mov	al, [ebx]
		cmpz	al, 0
		je	@@LL2
		call	COUT
		inc	ebx
		jmp	@@LL1

@@LL2:
		; Return to caller
		pop	ecx ebx eax
		ret

E_DTYPE:
DTYPE		endp


; ===============================================================
; = TYPED							=
; =                                                 		=
; = Code for the froth word:	TYPE-				=
; =								=
; = Stack:	( n addrz --- )					=
; =								=
; = REGISTERS ARE PRESERVED.					=
; ===============================================================

		align	4
TYPED		proc

		; Verify stack frame
		VSTACK	2

		; Save used regs
		push	eax ebx ecx

		; Loop until end of string
		mov	ebx, PARAM0
		clr	ecx
@@LL1:		mov	al, [ebx + ecx]
		cmpz	al, 0
		je	@@LL2
		call	COUT
		inc	ecx
		jmp	@@LL1
@@LL2:

		; Compute nb of spaces to print, if any
		mov	edx, PARAM1				; Get pad length
		DISCARD	2					; drop drop
		cmp	ecx, edx
		jge	@@Spaces
		
		DISPOSE	1
		sub	edx, ecx
		mov	PARAM0, edx
		SUMMON	SPACES
@@Spaces:

		; Return to caller
		pop	ecx ebx eax
		ret

E_TYPED:
TYPED		endp


; ===============================================================
; = CR								=
; =                                                 		=
; = Code for the forth word:	CR				=
; =                     	                            	=
; = Stack:	( --- )        	        	                =
; =                     	                            	=
; = REGISTERS ARE PRESERVED.					=
; ===============================================================

		align	4
CR		proc

		; Save registers
		push	ax

		; Print CR LF
		mov	al, 0Dh
		call	COUT
		mov	al, 0Ah
		call	COUT

		; Return to caller		
		pop	ax
		ret

E_CR:
CR		endp


; ===============================================================
; = SPACE							=
; =                                                 		=
; = Code for the forth word:	SPACE				=
; =                     	                            	=
; = Stack:	( --- )        	                           	=
; =                     	                            	=
; = REGISTERS ARE PRESERVED.					=
; ===============================================================

		align	4
SPACE		proc

		; Save registers
		push	ax

		; printf("%c", ' ')
		mov	al, ' '
		call	COUT

		; Return to caller		
		pop	ax
		ret

E_SPACE:
SPACE		endp


; ===============================================================
; = SPACES							=
; =                                                 		=
; = Code for the forth word:	SPACES				=
; =                     	                            	=
; = Stack:	( n --- )        	                      	=
; =                     	                            	=
; ===============================================================

		align	4
SPACES		proc

		; Save used regs
		push	ebx

		; Print <n> spaces
		mov	ebx, [esi]
		add	esi, @WordSize
@@Loop:		cmpz	ebx, 0
		je	@@Loop2
		mov	al, ' '
		call	COUT
		dec	ebx
		jmp	@@Loop
@@Loop2:

		; Return to caller
		pop	ebx
		ret
E_SPACES:

SPACES		endp


; ===============================================================
; = PDOTQ			                           	=
; =                                                         	=
; = Run time function for the forth word ."			=
; =                                                         	=
; = Regs modified:  none                  			=
; ===============================================================

		align	4
PDOTQ       	proc

		; Save used registers
		push	ax ebx ecx

		; Save then set '?UpdateCursor'
		mov	ecx, [edi + ddir_blkaddr + D_QUPDATECURSOR]
		mov	ebx, [ecx]					; Get actual value
		push	ebx
		mov	dword ptr [ecx], TRUE				; ?UpdateCursor 1!

		; Store into DI the string address
		mov	ecx, [esp + 2+4+4+4]

		; !string
		sub	esi, @WordSize
		mov	PARAM0, ecx
		SUMMON	STORESTRING

@@LL1:      	mov	al, [ecx]
		inc	ecx
		cmpz	al, 0
		jne	@@LL1

		; Move the IP after the string
		mov	[esp + 2+4+4+4], ecx

		; Restore '?UpdateCursor'
		mov	ecx, [edi + ddir_blkaddr + D_QUPDATECURSOR]
		pop	ebx
		mov	[ecx], ebx

		; Return to caller
		pop	ecx ebx ax
		ret

E_PDOTQ:
PDOTQ       	endp


; ===============================================================
; = PQUOTE                          				=
; =                                                         	=
; = Run time function for the forth word " : RT_"		=
; =                                                         	=
; = Regs modified:  BX ECX                  			=
; ===============================================================

		align	4
PQUOTE		proc

		; Save used registers
		push	ecx

		; Get the string address
		mov	ecx, [esp + @WordSize]

		; Push the string address onto the parameters stack
		add	ecx, size strentry					; Skip string header
		IPUSH	ecx
		
		; Move the IP after the string header and the string
		mov	ecx, [ecx.str_currlen - size strentry]
		add	dword  ptr [esp + @WordSize], size strentry+1		; Skip string header + '\0'
		add	dword  ptr [esp + @WordSize], ecx			; Skip string

		; Return to caller
		pop	ecx
		ret

E_PQUOTE:
PQUOTE 		endp


; ===============================================================
; = QUOTE                           				=
; =                                                         	=
; = Code for the forth word:    "				=
; =                                                         	=
; = Stack:  ( --- addrz )                 			=
; ===============================================================

		align	4
QUOTE       	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx edx
		

		; -------------------------------------------
		; Found next quote in the input stream buffer
		; -------------------------------------------
		
		mov	edx, [IBUFFPTR2]
		inc	edx

		; First, save current ptr: the start of the word
		mov	ecx, edx

		; End quote not forgetten ?
@@LL2:      	mov	al, [edx]
		cmpz	al, 0
		jne	@@LL3

		; "End quote forgetten"
		add	esp, 4 * @WordSize			; Clean stack
		mov	ah, ERR_LACKQUOTE
		clr	edx					; No msg
		clr	ecx					; No line #
		call	doError

@@LL3:      	cmp	al, '"'
		je 	@@LL4
		inc	edx
		jmp	@@LL2

		; Found the end of the word
@@LL4:      	mov	byte ptr [edx], 0
		mov	[IBUFFPTR2], edx        		; Current pointer into the buffer
		

		; -------------------------
		; Mode execute or compile ?
		; -------------------------

		cmp	[CompileMode], TRUE		
		je 	@@Compile
		
		
		; -----------------------------
		; Mode EXECUTE: move the string
		; -----------------------------
		
		mov	edx, [QuoteStrBuff]
		add	edx, size strentry
		mov	[edx.str_incsize - size strentry], 0	; Increment size (dynamic str) else 0 (fixed str)

		; Move string	
		; -----------
			
		clr	ebx
		dec	ebx
@@MovStr:  	inc	ebx
		mov	al, [ecx + ebx]
		cmp	al, ESCCHAR
		jne	@@MovStr2

		; Next char
		inc	ecx
		mov	ah, [ecx + ebx]				; Next character
		cmpz	ah, 0
		je	@@InvHexNb
		cmp	ah, ESCCHAR
		je	@@MovStr2
		TOUPPER	ah
		cmp	ah, '0'
		jb	@@InvHexNb
		cmp	ah, 'F'
		ja	@@InvHexNb
		cmp	ah, '9'
		ja	@@digit1
		sub	ah, '0'
		jmp	@@digit2
@@digit1:	sub	ah, 55
		
		; Next char
@@digit2:	inc	ecx
		mov	al, [ecx + ebx]				; Next character
		cmpz	al, 0
		je	@@InvHexNb
		TOUPPER	al
		cmp	al, '0'
		jb	@@InvHexNb
		cmp	al, 'F'
		ja	@@InvHexNb
		cmp	al, '9'
		ja	@@digit3
		sub	al, '0'
		jmp	@@digit4
@@digit3:	sub	al, 55
@@digit4:	
		shl	ah, 4
		or	al, ah
		mov	[edx + ebx], al
		jmp	@@MovStr					; Because \00 is not string terminal

@@MovStr2:	mov	[edx + ebx], al
		cmpz	al, 0
		jne	@@MovStr

		; Update string header
		mov	[edx.str_currlen - size strentry], ebx		; Current length
		mov	[edx.str_currsize - size strentry], IBUFFMAXS	; Max length (fixed string)
		mov	[edx.str_incsize - size strentry], 0		; Inc size (if dynamic string), else 0 (if fixed string)
		mov	[edx.str_memblknb - size strentry], 0		; Memory block nb (ref to DDIR)
		IPUSH	edx

		; Done
		jmp 	@@Exit

		; "Invalid hex number given in a string (\xx)."
@@InvHexNb:	add	esp, 4 * @WordSize				; Clean stack
		mov	ah, ERR_INVHEXNBSTR				; Error #
		clr	edx						; No msg
		clr	ecx						; No line #
		call	doError

		
		; -----------------------------------------
		; Mode Compile: compile (") and the string
		; -----------------------------------------
		
@@Compile:

		; Put 'CALL rel32' code operation       
		compCALL_rel32	PQUOTE

		; Alloc space for string header
		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ah, size strentry
		clr	al
@@StoreStrHdr:	call	CodeCComma
		dec	ah
		jne	@@StoreStrHdr
		
		; Put the string
		; --------------

		clr	ebx					; Compute string length
		dec	ebx
		dec	ecx
@@PutStr:    	inc	ecx
		mov	al, [ecx]
		cmp	al, ESCCHAR
		jne	@@PutStr2

		; Next char
		inc	ecx
		mov	ah, [ecx]				; Next character
		cmpz	ah, 0
		je	@@InvHexNb
		cmp	ah, ESCCHAR
		je	@@PutStr2
		TOUPPER	ah
		cmp	ah, '0'
		jb	@@InvHexNb
		cmp	ah, 'F'
		ja	@@InvHexNb
		cmp	ah, '9'
		ja	@@dig1
		sub	ah, '0'
		jmp	@@dig2
@@dig1:		sub	ah, 55
		
		; Next char
@@dig2:		inc	ecx
		mov	al, [ecx]				; Next character
		cmpz	al, 0
		je	@@InvHexNb
		TOUPPER	al
		cmp	al, '0'
		jb	@@InvHexNb
		cmp	al, 'F'
		ja	@@InvHexNb
		cmp	al, '9'
		ja	@@dig3
		sub	al, '0'
		jmp	@@dig4
@@dig3:		sub	al, 55
@@dig4:	
		shl	ah, 4
		or	al, ah
		call	CodeCComma
		inc	ebx
		jmp	@@PutStr				; Because \00 is not string terminal

@@PutStr2:	call	CodeCComma
		inc	ebx
		cmpz	al, 0
		jne	@@PutStr


		; Update the string header
		; ------------------------

		mov	[edx.str_currlen], ebx			; Current length
		mov	[edx.str_currsize], ebx			; Current size allocated
		mov	[edx.str_incsize], 0			; Increment size (if dynamic string), else 0 (if fixed string)
		mov	[edx.str_memblknb], 0			; Memory block nb (ref to DDIR)


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:     	pop	edx ecx ebx eax
		ret

E_QUOTE:
QUOTE       	endp


; ===============================================================
; = DOTQUOTE                           				=
; =                                                         	=
; = Code for the forth word:    ."				=
; =                                                         	=
; = Stack:  ( --- )                 				=
; ===============================================================

		align	4
DOTQUOTE       	proc

		; Mode execute or compile ?
		cmp		[CompileMode], TRUE		
		je 		@@Compile

		SUMMON		QUOTE
		SUMMON		_TYPE
		ret					; jmp	@@exit

@@Compile:	SUMMON		QUOTE
		compCALL_rel32	_TYPE

@@exit:		ret

E_DOTQUOTE:
DOTQUOTE       	endp


; ===============================================================
; = CHARUPCASE							=
; =                                                 		=
; = Code for the forth word:	charuppercase			=
; =                     	                            	=
; = Stack:	( c1 --- c2 )					=
; ===============================================================

		align	4
CHARUPCASE	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ----------------------		
		; Is char in 1st table ?
		; ----------------------		

		mov	eax, [esi]
		clr	ecx
		lea	edx, @@Table1
@@Search:	mov	bl, [edx + ecx]
		cmpz	bl, 0
		je	@@Exit
		cmp	bl, al
		je	@@Found
		inc	ecx
		jmp	@@Search
		
@@Found:		movzx	eax, [@@Table2 + ecx]
		mov	[esi], eax		
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

		align	4
@@Table1	db	"abcdefghijklmnopqrstuvwxyz", 0
		align	4
@@Table2	db	"ABCDEFGHIJKLMNOPQRSTUVWXYZCUEAAAACEEEIIIAAEAAOOOUUYOUAIOUNNAO", 0

E_CHARUPCASE:
CHARUPCASE	endp


; ===============================================================
; = CHARLOCASE							=
; =                                                 		=
; = Code for the forth word:	charlowercase			=
; =                     	                            	=
; = Stack:	( c1 --- c2 )					=
; ===============================================================

		align	4
CHARLOCASE	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------------
		; Conversion, if any
		; ------------------
		
		mov	eax, [esi]
		cmp	al, 'a'
		jb	@@Exit
		cmp	al, 'z'
		ja	@@Exit
		
		sub	al, 020h
		mov	[esi], eax
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_CHARLOCASE:
CHARLOCASE	endp


; ===============================================================
; = FSTRING							=
; =                                                 		=
; = Code for the forth word:	FSTRING				=
; =	(fixed string)						=
; =                     	                            	=
; = Stack:	( n --- )					=
; = Input stream:	n FSTRING myString			=
; ===============================================================

		align	4
FSTRING		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------
		
		; dup 12 1+ +  ballot  dup 0!  WordSize + !
		mov	eax, PARAM0
		add	eax, size strentry + 1
		sub	esi, @WordSize
		mov	PARAM0, eax
		SUMMON	RT_BALLOT			; BALLOT
		mov	edx, PARAM0			; EDX is data address
		mov	ecx, PARAM1			; Length
		DISCARD	2
		mov	[edx.str_currlen], 0		; Current length
		mov	[edx.str_currsize], ecx		; Current size allocated
		mov	[edx.str_incsize], 0		; Increment size (if dynamic string), else 0 (if fixed string)
		mov	[edx.str_memblknb], eax		; Memory block nb (ref to DDIR)

		; Zero the string
		mov	byte ptr [edx + size strentry], 0

		; DOES>		(this ends the compilation sequence)
		SUMMON	PDOES
		dd	R_FSTRING


		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_FSTRING:

		; 4+4+2+2
		add	PARAM0, size strentry
		
		; Return to caller
		ret

E_FSTRING:
FSTRING		endp


; ===============================================================
; = DSTRING							=
; =                                                 		=
; = Code for the forth word:	DSTRING				=
; =	(dynamic string)					=
; =                     	                            	=
; = Stack:	( n1 n2 --- )					=
; =		n1 : initial max length of the string		=
; =		n2 : increment size (dynamic allocation)	=
; = Input stream:	n1 n2 DSTRING myString			=
; ===============================================================

		align	4
DSTRING		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------

		; dup 12 1+ +  ballot  dup 0!  WordSize + !
		mov	eax, PARAM1			; Get initial max length
		add	eax, size strentry + 1
		IPUSH	eax
		SUMMON	RT_BALLOT			; BALLOT
		mov	ebx, eax			; Memory block nb
		mov	edx, PARAM0			; EDX is data address
		mov	[edx.str_currlen], 0		; Current length
		mov	eax, PARAM2			; Max length
		mov	[edx.str_currsize], eax		; Current size allocated
		mov	eax, PARAM1			; Get increment size
		mov	[edx.str_incsize], ax		; Increment size (if dynamic string), else 0 (if fixed string)
		mov	[edx.str_memblknb], ebx		; Memory block nb (ref to DDIR)

		; Store the word NIA address into the string header
		mov	eax, [VOC_CURRENT]
		add	eax, [VOC_START]
		mov	eax, [eax.voc_lastword]		; NIA pointer to caller word
		push	edx
		clr	edx
		mov	ebx, size wordentry
		idiv	ebx				; edx:eax / ebx
		pop	edx
		mov	[edx.str_wordref], ax		; Reference to the definition (NIA)
		DISCARD	3

		; Zero the string
		mov	byte ptr [edx + size strentry], 0

		; DOES>		(this ends the compilation sequence)
@@done:		SUMMON	PDOES
		dd	R_DSTRING


		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_DSTRING:

		; 4+4+2+2
		add	PARAM0, size strentry
		
		; Return to caller
		ret

E_DSTRING:
DSTRING		endp


; ===============================================================
; = CSTRING							=
; =                                                 		=
; = Code for the forth word:	CSTRING				=
; =                     	                            	=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
CSTRING		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ---------------------
		; Compute string length
		; ---------------------
		
		mov	edx, PARAM0
		clr	ecx
		dec	ecx
@@CompStrLen:	inc	ecx
		cmp	byte ptr [edx + ecx], 0
		jne	@@CompStrLen


		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------
		
		; dup 10 1+ +  ballot  dup 0!  WordSize + !
		lea	ebx, [ecx + size strentry + 1]
		sub	esi, 2 * @WordSize
		mov	PARAM1, ecx
		mov	PARAM0, ebx
		SUMMON	RT_BALLOT			; BALLOT
		mov	edx, PARAM0			; Data address
		mov	ecx, PARAM1			; String length
		mov	[edx.str_currlen], ecx		; Current length
		mov	[edx.str_currsize], ecx		; Current size allocated
		mov	[edx.str_incsize], 0		; Increment size (if dynamic string), else 0 (if fixed string)

		; Copy the string
		mov	ecx, PARAM2			; ECX is source address
		add	edx, size strentry		; EDX is destination address
		DISCARD	3
		clr	ebx
		dec	ebx
@@CopyString:	inc	ebx
		mov	al, [ecx + ebx]
		mov	[edx + ebx], al
		cmpz	al, 0
		jne	@@CopyString

		; DOES>		(this ends the compilation sequence)
		SUMMON	PDOES
		dd	R_CSTRING


		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_CSTRING:

		; 4+4+2
		add	PARAM0, size strentry
		
		; Return to caller
		ret

E_CSTRING:
CSTRING		endp


; ===============================================================
; = NULLSTR							=
; =                                                 		=
; = Code for the forth word:	NULL$				=
; =                     	                            	=
; = Stack:	( --- addrz )					=
; ===============================================================

		align	4
NULLSTR		proc

		call	@@next
@@next:		add	dword ptr [esp], 4+6+3+2+1
		mov	eax, [edi.ddir_blkaddr + D_NULLSTR]
		IPUSH	eax
		ret
		
		; 4+4+2+2
		add	PARAM0, size strentry
		
		; Return to caller
@@Exit:		ret

E_NULLSTR:
NULLSTR		endp


; ===============================================================
; = STR_LEN							=
; =                                                 		=
; = Code for the forth word:	$LEN				=
; =                     	                            	=
; = Stack:	( addrz --- n )					=
; ===============================================================

		align	4
STR_LEN		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -----------------
		; Get string length
		; -----------------
		
		mov	edx, PARAM0				; Address
		mov	edx, [edx.str_currlen - size strentry]
		mov	PARAM0, edx
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_STR_LEN:
STR_LEN		endp


; ===============================================================
; = STR_MLEN							=
; =                                                 		=
; = Code for the forth word:	$MLEN				=
; =                     	                            	=
; = Stack:	( addrz --- n )					=
; ===============================================================

		align	4
STR_MLEN		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -----------------
		; Get string length
		; -----------------
		
		mov	edx, PARAM0				; Address
		mov	edx, [edx.str_currsize - size strentry]
		mov	PARAM0, edx
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_STR_MLEN:
STR_MLEN		endp


; ===============================================================
; = _CopyString							=
; =                                                 		=
; = Usage:	Copy a given string to an other. Handle all 	=
; =             cases (fixed string overflow, dynamic string	=
; =		growing).					=
; =								=
; =	I) ESI : Source address					=
; =	   EBX : Offset in source				=
; =	   EDI : Destination address				=
; =	   ECX : Offset in destination				=
; =	   EDX : Nb of bytes to move				=
; =	    AL : 1 if store '\0' ad the end of the copy		=
; =								=
; = Note:	ESI and EDI points to the beginning of the	=
; =		string (not string header).			=
; ===============================================================

		align	4
_CopyString	proc

		; -----------------------------------------
		; Save input parameters into Temp variables
		; -----------------------------------------

t_SrcAddr	equ	[ebp - 4]
t_SrcOfs	equ	[ebp - 8]
t_DstAddr	equ	[ebp - 12]
t_DstOfs	equ	[ebp - 16]
t_Length	equ	[ebp - 20]
t_StoreZero	equ	byte ptr [ebp - 24]
		enter	6 * @WordSize, 0
		mov	[t_SrcAddr], esi
		mov	[t_SrcOfs],  ebx
		mov	[t_DstAddr], edi
		mov	[t_DstOfs],  ecx
		mov	[t_Length],  edx
		mov	[t_StoreZero], al


		; ------------
		; Fixed string
		; ------------

		; Fixed vs Dynamic string ?
		mov	edx, [t_DstAddr]
		cmp	[edx.str_incsize - size strentry], 0
		jne	@@DynamicString

		; Length overflow ?
		mov	eax, [t_DstOfs]
		add	eax, [t_Length]
		cmp	eax, [edx.str_currsize - size strentry]
		jbe	@@LengthOk
		
		; "String overflow."
		mov	ah, ERR_STRINGOVERFLOW				; Error #
		clr	edx						; No msg
		clr	ecx						; No line #
		jmp	doError

@@LengthOk:	

		; Copy after the end of the destination string ?
		mov	eax, [t_DstOfs]
		cmp	eax, [edx.str_currlen - size strentry]
		jbe	@@LengthOk2

		; "Copy after the end of the destination string."
		mov	ah, ERR_STRCOPYAFTEREND				; Error #
		clr	edx						; No msg
		clr	ecx						; No line #
		jmp	doError

@@LengthOk2:	

		; Copy now the string
		mov	esi, [t_SrcAddr]				; Source address
		add	esi, [t_SrcOfs]	
		mov	edi, [t_DstAddr]				; Destination address
		add	edi, [t_DstOfs]
		mov	ecx, [t_Length]					; String length
		cld                     				; Inc ESI and EDI
		rep	movsb  		            			; Move the block

		; Update string length
		mov	eax, [t_DstOfs]
		add	eax, [t_Length]
		cmp	eax, [edx.str_currlen - size strentry]
		jb	@@NotAdd
		mov	[edx.str_currlen - size strentry], eax
@@StoreZero:	mov	byte ptr [edx + eax], 0				; Zero terminal (end of string)
		jmp	@@StoZDone
@@NotAdd:	cmp	[t_StoreZero], 1
		je	@@StoreZero
@@StoZDone:	

		; Done		
		jmp	@@exit
		
		
		; --------------
		; Dynamic string
		; --------------
		
@@DynamicString:	
		; Is there enought space allocated to hold the string ?
		mov	ebx, PARAM1
		sub	ebx, size strentry
		mov	eax, [ebx.str_currlen]
		cmp	eax, [edx.str_currsize]
		jb	@@LengthOk

		; Compute new amount of memory to get
		mov	eax, [ebx.str_currlen]
		movzx	ebx, [edx.str_incsize]
		clr	edx
		idiv	ebx
		inc	eax
		mov	edx, PARAM0
		sub	edx, size strentry
		movzx	ebx, [edx.str_incsize]
		imul	eax, ebx
		add	eax, size strentry

		; Save string infos values
		movzx	ebx, [edx.str_incsize]
		push	eax ebx
		
		; Realloc the memory used
		mov	edx, [edx.str_memblknb]			; Memory block nb (ref to DDIR)
		call	ReAlloc
		push	eax					; Memory block nb (ref to DDIR)

		; Copy the string
		mov	edx, ebx				; Destination addr
		mov	ebx, PARAM1				; Source addr
		sub	ebx, size strentry
		mov	eax, [ebx.str_currlen]			; Source string length

		; Copy string
		push	esi edi
		lea	esi, [ebx + size strentry]		; Source address
		lea	edi, [edx + size strentry]		; Destination address
		lea	ecx, [eax + 1]				; String length + 1
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update string length
		mov	[edx.str_currlen], eax
		pop	eax					; Memory block nb
		push	edx
		clr	edx
		mov	ebx, size ddirentry
		idiv	ebx					; To save space (16-bits vs 32-bits)
		pop	edx
		mov	[edx.str_memblknb], eax			; Memory block nb (ref to DDIR)
		pop	ebx eax
		mov	[edx.str_incsize], bx
		sub	eax, (size strentry + 1)
		mov	[edx.str_currsize], eax


		; ----------------
		; Return to caller
		; ----------------

@@exit:		leave
		ret

_CopyString	endp


; ===============================================================
; = STR_EQU							=
; =                                                 		=
; = Code for the forth word:	$==				=
; =                     	                            	=
; = Stack:	( addrz1 addrz2 --- f )				=
; ===============================================================

		align	4
STR_EQU		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Test string equality
		; --------------------

		mov	ecx, PARAM1				; Get addrz1
		mov	edx, PARAM0				; Get addrz2
		DISCARD	1					; DROP
		mov	eax, [ecx.str_currlen - size strentry]	; Get current length
		cmp	eax, [edx.str_currlen - size strentry]
		jne	@@NotOk

		clr	ebx
		dec	ebx
@@Test:		inc	ebx
		mov	al, [ecx + ebx]
		cmpz	al, 0
		je	@@Ok
		cmp	al, [edx + ebx]
		je	@@Test


		; Return to caller
		; ----------------

@@NotOk:	mov	PARAM0, FALSE
		ret
@@Ok:		mov	PARAM0, TRUE
		ret


E_STR_EQU:
STR_EQU	endp


; ===============================================================
; = STR_LESS							=
; =                                                 		=
; = Code for the forth word:	$<				=
; =                     	                            	=
; = Stack:	( adrz1 adrz2 --- f )				=
; ===============================================================

		align	4
STR_LESS	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ----------------
		; Return to caller
		; ----------------

		DISCARD	1
		ret

E_STR_LESS:
STR_LESS	endp


; ===============================================================
; = STR_LESSEQ							=
; =                                                 		=
; = Code for the forth word:	$<=				=
; =                     	                            	=
; = Stack:	( adrz1 adrz2 --- f )				=
; ===============================================================

		align	4
STR_LESSEQ	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ----------------
		; Return to caller
		; ----------------

		DISCARD	1
		ret

E_STR_LESSEQ:
STR_LESSEQ	endp


; ===============================================================
; = STR_GREAT							=
; =                                                 		=
; = Code for the forth word:	$>				=
; =                     	                            	=
; = Stack:	( adrz1 adrz2 --- f )				=
; ===============================================================

		align	4
STR_GREAT	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ----------------
		; Return to caller
		; ----------------

		DISCARD	1
		ret

E_STR_GREAT:
STR_GREAT	endp


; ===============================================================
; = STR_GREATEQ							=
; =                                                 		=
; = Code for the forth word:	$>=				=
; =                     	                            	=
; = Stack:	( adrz1 adrz2 --- f )				=
; ===============================================================

		align	4
STR_GREATEQ	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ----------------
		; Return to caller
		; ----------------

		DISCARD	1
		ret

E_STR_GREATEQ:
STR_GREATEQ	endp


; ===============================================================
; = STR_STORE							=
; =                                                 		=
; = Code for the forth word:	$!				=
; =                     	                            	=
; = Stack:	( addrz1 addrz2 --- )				=
; =								=
; =	addrz1 : source address					=
; =	addrz2 : destination address				=
; ===============================================================

		align	4
STR_STORE	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ------------
		; Fixed string
		; ------------

		; Fixed vs Dynamic string ?
		mov	edx, PARAM0
		TSTSIGN	edx
		sub	edx, size strentry
		cmp	[edx.str_incsize], 0
		jne	@@DynamicString

		; Length overflow ?
		mov	ebx, PARAM1
		TSTSIGN	ebx
		sub	ebx, size strentry
		mov	eax, [ebx.str_currlen]
		cmp	eax, [edx.str_currsize]
		jb	@@LengthOk
		
		; "String overflow."
		mov	ah, ERR_STRINGOVERFLOW			; Error #
		clr	edx					; No msg
		clr	ecx					; No line #
		call	doError

@@LengthOk:	
		; Copy string
		push	esi edi
		mov	esi, ebx				; Source address
		add	esi, size strentry
		lea	edi, [edx + size strentry]		; Destination address
		lea	ecx, [eax + 1]				; String length + 1
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update string length
		mov	[edx.str_currlen], eax

		; Done		
		jmp	@@exit
		
		
		; --------------
		; Dynamic string
		; --------------
		
@@DynamicString:	
		; Is there enought space allocated to hold the string ?
		mov	ebx, PARAM1
		sub	ebx, size strentry
		mov	eax, [ebx.str_currlen]
		cmp	eax, [edx.str_currsize]
		jb	@@LengthOk

		; Compute new amount of memory to get
		mov	eax, [ebx.str_currlen]
		movzx	ebx, [edx.str_incsize]
		clr	edx
		idiv	ebx
		inc	eax
		mov	edx, PARAM0
		sub	edx, size strentry
		movzx	ebx, [edx.str_incsize]
		imul	eax, ebx
		add	eax, size strentry

		; Save string infos values
		movzx	ebx, [edx.str_incsize]
		push	eax ebx
		
		; Realloc the memory used
		mov	edx, [edx.str_memblknb]			; Memory block nb (ref to DDIR)
		call	ReAlloc
		push	eax					; Memory block nb (ref to DDIR)

		; Copy the string
		mov	edx, ebx				; Destination addr
		mov	ebx, PARAM1				; Source addr
		sub	ebx, size strentry
		mov	eax, [ebx.str_currlen]			; Source string length

		; Copy string
		push	esi edi
		lea	esi, [ebx + size strentry]		; Source address
		lea	edi, [edx + size strentry]		; Destination address
		lea	ecx, [eax + 1]				; String length + 1
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update string length
		mov	[edx.str_currlen], eax
		pop	eax					; Memory block nb
		push	edx
		clr	edx
		mov	ebx, size ddirentry
		idiv	ebx					; To save space (16-bits vs 32-bits)
		pop	edx
		mov	[edx.str_memblknb], eax			; Memory block nb (ref to DDIR)
		pop	ebx eax
		mov	[edx.str_incsize], bx
		sub	eax, (size strentry + 1)
		mov	[edx.str_currsize], eax


		; ----------------
		; Return to caller
		; ----------------

@@exit:		DISCARD	2		
		ret

E_STR_STORE:
STR_STORE	endp


; ===============================================================
; = STR_ADD							=
; =                                                 		=
; = Code for the forth word:	$ADD  or  $+			=
; =                     	                            	=
; = Stack:	( adrz1 adrz2 --- addrz3 )			=
; =								=
; =	addrz1 : source address					=
; =	addrz2 : destination address				=
; =	addrz3 : new destination address (may had changed)	=
; ===============================================================

		align	4
STR_ADD		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ------------
		; Fixed string
		; ------------

		; Fixed vs Dynamic string ?
		mov	edx, PARAM0
		sub	edx, size strentry
		cmp	[edx.str_incsize], 0
		jne	@@DynamicString

		; Length overflow ?
		mov	ebx, PARAM1
		sub	ebx, size strentry
		mov	eax, [ebx.str_currlen]
		add	eax, [edx.str_currlen]
		cmp	eax, [edx.str_currsize]
		jb	@@LengthOk
		
		; "String overflow."
		mov	ah, ERR_STRINGOVERFLOW			; Error #
		clr	edx					; No msg
		clr	ecx					; No line #
		call	doError

@@LengthOk:	
		; Copy string
		push	esi edi
		lea	esi, [ebx + size strentry]		; Source address
		lea	edi, [edx + size strentry]		; Destination address
		add	edi, [edx.str_currlen]			; is after the current string
		mov	ecx, [ebx.str_currlen]			; String length + 1
		inc	ecx
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update string length
		mov	[edx.str_currlen], eax

		; Done
		push	edx					; String header address
		jmp	@@exit
		
		
		; --------------
		; Dynamic string
		; --------------
		
@@DynamicString:	

		; Is there enought space allocated to hold the string ?
		mov	ebx, PARAM1
		sub	ebx, size strentry
		mov	eax, [ebx.str_currlen]
		add	eax, [edx.str_currlen]
		cmp	eax, [edx.str_currsize]
		jb	@@LengthOk

		; Compute new amount of memory to get
		mov	eax, [ebx.str_currlen]			; Source address length
		add	eax, [edx.str_currlen]			; + Destination address length
		movzx	ecx, [edx.str_incsize]
		clr	edx
		idiv	ecx
		inc	eax
		mov	edx, PARAM0				; Destination string addr
		sub	edx, size strentry
		imul	eax, ecx
		add	eax, size strentry
		mov	ecx, eax				; Save value

		; Save string infos values
		movzx	ebx, [edx.str_incsize]
		movzx	eax, [edx.str_wordref]			; Reference to the definition (NIA)
		push	eax					; Save for future use
		push	ecx ebx
		
		; Realloc the memory used
		mov	eax, ecx				; Asked memory
		mov	ecx, edx				; ECX = ptr to dest string header
		push	ecx
		mov	edx, [edx.str_memblknb]			; Memory block nb (ref to DDIR)
		call	ReAlloc
		pop	ecx
		push	eax					; Memory block nb (ref to DDIR)

		; Prepare the Copy of the string
		mov	edx, ebx				; Destination addr
		mov	ebx, PARAM1				; Source addr
		sub	ebx, size strentry
		mov	eax, [ebx.str_currlen]			; Source string length
		add	eax, [ecx.str_currlen]
		push	eax
		mov	eax, [ebx.str_currlen]			; Source string length

		; Copy the previous string
		push	ecx esi edi
		lea	esi, [ecx + size strentry]		; Source address
		lea	edi, [edx + size strentry]		; Destination address
		mov	ecx, [ecx.str_currlen]			; String length + 1
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi ecx

		; Copy the added string
		push	esi edi
		lea	esi, [ebx + size strentry]		; Source address
		lea	edi, [edx + size strentry]		; Destination address
		add	edi, [ecx.str_currlen]			; is after current string
		lea	ecx, [eax + 1]				; String length + 1
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		; Update string header
		pop	eax
		mov	[edx.str_currlen], eax
		push	edx
		clr	edx
		mov	ebx, size ddirentry
		idiv	ebx					; To save space (16-bits vs 32-bits)
		pop	edx
		pop	eax					; Memory block nb
		mov	[edx.str_memblknb], eax			; Memory block nb (ref to DDIR)
		pop	ebx ecx
		mov	[edx.str_incsize], bx
		sub	ecx, (size strentry + 1)
		mov	[edx.str_currsize], ecx
		pop	ebx					; Restore reference to the definition (NIA)
		mov	[edx.str_wordref], bx			; Reference to the definition (NIA)
		push	edx

		; Update the reference to the block nb (DDIR) into the word (defined by DSTRING)
		imul	ebx, size wordentry
		mov	edx, [NIA_START]
		mov	edx, [ebx + edx.word_realaddr]		; Real address of the code
		add	eax, ddir_blkaddr			; Block address (relative to the start of Data Dict)
		mov	[edx + 5 + 4  + 2], eax			; Store NIA block ref nb
								; 5: E8 00 00 00 00	(call $)
								; 4: 82 04 24 10	(add dword ptr [esp], 10h)
								; 2: 8b 87		(mov eax, [edi + ....])

		; ----------------
		; Return to caller
		; ----------------

@@exit:		DISCARD	1
		pop	eax					; String header address
		add	eax, size strentry
		mov	PARAM0, eax
		ret

E_STR_ADD:
STR_ADD		endp


; ===============================================================
; = ADD_STR							=
; =                                                 		=
; = Code for the forth word:	ADD$  or  +$			=
; =                     	                            	=
; = Stack:	( adrz1 adrz2 --- addrz3 )			=
; =								=
; =	addrz1 : source address					=
; =	addrz2 : destination address				=
; =	addrz3	  : nez=w destination address (may had chaged)	=
; ===============================================================

		align	4
ADD_STR		proc

		; SWAP $+
		SUMMON	SWAP			; SWAP
		SUMMON	STR_ADD			; $+

		; Return to caller
		ret

E_ADD_STR:
ADD_STR		endp


; ===============================================================
; = STR_REPL							=
; =                                                 		=
; = Code for the forth word:	$REPLACE or $REPL		=
; =                     	                            	=
; = Stack:	( adrz1 adrz2 n --- )				=
; =								=
; =	addrz1    : source address				=
; =	addrz2, n : destination address				=
; ===============================================================

		align	4
STR_REPL	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; ---------------------------------------------------------
		; Test if string to replace can fit into destination string
		; ---------------------------------------------------------

		; Length of source string
		mov	ecx, PARAM2
		mov	ebx, [ecx.str_currlen - size strentry]

		; Test
		mov	edx, PARAM1
		add	edx, PARAM0
		clr	eax
		dec	eax
		dec	edx
@@LL1:		inc	eax
		cmp	eax, ebx
		jge	@@Ok
		inc	edx
		cmp	byte ptr [edx], 0
		jne	@@LL1

		; "String overflow."
		mov	ah, ERR_STRINGOVERFLOW			; Error #
		clr	edx					; No msg
		clr	ecx					; No line #
		call	doError

@@Ok:

		; ---------------
		; Do the copy now
		; ---------------

		mov	edx, PARAM1
		add	edx, PARAM0
		push	esi edi
		mov	esi, ecx				; Source address
		mov	edi, edx				; Destination address
		mov	ecx, ebx				; String length
		cld                     			; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi


		; ----------------
		; Return to caller
		; ----------------

@@exit:		DISCARD	3
		ret

E_STR_REPL:
STR_REPL	endp


; ===============================================================
; = STR_SHRINK							=
; =                                                 		=
; = Code for the forth word:	STRING.SHRINK			=
; =                     	                            	=
; = Stack:	( adrz --- )					=
; =								=
; =	addrz : destination address				=
; ===============================================================

		align	4
STR_SHRINK	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ----------------
		; Return to caller
		; ----------------

@@exit:		DISCARD	1
		ret

E_STR_SHRINK:
STR_SHRINK	endp


; ===============================================================
; = STR_FIND							=
; =                                                 		=
; = Code for the forth word:	$FIND				=
; =                     	                            	=
; = Stack:	( addrz1 addrz2 --- addrz2 n )			=
; =								=
; =	addrz1 : What is search ?				=
; =	addrz2 : Source address					=
; =	n      : -1 if not found, else position 1st char.	=
; ===============================================================

		align	4
STR_FIND	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ------------------------------
		; Fist, check the strings length
		; ------------------------------

		mov	ecx, PARAM1				; Get addrz1 (what to find ?)
		mov	edx, PARAM0				; Get addrz2 (where to find ?)
		mov	PARAM1,edx

		mov	eax, [ecx.str_currlen - size strentry]	; Length of what to find
		cmp	eax, [edx.str_currlen - size strentry]	; ?? length of where to find
		jg	@@NotFound


		; ---------------
		; Make the search
		; ---------------

		clr	ebx

@@Find:		mov	al, [ecx + ebx]
		cmpz	al, 0
		je	@@Found

		mov	ah, [edx + ebx]
		cmpz	ah, 0
		je	@@NotFound

		cmp	al, ah
		jne	@@Find2

		; Go to next char into source
		inc	ebx
		jmp	@@Find

		; Go to next char into destination
@@Find2:	inc	edx
		clr	ebx
		jmp	@@Find


		; ----------------
		; Return to caller
		; ----------------

@@NotFound:	mov	PARAM0, -1
		ret

@@Found:	sub	edx, PARAM1
		mov	PARAM0, edx
		ret

E_STR_FIND:
STR_FIND	endp


; ===============================================================
; = STR_COPY							=
; =                                                 		=
; = Code for the forth word:	$COPY				=
; =                     	                            	=
; = Stack:	( addrz1 n1 n2 addrz2 --- )			=
; =								=
; =	addrz1 : Source address					=
; =	n1, n2 : from pos. n1, copy n2 char.			=
; =	addrz2 : Destination address				=
; ===============================================================

		align	4
STR_COPY	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ----------------
		; Return to caller
		; ----------------

@@exit:		DISCARD	1
		ret

E_STR_COPY:
STR_COPY	endp


; ===============================================================
; = STR_COPY_TOC						=
; =                                                 		=
; = Code for the forth word:	$CopyToChar			=
; =                     	                            	=
; = Stack:	( addrz1 n c addrz2 --- )			=
; =								=
; =	addrz1 : Source address					=
; =	n, c   : from pos. n1, to char c			=
; =	addrz2 : Destination address				=
; ===============================================================

		align	4
STR_COPY_TOC	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	4


		; -------------------------------
		; Compute length of source string
		; -------------------------------

		mov	ecx, PARAM3				; Get addrz1 (source string)
		add	ecx, PARAM2				; Where to start ?
		mov	eax, PARAM1				; Get char which stops copy (in AL)
		clr	ebx
@@CompLen:	mov	ah, [ecx + ebx]
		cmpz	ah, 0
		je	@@CompLenOk
		cmp	ah, al
		je	@@CompLenOk
		inc	ebx
		jmp	@@CompLen

@@CompLenOk:	


		; -------------
		; Make the copy
		; -------------

		push	esi edi
		mov	edx, ebx				; EDX: # of char to copy
		mov	ebx, PARAM2				; EBX: Offset in source
		mov	edi, PARAM0				; EDI: Destination address
		mov	esi, PARAM3				; ESI: Source address
		clr	ecx					; ECX: Offset in destination
		mov	al, 1					; AL : 1 if store '\0' ad the end of the copy
		call	_CopyString
		pop	edi esi


		; ----------------
		; Return to caller
		; ----------------

@@exit:		DISCARD	4
		ret

E_STR_COPY_TOC:
STR_COPY_TOC	endp


; ===============================================================
; = STR_CONVNB							=
; =                                                 		=
; = Code for the forth word:	$ConvNb				=
; =                     	                            	=
; = Stack:	( addrz n1 --- n2 n3 )				=
; =	    or	( addrz n1 --- -1 )				=
; =								=
; =	addrz  : destination address				=
; =	n1, n2 : offset						=
; =	n3     : the number					=
; ===============================================================

		align	4
STR_CONVNB	proc

		; Save used registers
		push	edi ebp

		; Get the base in ECX
		GETFVAR	ecx, D_BASE				; Get base
		mov	dl, cl					; Save current base in DL
		sub	ecx, 2
		mov	ecx, [POWERTAB + ecx * 4]

		; Is the 1st character '-' ?
		clr	dh					; Flag: dh=1 if '-'
		mov	ebp, PARAM1				; Get <addrz>
		add	ebp, PARAM0				; Add <n1>
		cmp	byte ptr [ebp], '-'
		jne	@@NotMinus
		inc	ebp
		inc	dh
@@NotMinus:

		; EBP: address of first char of the number
		; EDI: address of last  char of the number
		mov	edi, ebp
@@LL1:		mov	al, [edi]
		cmpz	al, 0
		je	@@LL2
		TOUPPER	al
		TODIGIT	al
		cmp	al, -1					; Not a digit ?
		je	@@LL2
		cmp	al, dl
		jae	@@LL2					; Not a valid digit ?
		inc	edi
		jmp	@@LL1
@@LL2:		dec	edi

		; Conversion failed ?
		cmp	ebp, edi
		ja	@@NOK

		; Store index of stop number conversion
		mov	eax, edi
		sub	eax, PARAM1
		inc	eax
		mov	PARAM1, eax

		; Init number being converted
		mov	PARAM0, 0

		; Convert next digit
@@LL3:		clr	eax
		mov	al, [edi]

		; If lowercase, convert to uppercase, then convert to digit
		TOUPPER	al
		TODIGIT	al
		
		; digit * power
		mov	ebx, [ecx]
		imul	eax, ebx
		add	PARAM0, eax

		; Test overflow error
		jno	@@NotOverflow

		; "Constant integer out of range."		
		add	esp, 2 * @WordSize			; Clean stack
		mov	ah, ERR_CONSTINTOUTOFRANGE		; Error msg #
		mov	edx, ebp				; Additionnal msg
		clr	ecx					; No line #
		call	doError

@@NotOverflow:	
		; Next digit
		cmp	edi, ebp
		je	@@OK
		dec	edi
		add	ecx, 4
		jmp	@@LL3

		; Not a valid number
@@NOK:		DISCARD	1					; drop
		mov	PARAM0, -1				; ( n --- -1 )
		jmp	@@LL5

		; A valid number
@@OK:		cmpz	dh, FALSE
		jz	@@NotNeg
		neg	PARAM0
@@NotNeg:


		; End of proc
		; -----------

@@LL5:
		; Restore EBP and EDI which are used localy
		pop	ebp edi
		cld

		; Return to caller
		ret

E_STR_CONVNB:
STR_CONVNB	endp


; ===============================================================
; = STR_SRCHCHR							=
; =                                                 		=
; = Code for the forth word:	$SearchChr			=
; =                     	                            	=
; = Stack:	( n c addrz --- )				=
; ===============================================================

		align	4
STR_SRCHCHR	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; ----------------
		; Return to caller
		; ----------------

@@exit:		DISCARD	1		
		ret

E_STR_SRCHCHR:
STR_SRCHCHR	endp


; ===============================================================
; = STR_SRCHSTR							=
; =                                                 		=
; = Code for the forth word:	STRING.SrchStr			=
; =                     	                            	=
; = Stack:	( n c addrz --- )				=
; ===============================================================

		align	4
STR_SRCHSTR	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; ----------------
		; Return to caller
		; ----------------

@@exit:		DISCARD	1		
		ret

E_STR_SRCHSTR:
STR_SRCHSTR	endp


; ===============================================================
; = INPUT							=
; =                                                 		=
; = Code for the forth word:	INPUT				=
; =                     	                            	=
; = Stack:	( addrz n --- )					=
; ===============================================================

		align	4
INPUT		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; -----------------------------------------
		; Verify that string can contains <n> chars
		; -----------------------------------------
		
		mov	eax, [esi + @WordSize]			; String address
		mov	ebx, [eax - 4]				; Max length given for the string
		mov	ecx, [esi]				; Max length asked
		cmp	ecx, ebx
		jbe	@@LengthOk

		; "String overflow"
		mov	ah, ERR_STRINGOVERFLOW			; String overflow
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@LengthOk:


		; ----
		; Loop
		; ----
		
		; Initialize
		clr	ecx					; Char counter

		; Read a key
@@ReadKey:	SUMMON	KEY
		
		; CR ?
		mov	eax, [esi]
		add	esi, @WordSize
		cmp	ax, EKEY_CR
		je	@@EndInput

		; BS ?
		cmp	ax, EKEY_BS
		jne	@@NotBS
		cmpz	cx, 0
		je	@@ReadKey
		SUMMON	PDOTQ
		db	8, ' ', 8, 0
		dec	ecx
		jmp	@@ReadKey
@@NotBS:
		
		; A key to take: length too big ?
		cmp	ecx, [esi]
		je	@@ReadKey

		; Char visible ?
		cmp	ax, 020h
		jb	@@ReadKey
		cmp	ax, 100h
		jae	@@ReadKey

		; Take the key
		call	COUT
		mov	ebx, [esi + @WordSize]
		mov	[ebx + ecx], al
		inc	ecx
		jmp	@@ReadKey
				
				
		; ------------
		; End of input
		; ------------

@@EndInput:	

		; Put a zero at the end of the string
		mov	ebx, PARAM1				; Get addr
		mov	byte ptr [ebx + ecx], 0

		; Clean stack
		DISCARD	2
		
		
		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_INPUT:
INPUT		endp


code32		ends
		end
