; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	USE.ASM							!
; 									!
; Description:	Definitions for playing with one-line help.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	USE		USE						!
;	USINGQ		USING?						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Ed.inc                      		; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = USE                              				=
; =                                                         	=
; = Code for the forth word:    USE              		=
; =                                                         	=
; = Stack:  ( --- )                     			=
; =                                                         	=
; = Input stream: USE name					=
; ===============================================================

		align	4
USE		proc

		; ------------------------------------------------
		; Issue an error is interpreted from a loaded file
		; ------------------------------------------------

		SUMMON	QLOAD
		

		; -------------------------------------
		; Isolate next word in the input stream
		; -------------------------------------

		; Found next word in input stream buffer
		mov	edx, [IBUFFPTR2]
		inc	edx

		; Strip space chars: goto on first not blank char
@@LL2:		mov	al, [edx]
		cmpz	al, 0
		jne	@@LL6

		; NULL word
		SUMMON	PDOTQ
		db	13, 10
		db	" Usage: use word_name", 13, 10
		db	"Object: give quick help about a given froth word."
		db	0
		jmp	@@Exit

@@LL6:		cmp	al, 09h
		je	@@LL7
		cmp	al, ' '
		jne	@@LL3
@@LL7:		inc	edx
		jmp	@@LL2

@@LL3:
		; First, save current ptr: the start of the word
		mov	ecx, edx

		; Goto now on the first blank, or end of buffer
@@LL4:		mov	al, [ecx]
		cmpz	al, 0
		je	@@LL5
		cmp	al, 09h
		je	@@LL5
		cmp	al, ' '
		je	@@LL5
		inc	ecx
		jmp	@@LL4

		; Found the end of the word
@@LL5:		mov	byte ptr [ecx], 0
		mov	[IBUFFPTR2], ecx			; Current pointer into the buffer


		; -------------------------------------------
		; Is this word defined into the dictionnary ?
		; -------------------------------------------

		IPUSH	edx		
		call	PFind
		mov	eax, PARAM0				; Get flag (word found ?)
		DISCARD	1					; drop
		cmpz	eax, 0
		jne	@@WordDef
		
		; Error: word not defined
		mov	ah, ERR_WORDNDEF			; Word not defined
		clr	ecx					; No line #
		call	doError
		
@@WordDef:


		; -----------------------
		; Is this word 'system' ?
		; -----------------------

		; Get word NIA address (in EDX)
		DISCARD	1
		mov	edx, ebx
		add	edx, [NIA_START]
		
		; Word 'system' ?
		test	[edx.word_flags], FSYSTEM
		jnz	@@System
		
		; Error: word is not system
		mov	ah, ERR_WORDHLPNSYS			; No help available on user words
		add	edx, word_name				; EDX: word name
		clr	ecx					; No line #
		call	doError
		
@@System:

		; ----------------------------------------------
		; Search the asked word name into help directory
		; ----------------------------------------------
		
		add	edx, word_name				; EDX: word name
@@SearchWord:	mov	ebp, [HelpDir]
		add	ebp, 4
		
@@SrchWord:	
		; End of list ?
		cmp	dword ptr [ebp], -1
		je	@@NotFound
	
		; Compare word names
		clr	ecx					; Char counter
@@CmpWrdNames:	mov	al, [ebp + ecx]
		TOUPPER	al
		mov	ah, [edx + ecx]
		TOUPPER	ah
		cmp	al, ah
		jne	@@NextWord
		or	ah, al
		je	@@Found
		inc	ecx
		jmp	@@CmpWrdNames
		
@@NextWord:	add	ebp, size hlpentry
		jmp	@@SrchWord
		

		; -------------------------------------
		; Word name is not found into directory
		; -------------------------------------
		
@@NotFound:	SUMMON	PDOTQ
		db	13, 10
		db	"No help available for this word."
		db	0
		jmp	@@Exit


		; ---------------------------------
		; Word name is found into directory
		; ---------------------------------

@@Found:
		; Pointer to text into file
		mov	ebp, [ebp - 4]
		
		; Move the file pointer
		mov	al, SEEK_SET				; Move file pointer from beginning of file
		mov	bx, [HelpFileHdl]			; File handle
		mov	ecx, ebp				; Current position in the file
		call	doSeekFile
		jc	@@Exit

		; Read help text into buffer
		mov	bx, [HelpFileHdl]			; File handle
		mov	ecx, MAXSZ_HELP_BUFFER		 	; Bytes to read
		mov	edx, [HelpBuff]				; Buffer address
		call	doReadFile				; in EBX: bytes read from the file
		jc	@@Exit
		
		; Is-it a reference to another word ?
		mov	edx, [HelpBuff]				; Buffer address
		cmp	byte ptr [edx], 015h
		jne	@@NoRefWord
		inc	edx					; Pointer to new word name to found
		jmp	@@SearchWord
@@NoRefWord:


		; ---------------------------
		; Ask the width od the screen
		; ---------------------------
		
		call	EdGetVideoMode
		shr	ax, 8
		movzx	eax, ax
		sub	eax, 20
		mov	[@@nbcols], eax				; Nb of columns


		; -------------------
		; Print the help text
		; -------------------

		SUMMON	CR		
		clr	ebx					; Left margin
		clr	ecx					; Chars on the line
		mov	ebp, [HelpBuff]
		
@@ReadLine:		
		mov	al, [ebp]
		cmpz	al, 0
		je	@@End
		cmp	al, 014h				; Force CR ?
		jne	@@RD2

		; Force a CR
		SUMMON	CR
		clr	ebx					; Left margin
		clr	ecx					; Chars on the line
		jmp	@@NextChar

		; Left margin ?		
@@RD2:		cmp	al, 010h
		jne	@@RD3
		cmpz	ebx, 0
		je	@@RD5
		add	ebp, ebx				; Skip margin
		dec	ebp
		mov	al, ' '
		jmp	@@OK
@@RD5:		mov	ebx, ecx
		jmp	@@NextChar
	
		; Break the line ?		
@@RD3:		cmp	ecx, [@@nbcols]
		jb	@@OK
		cmp	al, ' '
		jne	@@OK
		
		; CR
		SUMMON	CR
		clr	ecx					; Chars on the line
		clr	edx
@@RD4:		cmp	edx, ebx
		je	@@NextChar
		mov	al, ' '
		call	COUT
		inc	edx
		jmp	@@RD4
		
		; Print char
@@OK:		call	COUT
		inc	ecx
		
@@NextChar:	inc	ebp
		jmp	@@ReadLine
		
@@End:


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

; ------------------------
; --> Local variables --->
; ------------------------

		align	4
@@nbcols	dd	?

E_USE:
USE      	endp


ifdef UNDERDEV

; ===============================================================
; = USINGQ                              			=
; =                                                         	=
; = Code for the forth word:    USING?              		=
; =                                                         	=
; = Stack:  ( --- )                     			=
; ===============================================================

		align	4
USINGQ		proc

		; ---------------
		; Local variables
		; ----------------

t_LstFileHdl	equ	word ptr [ebp - 4]
t_NbWords	equ	word ptr [ebp - 8]
t_Char		equ	word ptr [ebp - 12]
		enter	3 * @WordSize, 0


		; --------------------------------------------------
		; Scan each defined word, see if there is help on it
		; --------------------------------------------------
		
		mov	ecx, [NIA_START]


		; -------------------------------------
		; Create the file '\froth\WordsLst.txt'
		; -------------------------------------
		
		; Try to create 
		lea	ebx, LstWordsFileN			; Ptr to filename
		mov	ax, 00h					; File attribute : normal
		call	doCreateFile
		cmpz	ax, 0
		je	@@CreatOk

		; "Unable to create file '\FROTH\WORDSLST.TXT'"
		SUMMON	EdPDotQ
		db	13, 10
		db	"Unable to create file '"
		db	WORDSLIST_FILENAME
		db	" '."
		db	13, 10, 0
		jmp	@@Exit
		
@@CreatOk:	mov	[t_LstFileHdl], bx			; File handle
		mov	[t_NbWords], 0				; Nb of words wrote to file


		; -------------
		; End of list ?		
		; -------------
		
@@loop:		mov	edx, [NIA_START]
		add	edx, [NIA_SIZE]
		cmp	ecx, edx
		je	@@Exit
		
		; Write the word name into the file '\FROTH\WORDSLST.TXT'
		pushad
		;	Word length
		lea	edx, [ecx.word_name]			; EDX: word name
		clr	ecx
@@WordLen:	inc	ecx
		cmp	byte ptr [edx + ecx - 1], 0
		jne	@@WordLen
		mov	byte ptr [edx + ecx - 1], 09h		; TAB char
		;	Write to file
		pushad
		mov	bx, [t_LstFileHdl]			; File handle
;		mov	ecx, ecx				; Bytes to write
;		mov	edx, edx				; Buffer address
		call	doWriteFile
		popad
		;	End
		mov	byte ptr [edx + ecx - 1], 0
		;	Skip a line each 6 word
		inc	[t_NbWords]
		cmp	[t_NbWords], 5
		jne	@@Not6W
		mov	[t_Char], 0A0Dh				; CR LF
		mov	bx, [t_LstFileHdl]			; File handle
		mov	ecx, 2					; Bytes to write
		lea	edx, t_Char				; Buffer address
		call	doWriteFile
		mov	[t_NbWords], 0
@@Not6W:
		popad
		
		; Skip run-time words (no help on these words)
		test	byte ptr [ecx.word_flags], FRUNTIME
		jnz	@@NextWord
	
		; ----------------------------------------------
		; Search the asked word name into help directory
		; ----------------------------------------------

		lea	edx, [ecx.word_name]			; EDX: word name
		mov	ebx, [HelpDir]
		add	ebx, 4
		
@@SrchWord:	
		; End of list ?
		cmp	dword ptr [ebx], -1
		je	@@NotFound
	
		; Compare word names
		push	ecx
		clr	ecx					; Char counter
@@CmpWrdNames:	mov	al, [ebx + ecx]
		TOUPPER	al
		mov	ah, [edx + ecx]
		TOUPPER	ah
		cmp	al, ah
		jne	@@NextHWord
		or	ah, al
		je	@@NextWord0
		inc	ecx
		jmp	@@CmpWrdNames
		
@@NextHWord:	pop	ecx
		add	ebx, size hlpentry
		jmp	@@SrchWord
		
		; Word is not found: show it
@@NotFound:	lea	eax, [ecx.word_name]			; EDX: word name
		IPUSH	eax
		SUMMON	_TYPE
		SUMMON	SPACE
		jmp	@@NextWord
		

		; ---------
		; Next word
		; ---------

@@NextWord0:	pop	ecx		
@@NextWord:	add	ecx, size wordentry
		jmp	@@loop


		; ------------------------------------
		; Close the file '\froth\WordsLst.txt'
		; ------------------------------------

@@Exit:		mov	bx, [t_LstFileHdl]
		call	doCloseFile


		; ----------------
		; Return to caller
		; ----------------
		
		leave
		ret

E_USINGQ:
USINGQ		endp

endif

code32		ends
		end
